/*****************************************************************************
 *
 * Program Name:  NetWare 4X ODI Logical Device Enumerator .h file
 *
 * Filename:	  odildec.h
 *
 * Date Created:  Sept 18, 1996
 *
 * Version:		  1.0
 *
 * Programmers:	  Mark Dakins
 *
 * Files used:
 *
 * Date Modified: 
 *
 * Modifications: 
 *
 * 961025 MCD added ODINEB_PARM_BLOCK so MSM can call into this module
 *
 * 970207 JRT modified ODINEB_PARM_BLOCK to match ODI_NESL.H definitions
 *
 * Comments:	  
 *
 * COPYRIGHT (c) 1987 by Novell, Inc.  All Rights Reserved.
 *
 ****************************************************************************/

#define MSM_ODILDE_MODULE_NAME "MSM.NLM"
#define MSM_ODILDE_PHYS_DEV_UNLOAD "HW.PhysicalDeviceUnload"
#define MSM_ODILDE_PHYS_DEV_PWR_CYLCE_REPLY "HW.PhysDevPwrCycleReply"
#define MSM_ODILDE_LOG_DEV_UNLOAD "HW.LogicalDeviceUnload"
#define MSM_ODILDE_PHYS_DEV_UNLOAD_REPLY "HW.PhysicalDeviceUnloadReply"
#define MSM_ODILDE_PHYS_DEV_FAIL "HW.PhysicalDeviceFailure"
#define MSM_ODILDE_PHYS_DEV_PWR_CYCLE "HW.PhysDevPwrCycle"


typedef struct _ODINEB_STACK_
{
   struct _MODULE_HANDLE_  *ModuleHandle;
   void                    *ScreenHandle;
   MEON                    *CommandLine;
   MEON                    *ModuleLoadPath;
   UINT32                  UnitializedDataLength;
   void                    *CustomDataFileHandle;
   UINT32                  (* FileRead)(
                              void      *FileHandle,
                              UINT32   FileOffset,
                              void      *FileBuffer,
                              UINT32   FileSize);
   UINT32                  CustomDataOffset;
   UINT32                  CustomDataSize;
   UINT32                  NumMsgs;
   MEON                    **Msgs;
}
ODINEB_STACK;


typedef struct   _ODINEB_PARM_BLOCK_
{
   UINT32                  ODINEBParameterSize;
   UINT32                  ODINEBParameterVersion;
   struct _ODINEB_STACK_   *ODINEBInitParmPointer;
   struct _MODULE_HANDLE_  *ODINEBModuleHandle;
   MEON                    ***ODINEBMessagesPtr;

   UINT32 (_cdecl *NESLRegisterProducer)( NESL_ECB *Producer );
   UINT32 (_cdecl *NESLDeRegisterProducer)( NESL_ECB *Producer );
   UINT32 (_cdecl *NESLRegisterConsumer)( NESL_ECB *Consumer );
   UINT32 (_cdecl *NESLDeRegisterConsumer)( NESL_ECB *Consumer );
   UINT32 (_cdecl *NESLProduceEvent)( NESL_ECB *Producer, NESL_ECB **Consumer, void *EventData );
   UINT32 (_cdecl *NESLProduceMLIDEvent)( NESL_ECB *Producer, NESL_ECB **Consumer, 
                              void *EventData, struct _DRIVER_DATA_ *driverData );

}
ODINEB_PARM_BLOCK;

/****************************************************************\
*                                                                *
* LoadDefinitionStructure - Fields accessable from               *
* ModuleHandle, which is passed in to the NLMs/LANs              *
* initialization procedure. It is normally passed to OS          *
* routines during NLM/LAN initialization, like                   *
* AllocateResourceTag.                                           *
* The HSM is not to modify any of the fields in this structure.  *
*                                                                *
\****************************************************************/

/* LoadDefinitionStructure */

typedef struct _MODULE_HANDLE_
{
   struct         _MODULE_HANDLE_ *LDLink;
   struct         _MODULE_HANDLE_ *LDKillLink;
   struct         _MODULE_HANDLE_ *LDScanLink;
   struct         _RTAG_HANDLE_ **LDResourceList;
   UINT32         LDIdentificationNumber;
   UINT8          *LDCodeImageOffset;
   UINT32         LDCodeImageLength;
   UINT8          *LDDataImageOffset;
   UINT32         LDDataImageLength;
   UINT32         LDUninitializedDataLength;
   UINT8          *LDCustomDataOffset;
   UINT32         LDCustomDataSize;
   UINT32         LDFlags;
   UINT32         LDType;
   UINT32         (_cdecl * LDInitializationProcedure)(
                     struct   _MODULE_HANDLE_ *LoadRecord,
                     struct   ScreenStruct *screenID,
                     MEON       *CommandLine,
                     MEON      *loadDirectoryPath,
                     UINT32   uninitializedDataLength,
                     UINT32   fileHandle,
                     UINT32   (*ReadRoutine)(
                                 UINT32    fileHandle,
                                 UINT32    offset,
                                 void      *buffer,
                                 UINT32   numberOfBytes),
                     UINT32 customDataOffset,
                     UINT32 customDataSize,
                     UINT32 NumMsgs,
                     MEON    **Msgs);
   void           (_cdecl * LDExitProcedure)(void);
   UINT32         (_cdecl * LDCheckUnloadProcedure)(
                     struct ScreenStruct *screenID);
   void           *LDPublics;
   MEON           LDFileName[36];
   MEON           LDName[128];
   UINT32         *LDCLIBLoadStructure;
   UINT32         *LDNLMDebugger;
   UINT32         LDParentID;
   UINT32         LDReservedForCLIB;
   UINT32         LDNumberOfReferencedModules;                /* Used in NT. */
   struct         _MODULE_HANDLE_   **LDReferencedModules;   /* Used in NT. */
   void           *LDModuleObjectHandle;
   UINT32         LDMajorVersion;
   UINT32         LDMinorVersion;
   UINT32         LDRevision;
   UINT32         LDYear;
   UINT32         LDMonth;
   UINT32         LDDay;
   MEON           *LDCopyright;
   UINT32         LDAllocAvailBytes;
   UINT32         LDAllocFreeCount;
   UINT32         LDLastGarbCollect;
   UINT32         LDAlloc16Lists[64];
   UINT32         LDAlloc256Lists[12];
   UINT32         LDAlloc4kList;
   void           *LDDomainID;
   struct         _MODULE_HANDLE_ *LDEnvLink;
   void           *LDAllocPagesListHead;
   void           *LDTempPublicList;
   UINT32         LDMessageLanguage;       /* for enabling */
   MEON           **LDMessages;            /* for enabling */
   UINT32         LDMessageCount;          /* for enabling */
   MEON           *LDHelpFile;             /* for enabling */
   UINT32         LDMessageBufferSize;     /* for enabling */
   UINT32         LDHelpBufferSize;        /* for enabling */
   UINT32         LDSharedCodeOffset;      /* for protection */
   UINT32         LDSharedCodeLength;      /* for protection */
   UINT32         LDSharedDataOffset;      /* for protection */
   UINT32         LDSharedDataLength;      /* for protection */
   UINT32         (_cdecl * LDSharedInitProcedure)(
                     struct   _MODULE_HANDLE_ *LoadRecord,
                     struct   ScreenStruct *screenID,
                     MEON      *CommandLine);
   void            (_cdecl * LDSharedExitProcedure)(void);
   UINT32         LDRPCDataTable;
   UINT32         LDRealRPCDataTable;
   UINT32         LDRPCDataTableSize;
   UINT32         LDNumberOfReferencedPublics;
   void           **LDReferencedPublics;
   UINT32         LDNumberOfReferencedExports;
   void           *LDUnresolvedExternals;
   void           *LDSharedUnresolvedExternals;
}
MODULE_HANDLE;


