//++=======================================================================
// Program Name:        NetWare Bus Interface
// File Name:           nbintmac.h
// Version:             v0.01
// Author:              KRA (960202)
//
// Abstract:				Defines macors for other driver to access the NBI.
//
// Notes:               It is assumed that pNBISupportAPI is defined by the
//                      user of this file and the name will always be the same.
//
// Revision History:
// 960906	KRA
// Changed prototypes for RdConfigSpace16 and RdConfigSpace to return
// UINT16 and UINT32 respectiverly rather than UINT8.
//
// 970305 KRA
// Added NBIRequestIdx and the NBIRequest macro.
//
// 970331 KRA
// Added MapBusMemoryIdx, FreeBusMemoryIdx, NBIMapBusMemroy macro and
// NBIFreeBusMemory macro.
//
// (C) Unpublished Copyright of Novell, Inc.  All Rights Reserved.
//
// No part of this file may be duplicated, revised, translated, localized,
// or modified in any manner or compiled, linked or uploaded or downloaded
// to or from any computer system without the prior written consent of
// Novell, Inc.
//=======================================================================--
extern UINT32 (**pNBISupportAPI)(void);

#define GetAlignmentIdx								0x00
#define GetBusInfoIdx								0x01
#define GetBusNameIdx								0x02
#define GetBusSpecificInfoIdx						0x03
#define GetBusTagIdx									0x04
#define GetBusTypeIdx								0x05
#define GetCardConfigInfoIdx						0x06
#define GetInstanceNumberIdx						0x07
#define GetInstanceNumberMappingIdx				0x08
#define GetNBIConfigurationIdx					0x09
#define GetSlotIdx									0x0a
#define GetSlotNameIdx								0x0b
#define GetUniqueIdentifierIdx					0x0c
#define GetUniqueIdentifierParametersIdx		0x0d
#define RdConfigSpace8Idx							0x0e
#define RdConfigSpace16Idx							0x0f
#define RdConfigSpace32Idx							0x10
#define ScanBusInfoIdx								0x11
#define SearchAdapterIdx							0x12
#define WrtConfigSpace8Idx							0x13
#define WrtConfigSpace16Idx						0x14
#define WrtConfigSpace32Idx						0x15
#define DMACleanupIdx								0x16
#define DMAStartIdx									0x17
#define DMAStatusIdx									0x18
#define SlowIdx										0x19
#define NBIRequestIdx								0x1a
#define MapBusMemoryIdx								0x1b
#define FreeBusMemoryIdx							0x1c

#define GetAlignment(x)										\
	((UINT32 (*)(UINT32))									\
	pNBISupportAPI[GetAlignmentIdx])((x))

#define GetBusInfo(x, y, z)								\
   ((ODI_NBI (*)(void *,UINT32 *,UINT32 *))			\
	pNBISupportAPI[GetBusInfoIdx])((x), (y), (z))

#define GetBusName(x, y)									\
   ((ODI_NBI (*)(void *,MEON_STRING **))				\
	pNBISupportAPI[GetBusNameIdx])((x), (y))

#define GetBusSpecificInfo(x, y, z)						\
   ((ODI_NBI (*)(void *,UINT32, void *))				\
	pNBISupportAPI[GetBusSpecificInfoIdx])((x), (y), (z))

#define GetBusTag(x, y)										\
   ((ODI_NBI (*)(const MEON_STRING *,void **))			\
	pNBISupportAPI[GetBusTagIdx])((x), (y))

#define GetBusType(x, y)									\
   ((ODI_NBI (*)(void *,UINT32 *))						\
	pNBISupportAPI[GetBusTypeIdx])((x), (y))

#define GetCardConfigInfo(u, v, w, x, y, z)			\
   ((ODI_NBI (*)(void *,UINT32, UINT32, UINT32,UINT32,void *))	\
	pNBISupportAPI[GetCardConfigInfoIdx])((u), (v), (w), (x), (y), (z))

#define GetInstanceNumber(x, y, z)						\
   ((ODI_NBI (*)(void *, UINT32, UINT16 *))		       	\
	pNBISupportAPI[GetInstanceNumberIdx])((x), (y), (z))

#define GetInstanceNumberMapping(x, y, z)				\
   ((ODI_NBI (*)(UINT16, void **, UINT32 *))			\
	pNBISupportAPI[GetInstanceNumberMappingIdx])((x), (y), (z))

#define GetNBIConfiguration(x, y)  						\
   ((ODI_NBI (*)(UINT32, NBIConfiguration *))  			\
	pNBISupportAPI[GetNBIConfigurationIdx])((x), (y))

#define GetSlot(x, y, z)									\
   ((ODI_NBI (*)(void *, MEON_STRING *, UINT16 *))		\
	pNBISupportAPI[GetSlotIdx])((x), (y), (z))

#define GetSlotName(x, y, z)						       		\
	((ODI_NBI (*)(void *, UINT16, MEON_STRING **))		\
	pNBISupportAPI[GetSlotNameIdx])((x), (y), (z))

#define GetUniqueIdentifier(w, x, y, z)				\
	((ODI_NBI (*)(void *, UINT32 *, UINT32, UINT32 *))		\
	pNBISupportAPI[GetUniqueIdentifierIdx])((w), (x), (y), (z))

#define GetUniqueIdentifierParameters(w, x, y, z)		\
	((ODI_NBI (*)(void *, UINT32, UINT32, UINT32 *))		\
	pNBISupportAPI[GetUniqueIdentifierParametersIdx])((w), (x), (y), (z))

#define RdConfigSpace8(x, y, z)							\
	((UINT8 (*)(void *, UINT32, UINT32))				\
	pNBISupportAPI[RdConfigSpace8Idx])((x), (y), (z))

#define RdConfigSpace16(x, y, z)							\
	((UINT16 (*)(void *, UINT32, UINT32))				\
	pNBISupportAPI[RdConfigSpace16Idx])((x), (y), (z))

#define RdConfigSpace32(x, y, z)							\
	((UINT32 (*)(void *, UINT32, UINT32))				\
	pNBISupportAPI[RdConfigSpace32Idx])((x), (y), (z))

#define ScanBusInfo(w, x, y, z)							\
	((ODI_NBI (*)(UINT32 *, void **, UINT32 *, MEON_STRING **))	\
	pNBISupportAPI[ScanBusInfoIdx])((w), (x), (y), (z))

#define SearchAdapter(u, v, w, x, y, z)				\
	((ODI_NBI (*)(UINT32 *, UINT32, UINT32,const MEON  *, void **, UINT32 *)) \
	pNBISupportAPI[SearchAdapterIdx])((u), (v), (w), (x), (y), (z))

#define WrtConfigSpace8(w, x, y, z)						\
	((void (*)(void *, UINT32, UINT32, UINT8))			\
	pNBISupportAPI[WrtConfigSpace8Idx])((w), (x), (y), (z))

#define WrtConfigSpace16(w, x, y, z)					\
	((void (*)(void *, UINT32, UINT32, UINT16))			\
	pNBISupportAPI[WrtConfigSpace16Idx])((w), (x), (y), (z))

#define WrtConfigSpace32(w, x, y, z)					\
	((void (*)(void *, UINT32, UINT32, UINT32))			\
	pNBISupportAPI[WrtConfigSpace32Idx])((w), (x), (y), (z))

#define NBIDMACleanup(x)	  								\
	((UINT32 (_cdecl *)(UINT32))									\
	pNBISupportAPI[DMACleanupIdx])((x))

#define NBIDMAStart(q, r, s, t, u, v, w, x, y, z)											\
	((UINT32 (_cdecl *)(void *, UINT32, const void *, void *, UINT32,	\
		const void *, UINT32, UINT32, UINT32, UINT32))		\
	pNBISupportAPI[DMAStartIdx])((q), (r), (s), (t), (u), (v),	\
		(w), (x), (y), (z))

#define NBIDMAStatus(x)					  					\
	((UINT32 (_cdecl *)(UINT32))									\
   pNBISupportAPI[DMAStatusIdx])((x))

#define NBISlow()												\
	((void (*)(void))											\
	pNBISupportAPI[SlowIdx])()

#define NBIRequest(x)									\
   ((ODI_NBI (*)(void *))								\
	pNBISupportAPI[NBIRequestIdx])((x))

#define NBIMapBusMemory(v, w, x, y, z)									\
   ((ODI_NBI (_cdecl *)(void *, const void *, void *, void **, UINT32))	\
	pNBISupportAPI[MapBusMemoryIdx])((v), (w), (x), (y), (z))

#define NBIFreeBusMemory(v, w, x, y, z)									\
   ((ODI_NBI (_cdecl *)(void *, const void *, void *, const void *, UINT32))	\
	pNBISupportAPI[FreeBusMemoryIdx])((v), (w), (x), (y), (z))
