ifdef TASM
%PAGESIZE 66,130
else
PAGE 66, 130
endif
;*****************************************************************************
; ETHERNET.ASM
;	(C) Unpublished Work of Novell, Inc.  All Rights Reserved.
;
;	THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL, 
;	PROPRIETARY AND TRADESECRET INFORMATION OF NOVELL, INC.  ACCESS 
;	TO THIS WORK IS RESTRICTED TO (I) NOVELL EMPLOYEES WHO HAVE A
;	NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE OF THEIR ASSIGNMENTS 
;	AND (II) ENTITIES OTHER THAN NOVELL WHO HAVE ENTERED INTO APPROPRIATE 
;	LICENSE AGREEMENTS.  NO PART OF THIS WORK MAY BE USED, PRACTICED,
;	PERFORMED, COPIED, DISTRIBUTED, REVISED, MODIFIED, TRANSLATED,
;	ABBRIDGED, CONDENSED, EXPANDED, COLLECTED, COMPILED, LINKED, RECAST,
;	TRANSFORMED OR ADAPTED WITHOUT THE PRIOR WRITTEN CONSENT OF NOVELL. 
;	ANY USE OR EXPLOITATION OF THIS WORK WITHOUT AUTHORIZATION COULD 
;	SUBJECT THE PERPERTRATOR TO CRIMINAL AND CIVIL LIABILITY.
;
;----------------------------------------------------------------------------
;
;	DOS ODI Topology Specific Module (TSM) for Ethernet Media
;
;	Written by:	DCD
;	Date:		6/12/90
;
;	ODI Specification Version:	4.00
;
;    Modification History:
; Date		By	Modification Description
; ------------	---	------------------------------------------
; 01 Aug. 1994	JWR,WTT  v2.00 Removed Previous history, 
; 			See DOS ODI Specifcation Changes Doc for changes.
;
; 19 Sep. 1994	JWR	TxTooBig path missed restoring the TxFreeCounter
;			Snap Protocol ID byte 0 and 1 were swapped.
;
; 14 Oct. 1994  JWR	Malformed Packets would be detected prior to 
;			determining the Frame Type, and did not get 
;			the DS:DI pointer to a LookAheadStruc.
;
; 09 Jan. 1995	JWR	Checkin New ToolKit. v2.02
;			Pad Bytes on Small Frames needs to be passed 
;			up when the Driver is in Promiscous Mode only.
;
; 27 March 1995	JWR	NewToolKit. v2.03.
;			TxTooBig error handling failed to load ES:SI correctly.
;			SPD#94304 -- Problem in MSMupdateMulticast was in not
;			settting the Promiscous filter correctly mode correctly
;			MSMRcvCompleteStatus not handling Ethernet802.3 and EII 
;			correctly for small padded frames.  Also Evenization not
;			handled correctly.
;
; 3 April 1995	JWR	New ToolKit. v2.03
;			Added MSMGetMulticastInfo (IOCTL 15)
;
; 19 Oct. 1995	JWR	Explicit 802.2 Header section was incrementing the 
;			TypeLength Field after it was swapped.
;
; 19 Nov. 1995  JWR	Explicit 802.2 type two failed to inc Hardware length.
;
; 23 Jan. 1996 JWR	v2.06
;			Packet Fragments didn't have the Runt bit set.
;			Now Or in Runt Bit to match the Server(4.1) code base.
;
; 1 April 1996	JWR	New ToolKit. v2.10
;		SPD# 119818 Multicast Packets pass filter after deleted from 
;			table in Non-Promiscuous Mode.
;
; 15 SEP. 97    CKP     V2.11
;                       Donot increment padded Hardware len for 802.2 type II
;                       packet, if it was previously padded. SPD#164800.
;
; 21 Nov. 97	SMN		New ToolKit. v2.12
;*****************************************************************************
.8086
%NOINCL				; See listings of Include files in MSM.asm
include	msm.inc

ifdef TASM
%NEWPAGE ;
endif
;----------------------------------------------------------------------------
EthernetTSMVersion      equ     'v2.12 (971121) '; Set November 21, 1997 SMN

CGroup	group	Code,Init
	assume	cs:CGroup, ds:CGroup, es:nothing, ss:nothing

Code	segment	para public 'CODE'
;----------------------------------------------------------------------------
;	Variables Provided by the TSM.
;----------------------------------------------------------------------------
	public	MSMTxFreeCount
	public	MSMLookAheadSegment
	public	MSMMaxMulticastAddr

ifdef DEBUG		;Internal Use Only. 
	public	TCBFreeList
	public	HeaderSizeTable
	public	BoardFlags
	public	MaxVirtBoards
	public	BuildHeaderHandlers
	public	MulticastTable
	public	ItIsA8022Frame

	public	LookAhead8023	
	public	LookAhead8022
	public	LookAheadSNAP
	public	LookAheadEII
endif

;----------------------------------------------------------------------------
;	Functions Provided by the TSM.
;----------------------------------------------------------------------------
	public	MSMSendPacket
	public	MSMSendComplete
	public	MSMGetNextSend
	public	MSMGetRCB
	public	MSMReturnRCB
	public	MSMRcvComplete
	public	MSMRcvCompleteStatus
	public	MSMGetMulticastInfo
	public	MSMAddMulticastAddress
	public	MSMDelMulticastAddress
	public	MSMUpdateMulticast
	public	MSMMediaShutdown		; Unused for Ethernet
	public	MSMMediaReset			; Unused for Ethernet
	public	MSMMediaConfigUpdate		; Called after Hardware Initializes
	public	MSMBuildTransmitControlBlock	; Called for Priority Support by the HSM.

ifdef DEBUG		;Internal Use Only. 
	public	FindAddressInMCTable, FindAddressInMCTableNotLoaded
endif

;----------------------------------------------------------------------------
;	Variables Provided by the MSM
;----------------------------------------------------------------------------
	extrn	LSLSupport		:dword
	extrn	MSMTxMonPtr		:dword
	extrn	MSMTxQueueHead		:word
	extrn	MSMTxQueueTail		:word
	extrn	MSMActiveBoardCount	:word
	extrn	MSMConfigTables		:word
	extrn	MSMActivePromisMask	:word
	extrn	MSMTableIndex		:byte	; Use with Upper Register = 0
	extrn	MSMTxFlags		:word
	extrn	MSMTxFlags0		:byte
	extrn	MSMTxFlags1		:byte
	extrn	MSMMaxActiveTx		:byte
	extrn	MSMPhysNodeAddress	:byte

;----------------------------------------------------------------------------
;	Functions Provided by the MSM
;----------------------------------------------------------------------------
	extrn	MSMEnqueueSend		:near
	extrn	MSMUpdateConfigTables	:near

;----------------------------------------------------------------------------
;	Variables Provided by the HSM
;----------------------------------------------------------------------------
	extrn	DriverStatTable		:byte
	extrn	DriverConfigTable	:byte

;----------------------------------------------------------------------------
;	Functions Provided by the HSM
;----------------------------------------------------------------------------
	extrn	DriverSend		:near
	extrn	DriverMulticastChange	:near
	extrn	DriverPromiscuousChange	:near

;----------------------------------------------------------------------------
;	The following equates are media specific.
;----------------------------------------------------------------------------
MSMMaxMulticastAddr	equ	16		; We support 16 MC addresses MAX
MaxVirtBoards		equ	4		; EthernetII, Raw8023, 8022, SNAP
MAX_TCB_ALLOCATED	equ	3		; Number of TCB's this Module will allocate.

MinMediaFrameSize	equ	60		; Absolute min. for Ethernet
MaxMediaFrameSize	equ	1514		; Absolute max. for Ethernet
MaxFrameHeaderLen	equ	14+3+5		; Max. possible frame header (Should be even)
MinFrameHeaderLen	equ	14		; Min Possible Frame Header

Min8023LenField		equ	03h		; Min value in 8023 LenField
MinRuntPacketSize	equ	22 		; MinRuntPacketSize <= Runt Frame < 60 
		; Runt frames may possibly be valid, but are too small.
		; Small Fragments also set Runt Bit.


;----------------------------------------------------------------------------
;	Media ID's Supported in this module.
;----------------------------------------------------------------------------
MediaIDEthernetTyped	equ	2	; Well known Media ID's
MediaID8022		equ	3	; defined by ODI specification
MediaIDRaw8023		equ	5
MediaIDSNAP		equ	10

;----------------------------------------------------------------------------
;	Media Type Strings
;----------------------------------------------------------------------------
Message	MediaName2	'ETHERNET_II'
Message	MediaName3	'ETHERNET_802.2'
Message	MediaName5	'ETHERNET_802.3'
Message	MediaName10	'ETHERNET_SNAP'

;----------------------------------------------------------------------------
;	Ethernet frame format structure
;----------------------------------------------------------------------------
EthernetFrameFormat	struc
	DestAddr	db	6 dup (?)
	SrcAddr		db	6 dup (?)
	LenTypeField	dw	?
	DSAP		db	?
	SSAP		db	?
	Control		db	?
	SNAP		db	5 dup (?)
EthernetFrameFormat	ends

;----------------------------------------------------------------------------
;	BoardFlags bit definitions.
;----------------------------------------------------------------------------
LOADED		equ	01h		; Set if This Frame Is Loaded.	
EVENIZE		equ	02h		; Set if this frame needs to be evenized

;----------------------------------------------------------------------------
;	802.2 Control Fields LLC Frame Type 
;----------------------------------------------------------------------------
SU_FormatBit		equ	01	;if Bit 0 = 0 then I_Format else S or U_Format
U_FormatBit		equ	02	;if Bit 1 = 1 then U_format 

;----------------------------------------------------------------------------
;	Bit definitions used in LoadedFrameByte
;----------------------------------------------------------------------------
FRAME_IS_8022		equ	01	; Flag this frame is 802.2
FRAME_IS_SNAP		equ	02	; Flag this frame is SNAP
FRAME_IS_8023		equ	04	; Flag this Frame is 802.3
FRAME_IS_EII		equ	08	; Flag this frame is EII

;----------------------------------------------------------------------------
;	Some general variables
;----------------------------------------------------------------------------
	even
TCBFreeList		dw	0	; Points to a list of TCB's
DestAddrTypeFlag	dw	0	;	 Defined in ODI.INC
HardwareRcvLength	dw	0	; Used to store the hardware reported
					;  length of a received packet.

;----------------------------------------------------------------------------
;	Frame Information
;	Bit definitions used in Frame Info
;----------------------------------------------------------------------------
FrameInfo		db	0	; Particular Frame Data.
FRAME_COULD_BE_PADDED	equ	10	; Flag if the FrameLength might included Padding

;----------------------------------------------------------------------------
;	Variables initialized by the HSM
;----------------------------------------------------------------------------
MSMTxFreeCount		db	?	; Number of Tx resources available
MSMLookAheadSegment	dw	0	; Where the Look Ahead Buffer is.

;----------------------------------------------------------------------------
;	MSMGetRCB procedure variables.
;----------------------------------------------------------------------------
MSMRcvPacketStatus	dw	0	; Packet Status for MSMGetRCB call.
MSMRcvHeaderLen		dw	0	; Packet Header length for MSMGetRCB call. (Max = 22)
MSMRcvProtSkipCount	db	0	; Protocol Skip Count value (MAX = 128)

;----------------------------------------------------------------------------
;	The following tables are used to look up media specific information
;	about each logical ODI board number.
;----------------------------------------------------------------------------
HeaderSizeTable		db	MaxBoards dup (0)  ; Each boards media header size is here
BoardFlags		db	MaxBoards dup (0)  ; Possible Config Flags.

;----------------------------------------------------------------------------
;
; Using Equates to place memory in PSP Area of Memory
;
;	The following tables are used to look up media specific information
;	about each logical ODI board number.
;
;	even
;BuildHeaderHandlers	dw	MaxBoards dup (0)	;Contains the addr of a routine that builds the media header
;MulticastTable		db	(MSMMaxMulticastAddr * 8) dup (0)
;

	align	4
LookAhead8023		LookAheadStruc	<>
LookAhead8022		LookAheadStruc	<>
LookAheadSNAP		LookAheadStruc	<>
LookAheadEII		LookAheadStruc	<>


BuildHeaderHandlers	equ	(((PSPTableStart+1) shr 1) shl 1)  ; ensure even
MulticastTable		equ	(BuildHeaderHandlers + (2 * MaxBoards))	

;Lookahead tables got too big.  couldn't fit in the PSP area.
;LookAhead8023	equ	(MulticastTable + (MSMMaxMulticastAddr * 8))
;LookAheadSNAP	equ	(LookAhead8023 + size LookAheadStruc) 
;LookAhead8022	equ	(LookAheadSNAP + size LookAheadStruc) 
;LookAheadEII	equ	(LookAhead8022 + size LookAheadStruc) 


ifdef TASM
%NEWPAGE ;
endif
even
;******************************************************************************
;		F R A M E	T R A N S M I T		C O D E
;******************************************************************************
;
;	MSMSendPacket
;
;	Called directly by the LSL when it wants us to send a packet.
;
;	assumes:
;		es:si	-> ECB
;		Interrupts are disabled
;
;	returns:
;		ax	=0 if no initial errors
;			=LSLERR_NO_SUCH_DRIVER if MLID is shut down
;			=LSLERR_PACKET_ERRORED  if Packet is Too BIG
;		z flag	set according to ax.
;		Interrupts disabled
;		DS,BP,SS,SP preserved
;****************************************************************************
;----------------------------------------------------------------------------
;    We are Shutdown, or Already Queued up or in a Critical Section.
;----------------------------------------------------------------------------
CantSendRightNow:
	test	word ptr MSMTxFlags0, (TXQUEUED*100h) or CRITICAL
	jz	ShutDownCase
	jmp short QueueItAndExit

;----------------------------------------------------------------------------
;	Tried to Send one too many, reset and go on.
;----------------------------------------------------------------------------
MLIDIsSaturated:
	mov	MSMTxFreeCount, 0		;Don't go negative

;----------------------------------------------------------------------------
;	Can't send right now, just Queue it up.
;----------------------------------------------------------------------------
QueueItAndExit:
	call	MSMEnqueueSend			;Queue it for later
	xor	ax, ax

;----------------------------------------------------------------------------
;	Most come here to restore the Regs off the stack and exit.
;----------------------------------------------------------------------------
ExitSendPacket:
	pop	ds
	pop	bp
	retf			; Return far to caller

;----------------------------------------------------------------------------
;	If we are in a Shutdown mode, we simply give ECB back 
;----------------------------------------------------------------------------
ShutDownCase:
	mov	ax, LSLERR_NO_SUCH_DRIVER	; Shutdown, set nodriver go figure.
ReturnTooBigECB:
	push	ax				; Save the Error code for later
	mov	es:[si].Status, ax		; Set ECB Status field
	call	es:[si].ESR			; Give ECB back to Caller.
	pop	ax				; AX = Error code
	or	ax, ax				; Set flags accordingly
	jmp short ExitSendPacket

;----------------------------------------------------------------------------
;	MSMSendPacket  Starts HERE:
;----------------------------------------------------------------------------
MSMSendPacket	proc	far
	push	bp				; Must save for ODI spec.
	push	ds
	mov	ax, cs
	mov	ds, ax				; DS = CGroup

;----------------------------------------------------------------------------
;    Check if we are Shutdown, or Already Queued up or in a Critical Section.
;----------------------------------------------------------------------------
	test	word ptr MSMTxFlags, ((SHUTDOWN or TXQUEUED)*100h) or CRITICAL
	jnz	CantSendRightNow

;----------------------------------------------------------------------------
;	See if the HSM has a free Resource
;----------------------------------------------------------------------------
	dec	MSMTxFreeCount
	js	MLIDIsSaturated			; Not ready yet, Get it Queued.

;----------------------------------------------------------------------------
;	Get the Header built
;----------------------------------------------------------------------------
	mov	di, si				; ES:DI -> ECB
	cld					; Required for DriverSend
	call	TSMBuildTransmitControlBlock	; get header built

;----------------------------------------------------------------------------
;	Guarentee that the HSM never Gets a Too Big Packet.
;----------------------------------------------------------------------------
	cmp	cx, MaxMediaFrameSize
	ja	TXTooBig		; a=> Frame is too large for Ethernet.

;----------------------------------------------------------------------------
;	Enter a critical section.
;----------------------------------------------------------------------------
	mov	bx, MLIDSUP_START_CRITICAL_SECTION
	call	LSLSupport
	inc	MSMTxFlags0			; Set critical section flag

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;	DriverSend assumes the following
;
;	assumes: 
;		si	-> TCB
;		ds	is set to CGroup
;		cx	hardware frame length
;		CLD has been executed
;		Interrupts are disabled
;		A critical section has been started
;
;	returns:
;		Can destroy all registers
;		Interrupts must be DISABLED!
;
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	call	DriverSend			;Send the frame

;----------------------------------------------------------------------------
;	End the Critical section, and return
;----------------------------------------------------------------------------
	pop	ds
	pop	bp
	dec	cs:MSMTxFlags0			; Exit critical section
	mov	bx, MLIDSUP_END_CRITICAL_SECTION
	call	cs:LSLSupport			; Also calls ServiceEvents
	xor	ax, ax				; Set flags, and ax.
   	retf					; Return far to caller

;----------------------------------------------------------------------------
;	Packet was too Big to send, Give an Error indication and return it.
;----------------------------------------------------------------------------
TXTooBig:
	inc	DriverStatTable.MTxTooBig
	inc	MSMTxFreeCount			; Restore -- We didn't take it after all.
	mov	cx, TCBFreeList			; Free up this TCB
	mov	[si].NextLink, cx
	mov	TCBFreeList, si
	les	si, [si].TCBFragStrucPtr	; Get ptr to ECB.fragstruc
	sub	si, FragCount			; ES:SI -> ECB
	mov	ax, LSLERR_PACKET_ERRORED
	jmp short ReturnTooBigECB

MSMSendPacket	endp


ifdef TASM
%NEWPAGE ;
endif
;****************************************************************************
;	MSMSendComplete
;
;	Called by the HSM when it has finished with a transmit event.
;	This routine will assume control of the TCB handed to this routine.
;
;	assumes:
;		DS	is set to CGroup
;		SI	-> TCB
;		Interrupts are disabled
;
;	returns:
;		DS,BP,SS,SP preserved
;		CLD preserved
;		Interrupts disabled
;
;****************************************************************************
MSMSendComplete	proc	near
;----------------------------------------------------------------------------
;	Count the Packets transmited
;----------------------------------------------------------------------------
	add	DriverStatTable.MTotalTxPackets+0, 1
	adc	DriverStatTable.MTotalTxPackets+2, 0

;----------------------------------------------------------------------------
;	See if the Transmit Monitor is around
;----------------------------------------------------------------------------
	cmp	word ptr MSMTxMonPtr+2, 00h	; ptr. to Rcv Monitor Tx Routine
	jz	MSMSendCompleteProceed

	call	dword ptr MSMTxMonPtr		; call Rcv Monitor Tx Routine

;----------------------------------------------------------------------------
;	Put TCB back on the TCBFreeList
;----------------------------------------------------------------------------
MSMSendCompleteProceed:
	mov	cx, TCBFreeList			; Free up this TCB
	mov	[si+0], cx
	mov	TCBFreeList, si

;----------------------------------------------------------------------------
;	Send the ECB Back to its owner
;----------------------------------------------------------------------------
	les	si, [si].TCBFragStrucPtr	; Get ptr to ECB.FragCount
	sub	si, FragCount			; ES:SI -> ECB
	mov	bx, MLIDSUP_SEND_COMPLETE	; Hand it to the LSL
	call	LSLSupport
	ret

MSMSendComplete	endp

ifdef TASM
%NEWPAGE ;
endif
;****************************************************************************
;	MSMGetNextSend
;
;	Called by the MLID when it is ready for another transmit 
;	request block.  This routine must check to make sure the HSM
;	can handle another transmit.
;
;	assumes:
;		ds	is set to CGroup
;		Interrupts are disabled
;
;	returns:
;		si	-> TCB if there is one
;		ds	is set to CGroup
;		cx	hardware frame length
;		All other registers destroyed
;		nz flag if TCB NOT available
;		Interrupts disabled
;
;****************************************************************************
;----------------------------------------------------------------------------
;	The HSM has no free Resources.
;----------------------------------------------------------------------------
DriverIsAlreadySaturated:
	mov	MSMTxFreeCount, 0		; Don't go negative

;----------------------------------------------------------------------------
;	Signal NO TCB available.
;----------------------------------------------------------------------------
QueueisEmpty:
	or	al, 1
	ret

;----------------------------------------------------------------------------
;	MSMGetNextSend starts HERE	
;----------------------------------------------------------------------------
MSMGetNextSend	proc	near
;----------------------------------------------------------------------------
;	See if there is one on the Q	
;----------------------------------------------------------------------------
	les	di, dword ptr MSMTxQueueHead
	mov	ax, es
	or	ax, di
	jz	QueueisEmpty

;----------------------------------------------------------------------------
;	We think that there is one
;----------------------------------------------------------------------------
	dec	MSMTxFreeCount
	js	DriverIsAlreadySaturated	; Not ready yet

;----------------------------------------------------------------------------
; 	Keep track of how deep the Queue is.	
;----------------------------------------------------------------------------
	dec	DriverStatTable.MQueueDepth	; Decrement Queue Depth

;----------------------------------------------------------------------------
;	Get an ECB off the MSM's QUEUE
;----------------------------------------------------------------------------
	mov	ax, es:[di].NextLink+0		; Unlink the ECB
	mov	cx, es:[di].NextLink+2
	mov	word ptr MSMTxQueueHead+0, ax
	mov	word ptr MSMTxQueueHead+2, cx
	or	ax, cx				; Did we take the last one?
	jnz	StillMoreOnQueue

;----------------------------------------------------------------------------
;	No ECB is left on the Queue
;----------------------------------------------------------------------------
	and	MSMTxFlags1, not TXQUEUED
	mov	word ptr MSMTxQueueTail+0, offset CGroup:MSMTxQueueHead
	mov	word ptr MSMTxQueueTail+2, cs

;----------------------------------------------------------------------------
;	Build the Header and return
;----------------------------------------------------------------------------
StillMoreOnQueue:
	call	TSMBuildTransmitControlBlock
MSMMediaShutdown:		; We don't need to do anything.
MSMMediaReset:
	xor	ax, ax
	ret
MSMGetNextSend	endp

ifdef TASM
%NEWPAGE ;
endif
;****************************************************************************
;	MSMBuildTransmitControlBlock /TSMBuildTransmitControlBlock
;
;	Takes a ODI ECB and converts into a Transmit Request Block (TCB).
;	It also builds the appropriate media header.
;
;	assumes:	
;		DS	is set to CGroup
;		ES:DI	-> ECB
;		Interrupts are disabled
;
;	returns:
;	(MSMBuildTransmitControlBlock will return si = 0 if no TCB is available.)
;	(TSMBuildTransmitControlBlock will always return a TCB)
;		DS:SI	-> TCB
;		CX	has the hardware frame length
;		DS, SS, SP preserved
;		All other Registers Destroyed.
;		Interrupts are disabled
;
;****************************************************************************
MSMBuildTransmitControlBlock	proc	near
	mov	si, TCBFreeList			; Get a TCB
	or	si, si			    	; Verify that there is one.
	jnz	TSMBuildTransmitControlBlock
	ret					; Return to HSM.

TSMBuildTransmitControlBlock:
;----------------------------------------------------------------------------
;	Get a free TCB
;----------------------------------------------------------------------------
	mov	si, TCBFreeList			; Get a TCB
	mov	ax, [si].NextLink		; Unhook it from the list
	mov	TCBFreeList, ax			; Keep track of who's next.

;----------------------------------------------------------------------------
;	Get the TCB pointers loaded
;----------------------------------------------------------------------------
	lea	ax, [di].FragCount			; AX -> ECB.FragCnt
	mov	word ptr [si].TCBFragStrucPtr+0, ax	; Fill in TCB
	mov	word ptr [si].TCBFragStrucPtr+2, es

	mov	bx, es:[di].BoardNum		; Get the board number
	xor	ax, ax				; Prepare for Zero HeaderLen
	mov	es:[di].Status, ax		; Get this out of the way

;----------------------------------------------------------------------------
;	Check for RAW send packets.  (Allow for Priority settings)
;					FFFFh -> FFF8h Raw Send Priority.
;					FFF7h -> FFF0h Non-Raw Send Priority.
;----------------------------------------------------------------------------
	cmp	es:[di].ProtoNum, 0FFF7h ; Raw send? 
	ja	DoRawSend			; a => Raw Send

	mov	al, HeaderSizeTable[bx]		; Get headerSize for this frame

;----------------------------------------------------------------------------
;	AX = Header Size
;----------------------------------------------------------------------------
DoRawSend:
	mov	[si].TCBMediaHeaderLen, ax	; Set the media header length
	add	ax, es:[di].DataLen		; Get total of frag's and media header
	mov	[si].TCBDataLen, ax		; Set it in the TCB

;----------------------------------------------------------------------------
;	EII and Raw802.3 packets must be evenized.
;----------------------------------------------------------------------------
	test	BoardFlags[bx], EVENIZE		; Does it need evenized?
	jz	DontEvenize			; z = don't evenize

	inc	ax				; Evenize Raw802.3 & EII frames
	and	al, 0FEh			

;----------------------------------------------------------------------------
;	Check for Minimum Size Frames
;----------------------------------------------------------------------------
DontEvenize:
	mov	cx, ax				; CX = total frame length
	cmp	cx, MinMediaFrameSize		; Padding required?
	ja	NoPaddingNeeded

	mov	cl, MinMediaFrameSize		; Ethernet Minimum size

;----------------------------------------------------------------------------
;	Check for Needing to do a Header.
;----------------------------------------------------------------------------
NoPaddingNeeded:
	cmp	[si].TCBMediaHeaderLen, 0	; See if a Header is needed.
	je	NoHeaderNeeded

;----------------------------------------------------------------------------
;    By Default set the LenTypeField up for 802.3 headers (3 out of 4 use it.)
;----------------------------------------------------------------------------
	sub	ax, 14				; Get 802.3 length field
	xchg	ah, al				; Get into high-low order
	mov	[si].TCBMediaHeader.LenTypeField, ax

;----------------------------------------------------------------------------
;	Now build the media header.
;----------------------------------------------------------------------------
	mov	ax, word ptr es:[di].ImmAddr+0	; Set destination address
	mov	word ptr [si].TCBMediaHeader.DestAddr+0, ax
	mov	ax, word ptr es:[di].ImmAddr+2
	mov	word ptr [si].TCBMediaHeader.DestAddr+2, ax
	mov	ax, word ptr es:[di].ImmAddr+4
	mov	word ptr [si].TCBMediaHeader.DestAddr+4, ax	; Source addr already set

;----------------------------------------------------------------------------
;	Each Board has its own header, BX = Board number
;----------------------------------------------------------------------------
	mov	bl, MSMTableIndex[bx]		; Get Word VirtBoard Index
	jmp	BuildHeaderHandlers[bx]		; jump to appropriate header builder.


;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;	Build Media Header Handler Routines
;
;	assumes:
;		ds:si	-> TCB to build the header in
;		es:di	-> ECB
;		cx	hardware length
;
;	returns:
;		si -> TCB
;		ds	CGroup
;		cx	has the hardware Frame Length
;
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BuildEthernetTypedHeader:
	mov	ax, word ptr es:[di].ProtID+4 		; Get EII Protocol ID
	mov	[si].TCBMediaHeader.LenTypeField, ax	; Overrite the Len

BuildRaw8023Header:		; 802.3 Raw is set.
NoHeaderNeeded:			; Raw Send needs no header
	ret

;----------------------------------------------------------------------------
;	802.2 headers can be given.  Check ProtID+0.
;----------------------------------------------------------------------------
Build8022Header:
	xor	ax, ax				; Clears need zero
	cmp	byte ptr es:[di].ProtID+0, al		
	jne	Explicit8022Header		; Header is given Explicitly

;----------------------------------------------------------------------------
;	Set up a UI Type 1 frame as default.
;----------------------------------------------------------------------------
	mov	dl, byte ptr es:[di].ProtID+5		; Get DSAP value
	mov	dh, dl					; DSAP=SSAP
	mov	word ptr [si].TCBMediaHeader.DSAP, dx 	; Store it.
	mov	byte ptr [si].TCBMediaHeader.Control, 03h ; UI control byte
	ret

;----------------------------------------------------------------------------
;	Build Header with Explicit values given.
;	Generically done so protocols can use all 5 bytes of the ProtoID field.
;	CX = Hardware Length
;	DS:SI -> TCB 
;	ES:DI -> ECB
;
;	Save off the Hardware Len and pointer to TCB
;----------------------------------------------------------------------------
Explicit8022Header:				
	xchg	ax, cx				; AX = HardwareLen, CX = 0
	mov	bp, si				; DS:BP -> TCB

;----------------------------------------------------------------------------
;	Get Number of Bytes in 802.2 header 0 based (2, or 3)
;----------------------------------------------------------------------------
	mov	cl, byte ptr es:[di].ProtID+0	; Get # of bytes zero based. (2, 3)
	cmp	cl, 02				; If 2 then we have already added it in.
	je	DataLenAdjusted			; ne => need to add one more byte.

	inc	[si].TCBMediaHeaderLen		; Add one more for Type 2 frames.
	inc	[si].TCBDataLen

; CKP 15 SEP. 97 Donot increment padded Hardware len for 802.2 type II packet
;               if it was previously padded. SPD#164800.
        cmp     [si].TCBDataLen, MinMediaFrameSize ; Was it previously padded?
        jbe     DontINCHardwareLen              ; Dont Increment H/W Len.
	inc	ax				;Don't forget the Hardware length
				    ; Remember LenTypField already Swapped.
				    ; to High Low Order.
DontINCHardwareLen:
	add	byte ptr [si].TCBMediaHeader.LenTypeField+1, 1
	adc	byte ptr [si].TCBMediaHeader.LenTypeField+0, 0

;----------------------------------------------------------------------------
;	Set up pointers to move header over.
;----------------------------------------------------------------------------
DataLenAdjusted:
	lea	si, es:[di].ProtID+6		; Move to end. (but Remember cx is 0 based)
	inc	cx				; Make CX 1 Based
	sub	si, cx				; Move to start of 8022Header Source
	lea	di, [bp].TCBMediaHeader.DSAP	; Start at DSAP to move it in.
	mov	bx, ds				; BX:BP ->TCB (Save off DS)
	mov	dx, es				; Swap ES <-> DS
	mov	es, bx				; Get ES:DI -> Target space						; 
	mov	ds, dx				; Get DS:SI -> Source of 8022 Header
 rep	movsb					; Move 8022 header.

;----------------------------------------------------------------------------
;	On Exit: 
;		Set DS:SI -> TCB
;		    CX = Hardware Length
;----------------------------------------------------------------------------
	mov	si, bp				
	mov	ds, bx		; Restore DS:SI -> TCB
	xchg	cx, ax		; Restore CX Hardware Length.
	ret			; End Explicit 802.2 header

;----------------------------------------------------------------------------
;	Build Snap Packet Header
;----------------------------------------------------------------------------
BuildSNAPHeader:
	mov	word ptr [si].TCBMediaHeader.DSAP, 0AAAAh
	mov	byte ptr [si].TCBMediaHeader.Control, 03h
	mov	al, es:[di].ProtID+1			; Get 1st byte of SNAP
	mov	[si].TCBMediaHeader.SNAP+0, al
	mov	ax, word ptr es:[di].ProtID+2		; 2nd and 3rd
	mov	word ptr [si].TCBMediaHeader.SNAP+1, ax
	mov	ax, word ptr es:[di].ProtID+4		; 4th and 5th
	mov	word ptr [si].TCBMediaHeader.SNAP+3, ax
	ret				; CX = Hardware Length
					; DS:SI -> TCB

MSMBuildTransmitControlBlock	endp


ifdef TASM
%NEWPAGE ;
endif
;*******************************************************************************
;		F R A M E	R E C E P T I O N	C O D E
;*******************************************************************************
;
;	MSMGetRCB
;
; Called by the MLID when it has a frame waiting for a buffer.
; The MLID has passed us a pointer to the frame data. (Poss. Errors)
;	1. Check for Proper Size
;	2. Determine Destination Address Type 
;		(Remote or Direct, MC or Direct etc.)
;	3. Parse the Media Header and fill in our LookAheadStructure.
;	4. Pass the frame up to the LSL.
;	5. The LSL determines which Protocol Stack to give the Packet to.
;	6. if an ECB is returned from the LSL, an RCB is passed to the HSM.
;	7. After the HSM has filled in the buffers, the HSM will call 
;	   MSMRcvComplete if the HSM passed in a valid Frame Length
;	   (bp!=-1) to MSMGetRCB. If the HSM (bp=-1) didn't know the 
;	   Frame length when it called MSMGetRCB, then it will call 
;	   MSMRcvCompleteStatus when the buffers are filled in.
;
;
;	On Entry:
;		AX	Zero		Normal Rx Modes
;			Non-Zero	Rx Frame Errors  (See LPacketAttrib definitions in ODI.INC)
;					PKT_CRC_ERR 	equ 0	CRC Error
;					PKT_ALIGN_ERR	equ 1	CRC/Alignment Error
;
;		ES:SI	-> LookAhead Frame Data
;		DS	is set to CGroup
;		BP	has the hardware reported frame length or
;			-1 for unknown packet length at the time of call.
;		CLD has been executed
;		Interrupts are Enabled.
;
;	On Return:
;		nz flag	if no buffer for this frame (Discard the frame)
;		ds	is set to CGroup
;		ax	has the number of bytes to skip from the
;			Start of the frame before starting to copy to
;			the RCB buffers.
;		bp	How much data needs to be copied or -1 
;			for unknown packet length at the time this 
;			call was made.
;		es:si	-> Receive Control Block (RCB) if z flag
;		All other registers destroyed
;		CLD has been executed
;		Interrupts Enabled, but have been disabled for a period.
;
;*******************************************************************************
;----------------------------------------------------------------------------
; Make the code think we have a hardware length for the packet.
;	BP must be as large as the largest possible Frame size.
;----------------------------------------------------------------------------
UnknownPacketLength:
	mov	bp, MaxMediaFrameSize 		; BP = Fake Hardware Length
	jmp short UnknownPacketLengthRet	; Check Len during RcvComplete

;----------------------------------------------------------------------------
;	Hardware Reported Length is larger than Media Max.
;----------------------------------------------------------------------------
FrmSizeTooBig:
	inc	DriverStatTable.MRxTooBig		; Inc. Rx Too Big count.
	or  byte ptr MSMRcvPacketStatus+0, PKT_BIG_ERR ; Set Too Big bit in packet
	jmp short FrmSizeErrorRtn			;  status.

;----------------------------------------------------------------------------
;	Check to see if the packet is directed to us.
;----------------------------------------------------------------------------
TestForOurNodeAddress:
	mov	ax, word ptr MSMPhysNodeAddress+4 ; Test last word
	cmp	word ptr es:[si].DestAddr+4, ax
	jnz	SetDestRemoteUnicastBit		  ; nz =>Go set dest. remote unicast bit.

	mov	ax, word ptr MSMPhysNodeAddress+2 ; Test middle word
	cmp	word ptr es:[si].DestAddr+2, ax
	jnz	SetDestRemoteUnicastBit		  ; nz => Go set dest. remote unicast bit.

	mov	ax, word ptr MSMPhysNodeAddress+0 ; Test first word
	cmp	word ptr es:[si].DestAddr+0, ax
	je	BackFromCheckMulticast		  ; e=> our address.

;----------------------------------------------------------------------------
;	Set the Remote unicast bit and skip over MC Tests on our return.
;----------------------------------------------------------------------------
SetDestRemoteUnicastBit:
	mov byte ptr DestAddrTypeFlag+0, DEST_REMOTE_UNICAST  ; Signal its a remote unicast
	jmp short BackFromCheckMulticast

;----------------------------------------------------------------------------
;MSMGetRCB Code Starts Here
;----------------------------------------------------------------------------
MSMGetRCB	proc	near
;----------------------------------------------------------------------------
;	AX = 	Rx Error Status on entry
;	BP = 	Hardware reported frame length or 
;		-1 unknown packet length at the time of call.
;----------------------------------------------------------------------------
	mov	MSMRcvPacketStatus, ax			; Set initial Error Status
	mov	HardwareRcvLength, bp			; Save for later use.
	cmp	bp, -1					; Is packet length known?
	je	UnknownPacketLength

;----------------------------------------------------------------------------
;	Is reported Frame size Larger than Media can Handle?
;----------------------------------------------------------------------------
	cmp	bp, MaxMediaFrameSize		; BP = Reported FrameSize
	ja	FrmSizeTooBig			; a => Too Big
							
;----------------------------------------------------------------------------
;	Is reported frame size smaller than Minimum media frame size?
;----------------------------------------------------------------------------
	cmp	bp, MinMediaFrameSize		; BP = Reported FrameSize 
	jb	FrmSizeTooSmall			; b => Too Small

;----------------------------------------------------------------------------
;	ES:SI-> look ahead frame data
;----------------------------------------------------------------------------
UnknownPacketLengthRet:
FrmSizeErrorRtn:
	xor	ax, ax					; Clear AX for size.
	mov	FrameInfo, al				; Clear Padding Flag.
	mov	word ptr DestAddrTypeFlag, DEST_DIRECT	; Flag Pkt is for us. (default)
	test	byte ptr es:[si].DestAddr, 01h	 	; Is this a Group Addressed frame?
	jnz	CheckMulticast				; nz = Group Address Bit Set

;----------------------------------------------------------------------------
;	See if the NIC is in Promiscuous Mode.
;	If Not Then It is our Directed Packet
;----------------------------------------------------------------------------
	cmp	MSMActivePromisMask, ax		; Test if in Promiscuous Mode. (AX = 0)
	jne	TestForOurNodeAddress		; ne => Yes, need to test

;----------------------------------------------------------------------------
;	See if the frame is 802.3 or EthernetTyped
;----------------------------------------------------------------------------
BackFromCheckMulticast:
	sub	bp, MinFrameHeaderLen		; Sub basic MAC header
	mov	ax, es:[si].LenTypeField	; Type or Length?
	xchg	ah, al				; Get LowHigh Order
	cmp	ax, 1500			; Typed are larger than 1500
	jbe	PacketIsEth8023Hdr		; be => 802.3 header
	jmp	PacketIsEthTyped		; else EII packet.

;----------------------------------------------------------------------------
;     Malformed Error Path.
;	The Packet doesn't look right to begin with.
;	Need to place in a default LookAhead Structure to pass up.
;	The Board number will make it so that it goes up to the active
;	Frames, if 802.2 is not loaded it will be an unsupported Frame and 
;	Malformed frame.
;
;----------------------------------------------------------------------------
MalformedPacketLoadRegs:
	mov	di, offset CGroup:LookAhead8022	; DI = LookAhead Struc.
	xor	cx, cx				; Clear CX Unsupported (Unknown Frame) Type
	mov byte ptr [di].LProtID+5, cl		; Guarentee ProtID not valid. 
	jmp	MalformedPacket			; Go do the normal Malformed Stuff.

;----------------------------------------------------------------------------
;	Determine if Broadcast or other Multicast packet.
;	and load AX, BX, DX for FindAddressinMcTable.
;----------------------------------------------------------------------------
CheckMulticast:
	mov	byte ptr DestAddrTypeFlag+0, DEST_BROADCAST+DEST_MULTICAST ; Set destination broadcast bit.
	mov	ax, word ptr es:[si].DestAddr	; See if a broadcast and 
	mov	bx, word ptr es:[si].DestAddr+2	; Set up for Call to FindAddresinMcTable.
	mov	dx, word ptr es:[si].DestAddr+4	; Get last word of addr.
	mov	cx, dx				; Check for Broadcast
	and	cx, bx		
	and	cx, ax
	inc	cx				; Does it go from -1 to 0?
	jz	BackFromCheckMulticast

;----------------------------------------------------------------------------
;	See if we can receive this multicast frame.
;	See if we need to examine this Multicast Addr. eg. Board has a CAM
;----------------------------------------------------------------------------
	mov	byte ptr DestAddrTypeFlag+0, DEST_MULTICAST	; Set MC Bit Only

	cmp	MSMActivePromisMask, 0		; Test if in Promiscuous Mode.
	jnz	SearchMCTable			; Must search table in order to
						;  set dest address type.

;----------------------------------------------------------------------------
;	See if the Adapter is completely filtering the Address
;----------------------------------------------------------------------------
	test byte ptr DriverConfigTable.MFlags+1, HIGH(GROUP_ADDRESSING_BIT+FILTER_MULTICAST_BIT) ; Isolate bits 10 & 9.
	jz	SearchMCTable			; Z - Search Multicast Table. (10+9 = 0)
	jpe	BackFromCheckMulticast		; NZ and Parity Even, NIC has ensured Multicast.

SearchMCTable:
	call	FindAddressInMCTable		; AX:BX:DX has address
	jz	BackFromCheckMulticast		; z => Found Address

;----------------------------------------------------------------------------
;	Not directed Multicast, Mark as Remote and continue.
;----------------------------------------------------------------------------
	mov	byte ptr DestAddrTypeFlag+0, DEST_REMOTE_MULTICAST	; Set destination remote multicast bit.
	cmp	MSMActivePromisMask, 0      ; Test if in Promiscuous Mode.
	jnz	BackFromCheckMulticast      ; Pass it on.

;----------------------------------------------------------------------------
;	Filter the Packet out.
;----------------------------------------------------------------------------
	or	al, 1		; NZ Flag
	ret			; Signal that We Don't Want this packet.

;----------------------------------------------------------------------------
;	Hardware Reported size is smaller than Minimum allowed.
;----------------------------------------------------------------------------
FrmSizeTooSmall:
	mov	di, offset CGroup:LookAhead8023	; Get 802.3 LookAhead.
	inc	DriverStatTable.MRxTooSmall	; Inc. Rx Too Small Count.
	or  byte ptr MSMRcvPacketStatus, PKT_RUNT_ERR ; Set Runt Packet bit.
	cmp	bp, MinRuntPacketSize		; Test have enough to qualify as Runt
	jb	TooSmallForRunt			; b => Too Small to qualify as Runt
						; Although too small still runt, just also 
						; Unknown type.
	jmp	FrmSizeErrorRtn			; Go report errors etc.

;----------------------------------------------------------------------------
;	Packet Has An 802.3 Frame Header.
;	Start with consistency checks.
;	AX = LenTypeField 
;	BP = FrameDataLen  (802.3 frame header subtracted from FrameLen)
;----------------------------------------------------------------------------
PacketIsEth8023Hdr:
	cmp	ax, bp			; 802.3 len greater then hard len?
	ja	MalformedPacketLoadRegs	; a => malformed frame

	cmp	ax, Min8023LenField	; Test for at least 3 bytes, ie. 802.2 or Raw 802.3.
	jb	MalformedPacketLoadRegs	; AX < 3 bytes => malformed frame

;----------------------------------------------------------------------------
;	Strip off the Pad Bytes if Not in Promiscous Mode.
;----------------------------------------------------------------------------
	cmp	MSMActivePromisMask, 00h	; Test if in Promis. Mode
	jne	PromiscousSavePad		; ne - Yes, skip down (Save Pad Bytes)

	mov	bp, ax				; Set Length to FieldLength (Trims off any padding)

;----------------------------------------------------------------------------
;	Check for RAW 802.3 Frame (IPX)
;----------------------------------------------------------------------------
PromiscousSavePad:
	mov	bx, word ptr es:[si].DSAP	; Get first two bytes of data.
	cmp	bx, -1				; Raw 802.3 == 0FFFFh.
	je	PacketIs8023			; If Raw Go Take Care of it.

;----------------------------------------------------------------------------
;	Could this Packet be a Snap frame?
;----------------------------------------------------------------------------
	cmp	bx, 0AAAAh			; Is it a SNAP?
	je	ItMaybeSNAP

;----------------------------------------------------------------------------
;	It is normal 802.2 Frame
;	Get DI ->Lookahead,
;	Fill in LProtID,
;	Determine 802.2 frame Type
;	BX = DSAP SSAP value
;----------------------------------------------------------------------------
ItIsA8022Frame:
	mov	di, offset CGroup:LookAhead8022	; DI = LookAhead Struc.
	mov byte ptr [di].LProtID+5, bl		; PID is the DSAP.
	mov	cl, FRAME_IS_8022		; CL = Flag which Frame Type.

;----------------------------------------------------------------------------
;	Set up for Packet is 802.2 frame Type I (Default)
;	Load AX with Data Pointer for 802.2 Type 1 header.
;	Load BP with FrameDataLength for type 1 format
;----------------------------------------------------------------------------
	inc	byte ptr DestAddrTypeFlag+1	; Signal 802.2 frame Type 1 (Default).
	sub	bp, 3				; BP = FrameDataLength (Type 1 frame Default)
	lea	ax, [si+17]			; Skip Past 802.2 Type 1 header.

;----------------------------------------------------------------------------
;	Check for Which Format
;	AX = Ptr to Start of Data
;	Type I = UI-format, and U-Format
;	Type II = S-Format and I-Format
;	Bit 0 & 1 both set, implies U-Format
;----------------------------------------------------------------------------
	test	word ptr es:[si].Control, SU_FormatBit + U_FormatBit
	jz	Type2Frame			; Z => I-format (Type 2)
	jpe	FrameTypeDetermined		; Parity Even => U-format (Type 1)
;	jpo	Type2Frame			; Parity Odd =>I-Format or S-Format (Type 2)

;----------------------------------------------------------------------------
;	Packet is 802.2 frame Type II 
;----------------------------------------------------------------------------
Type2Frame:			       
	inc	byte ptr DestAddrTypeFlag+1	; Signal 802.2 frame Type 2
	inc	ax				; Point past extra Control Byte
	dec	bp				; Subtract off 1 extra for 802.2 Type II header
	jmp short FrameTypeDetermined

;----------------------------------------------------------------------------
;   UnSupported Frame:
;	  HardwareReportedLen < MinRuntPacketSize 
;	OR
;	  Frame Not loaded, (allow the LSL to hand it up anyway.)
;----------------------------------------------------------------------------
TooSmallForRunt:
	mov	ax, si				; Get offset start of MAC Header.

UnSupportedFrame:
	or  byte ptr MSMRcvPacketStatus, PKT_TYPE_ERR ; Set not Board not installed bit in packet status.
	jmp short TestLookAheadSize

;----------------------------------------------------------------------------
;	Malformed Packet:
;	(HardwareReportedLen => MinRuntPacketSize) && 
;		(LengthField > (HardwareReportedLen - MinFrameHeaderLen))
;	OR
;	(HardwareReportedLen => MinFrameHeaderLen) && 
;		(LengthField < (HeaderLength - MinHeaderLen))
;
;	CL = Which Frame Type Lookahead Structure is being used.
;	ES:SI -> HSM LookAhead Data Buffer
;	DS:DI -> LookAhead Structure
;----------------------------------------------------------------------------
MalformedPacket:
	mov	bp, HardwareRcvLength	; Reset BP to Hardware Reported Len.
					; BP =-1 is not valid, Pipeline Cards
					; will give the size of this small packet.
	mov	ax, si					 ; Get offset start of MAC Header.
	inc	DriverStatTable.MRxMismatchError	 ; Count the error
	or  byte ptr MSMRcvPacketStatus, PKT_MALFORM_ERR ; Set malformed pkt flag.
	jmp short TestFrameType

;----------------------------------------------------------------------------
;	Packet is Frame Type 802.3
;----------------------------------------------------------------------------
PacketIs8023:
	mov	di, offset CGroup:LookAhead8023	; DI -> LookAheadStruc
	mov	cl, FRAME_IS_8023		; Set up Frame Value
	jmp short Get14byteHeaderDataPtr	; Get AX loaded with Data pointer.

;----------------------------------------------------------------------------
;	Frame maybe a SNAP packet.
;----------------------------------------------------------------------------	
ItMaybeSNAP:
	cmp	byte ptr es:[si].Control, 03h	; test for SNAP AA AA 03
	jne	ItIsA8022Frame			; ne - Its a Reg. 802.2 Frame

;----------------------------------------------------------------------------
;	Frame is a SNAP packet. Load ProtID and continue.
;----------------------------------------------------------------------------
	mov	di, offset CGroup:LookAheadSNAP
	sub	ax, ax				; Clear AL (First byte 0)
	mov	ah, byte ptr es:[si].SNAP+0	; Copy 5 bytes of SNAP PID
	mov	word ptr [di].LProtID+0, ax
	mov	ax, word ptr es:[si].SNAP+1
	mov	word ptr [di].LProtID+2, ax
	mov	ax, word ptr es:[si].SNAP+3
	mov	word ptr [di].LProtID+4, ax
	sub	bp, 8				; 802.2 Snap Header Length
	lea	ax, [si+22]			; AX -> Start of Data
	mov	cl, FRAME_IS_SNAP		; Set up Frame Value
	jmp short FrameTypeDetermined

;----------------------------------------------------------------------------
;	Packet is completely Contained in the LookAhead.
;----------------------------------------------------------------------------
FrameIsLessThanLookAhead:
	mov	dx, bp				; Use packet length instead
	jmp short LookAheadSizeAdjusted		

;----------------------------------------------------------------------------
;	Set Error Bit and Clear All Other DestAddrType Bits.
;----------------------------------------------------------------------------
SetDestTypeErrorBit:
	mov byte ptr DestAddrTypeFlag, DEST_ERRORED ; Exclusive no other bits should be set.
	jmp short SetDestTypeErrorBitRet

;----------------------------------------------------------------------------
;	Restore bp=-1 for unknown Lenght at time GetRCB was called (pipelined)
;----------------------------------------------------------------------------
SetLDataSizeUnknown:
	mov	bp, -1				; Total Packet length Unknown.
	jmp short LDataSizeUnknownRet

;----------------------------------------------------------------------------
;	The frame is Ethernet with a type field.
;----------------------------------------------------------------------------
PacketIsEthTyped:
	mov	cl, FRAME_IS_EII
	mov	di, offset CGroup:LookAheadEII	; DI = ptr LookAhead Struc
	xchg	ah, al				; Get back into high-low order
	mov	word ptr [di].LProtID+4, ax	; Set LProtID field up.

	cmp	bp, MinMediaFrameSize
	ja	Get14byteHeaderDataPtr		; Might have pad bytes.	

	mov	FrameInfo, FRAME_COULD_BE_PADDED ; Flag possible Pad

Get14byteHeaderDataPtr:
	lea	ax, [si+14]			; AX = ptr Data Starts (Skip 14 byte headers)

;----------------------------------------------------------------------------
;	Correct Frame Type is determined, and the LProtID field is filled.
;	AX = ptr LookAhead Data Starts
;	CL = Which Frame Type was Determined.
;	BP = Packet Data Length (Frame Size - Header len)
;	ES:SI -> HSM LookAhead Data Buffer
;	DS:DI -> LookAhead Structure
;----------------------------------------------------------------------------
FrameTypeDetermined:
	or	bp, bp		 ; Was header's ok?
	js	MalformedPacket	 ; Neg says there is malformed.

;----------------------------------------------------------------------------
;	Determine if this frame type is even loaded (supported).
;----------------------------------------------------------------------------
TestFrameType:			 ; Next Lines Critical.  Determine Supported Frame.
	test	cl, 0		 ; This is a runtime Modified Value
LoadedFrameByte:		 ; Point where the Loaded Bits will be stored
	jz	UnSupportedFrame ; z => Frame not loaded.

;----------------------------------------------------------------------------
;	Check if the Packet is completly contained in the LookAhead Struc.
;----------------------------------------------------------------------------
TestLookAheadSize:
	mov	dx, DriverConfigTable.MLookAheadSize
	cmp	dx, bp				; Is pckt len less than lookahead size?
	ja	FrameIsLessThanLookAhead

;----------------------------------------------------------------------------
;	Load LookAhead Struc
;----------------------------------------------------------------------------
LookAheadSizeAdjusted:
	mov	[di].LMediaHeaderPtr+0, si	  ; Set the offset of the look ahead data
	mov	[di].LookAheadPtr+0, ax
	mov	[di].LookAheadLen, dx		  ; LookAhead buffer size
	cmp	HardwareRcvLength, -1		  ; Is packet length known?
	je	SetLDataSizeUnknown

;----------------------------------------------------------------------------
;	Set Attributes and check for Errors.
;----------------------------------------------------------------------------
LDataSizeUnknownRet:
	mov	word ptr [di].LStartCopyOffset, 0 ; Clear The Protocol Skip Value 
	mov	[di].LDataSize, bp		; Hardware Reported length - MAC Header
						;  will be -1 if unknown.
	mov	cx, MSMRcvPacketStatus		; Get packet status bits.
	mov	[di].LPacketAttrib, cx 		; Set status in lookahead.
	or	cl, cl				; Was There any Errors?
	jnz	SetDestTypeErrorBit		; set bit and come back here

;----------------------------------------------------------------------------
;	Save Frame Header Len on the Stack
;	AX = LookAheadPtr
;	SI = MediaHeaderPtr
;	DX = LookAheadLen
;----------------------------------------------------------------------------
SetDestTypeErrorBitRet:	
	sub	ax, si				; Get media header length
	push	ax				; Save frame header length

;----------------------------------------------------------------------------
;	Set Destination Flag on Stack
;----------------------------------------------------------------------------
	mov	ax, DestAddrTypeFlag		; Get destination type bits.
	mov	[di].LDestType, ax		; Set destination type bits in lookahead.
	push	ax				; Save DestAddrTypeFlag for ECB

;----------------------------------------------------------------------------
;	Put the Immediate address in the LookAhead Struc and
;	Save the Immediate Address for Later
;----------------------------------------------------------------------------
	mov	ax, word ptr es:[si].SrcAddr+0	; Get the immediate address + 0
	mov	word ptr [di].LImmAddress+0, ax	; Set the immediate address + 0
	push	ax				; Save immediate address + 0

	mov	ax, word ptr es:[si].SrcAddr+2	; Get the immediate address + 2
	mov	word ptr [di].LImmAddress+2, ax	; Set the immediate address + 2
	push	ax				; Save immediate address + 2

	mov	ax, word ptr es:[si].SrcAddr+4	; Get the immediate address + 4
	mov	word ptr [di].LImmAddress+4, ax	; Set the immediate address + 4
	push	ax				; Save immediate address + 4

;----------------------------------------------------------------------------
;	See if the LSL can get an ECB from someone.
;----------------------------------------------------------------------------
	cli					; Hold on the Interrupts
	mov	bx, MLIDSUP_GET_STACK_ECB	; Attempt to get a Rx buffer
	call	LSLSupport
	cld					; Not guarenteed
	sti					; Allow Interrupts to fire.
	jnz	DiscardTheFrameCleanStack

;----------------------------------------------------------------------------
;	We got one, fill it out and return a Receive Control Block (RCB) to 
;	the MLID so it can disperse the frame data.
;----------------------------------------------------------------------------
;	First Set the Immediate address field in the ECB.
;	and set the Dest Type field
;----------------------------------------------------------------------------
	pop	word ptr es:[si].ImmAddr+4	; Set the immediate address
	pop	word ptr es:[si].ImmAddr+2
	pop	word ptr es:[si].ImmAddr+0
	pop	word ptr es:[si].DriverWS+0	; Set Dest Type and 802.2 type

;----------------------------------------------------------------------------
;     	Determine size of ECB buffers, Initialize some more ECB fields
;----------------------------------------------------------------------------
	xor	dx, dx
	mov	cx, es:[si].FragCount		; Number of Frags
	lea	bx, [si].Frag1Addr		; ES:BX -> ECB

TotalRxBuffer:
	add	dx, word ptr es:[bx].FragLen 	; Total length of Rx buffers
	add	bx, size FragStruc		; Next Frag Please
	loop	TotalRxBuffer			; One less Frag

;----------------------------------------------------------------------------
;	Initialize some more ECB fields
;----------------------------------------------------------------------------
	mov	word ptr es:[si].DriverWS+2, bp	   ; Set the FrameDataLen up.
	mov	word ptr es:[si].DataLen, dx	   ; initialize the packet Data length field,
	mov	word ptr es:[si].Status, cx	   ; Clear ECB Status Field (CX = Zero)

;----------------------------------------------------------------------------
;	Get Protocol's Skip Count
;----------------------------------------------------------------------------
	mov	cx, word ptr [di].LStartCopyOffset ; CX = The Protocol Skip Value 

;----------------------------------------------------------------------------
;	Was Packet Size Known at Time of Getting ECB?
;----------------------------------------------------------------------------
	cmp	bp, -1			  ; Is packet length known?
	je	BackFromOverflow	  ; No, then skip overflow test.
	
;----------------------------------------------------------------------------
;	Check for Buffer Overflow Conditions
;----------------------------------------------------------------------------
	sub	bp, cx			; Adjust Number of Bytes to Move.
	cmp	dx, bp			; Is there enough room?
	jb	PossibleOverflow	; b => BufferSize < DataLen

	mov	es:[si].DataLen, bp	; Set the packet Data length field,
					; to the amount of data to move 
;----------------------------------------------------------------------------
;	Calculate/Save the Skip Values
;----------------------------------------------------------------------------
BackFromOverflow:
	pop	ax				; Restore the TSM skip value
	mov	MSMRcvHeaderLen, ax		; Save for RxCompleteStatus
	mov	MSMRcvProtSkipCount, cl		; Save for RxCompleteStatus
	add	ax, cx				; AX = Total Skip Count

;----------------------------------------------------------------------------
;	Count the Packet as Received and Return to MLID
;----------------------------------------------------------------------------
	add	DriverStatTable.MTotalRxPackets+0, 1
	adc	DriverStatTable.MTotalRxPackets+2, 0
	xor	cx, cx				; Set Z flag.
	ret

;----------------------------------------------------------------------------
;	No one wants this packet, Clean off the Stack and ret.
;----------------------------------------------------------------------------
DiscardTheFrameCleanStack:
	add	DriverStatTable.MNoECBsAvailable+0, 1
	adc	DriverStatTable.MNoECBsAvailable+2, 0
	add	sp, 10		; Pop off DestType, Immed. Addr, and Skip value
				; Sets NZ flag appropriately
	ret

;----------------------------------------------------------------------------
;	The BufferSpace is less than data Available,
;	might indicate not be enough room for this packet.
;----------------------------------------------------------------------------
PossibleOverflow:
	mov	ax, bp			; AX = Hardware Reported Length - headers - Skip counts.
	mov	bp, dx			; BP = Size of Buffer Space to fill

;----------------------------------------------------------------------------
;	Check to see if there is an Extra Byte to Evenize the Packet
;----------------------------------------------------------------------------
	sub	ax, bp			; Get overflow amount
	dec	ax			; Was it evenized?
	jz	BackFromOverflow

;----------------------------------------------------------------------------
;	Was This Packet just a padded EII Frame 
;    if HardwareReportedLen was <= 60 then just a padded Frame for EII only.
;----------------------------------------------------------------------------
	test	FrameInfo, FRAME_COULD_BE_PADDED
	jnz	BackFromOverflow		 ; nz => Most likely Padded go on.

;----------------------------------------------------------------------------
;	Not enough room for the entire Frame
;----------------------------------------------------------------------------
DefinitelyOverflow:
	mov	es:[si].Status, LSLERR_RX_OVERFLOW ; Indicate Overflow only.
	jmp short BackFromOverflow		   ; fill buffer space provided

MSMGetRCB	endp
;****************************************************************************

ifdef TASM
%NEWPAGE ;
endif
;****************************************************************************
;	MSMReturnRCB
;
;	Called by the driver when it needs to return an allocated RCB that
;	it decided it really didn't need after all.
;
;	assumes:	es:si	-> RCB structure
;			ds	is set to CGroup
;			Interrupts are disabled
;
;	returns:	bp,ds,ss,sp preserved
;			Interrupts disabled
;
;****************************************************************************
MSMReturnRCB	proc	near
	mov	word ptr es:[si].PrevLink, 0	; Clear the first word for RxStatus.
						; 0 RxStatus, and Cancelled => ReturnRCB 

ChangePacketStatus:
	mov byte ptr es:[si].DriverWS+0, DEST_ERRORED ; Change dest. type as error.
	mov	es:[si].Status, LSLERR_CANCELLED	; Set the ECB's status field.
	jmp short MSMRcvCompleteProceed			; Return the ECB.

MSMReturnRCB	endp

;****************************************************************************
;	MSMRcvCompleteStatus
;
;	Called by the HSM when it has finished dispersing the frame
;	data into the RCB fragment buffers.
;
;	This call passes control of the RCB to the MSM. The HSM should
;	not modify any values in the RCB after calling this routine.
;
;	assumes:	
;		es:si	-> RCB structure
;		ax = Status
;		bp = Total Hardware reported Length of the Packet
;		ds	is set to CGroup
;		Interrupts are disabled
;
;	returns:
;		ds,ss,sp preserved
;		Interrupts disabled
;
;****************************************************************************
;----------------------------------------------------------------------------
;	The BufferSpace is less than data Available,
;	might indicate not be enough room for this packet.
;	AX = Size of Buffer Space to fill
;	BP = Bytes that could have moved. 
;----------------------------------------------------------------------------
RcvCompPossibleOverflow:
;----------------------------------------------------------------------------
;	Check to see if there is an Extra Byte to Evenize the Packet
;----------------------------------------------------------------------------
	sub	bp, ax			; Get overflow amount
	dec	bp			; Was it evenized?
	jz	RcvCompBackFromOverflow

;----------------------------------------------------------------------------
;	Was This Packet just a padded EII Frame 
;    if HardwareReportedLen was <= 60 then just a padded Frame for EII only.
;----------------------------------------------------------------------------
	test	FrameInfo, FRAME_COULD_BE_PADDED
	jnz	RcvCompBackFromOverflow		 ; nz => Most likely Padded go on.

;----------------------------------------------------------------------------
;	Not enough room for the entire Frame
;----------------------------------------------------------------------------
RcvCompDefinitelyOverflow:
	mov	es:[si].Status, LSLERR_RX_OVERFLOW ; Indicate Possible Overflow only.
	jmp short RcvCompBackFromOverflow	   ; buffer space provided was filled.

;----------------------------------------------------------------------------
;	Hardware Reported Length is larger than Media Max.
;----------------------------------------------------------------------------
RcvCmpltFrmSizeTooBig:
	inc	DriverStatTable.MRxTooBig		; Inc. Rx Too Big count.
	or  byte ptr MSMRcvPacketStatus, PKT_BIG_ERR	; Set Too Big bit in packet
	jmp short RcvCmpltFrmSizeErrorRtn		;  status.

;----------------------------------------------------------------------------
;MSMRcvCompleteStatus Starts HERE:
;----------------------------------------------------------------------------
MSMRcvCompleteStatus	proc near
;----------------------------------------------------------------------------
;	Set the Current Status field, (sets up for fall through code later.)
;----------------------------------------------------------------------------
	or	MSMRcvPacketStatus, ax		; Set the Current FrameStatus

;----------------------------------------------------------------------------
;	Check that the Frame Meets current regs.
;	Is reported Frame size Larger than Media can Handle?
;----------------------------------------------------------------------------
	cmp	bp, MaxMediaFrameSize		; BP = Reported FrameSize
	ja	RcvCmpltFrmSizeTooBig		; a => Too Big
							
;----------------------------------------------------------------------------
;     Determine the Frame size and number bytes that should have been copied.
;----------------------------------------------------------------------------
RcvCmpltFrmSizeErrorRtn:
	mov	ax, MSMRcvHeaderLen		; Get HeaderLen and set AH=0
	sub	bp, ax				; Subtract off the Header Value
	mov	word ptr es:[si].DriverWS+2, bp	; Store Total FrameDataLen
	mov	al, MSMRcvProtSkipCount		; Get Protocol Skip Count
	sub	bp, ax				; BP = Number of Bytes that Should have Moved
	mov	ax, word ptr es:[si].DataLen	; AX = Total Buffer Len
	cmp	ax, bp				; (AX=TotalBuffer Len) < (BP = Bytes That could have moved)
	jb	RcvCompPossibleOverflow		; TotalBufferLen < FrameDataLen == Overflow

	mov	word ptr es:[si].DataLen, bp	; Set to Total Bytes Moved.

RcvCompBackFromOverflow:			; Return for fall through keep code clean.

MSMRcvCompleteStatus	endp			; Fall into MSMRcvComplete to finish up.

;****************************************************************************
;	MSMRcvComplete
;
;	Called by the MLID when it has finished dispersing the frame
;	data into the RCB fragment buffers.
;
;	This call passes control of the RCB to us. The driver should
;	not modify any values in the RCB after calling this routine.
;
;	assumes:	es:si	-> RCB structure
;			ds	is set to CGroup
;			Interrupts are disabled
;
;	returns:	bp,ds,ss,sp preserved
;			Interrupts disabled
;
;****************************************************************************
MSMRcvComplete	proc	near
	mov	ax, MSMRcvPacketStatus
	mov	word ptr es:[si].PrevLink, ax

;----------------------------------------------------------------------------
;	Check for any error condition.
;----------------------------------------------------------------------------
	or	al, al		; Any errors?
	jnz	ChangePacketStatus

;----------------------------------------------------------------------------
;	Give the ECB back to the LSL
;----------------------------------------------------------------------------
MSMRcvCompleteProceed:
	mov	bx, MLIDSUP_HOLD_RECV_EVENT	; Give it to the LSL
 	call	LSLSupport
	ret

MSMRcvComplete	endp

ifdef TASM
%NEWPAGE ;
endif
;*******************************************************************************
;	E T H E R N E T		M U L T I C A S T	C O D E
;*******************************************************************************
;	MSMAddMulticastAddress
;
;	Called by a protocol stack to enable reception of a particular
;	multicast address.  In other words, we become a member of a 
;	multicast group.
;
;	assumes:
;		ax	Board Number
;		es:si	-> 6 byte buffer holding the MC address
;		ds	is set to CGroup
;		CLD has been executed
;		Interrupts are enabled
;			MSM disables them prior to calling TSM
;
;	returns:
;		ax	=0 if command was successful
;			=LSLERR_BAD_PARAMETER is addr was invalid
;			=LSLERR_OUT_OF_RESOURCES if max. already in use
;		Flags set according to AX
;		All other registers destroyed
;		Interrupt State is Preserved.
;
;*******************************************************************************
MSMAddMulticastAddress	proc	near
;----------------------------------------------------------------------------
;	Check for valid address
;----------------------------------------------------------------------------
	mov	ax, LSLERR_BAD_PARAMETER	; Set possible error code
	test	byte ptr es:[si+0], 01h		; Is MC bit set?
	jz	BadMCAddress

;----------------------------------------------------------------------------
;	Check if the address is already in the table.
;----------------------------------------------------------------------------
	call	FindAddressInMCTableNotLoaded	; See if already added (AX:BX:DX not loaded yet)
	jz	AddressWasFound			; Returns AX:BX:DX with Address.

;----------------------------------------------------------------------------
;	Not There Find an empty slot.
;----------------------------------------------------------------------------
	mov	di, MulticastTable		; DI = MulticastTable (in PSP)
	mov	cx, MSMMaxMulticastAddr		; Get max. possible inuse in table

FindEmptySlotLoop:
	cmp	word ptr [di+6], 0		;Slot in use?
	je	FoundEmptyMCSlot

	add	di, 8				;Advance to next slot
	loop	FindEmptySlotLoop

;----------------------------------------------------------------------------
;	Set an error flag and return.
;----------------------------------------------------------------------------
	mov	ax, LSLERR_OUT_OF_RESOURCES	; No free slots

BadMCAddress:
	ret					; PSW set after return.

;----------------------------------------------------------------------------
;	Found a slot
;----------------------------------------------------------------------------
FoundEmptyMCSlot:
	mov	word ptr [di+0], ax		; Save the address
	mov	word ptr [di+2], bx
	mov	word ptr [di+4], dx
	inc	word ptr [di+6]			; Increment use count
	mov	bx, ADD_MULTICAST_ADDRESS	; Indicate that the Address at DS:DI was added

;----------------------------------------------------------------------------
;	Tell the HSM the MC Table has changed.
;----------------------------------------------------------------------------
IssueChangeToDriver:
	call	MSMUpDateMulticastInternal
	jmp short MSMNoChange		; Finish out.

AddressWasFound:
	inc	word ptr [di+6]			; Increment use count
	jmp short MSMNoChange			; Dont Bother The HSM

MSMAddMulticastAddress	endp


ifdef TASM
%NEWPAGE ;
endif
;*******************************************************************************
;	MSMDelMulticastAddress
;
;	Called by a protocol stack to disable reception of a particular
;	multicast address.
;
;	assumes:	
;		ES:SI	-> 6 byte buffer holding the MC address
;		DS	is set to CGroup
;		CLD has been executed
;		Interrupts are enabled
;				Prior to calling TSM the MSM disables Ints
;					
;
;	returns:
;		ax	=0 if command was successful
;			=LSLERR_BAD_PARAMETER is addr was invalid
;			=LSLERR_ITEM_NOT_PRESENT if addr was not found
;		Flags set according to AX 
;		BP, DS, SS, SP registers preserved
;		Interrupt State is preserved.
;
;*******************************************************************************
MSMDelMulticastAddress	proc	near
;----------------------------------------------------------------------------
;	Check for Valid MC Address first.
;----------------------------------------------------------------------------
	mov	ax, LSLERR_BAD_PARAMETER	; Set possible error code
	test	byte ptr es:[si+0], 01h		; Is MC bit set?
	jz	BadMCAddress			; Exit out and set flags.

;----------------------------------------------------------------------------
;	Look for the Address in the table to delete.
;----------------------------------------------------------------------------
	call	FindAddressInMCTableNotLoaded	; See if already added (AX:BX:DX not loaded yet)
	jnz	AddressWasNotFound		

;----------------------------------------------------------------------------
;	Address was found, Decrement count and continue
;----------------------------------------------------------------------------
	mov	bx, DELETE_MULTICAST_ADDRESS	; Let the HSM Know we are Deleting address at DS:DI
	dec	word ptr [di+6]			; Adjust use count
	jz	IssueChangeToDriver

MSMNoChange:
	xor	ax, ax		; Signal Success
	ret

AddressWasNotFound:
	mov	ax, LSLERR_ITEM_NOT_PRESENT
	ret

MSMDelMulticastAddress	endp

ifdef TASM
%NEWPAGE ;
endif
;*******************************************************************************
;	MSMGetMulticastInfo
;
;	Called by a protocol stack to get pointers to the Multicast Table.
;
;	assumes:	(MSMControlHandler has been called)
;		
;		DS	is set to CGroup
;		CLD has been executed
;		Interrupts are enabled
;				Prior to calling TSM the MSM disables Ints
;					
;
;	returns:
;		ax	=LSL_SUCCESSFUL (0) if command was successful
;		ES:SI -> Multicast Table (Group Address)
;		ES:DI -> Functional Address (Not Used for Ethernet)
;		CX	Size of the Multicast Table.
;	      (MSMControlHandler takes care of any other return values.)
;
;		Flags set according to AX 
;		DS, BP, SS, SP registers preserved
;		Interrupt State is preserved.
;
;*******************************************************************************
MSMGetMulticastInfo	proc	near
	xor	ax, ax			; setup for return code.
	push	ds
	pop	es			; Get CGROUP
	mov	si, MulticastTable	; Located in PSP area.
	mov	di, ax			; Functional Addressing not used.
	mov	cx, MSMMaxMulticastAddr	; Set for table size.
	ret
MSMGetMulticastInfo	endp


ifdef TASM
%NEWPAGE ;
endif
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;	FindAddressInMCTable
;
;	Called to see if the specified Multicast address is present
;	in the multicast filter table.
;
;	Assumes:	
;		AX:BX:DX	has the 6-byte MC address
;		DS		is set to CGroup
;		CLD has been executed
;		Interrupts are disabled
;
;	Returns:	
;	 	z flag 	set if address was found
;		nz flag	is addr not found
;		ds:di	-> Addr entry if it was found
;		CLD preserved
;		CX, DI destroyed 
;		All other registers preserved
;		Interrupts disabled and not enabled
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;----------------------------------------------------------------------------
;	Assumes: ES:SI -> Address to find
;----------------------------------------------------------------------------
FindAddressInMCTableNotLoaded:
	mov	ax, word ptr es:[si+0]		; Load up the address
	mov	bx, word ptr es:[si+2]
	mov	dx, word ptr es:[si+4]

FindAddressInMCTable	proc	near
;----------------------------------------------------------------------------
;	Get Pointer to MC Table and count
;	Look for address in the Table
;----------------------------------------------------------------------------
	mov	di, MulticastTable	; DI = MulticastTable (in PSP)
	mov	cx, MSMMaxMulticastAddr	; Get max. possible being used in table

LookUpAddrLoop:
	cmp	word ptr [di+6], 0	; Entry in use?
	je	NotThisEntry

	cmp	word ptr [di+4], dx	; check the Address Word at a time.
	jne	NotThisEntry

	cmp	word ptr [di+2], bx
	jne	NotThisEntry

	cmp	word ptr [di+0], ax
	jne	NotThisEntry

;----------------------------------------------------------------------------
;	Address was Found it is in the Table.
;----------------------------------------------------------------------------
	xor	cx, cx			; Clear CX and set z-flag
	ret

NotThisEntry:
	add	di, 8			; Next table entry
	loop	LookUpAddrLoop
	
;----------------------------------------------------------------------------
;	Address not found Set up completion Code
;----------------------------------------------------------------------------
	or	cx, di			; Set NZ flag (Addr not found)
	ret

FindAddressInMCTable	endp
;============================================================================

;****************************************************************************
;	MSMUpdateMulticast
;
;	This routine is assumed to be in the media specific module and will
;	be called by the generic MSM module after a DriverReset IOCTL
;	has been issued to the driver.  By calling this routine after
;	DriverReset we guarantee the driver has the proper multicasts
;	setup after a reset.
;
;	MSMUpdateMulticast will not call DriverMulticastChange if in 
;	Promiscuous Mode as this may well affect a number of Multicast 
;	functions in a LAN Controller, eg. 8390.  But will call 
;	DriverPromiscuousChange instead, as this allows HSM's to be re-enabled
;	for Promiscuous Mode if internally reset, eg. after an internal
;	reset the HSM calls MSMUpdateMulticast to ensure correct settings
;
;
;	MSMUpdateMulticast is called by the following Procs:
;		MSMAddMulticastAddress
;		MSMDelMulticastAddress
;		MSMReset
;		MSMPromiscuousChange
;
;	assumes:
;		DS	CGroup
;		BX	= Function Indicator
;			 0 => DI not Valid
;			 DELETE_ADD_MULTICAST_ADDRESS
;			 ADD_MULTICAST_ADDRESS
;		DS:DI -> Address in Multicast Table if Function Indictor !=0
;
;		CLD has been executed
;		Interrupts are disabled
;
;	returns:	
;		Nothing Preserved
;****************************************************************************
MSMUpdateMulticast	proc	near
	xor	bx, bx		; Signal that this is not an Add/Delete Multicast call

;----------------------------------------------------------------------------
;	Check for Promiscuous State.
;----------------------------------------------------------------------------
MSMUpDateMulticastInternal:
	mov	cx, MSMActivePromisMask		; Get Current State of Promiscuous Mode
	jcxz short MSMUpdateMulticastCont	; Test if in Promis. Mode cxz =>NO, skip down

;----------------------------------------------------------------------------
;	We are in Promiscuous Mode, 
;----------------------------------------------------------------------------
	jmp	DriverPromiscuousChange		; CX = Not Zero -> IN Promiscuous Mode.

;----------------------------------------------------------------------------
;	Not in Promiscous Mode, tell the HSM the Current MC addresses.
;----------------------------------------------------------------------------
MSMUpdateMulticastCont:
	mov	cx, MSMMaxMulticastAddr
	mov	si, MulticastTable		; SI = MulticastTable (in PSP)
	jmp	DriverMulticastChange		; Tell the driver about the update

MSMUpdateMulticast	endp

;****************************************************************************
;  MSMMediaConfigUpdate
;
;	This function is called by the HSM after the hardware is fully
;	functional.  At Init time all HSMs must make this call to initially
;	setup the Configuration tables and Media Specific items.
;
;	At Run Time, an HSM would only call this function to update
;	the configuration tables after a HARDWARE change had occured.

;	The HSM must have the MSMTxFreeCount, MSMLookAheadSegment, and 
;	MSMPriorityQSupportPtr (if needed) initialized prior to calling.
;
;	On ENTRY:
;		DS = CS = CGroup
;		Interrupts disabled
;
;	On Return:
;		AX = Error code
;			0 = success,
;			MSMERR_TSM_FAILED => error occured.
;
;		DX = Message Offset of the Error.
;		     only the Init time can the Error Code be used to print
;		     the error message.
;
;		ES = CGroup
;
;		Interrupts disabled
;		DS,BP,SS,SP preserved.
;
;****************************************************************************
MSMMediaConfigUpdate	proc	near
;----------------------------------------------------------------------------
;	Guarentee that ES is set to CGroup
;----------------------------------------------------------------------------
	push	cs
	pop	es	; CGroup

;----------------------------------------------------------------------------
;	Check that the HSM has set the MSMTxFreeCount
;----------------------------------------------------------------------------
	mov	dx, offset CGroup:ForgotTxFreeCount	; Possible Error
	mov	al, MSMTxFreeCount		; Save a copy of the value
	mov	MSMMaxActiveTx, al
	or	al, al				; Check if even set?
	jz	HSMForgotExit

;----------------------------------------------------------------------------
;	Check that there aren't too many Tx Resources.
;----------------------------------------------------------------------------
	mov	dx, offset CGroup:TooManyTCBs	; Possible Error
 	cmp	al, MAX_TCB_ALLOCATED		; Does it exceed Max number of TCBs.
	ja	HSMForgotExit

;----------------------------------------------------------------------------
;	Now check to make sure the HSM set the MSMLookAheadSegment
;----------------------------------------------------------------------------
	mov	ax, MSMLookAheadSegment
	mov	dx, offset CGroup:ForgotLASegment
	or	ax, ax
	jz	HSMForgotExit

;----------------------------------------------------------------------------
;	Initialize the Look Ahead structures Segment Values
;----------------------------------------------------------------------------
	mov	LookAhead8023.LMediaHeaderPtr+2, ax
	mov	LookAhead8023.LookAheadPtr+2, ax

	mov	LookAhead8022.LMediaHeaderPtr+2, ax
	mov	LookAhead8022.LookAheadPtr+2, ax

	mov	LookAheadSNAP.LMediaHeaderPtr+2, ax
	mov	LookAheadSNAP.LookAheadPtr+2, ax

	mov	LookAheadEII.LMediaHeaderPtr+2, ax
	mov	LookAheadEII.LookAheadPtr+2, ax

;----------------------------------------------------------------------------
;	Set up the Node Address in all Config Tables.
;	Ethernet can only be LSB.  Simply need to copy to table.
;	Set up the Best and worst Data Size Fields also
;----------------------------------------------------------------------------
	push	cs
	pop	es				; ES = CGroup
	mov	bx, MSMActiveBoardCount		; Number of Loaded FrameTypes
	add	bx, bx				; Dword Offset

CheckNextBoardsAddress:
	mov	di, MSMConfigTables[bx-2]	; Get first Config Table
	mov	ax, [di].MMaxPacketSize		; Get Max for This Frame Type
	xor	cx, cx				; Clear CX
	mov	si, [di].MBoardNumber		; SI => Board Number
	mov	cl, HeaderSizeTable[si]		; CX = Header Size.
	sub	ax, cx				; AX = BestDataSize
	mov	[di].MBestDataSize, ax		; Store Size
	cmp	[di].MFrameID, MediaID8022	; Check for 802.2
	jne	SetWorstDataSize		; Worst case is same as best on Ethernet

	dec	ax				; Except in 802.2 case need to sub one more.

SetWorstDataSize:
	mov	[di].MWorstDataSize, ax		; Store the WorstDataSize.
	lea	si, MSMPhysNodeAddress		; We don't allow Node Address to be 
	lea	di, [di].MNodeAddress		; different. (LSB mode only)
	mov	cx, 3
 rep	movsw
	sub	bl, 2				; Set Index
	jnz	CheckNextBoardsAddress		; Check each Board

;----------------------------------------------------------------------------
;	Setup the source node address in all of the media headers.
;----------------------------------------------------------------------------
	mov	bx, TCBFreeList			;Get ptr first TCB
	lea	si, MSMPhysNodeAddress		;DS:SI -> Physical node address

DoNextTCB:
	push	si
	lea	di, [bx].TCBMediaHeader.SrcAddr	; ES:DI -> src addr field
	movsw					; Set the src address
	movsw
	movsw
	pop	si
	mov	bx, [bx+0]		; Next TCB
	or	bx, bx
	jnz	DoNextTCB

	jmp	MSMUpdateConfigTables	; No Known Errors to pass back out on ret.

;----------------------------------------------------------------------------
;	Error Exit
;----------------------------------------------------------------------------
HSMForgotExit:
	mov	ax, MSMERR_TSM_FAILED	; DX = MSG offset for init time
	or	ax, ax			; Set appropriate flags.
	ret		

MSMMediaConfigUpdate	endp

Code	ends

ifdef TASM
%NEWPAGE ;
endif
;*******************************************************************************
;		M E D I A	I N I T I A L I Z A T I O N
;*******************************************************************************
Init	segment	para public 'CODE'
	assume	cs:CGroup, ds:CGroup, es:nothing, ss:nothing

	public	InitializeMedia
	public	HSMProvideTCB
	public	SupportedMedias

;----------------------------------------------------------------------------
;	Provided by the MSM
;----------------------------------------------------------------------------
	extrn	MSMAlloc		:near
	extrn	MSMEatWhite		:near
	extrn	MSMGetStringLength	:near
	extrn	MSMHexStringToByteArray :near
	extrn	MSMPrintStringZero	:near

;----------------------------------------------------------------------------
;	Provided by the HSM
;----------------------------------------------------------------------------
	extrn	DriverInit		:near

;----------------------------------------------------------------------------
;	Messages from MSMtxt.inc
;----------------------------------------------------------------------------
	extrn	InvalidMediaAddress	:byte
	extrn	InvalidMediaAddress2	:byte
	extrn	ForgotTxFreeCount	:byte
	extrn	ForgotLASegment		:byte
	extrn	TooManyTCBs		:byte

;Internal Debug
ifdef DEBUG
	public	MediaNodeAddressCheck, CheckEthernetAddrMode
	public	PhysNodeInMSBForm, CheckSwapBitAddr
	public	MediaIndexToBoardMap, SetDefaultLBoardNum
endif

;----------------------------------------------------------------------------
;		M E S S A G E   C E N T E R
;		  Media Specific Messages.
;----------------------------------------------------------------------------
		db	'MEDIA = ',EthernetTSMVersion	; Must always be present
EthernetMsg	db	'Ethernet', 0

;----------------------------------------------------------------------------
;	This table is required for use with the generic MSM module.
;----------------------------------------------------------------------------
SupportedMedias	dw	MaxVirtBoards	 	  ;Num. medias supported
		; Ethernet_802.2
		dw	offset CGroup:MediaName3  ;Default is first one
		dw	MediaID8022		  ;Ethernet 802.2
		db	0,0,0,0,0,0E0h		  ;Default ProtocolID for IPX
		; Ethernet_802.3
		dw	offset CGroup:MediaName5
		dw	MediaIDRaw8023		  ;Raw 802.3
		db	0,0,0,0,0,0		  ;Default ProtocolID for IPX
		; Ethernet_II
		dw	offset CGroup:MediaName2  ;Ethernet Typed
		dw	MediaIDEthernetTyped
		db	0,0,0,0,81h,37h		  ;Default ProtocolID for IPX
		; Ethernet_SNAP
		dw	offset CGroup:MediaName10 ;SNAP
		dw	MediaIDSNAP
		db	0,0,0,0,81h,37h		  ;Default ProtocolID for IPX

;----------------------------------------------------------------------------
;	Must have each entry up to the Last Entry supported by this Media
;----------------------------------------------------------------------------
MediaToBoardIndexTable	dw	0,0		   ; Skip Media 0,1
			dw	EthernetTypedIndex ; Index for EthernetTyped (2)
			dw	Ethernet8022Index  ; Index for 802.2 (3)
			dw	0		   ; Skip Media 4
			dw	Ethernet8023Index  ; Index for Raw 802.3 (5)
			dw	0,0,0,0		   ; Skip Media 6,7,8,9
			dw	EthernetSNAPIndex  ; Index for SNAP (10)

;----------------------------------------------------------------------------
;	Index Equates for Accessing the following Tables.
;----------------------------------------------------------------------------
EthernetTypedIndex	equ	0		;Index for EthernetTyped
Ethernet8022Index	equ	2		;Index for Ethernet 802.2
Ethernet8023Index	equ	4		;Index for Ethernet Raw 802.3
EthernetSNAPIndex	equ	6		;Index for Ethernet 802.2/SNAP

;----------------------------------------------------------------------------
;	Index equates are used to access the Size Table.
;----------------------------------------------------------------------------
MediaHeaderSizeTable	dw	14			; Ethernet Typed
			dw	17			; Ethernet 802.2 (Default Type 1)
			dw	14			; Ethernet Raw 802.3
			dw	22			; Ethernet SNAP

;----------------------------------------------------------------------------
;	Index equates access this table.
;----------------------------------------------------------------------------
MediaTxHandlerTable	dw	offset CGroup:BuildEthernetTypedHeader
			dw	offset CGroup:Build8022Header
			dw	offset CGroup:BuildRaw8023Header
			dw	offset CGroup:BuildSNAPHeader

;----------------------------------------------------------------------------
;	Index Equates used to access this table 
;	Board number for Frame types, 
;	Default Board number is set to allow error packets to be passed up.
;	Default will be set to the initial logical board number.
;----------------------------------------------------------------------------
MediaIndexToBoardMap		dw	MaxVirtBoards dup (0)	;Holds the board nums for each virtual board

;----------------------------------------------------------------------------
;	Index Equates used to access this table 
;	Bit Table Used to Set the LoadedFrameByte Field
;----------------------------------------------------------------------------
LoadedFrame	dw	FRAME_IS_EII	
		dw	FRAME_IS_8022	
		dw	FRAME_IS_8023	
		dw	FRAME_IS_SNAP	

ifdef TASM
%NEWPAGE ;
endif
;****************************************************************************
;	InitializeMedia
;
;	Called by the main MSM to do any initialization of variables
;	and tables that will be used in building and stripping our media 
;	headers.  We also call the HSM to initialize.
;
;	assumes:
;		DS	is set to CGroup
;		CLD has been executed
;		Interrupts are enabled
;
;	returns:
;		AX	= 0 if no errors
;			!= 0 if error (DX is valid)
;		DX	-> err message offset if AX != 0
;		DS = ES = CGroup
;		Flags set according to AX
;		CLD has been executed
;		Interrupts are enabled.
;
;****************************************************************************
InitializeMedia	proc	near
;----------------------------------------------------------------------------
;	Test Node Address override if any correct for media.
;----------------------------------------------------------------------------
	push	cs
	pop	es			; ES = CGroup
	call	MediaNodeAddressCheck	; Check for Node Address Override.

;----------------------------------------------------------------------------
;	Enable the counter bits that we increment.
;	Prior to the AND Need to invert our logic, set => not used.
;	Remember this 32 bit number high order bits were used first.
;----------------------------------------------------------------------------
	and	word ptr DriverStatTable.MValidCounterMask+2, not 1111001100001100b

;----------------------------------------------------------------------------
;    Reset PSP Area to zero to allow for our use as no longer need Swap Table
;----------------------------------------------------------------------------
	mov	di, PSPTableStart	; get start of Swap Table Area
	mov	cx, 100h
	xor	ax, ax			; Provides zero default values.
	rep stosb			; reset area to zero.

;----------------------------------------------------------------------------
;	Initialize the LookAhead Board Number to the First Logical board
;	supported by the Driver as a default
;----------------------------------------------------------------------------
	mov	bx, DriverConfigTable.MBoardNumber ; Get Default Board Num
	mov	bp, (MaxVirtBoards-1)		   ; DI = Num of possible Virt Boards (0 based)

SetDefaultLBoardNum:
	mov	ds:MediaIndexToBoardMap[bp], bx	; Save in look up table 
	dec	bp				; count down till done.
	jnz	SetDefaultLBoardNum		; nz => not done

;----------------------------------------------------------------------------
;	Initialize The Media Tables.
;	BP = 0
;----------------------------------------------------------------------------
	mov	cx, MSMActiveBoardCount	; Number of config tables used

SetupMediaTablesLoop:
	mov	di, ds:MSMConfigTables[bp]	; DI -> config table
	mov	bx, [di].MBoardNumber	; BX =  Actual Board Number	

;----------------------------------------------------------------------------
;	Determine Frame type, and Load Table with board number
;----------------------------------------------------------------------------
	mov	si, [di].MFrameID		; Get the media of this virtual board
	add	si, si				; Convert to word index
	mov	si, MediaToBoardIndexTable[si]	; Convert to EthernetXXXXIndex value
	mov	MediaIndexToBoardMap[si], bx	; Save in look up table For Setting Lookahead Strucs.

;----------------------------------------------------------------------------
;	See if this media needs to be evenized.
;----------------------------------------------------------------------------
	cmp	si, EthernetTypedIndex		; EII needs evenized
	je	EvenizeMedia

	cmp	si, Ethernet8023Index		; Raw 802.3 needs evenized
	jne	DontEvenizeMedia		; Everyone else don't

EvenizeMedia:
	or	BoardFlags[bx], EVENIZE		; Set evenize flag

;----------------------------------------------------------------------------
;	Initialize the Header size Table.
;----------------------------------------------------------------------------
DontEvenizeMedia:
	mov	ax, MediaHeaderSizeTable[si]	; Get this headers size
	mov	HeaderSizeTable[bx], al		; Initialize table
	or	BoardFlags[bx], LOADED		; Mark Media (Board) Loaded

;----------------------------------------------------------------------------
;	Set the Loaded Frame Bit
;	SI = Frame type Index
;----------------------------------------------------------------------------
	mov	dx, LoadedFrame[si]	 ; Get Loaded Frame Bit
	or	byte ptr LoadedFrameByte - 1, dl ; Mark Loaded Frame in Code.

;----------------------------------------------------------------------------
;	Now adjust the packet size fields in the config table to reflect 
;	the size of this media header.
;	AX = Header Size
;	DI -> Config Table
;	BX = Board Number
;	SI = Frame type Index
;----------------------------------------------------------------------------
	mov	dx, MaxMediaFrameSize	; Get absolute max. for Ethernet
	mov	[di].MMaxPacketSize, dx	; Save in config table
	sub	dx, ax			; Subtract off length of media header
	mov	[di].MBestDataSize, dx	; Set best case size
	mov	[di].MWorstDataSize, dx	; Worst case is same as best on Ethernet
	cmp	si, Ethernet8022Index	; Except in 802.2 case need to sub one more.
	jne	Not8022Index

	dec	[di].MWorstDataSize	; Take one more for type 2 case.

;----------------------------------------------------------------------------
;	Setup MSMTableIndex table.
;	Save off the BuildHeaderHandlers
;----------------------------------------------------------------------------
Not8022Index:
	mov	dx, MediaTxHandlerTable[si]	   ; Get -> Txhandler routine for this media type
	mov	si, bx				   ; SI = Board Number
	sub	bx, DriverConfigTable.MBoardNumber ; Convert to Virtual value
	add	bx, bx				   ; Need a word Virtboard index
	mov	MSMTableIndex[si], bl		   ; Save MSMTableIndex table value
	mov	BuildHeaderHandlers[bx], dx	   ; Save TxHandler in board lookup table

;----------------------------------------------------------------------------
;	Initialize ModeFlag bits in Config Table.
;----------------------------------------------------------------------------
	or	[di].MModeFlags, MRawSendBit or MLDataSizeBit ; Set raw send bit and lookahead Data Size support field
	and	[di].MModeFlags, NOT MNonCanonicalBit	      ; ensure bit is cleared

;----------------------------------------------------------------------------
;	Move to next Config Table.
;----------------------------------------------------------------------------
	add	bp, 2				;Advance to next config table slot
	loop	SetupMediaTablesLoop

;----------------------------------------------------------------------------
;	Initialize the Board Number Initialized for each LookAhead struc.
;----------------------------------------------------------------------------
	mov	ax, MediaIndexToBoardMap[ EthernetTypedIndex ]
	mov	LookAheadEII.LBoardNum, ax

	mov	ax, MediaIndexToBoardMap[ Ethernet8022Index  ]
	mov	LookAhead8022.LBoardNum, ax

	mov	ax, MediaIndexToBoardMap[ Ethernet8023Index  ]
	mov	LookAhead8023.LBoardNum, ax

	mov	ax, MediaIndexToBoardMap[ EthernetSNAPIndex  ]
	mov	LookAheadSNAP.LBoardNum, ax

;----------------------------------------------------------------------------
;	Allocate and Initialize Max number of TCB structures.
;----------------------------------------------------------------------------
	mov	cx, MAX_TCB_ALLOCATED	; Allocate the Max
	mov	ax, size TCBStruc + MaxFrameHeaderLen
	mov	di, offset CGroup:TCBFreeList

LinkHeaderBufLoop:
	call	MSMAlloc		; Allocate a TCB
	mov	[di], bx		; Link it in the list
	mov	di, bx			; DI = prev Link
	loop	LinkHeaderBufLoop

	mov	[di], cx		; Terminate the list

;----------------------------------------------------------------------------
;	Call the HSM to Initialize the Card.
;	DriverInit
;	On Entry:
;		DS=ES=CGroup
;		Interrupts are enabled
;		cld in effect
;	On Return:
;		AX = Completion Code:
;			MSM_SUCCESSFUL (0)  successful, (MSMMediaConfigUpdate has been Called)
;			MSMERR_HSM_FAILED (1) The HSM was Unable to INit the Card,
;			MSMERR_TSM_FAILED (2) the TSM was unable to Initalize something,
;			MSMERR_MSM_FAILED (3) the MSM was unable to Initalize something.
;			MSMERR_CARDNOTPRESENT (4) The Software was able to INit,
;					but the Hardware is not present. TSR and Stay In Shutdown State.
;		DX = ptr to Error Message if AX != 0 
;		Interrupts are disabled
;		Flags set according to AX
;		SS,SP preserved
;					
;----------------------------------------------------------------------------
	push	cs
	pop	es
	call	DriverInit		; is not required to Preserve anything.
	cld
	push	cs
	push	cs
	pop	ds			; CGroup
	pop	es
	ret

InitializeMedia	endp

;****************************************************************************
;****************************************************************************
;	HSMProvideTCB
;	This procedure is only available at DriverInit Time.  The HSM must
;	Provide all TCBs with in CGroup.   The TCB buffer doesn't have to be
;	initialized, but they must be large enough to hold the TCB and the 
;	MAX Media Header. (TCBStruc + MaxFrameHeaderLen)
;
;	On Entry:
;		DS:SI -> TCB 
;		Interrupts are disabled.
;
;	On Exit:
; 		CX is destroyted.
;		Everything else is preserved.
;
;	All the TCB's are initialized when MSMConfigMediaUpdate is Called.
;****************************************************************************
HSMProvideTCB	proc near
	mov	cx, TCBFreeList			; Free up this TCB
	mov	[si].NextLink, cx
	mov	TCBFreeList, si
	ret
HSMProvideTCB	endp

ifdef TASM
%NEWPAGE ;
endif
;****************************************************************************
;
; NAME:		MediaNodeAddressCheck
;
; DESCRIPTION:	
;	If Node Address Override has taken Place then
;	check the node address in MSMPhysicalNodeAddress 
;	Also fill in the Node Address in the DriverConfigTable 
;	in the form Specified by ModeFlags bit 15 and 14.
;
;
; ON ENTRY:	AX	N/A
;		BX	N/A
;		CX	N/A
;		DX	N/A
;		BP	N/A
;		SI	N/A
;		DI	N/A
;		CS	CGroup
;		DS	CGroup
;		ES	CGroup
;	
;		NOTE:	cld in effect, Interrupts are enabled
;
;
; ON RETURN:	AX	Destroyed
;		BX	Destroyed
;		CX	Destroyed
;		DX	Destroyed
;		BP	Preserved (Not Used)
;		SI	Destroyed
;		DI	Destroyed
;		CS	CGroup
;		DS	CGroup
;		ES	CGroup
;
;
;		Interrupt state is not changed during function.
;	
;
; REMARKS:
;	The DoNodeKeyWord in the Generic MSM.asm sets the LSB/MSB mode bits
;	if nessasary.
;	Bit 15 signals the support of the particular driver (Set by the HSM)
;			Bit 15	1 = LSB/MSB supported. 
;				0 = Not Supported (old driver)
;
;	Bit 14 describes the form of Node Address in the DriverconfigTable 
;				(Set by the MSM)
;			Bit 14 	1 = MSB (Non-Canonical Form)
;				0 = LSB (Canonical Form)
;
; SEE ALSO:
;
;****************************************************************************
;****************************************************************************
MediaNodeAddressCheck	proc	near
	mov	ax, word ptr MSMPhysNodeAddress
	and	ax, word ptr MSMPhysNodeAddress+2
	and	ax, word ptr MSMPhysNodeAddress+4	; Check for Not Set
	inc	ax					
	jz	DoNodeExit				; z=>no node override

;----------------------------------------------------------------------------
;	We are Ethernet make sure it is a Valid Address.
;----------------------------------------------------------------------------
CheckEthernetAddrMode:
	cmp	byte ptr MSMPhysNodeAddress+6, 'M' ; If MSB form need to Swap it.
	jne	PhysNodeInLSBForm

;----------------------------------------------------------------------------
;	MSMPhysNodeAddress is in MSB form.   For Ethernet it must be Swapped.
;----------------------------------------------------------------------------
PhysNodeInMSBForm:
	mov	si, Offset CGroup:MSMPhysNodeAddress
	mov	di, si					; Do Swap in Place.
	mov	cx, 3
	mov	bx, PSPTableStart			; BitSwapTable @

CheckSwapBitAddr:
	lodsw				; Get a word
	xlat				; Swap it
	xchg	ah, al			; AH gets result, AL is Next look up
	xlat				; Swap it
	xchg	ah, al			; straighten back for save
	stosw				; Store it.
	dec	cx
	jnz	CheckSwapBitAddr	; get all 3 words of Address.

	mov	al, 'L'			; Flag it correctly.
	stosb

;----------------------------------------------------------------------------
;	MSMPhysNodeAddress is in LSB form.   For Ethernet No Action is needed
;----------------------------------------------------------------------------
PhysNodeInLSBForm:
	mov	ax, DriverConfigTable.MFrameID	; Get FrameID
	cmp	ax, MediaID8022			; ETHERNET_802.2?
	je	CheckEthernetAddr

	cmp	ax, MediaIDSNAP			; ETHERNET_SNAP?
	je	CheckEthernetAddr

	cmp	ax, MediaIDEthernetTyped	; ETHERNET_II?
	je	CheckEthernetAddr

	cmp	ax, MediaIDRaw8023		; ETHERNET_802.3?
	jne	DoNodeExit			; Not An Ethernet address

;----------------------------------------------------------------------------
;    Check the Address that the Local AdminBit is set and GroupBit not set.
;----------------------------------------------------------------------------
CheckEthernetAddr:
	test	byte ptr MSMPhysNodeAddress+0, 02h 	;LocalAdmin bit set?
	jz	AddressHasBitError

	test	byte ptr MSMPhysNodeAddress+0, 01h 	;MC bit set?
	jz	DoNodeExit

;----------------------------------------------------------------------------
;	 Some one left the MC Bit Set, clear it, and Inform the user.
;	or
;	 Some one left the LA Bit clear, set it, and Inform the user.
;----------------------------------------------------------------------------
AddressHasBitError:
	or	byte ptr MSMPhysNodeAddress+0, 02h	; Set Local AdminBit.
	and	byte ptr MSMPhysNodeAddress+0, 0FEh	; Clear MC-Bit for the user.
	mov	dx, offset CGroup: InvalidMediaAddress	; get error message
	call	MSMPrintStringZero
	mov	dx, offset CGroup: EthernetMsg		;get Media Name
	call	MSMPrintStringZero
	mov	dx, offset CGroup: InvalidMediaAddress2
	call	MSMPrintStringZero

;----------------------------------------------------------------------------
;	Finished Checking the Address.
;----------------------------------------------------------------------------
DoNodeExit:
	mov	si, offset CGroup:MSMPhysNodeAddress	; DS:SI -> Source
	lea	di, DriverConfigTable.MNodeAddress	; ES:DI -> Target
	mov	cx, 3					; 3 words in address.
 rep	movsw
	ret

MediaNodeAddressCheck	endp

Init	ends
	end

