;****************************************************************************
; ODI.INC
;
;  Copyright Unpublished Work of Novell, Inc. All Rights Reserved
;  
;  THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
;  PROPRIETARY AND TRADE SECRET INFORMATION OF NOVELL, INC.   
;  ACCESS TO THIS WORK IS RESTRICTED TO (i) NOVELL EMPLOYEES  
;  WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE  
;  OF THEIR ASSIGNMENTS AND (ii) ENTITIES OTHER THAN NOVELL   
;  WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.      
;  NO PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED,
;  DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED,      
;  CONDENSED, EXPANDED, COLLECTED, COMPILED, LINKED, RECAST,  
;  TRANSFORMED OR ADAPTED WITHOUT THE PRIOR WRITTEN CONSENT OF
;  NOVELL.  ANY USE OR EXPLOITATION OF THIS WORK WITHOUT AUTHORIZATION
;  COULD SUBJECT THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY.
;****************************************************************************
;   DOS VERSION
;
;	General definitions used by Protocol Stacks, LSL, and MLIDs.
;	
;	Written by:	DCD
;	Date:		7/31/89
;
;	ODI Specification Version: 4.00
;	
;    Modification History:
;	Date		By	Modification Description
;	---------------	---	------------------------------------------
;	16 Aug 1994	JWR	Removed Previous history, 
;				See DOS ODI Specifcation Changes Doc for changes.
;
;	9 Sept. 1994	JWR	Added Protocol Control Function Numbers 6-7.
;	
;	21 Feb. 1995	JWR	Added ECBCancelOK count to LSL Stat Table v1.1,
;				Added LSL_SUCCESSFUL, LSL_PENDING_SUCCESS, and PROM_SMT equate
;				Added StackID equate to migrate off ProtoNum naming schemes.
;				Corrected MaxECBSize Equate to 17954.
;
;*****************************************************************************

ifdef TASM
%NEWPAGE ;
endif
;----------------------------------------------------------------------------
;	Macro used for generating length preceded, 0 terminated strings.
;
;	Example:	Message	CustomString1	'TotalTxCount'
;
;----------------------------------------------------------------------------
Message	macro	name,stringg
	local	tail

name	db	(tail-2) - name
	db	stringg, 0
tail	equ	$
	
	endm

;----------------------------------------------------------------------------
;	Event Control Block Structure
;----------------------------------------------------------------------------
ECB struc
	NextLink	dw	0,0		;00h
	PrevLink	dd	0		;04h
	Status		dw	0		;08h
	ESR		dd	?		;0Ah
	StackID		dw	?		;0Eh
	ProtID		db	6 dup (?)	;10h
	BoardNum	dw	?		;16h
	ImmAddr		db	6 dup (?)	;18h
	DriverWS	db	4 dup (?)	;1Eh
	ProtocolWS	dw	4 dup (?)	;22h
	DataLen		dw	0		;2Ah
	FragCount	dw	1		;2Ch
	Frag1Addr	dd	?		;2Eh
	Frag1Len	dw	?		;32h
ECB	ends					;34h Size

ProtoNum	equ	StackID		; Setup offset 0Eh for usage equate.

FragStruc	struc
	FragAddr	dd	?
	FragLen		dw	?
FragStruc	ends

;----------------------------------------------------------------------------
;	AES ECB Structure
;----------------------------------------------------------------------------
AESECB	struc
	AESLink		dd	0
	MSecondValue	dd	?
	AESStatus	dw	0
	AESESR		dw	?, ?
AESECB	ends


ifdef TASM
%NEWPAGE ;
endif
;----------------------------------------------------------------------------
;	Look Ahead Structure
;----------------------------------------------------------------------------
LookAheadStruc	struc
	LMediaHeaderPtr	dw	2 dup (?)   ;00h Complete Low Level Media Header
	LookAheadPtr	dw	2 dup (?)   ;04h first 'n' bytes of the packet
	LookAheadLen	dw	?	    ;08h amount of data pointed to
	LProtID		db	6 dup (?)   ;0Ah Protocol ID value for this packet.
	LBoardNum	dw	0	    ;10h Logical Board Number (Default to 0, Handles Error packets)
	LDataSize	dw	?	    ;12h Packet data length (no headers)
	LImmAddress	db	6 dup (?)   ;14h Logical Node address of the Sending Node 
	LPacketAttrib	dw	0	    ;1Ah Same as PreviousLink Field in Rx ECB
	LDestType	dw	0	    ;1Ch Value to test against StkChnMask
	LStartCopyOffset dw	0	    ;1Eh Skip value from LookAheadPtr
	LPriorityLevel	db	0	    ;20h Priority Level (0-7, 0 => no priority)
					    ;    Valid only if RX_PRIORITY_FRAME bit set in LDestType.
	LRESERVED	db	3 dup(?)    ;21h Reserved
LookAheadStruc	ends			    ;24h Size	

;----------------------------------------------------------------------------
;	LPacketAttrib definitions
;	PreviousLink Field definitions (First word)
;----------------------------------------------------------------------------
PKT_CRC_ERR	  equ	0001h	; CRC Error, (FrameCheckSequence (FCS) Error).
PKT_ALIGN_ERR	  equ	0002h	; CRC/Frame Alignment Error
PKT_RUNT_ERR	  equ	0004h	; Runt Packet
		  ;    07F88h	; Reserved
PKT_BIG_ERR	  equ	0010h	; Packet Larger than allowed by Media
PKT_TYPE_ERR	  equ	0020h	; Packet for unsupported Frame Type
PKT_MALFORM_ERR	  equ	0040h	; Malformed Packet
IMM_ADR_MSB_FORM  equ	8000h	; Immediate Address is in MSB format. (Non-Cannonical form)

;----------------------------------------------------------------------------
;	LDestType Definitions
; 	   Rx ECB Destination Address Type (First word of ECB.DriverWS)
;	   Rx ECB 802.2 type field  (ECB.DriverWS + 1)
;----------------------------------------------------------------------------
DEST_MULTICAST		equ	0001h		; Multicast destination address
DEST_BROADCAST		equ	0002h		; Broadcast destination address
DEST_REMOTE_UNICAST	equ	0004h		; Remote Unicast Dest Addr.
DEST_REMOTE_MULTICAST	equ	0008h		; Remote Multicast Dest Addr.
DEST_SOURCE_ROUTE	equ	0010h		; RII Bit Set in Packet
DEST_ERRORED		equ	0020h		; Packet is errored
DEST_MAC_FRAME		equ	0040h		; Packet is a MAC Frame
DEST_DIRECT		equ	0080h		; Physical destination address
RX_NOT_8022		equ	0000h		; Non 802.2 packet.
RX_8022_TYPE1		equ	0100h		; 802.2 Type 1 packet
RX_8022_TYPE2		equ	0200h		; 802.2 Type 2 packet.
RX_PRIORITY_FRAME	equ	0400h		; Priority Level greater than normal.
						;    Level is stored in LPriorityLevel.
DEST_PROMISCUOUS	equ    0FFFFh		; All Packets for Promiscuous

ifdef TASM
%NEWPAGE ;
endif
;----------------------------------------------------------------------------
;	General definitions.
;----------------------------------------------------------------------------
DefaultNumECBs		equ	0		; By default the LSL doesn't allocate ECB's
MinNumECBs		equ	0		; LSL is not required to allocate ECB's
MaxNumECBs		equ	50		; Server Max / 40

DefaultECBDataSize	equ	1514		; 1388 data + 74 Protocol Headers + 52 MAC Headers
MinECBDataSize		equ	512+74+52	; 512 Min data + 74 Protocol Headers + 52 MAC Headers
MaxECBDataSize		equ	17954		; Max ECB Data size (Token Ring Largest Known is 17K)

DefaultNumMaxBoards	equ	4		; max is 16!
DefaultNumMaxStacks	equ	4		; max is 16!

MaxBoards		equ	16
MaxStacks		equ	16

MaxStackNameLength	equ	15		; Not including length & NULL in bytes

DefaultLookAheadLength	equ	18		; 
MaxLookAheadLength	equ	128		; Maximum Data length for LookAhead.

;----------------------------------------------------------------------------
;	Promiscous State Flag bit definitions
;----------------------------------------------------------------------------
PROM_OFF		equ	00h		; Promiscous mode off
PROM_MAC		equ	01h		; All MAC Frames
PROM_NON_MAC		equ	02h		; All non-MAC Frames
PROM_SMT		equ	04h		; All SMT type frames

;----------------------------------------------------------------------------
;	System Error Code Definitions
;----------------------------------------------------------------------------
LSL_SUCCESSFUL		equ	0000h
LSL_PENDING_SUCCESS	equ	0001h
LSLERR_OUT_OF_RESOURCES	equ	8001h
LSLERR_BAD_PARAMETER	equ	8002h
LSLERR_NO_MORE_ITEMS	equ	8003h
LSLERR_ITEM_NOT_PRESENT	equ	8004h
LSLERR_FAIL		equ	8005h
LSLERR_RX_OVERFLOW	equ	8006h
LSLERR_CANCELLED	equ	8007h
LSLERR_BAD_COMMAND	equ	8008h
LSLERR_DUPLICATE_ENTRY	equ	8009h
LSLERR_NO_SUCH_HANDLER	equ	800Ah
LSLERR_NO_SUCH_DRIVER	equ	800Bh
LSLERR_PACKET_ERRORED	equ	800Ch

ifdef TASM
%NEWPAGE ;
endif
;----------------------------------------------------------------------------
;	LSL Initialization Entry Function Codes
;----------------------------------------------------------------------------
LSLINIT_MLID_REG		equ	1	; Registers MLID, DS:DI -> MLIDRetInfoBlockStruc
LSLINIT_GET_PROTSUP_ENTRY	equ	2	; Get ES:SI -> LSL Entry point info Block
LSLINIT_GET_MLID_ENTRY		equ	3	; Get ES:SI -> MLID support Entry point.
LSLINIT_GET_ENTRY_POINTS	equ	4	; Get ES:SI -> EntryPoint Block (LSL v2.10 and above)

;----------------------------------------------------------------------------
;	For LSLINIT_GET_PROTSUP_ENTRY:
;	   LSL Entry point Info Block is a Buffer provided by the caller:
;			dd	Protocol Stack Support Entry Point
;			dd	General Service Entry Point	
;----------------------------------------------------------------------------
;----------------------------------------------------------------------------
;	For LSLINIT_GET_ENTRY_POINTS -- Returns ES:SI filled in with
;----------------------------------------------------------------------------
LSLInitEntryPointBlock	struc
	LSLProtSupEntryPt	dd	0	; Protocol Support Entry Point (MPI API)
	LSLGenSupEntryPt	dd	0	; General Services Entry Point 
	LSLMLIDSupEntryPt	dd	0	; MLID Support Entry Point (MLI API)
LSLInitEntryPointBlock	ends

;----------------------------------------------------------------------------
;	MLID Registration Blocks
;----------------------------------------------------------------------------
MLIDInfoBlockStruc	struc
	MIBS_SendEntry		dd	?	; MLID's Tx Handler
	MIBS_ControlEntry	dd	?	; MLID's Control Handler
	MIBS_ConfigTable	dd	?	; ptr to ConfigTable.
MLIDInfoBlockStruc ends

MLIDRetInfoBlockStruc struc
	MRIBS_MLID_SUP		dd	?	; LSL's MLID Support Entry point.
	MRIBS_BoardNum		dw	?	; LSL Assigned Board Num.
	MRIBS_ECB_DataSize	dw	?	; LSL's Configured ECB size.
MLIDRetInfoBlockStruc ends			

ifdef TASM
%NEWPAGE ;
endif
;----------------------------------------------------------------------------
;	LSL General Services Function Codes
;----------------------------------------------------------------------------
GENSERV_ALLOC_MEMORY		equ	0	
GENSERV_FREE_MEMORY		equ	1	
GENSERV_REALLOC_MEMORY		equ	2	
GENSERV_MEMORY_STATISTICS	equ	3	
GENSERV_ADD_MEMORY_TO_POOL	equ	4	
GENSERV_ADD_GENERAL_SERVICE	equ	5
GENSERV_REMOVE_GENERAL_SERVICE	equ	6
GENSERV_GET_NETCFG_PATH		equ	7
GENSERV_RESERVED		equ	8
GENSERV_GET_SERVICE_CHAIN	equ	9

;----------------------------------------------------------------------------
;	LSL General Services Description Record for 8000h-FFFFh services.
;----------------------------------------------------------------------------
GenServiceDescription	struc
	GenServName    	db	11 dup (0)	; ASCII-Z string
	GenServMonth   	db	?
	GenServDay     	db	?
	GenServYear    	db	?
	GenServMajorVer	db	?
	GenServMinorVer	db	?
GenServiceDescription	ends

;----------------------------------------------------------------------------
;	LSL General Services Control Block
;----------------------------------------------------------------------------
GenServiceControlBlock	struc
	GenServiceLink		dd	?	
	GenServiceAddress	dd	?
	GenServiceCommand	dw	?   ; 8000h-FFFFh available to Developers
GenServiceControlBlock	ends			

;----------------------------------------------------------------------------
;	LSL General Service Memory Status Block
;----------------------------------------------------------------------------
MemStatStruc	struc
	MemAvail	dw	0		; in paragraphs
	MemInUse	dw	0		; in paragraphs
	LargestAvailBlk	dw	0		; in paragraphs
	NumAvailBlocks	dw	0		
	MemOverhead	dw	0		; in bytes
	MinAllocation	dw	0		; in bytes
MemStatStruc	ends

ifdef TASM
%NEWPAGE ;
endif
;----------------------------------------------------------------------------
;	LSL Protocol Stack Services Function Codes
;----------------------------------------------------------------------------
PROTSUP_GET_ECB				equ	0	 
PROTSUP_RETURN_ECB			equ	1
PROTSUP_DEFRAG_ECB			equ	2
PROTSUP_SCHEDULE_AES_EVENT 		equ	3
PROTSUP_CANCEL_EVENT			equ	4
PROTSUP_GET_INTERVAL_MARK		equ	5
PROTSUP_REGISTER_STACK			equ	6  
PROTSUP_DEREGISTER_STACK		equ	7  
;----------------------------------------------------------------------------
;Reserved, please migrate off functions 8-11.
;PROTSUP_REGISTER_DEFAULT_STACK		equ	8  	; Use PROTSUP_REGISTER_DEFAULT_CHAIN instead.
;PROTSUP_DEREGISTER_DEFAULT_STACK	equ	9  	; Use PROTSUP_DEREGISTER_DEFAULT_CHAIN instead.
;PROTSUP_REGISTER_PRESCAN_STACK		equ	10 	; 0Ah ;Use PROTSUP_REGISTER_PRESCAN_RX_CHAIN
;PROTSUP_DEREGISTER_PRESCAN_STACK	equ	11 	; 0Bh ;Use PROTSUP_DEREGISTER_PRESCAN_RX_CHAIN
;----------------------------------------------------------------------------
;----------------------------------------------------------------------------
;	Inorder to prevent Loss of RPL functionality 8 & 9 have been renamed.
;----------------------------------------------------------------------------
PROTSUP_REGISTER_RPL_BOOTROM		equ	8	; RPL BOOT ROMS use these
PROTSUP_DEREGISTER_RPL_BOOTROM		equ	9	; to complete the boot up process.
	;RESERVED			equ	10	; 0Ah  ; DO NOT USE
	;RESERVED			equ	11	; 0Bh  ; DO NOT USE
PROTSUP_SEND_PACKET			equ	12 	; 0Ch
PROTSUP_HOLD_PACKET			equ	13 	; 0Dh
PROTSUP_GET_HELD_PACKET			equ	14 	; 0Eh
PROTSUP_SCAN_PACKET			equ	15 	; 0Fh
PROTSUP_GET_PROTNUM_FROM_NAME		equ	16 	; 10h StackIDFromName
PROTSUP_GET_PID_PROTNUM_MLIDNUM		equ	17 	; 11h PIDfromStackIDBoard
PROTSUP_GET_MLID_CTL_ENTRY		equ	18 	; 12h 
PROTSUP_GET_PROTO_CTL_ENTRY		equ	19 	; 13h 
PROTSUP_GET_LSL_STATS			equ	20 	; 14h
PROTSUP_BIND_STACK_TO_MLID		equ	21 	; 15h
PROTSUP_UNBIND_STACK_FROM_MLID		equ	22 	; 16h
PROTSUP_ADD_PID				equ	23 	; 17h
PROTSUP_RELINQUISH_CONTROL		equ	24	; 18h 
PROTSUP_GET_LSL_CONFIG			equ	25	; 19h
PROTSUP_GET_TICK_MARKER			equ	26	; 1Ah
PROTSUP_REGISTER_DEFAULT_CHAIN		equ	27	; 1Bh
PROTSUP_REGISTER_PRESCAN_RX_CHAIN	equ	28	; 1Ch
PROTSUP_REGISTER_PRESCAN_TX_CHAIN	equ	29	; 1Dh
PROTSUP_DEREGISTER_DEFAULT_CHAIN	equ	30	; 1Eh
PROTSUP_DEREGISTER_PRESCAN_RX_CHAIN	equ	31	; 1Fh
PROTSUP_DEREGISTER_PRESCAN_TX_CHAIN	equ	32	; 20h
PROTSUP_GET_START_CHAIN			equ	33	; 21h
PROTSUP_RESUBMIT_DEFAULT		equ	34	; 22h
PROTSUP_RESUBMIT_PRESCAN_RX		equ	35	; 23h
PROTSUP_RESUBMIT_PRESCAN_TX		equ	36	; 24h
PROTSUP_HOLD_EVENT	   		equ	37	; 25h
PROTSUP_START_CRITICAL_SECTION		equ	38	; 26h
PROTSUP_END_CRITICAL_SECTION   		equ	39	; 27h
PROTSUP_CRITICAL_SECTION_STATUS		equ	40	; 28h
PROTSUP_SERVICE_EVENTS	      		equ	41	; 29h
PROTSUP_GET_BOUND_BOARD_INFO		equ	42	; 2Ah
PROTSUP_MODIFY_STACK_FILTER		equ	43	; 2Bh
ifdef TASM
%NEWPAGE ;
endif
;----------------------------------------------------------------------------
;	LSL MLID Services Function Codes
;----------------------------------------------------------------------------
MLIDSUP_GET_ECB	       		equ	0
MLIDSUP_RETURN_ECB     		equ	1
MLIDSUP_DEFRAG_ECB	   	equ	2
MLIDSUP_SCHEDULE_AES_EVENT 	equ	3
MLIDSUP_CANCEL_AES_EVENT   	equ	4
MLIDSUP_GET_INTERVAL_MARKER	equ	5
MLIDSUP_DEREGISTER_MLID	   	equ	6  
MLIDSUP_HOLD_RECV_EVENT	   	equ	7  
MLIDSUP_START_CRITICAL_SECTION	equ	8  
MLIDSUP_END_CRITICAL_SECTION   	equ	9  
MLIDSUP_CRITICAL_SECTION_STATUS	equ	10 	 ; 0Ah
MLIDSUP_SERVICE_EVENTS	      	equ	11	 ; 0Bh
	;RESERVED	      	equ	12	 ; 0Ch	; DO NOT USE
	;RESERVED      		equ	13 	 ; 0Dh	; DO NOT USE
MLIDSUP_SEND_COMPLETE	      	equ	14 	 ; 0Eh
MLIDSUP_ADD_PID			equ	15	 ; 0Fh
MLIDSUP_GET_STACK_ECB		equ	16	 ; 10h
MLIDSUP_CONTROL_STACK_FILTER	equ	17	 ; 11h
						 
						 
;----------------------------------------------------------------------------
;	Protocol Control Commands
;----------------------------------------------------------------------------
GET_STACK_CONFIGURATION		equ	0
GET_STACK_STATISTICS		equ	1
BIND_TO_MLID			equ	2
UNBIND_FROM_MLID		equ	3
INFORM_MLID_DEREGISTERED	equ	4
INFORM_PROMISCUOUS_CHANGE	equ	5
	;Reserved	 	equ	6	; SFTIIIExchange
	;Reserved		equ	7	; GetProtocolStringForBoard
PROTOCOL_MANAGEMENT		equ	8	; New User Defined Management Entry Point.

;----------------------------------------------------------------------------
;	MLID Control Commands
;----------------------------------------------------------------------------
GET_MLID_CONFIGURATION		equ	0
GET_MLID_STATISTICS		equ	1
ADD_MULTICAST_ADDRESS		equ	2
DELETE_MULTICAST_ADDRESS	equ	3
REGISTER_TX_MONITOR		equ	4	
MLID_SHUTDOWN			equ	5
MLID_RESET			equ	6
	;RESERVED 		equ	7	; DO NOT USE
	;RESERVED		equ	8	; DO NOT USE
SET_LOOK_AHEAD_SIZE		equ	9
PROMISCUOUS_CHANGE		equ	10
	;RESERVED		equ	11	; DO NOT USE
DRIVER_POLL			equ	12
	;RESERVED		equ	13	; Custom IOCTL-13(IBM Token-Ring)
DRIVER_MANAGEMENT		equ	14	; New User Defined Management Entry point.
GET_MULTICAST_INFO		equ	15	; Returns pointer to Multicast tables.

ifdef TASM
%NEWPAGE ;
endif
;----------------------------------------------------------------------------
;	MLID Configuration Table Structure
;----------------------------------------------------------------------------
MLIDConfigurationStructure	struc
	MSignature		db	'HardwareDriverMLID',8 dup (' ')
	MConfigTableMajorVer	db	1
	MConfigTableMinorVer	db	13
	MNodeAddress		db	6 dup (?)
	MModeFlags		dw	0		; Unused Bits set to zero.
	MBoardNumber		dw	?
	MBoardInstance		dw	?
	MMaxPacketSize		dw	?
	MBestDataSize		dw	?
	MWorstDataSize		dw	?
	MCardLongName		dd	?
	MCardShortName		dd	?
	MFrameString		dd	?
	MReserved0		dw	0		; Must be set to 0
	MFrameID		dw	?
	MTransportTime		dw	1		; Nomimally set to 1
	MRouteHandler		dd	?		; For Source Routing Support 
	MLookAheadSize		dw	18		; Default Value
	MLineSpeed		dw	?		; In Mbps, if MSb set then Kbps
	MReserved1		db	6 dup (0)	; Must be set to 0
	MPrioritySup		db	0		; Number of Priorities supported (Max 8)
	MBusID			db	0		; Which Bus is Card on
	MMLIDMajorVer		db	?
	MMLIDMinorVer		db	?
	MFlags			dw	0		; Unused Bits set to zero
	MSendRetries		dw	?
	MLink			dd	0		; HSM Must not Modify
	MSharingFlags		dw	1		; Initially a Driver is ShutDown
	MSlot			dw	0		; Default: HSM should Scan for slot.
	MIOAddress1		dw	0		; if not used set to 0
	MIORange1		dw	0		
	MIOAddress2		dw 	0
	MIORange2		dw	0
	MMemoryAddress1		dd	0
	MMemorySize1		dw	0
	MMemoryAddress2		dd	0
	MMemorySize2		dw	0
	MIRQLine1		db	0FFh		; if not used set to 0FFh
	MIRQLine2		db	0FFh
	MDMALine1		db	0FFh
	MDMALine2		db	0FFh
MLIDConfigurationStructure	ends

ifdef TASM
%NEWPAGE ;
endif
;----------------------------------------------------------------------------
;	MLID Config Table 'MFlags' bit definitions.
;----------------------------------------------------------------------------
	;RESERVED	equ	0100h		; Hub-Management on server
FILTER_MULTICAST_BIT	equ	0200h		; TSM Filtering Not Needed
GROUP_ADDRESSING_BIT	equ	0400h		; Group Addressing Supported
NESL_REQUIRED_BIT	equ	0800h		; MLID Requires NESL 
PRIORITY_SUPPORT_BIT	equ	1000h		; Priority Support Enabled.
	;RESERVED	equ    0E0FFh		; DO NOT USE

;----------------------------------------------------------------------------
;	MLID Config Table 'MModeFlags' bit definitions.
;----------------------------------------------------------------------------
MRealDriverBit		equ	0001h	; Should always be set to 1
MUsesDMABit		equ	0002h	; Set if Physical Board uses DMA
	; DO NOT USE	equ	0004h	; RESERVED (Server DXFER Dependibility Bit)
MMulticastBit		equ	0008h	; Supports Multicasting 
MPointToPoint		equ	0010h	; (Server PointToPointBit)
MNeedsPollingBit	equ	0020h	; Set if MLID needs to be Polled
MRawSendBit		equ	0040h	; Supports Raw Sends
MLDataSizeBit		equ	0080h	; RESERVED SET to ONE (Backward Compatibility only)
	; DO NOT USE	equ	1F00h	; RESERVED Set to Zero	
MPromiscuousModeBit	equ	2000h	; Supports Promiscuous mode
MNonCanonicalBit	equ	4000h	; (Valid only if MSupportLSBMSB is set)
MSupportLSBMSB		equ	8000h	; MLID supports Octet Bit Reversal

;----------------------------------------------------------------------------
;	MLID Config Table 'MSharingFlags' bit definitions.
;----------------------------------------------------------------------------
MShutDownBit		equ	0001h
MCanShareIO1		equ	0002h
MCanShareIO2		equ	0004h
MCanShareMemory1	equ	0008h
MCanShareMemory2	equ	0010h
MCanShareIRQ1		equ	0020h
MCanShareIRQ2		equ	0040h
MCanShareDMA1		equ	0080h
MCanShareDMA2		equ	0100h
	;DO NOT USE 	equ    0FE00h	; Reserved

;----------------------------------------------------------------------------
;	MLID Config Table 'MBusID' Value definitions.
;----------------------------------------------------------------------------
BUS_ID_ISA		equ	00h
BUS_ID_MCA		equ	01h
BUS_ID_EISA		equ	02h
BUS_ID_PCMCIA		equ	03h
BUS_ID_PCI		equ	04h
BUS_ID_VESA		equ	05h
BUS_ID_HSM_DEFAULT	equ	0FFh
	
ifdef TASM
%NEWPAGE ;
endif
;----------------------------------------------------------------------------
;	MLID Statistics Table
;----------------------------------------------------------------------------
MLIDStatStructure	struc
	MStatTableMajorVer	db	01
	MStatTableMinorVer	db	01		; v1.01
	MNumGenericCounters	dw	14
	MValidCounterMask	dd	?		; 0 => used, 1 => NOT used.
	MTotalTxPackets		dw	2 dup (0)
	MTotalRxPackets		dw	2 dup (0)
	MNoECBsAvailable	dw	2 dup (0)
	MTxTooBig		dw	2 dup (0)
	MTxTooSmall		dw	2 dup (0)
	MRxOverflow		dw	2 dup (0)
	MRxTooBig		dw	2 dup (0)
	MRxTooSmall		dw	2 dup (0)
	MTxMiscError		dw	2 dup (0)
	MRxMiscError		dw	2 dup (0)
	MTxRetryCount		dw	2 dup (0)
	MRxChecksumError	dw	2 dup (0)
	MRxMismatchError	dw	2 dup (0)
	MQueueDepth		dw	2 dup (0)
	MNumCustomCounters	dw	0
MLIDStatStructure	ends

;----------------------------------------------------------------------------
;	Protocol Configuration Table
;----------------------------------------------------------------------------
ProtocolConfigStructure	struc
	PConfigTableMajorVer	db	1
	PConfigTableMinorVer	db	00	; v1.00
	PProtocolLongName	dw	?,?
	PProtocolShortName	dw	?,?
	PProtocolMajorVer	db	?
	PProtocolMinorVer	db	?
	PConfigTableReserved	db	16 dup (0)
ProtocolConfigStructure	ends

;----------------------------------------------------------------------------
;	Protocol Statistics Table
;----------------------------------------------------------------------------
ProtocolStatStructure	struc
	PStatTableMajorVer	db	1
	PStatTableMinorVer	db	00	; v1.00
	PNumGenericCounters	dw	3
	PValidCounterMask	dd	00011111111111111111111111111111b
	PTotalTxPackets		dw	2 dup (0)
	PTotalRxPackets		dw	2 dup (0)
	PIgnoredRxPackets	dw	2 dup (0)
	PNumCustomCounters	dw	0
ProtocolStatStructure	ends

ifdef TASM
%NEWPAGE ;
endif
;----------------------------------------------------------------------------
;	Stack Chain Definitions
;----------------------------------------------------------------------------
;	Positional notations based on directional flow
;
;	Default Transmit
; First Stack -> LSL TX 
;
;	Prescan Transmit
; LSL Tx -> 0 -> 1a -> 1b -> 2c -> 2b -> 2a -> 3b -> 3a -> 4 -> MLID (NIC)
;
;	Prescan Receive
; LSL Rx <- 4 <- 3a <- 3b <-2c <- 2b <- 2a <- 1b <- 1a <- 0 <- MLID (NIC)
;
;	Default Receive
; Last Stack <- 4 <- 3a <- 3b <-2c <- 2b <- 2a <- 1b <- 1a <- 0 <- LSL Rx
;
STACK_REQ_FIRST		equ	0	; The Stack must start the Chain.
STACK_REQ_NEXT_FIRST	equ	1	; Load at next available postion from front of chain
STACK_REQ_DEPENDS	equ	2	; Chain position dependent on load order
STACK_REQ_NEXT_LAST	equ	3	; Load at next available position from end of chain
STACK_REQ_LAST		equ	4	; The Stack must End the Chain.

;----------------------------------------------------------------------------
; Stack Chain Structure for Chained Protocol Stacks
;----------------------------------------------------------------------------
StackChainStruc		struc			;Offset
	StkChnLink	dd	0		;  0h  Link Field filled out by LSL to chain Stacks
	StkChnBoardNum	dw	-1		; 04h  Board Number, to register for
	StkChnPositReq	dw STACK_REQ_DEPENDS	; 06h  Chain Position requested
	StkChnHandler	dd	?		; 08h  Stack's Transmit or Receive Handler
	StkChnControl	dd	?		; 0Ch  Stack's Control Entry Point
	StkChnID	dw	0		; 10h  Stack's ID, filled by LSL
	StkChnMask	dw	0		; 12h  Stack's Packet Filter Mask
	StkChnReserved	dd	0		; 14h  Reserved
StackChainStruc	ends				; 18h  Size


ifdef TASM
%NEWPAGE ;
endif
;----------------------------------------------------------------------------
;	LSL Configuration Table
;----------------------------------------------------------------------------
LSLConfigurationStructure	struc
	LConfigTableMajorVer	db	1
	LConfigTableMinorVer	db	11	; v1.11
	LSLNumECB		dw	0	; LSL has no ECBs by default
	LSLConfigReserved0	dw	0
	LSLECBDataSize		dw	0	; Data Buffer Size
	LSLConfigReserved1	dw	0
	LMajorVersion		db	?	; Module Version
	LMinorVersion		db	?
	LMaxNumBoards		dw	DefaultNumMaxBoards
	LMaxNumStacks		dw	DefaultNumMaxStacks
	LMachineType		db	0	; 0=IBM 1=NEC 2=Fujitsu
	LConfigTableReserved	db	11 dup (0)
LSLConfigurationStructure	ends

;----------------------------------------------------------------------------
;	LSL Statistics Table
;----------------------------------------------------------------------------
LSLStatStructure	struc
	LStatTableMajorVer	db	01
	LStatTableMinorVer	db	01	; v1.01
	LNumGenericCounters	dw	10
	LValidCounterMask	dd	00000000001111111111111111111111b
	LTotalTxPackets		dw	2 dup (0)    ; Used	
	LGetECBRequests		dw	2 dup (0)    ; Used
	LGetECBFailures		dw	2 dup (0)    ; Used
	LAESEventCount		dw	2 dup (0)    ; Used
	LPostponedEvents	dw	2 dup (0)    ; Used
	LECBCancelFailures	dw	2 dup (0)    ; Used
	LBuffersReused		dw	2 dup (0)    ; Used
	LECBCancelOK		dw	2 dup (0)    ; Used
	LTotalRxPackets		dw	2 dup (0)    ; Used
	LUnclaimedPackets	dw	2 dup (0)    ; Used
	LNumCustomCounters	dw	0	     ; Currently none
LSLStatStructure	ends

;----------------------------------------------------------------------------
;	Machine Type values
;----------------------------------------------------------------------------
MACHINE_IBM_TYPE	equ	0
MACHINE_NEC_TYPE	equ	1
MACHINE_FMR_TYPE	equ	2



