%PAGESIZE 66,130
;*****************************************************************************
; NE2000.ASM
;	(c) Copyright Unpublished Work of Novell, Inc.  All Rights Reserved.
;
;       THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL, 
;       PROPRIETARY AND TRADESECRET INFORMATION OF NOVELL, INC.  ACCESS 
;       TO THIS WORK IS RESTRICTED TO (I) NOVELL EMPLOYEES WHO HAVE A
;       NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE OF THEIR ASSIGNMENTS 
;       AND (II) ENTITIES OTHER THAN NOVELL WHO HAVE ENTERED INTO APPROPRIATE 
;       LICENSE AGREEMENTS.  NO PART OF THIS WORK MAY BE USED, PRACTICED,
;       PERFORMED, COPIED, DISTRIBUTED, REVISED, MODIFIED, TRANSLATED,
;       ABBRIDGED, CONDENSED, EXPANDED, COLLECTED, COMPILED, LINKED, RECAST,
;       TRANSFORMED OR ADAPTED WITHOUT THE PRIOR WRITTEN CONSENT OF NOVELL. 
;       ANY USE OR EXPLOITATION OF THIS WORK WITHOUT AUTHORIZATION COULD 
;       SUBJECT THE PERPERTRATOR TO CRIMINAL AND CIVIL LIABILITY.
;
;------------------------------------------------------------------------------
;
;       Novell NE2000 MLID for DOS ODI.
;
;       Written by:     DCD
;       Date:           8/13/90
;
;       ODI Specification Version:      4.00
;
; Modification History:
;       By      Date        Ver.           Modification Description
;       ------ ---------   -----   --------------------------------------
;       JWR,WTT 01 August 94 2.00    Removed Old History, 
;                                  Remade to confirm to ODI SPEC 4.00
;
;       JWR,WTT 22 Sept. 94 2.01   Re-linked to include latest tool kit.
;                               
;       JWR,WTT,AD 10 Oct. 94 2.02   Re-linked to include latest tool kit.
;                                    Added one byte to LookAheadBuf.  
;                                    Buffer was one byte too small.
;
;       JWR     7 Dec. 94 2.03 1    Update Copyright date, New MSM Keep Pad
;                                    When in Promiscous Mode.
;
;       JWR     9 Jan. 95 2.03       Re-Link Latest ToolKit.
;
;       TNL     8 Feb. 95          Added "slow" code to DoRAMTest write/read 
;                                  loops to accomodate Pentium machines which
;                                  occasionally report RAM Failures.
;
;       EDR     10 Mar. 95              Added code to assert page 0 before 
;                                       any io to card at the beginning of 
;                                       DriverInit. SPD 93236
;
;       EDR     14 Mar. 95              Added hardware configuration conflict
;                                       checks. (As implemented in OS driver)
;
;       JWR     5 April 95 2.04     Re-Link Latest ToolKit.
;
;	JWR	3 May 95   2.05		Hardware conflict Check code leaves the
;					Hardware in Promiscous Mode. Added Call
;					to Driver Reset to clear Mode.
;
;       JWR     7 Sept. 95 2.06     Re-Link Latest ToolKit.
;
;       JWR     25 Oct. 95 2.07     Re-Link Latest ToolKit.
;
;	JWR	4 Jan.  96 2.07	    Update CopyRight. 
;
;	JWR	9 Feb.  96 2.09	    Re-Link Latest ToolKit.
;
;	JWR	1 April 96 2.10	    Re-Link Latest ToolKit.
;
;	SMN 21 Nov. 97 2.12		Re-Link Latest ToolKit. v2.12, Update Copyright date
;
;*****************************************************************************
include driver.inc
.8086

Init    segment para public 'CODE'
;----------------------------------------------------------------------------
;       Version Control Center
;----------------------------------------------------------------------------
MajorVersion    equ     02
MinorVersion    equ     12              ; Last Release November 21,1997 (v2.12)
;AlphaVersion   equ     1               ; Comment in and out as needed for
;BetaVersion    equ     0               ; Prerelease (Set November 21, 1997)

;----------------------------------------------------------------------------
;       Signon Message
;----------------------------------------------------------------------------
HSMSPEC                 db      'HSM_SPEC_VERSION: 4.00', 0
			db      'VeRsIoN='
DriverSignOnMessage     db      'Novell NE2000 Ethernet MLID  v'
			db      '0' or (MajorVersion MOD 10),'.'
			db      '0' or (MinorVersion/10)
			db      '0' or (MinorVersion MOD 10)
ifdef AlphaVersion      
		db      ' Alpha ',((AlphaVersion/10) OR '0')
		db      ((AlphaVersion MOD 10) OR '0') ; Only for Alpha builds.
endif

ifdef BetaVersion       
		db      ' BETA ',(BetaVersion + '0')   ; Only for Beta Builds.
endif
ifdef DEBUG
		db      ' *DEBUG* '
endif
			db      ' (971121) ',0dh,0ah
			db      '(c) Copyright 1991 - 1997, by Novell, Inc. '
			db      'All rights reserved.',0

DriverMainSectionText   db      'LINK DRIVER NE2000',0

Init    ends

	
Code    segment para public 'CODE'
;----------------------------------------------------------------------------
;       DEBUG Publics
;----------------------------------------------------------------------------
ifdef DEBUG
	public  RunTimeCode0, RunTimeCode1, RunTimeCode2, RunTimeCode3
	public  RunTimeCode4
	public  OddByteToDo, DoLastOddByte, DontStartThisOne, DMAFinished
	public  RawSend, FragmentOutLoop, DoTheCurrentFrag, AdvanceToNextFrag
	public  LookAheadBuf, SharedRAMFlag, Control1Value
	public  AsyncECB, RAMSegment
endif

SendRetryCount          equ     10

;----------------------------------------------------------------------------
;       Miscellaneous Equates
;----------------------------------------------------------------------------
INT_MASK                equ     38h

;----------------------------------------------------------------------------
;       NE2000 memory organization I/O mode.
;----------------------------------------------------------------------------
TXPAGE                  equ     40h             ;Tx buffer at 4000h
TXPAGE1                 equ     46h             ;Tx buffer at 4600h

;----------------------------------------------------------------------------
;       8390 Registers
;----------------------------------------------------------------------------
PhysicalReg0            equ     NICPageStart
Current                 equ     NICInterruptStatus
MulticastRegister0      equ     8

;----------------------------------------------------------------------------
;       8390 Commands
;----------------------------------------------------------------------------
NICTransmit             equ     26h
NICUnmaskByte           equ     1fh
NICPage0                equ     22h
NICPage1                equ     62h
NICPage0Stop            equ     21h
NICPage0DMAWrite        equ     12h
NICPage2DMAWrite        equ     92h
NICPage0DMARead         equ     0ah

;----------------------------------------------------------------------------
;       8390 Tx Status bits
;----------------------------------------------------------------------------
NIC_TxdPkt              equ     01h             ;Transmitted Packet
NIC_TxCollid            equ     02h             ;Transmit Collided
NIC_TxAborted           equ     04h             ;Transmit Aborted, Excess Collisions
NIC_CarrSenseLost       equ     08h             ;Carrier Sense Lost
NIC_TxFIFOUnderRun      equ     10h             ;FIFO UnderRun
NIC_TXLateColl          equ     80h             ;Transmit Late Collision

;----------------------------------------------------------------------------
;       8390 Rx Status bits
;----------------------------------------------------------------------------
NIC_RxPktIntact         equ     01h             ;Received Packet intact
NIC_RxCRCError          equ     02h             ;CRC Error
NIC_RxFrameAlignError   equ     04h             ;Frame Alignment Error (inc. CRC Error)
NIC_RxRxFIFOUnderRun    equ     08h             ;Rx FIFO UnderRun
NIC_RxMissedPkt         equ     10h             ;Missed Packet
NIC_RxDisabled          equ     40h             ;Receiver Disabled
NIC_RxDeffering         equ     80h             ;Receive Defferring

;----------------------------------------------------------------------------
;       Receive header.
;----------------------------------------------------------------------------
RxHeaderStructure       struc
	RxStatus                db      ?
	NextPacket              db      ?
	RxCount                 dw      ?
RxHeaderStructure       ends

;----------------------------------------------------------------------------
;       NIC port definitions.
;----------------------------------------------------------------------------
NICPortsStructure       struc
	NICCommand              db      ?
	NICPageStart            db      ?
	NICPageStop             db      ?
	NICBoundary             db      ?
	NICTxStatus             db      ?
	NICTxByteCountLow       db      ?
	NICTxByteCountHigh      db      ?
	NICInterruptStatus      db      ?
	NICStartAddressLow      db      ?
	NICStartAddressHigh     db      ?
	NICRemoteByteCountLow   db      ?
	NICRemoteByteCountHigh  db      ?
	NICRxConfiguration      db      ?
	NICTxConfiguration      db      ?
	NICDataConfiguration    db      ?
	NICInterruptMask        db      ?
	NICDataPort             dw      ?
	NICNotUsed              db      13 dup (?)
	NICReset                db      ?
NICPortsStructure       ends

NICSharedRAM    struc
	IControl1                       db      ?               ; 00h
	IATDetect                       db      ?               ; 01h
					db      3 dup (?)       ; 02h
	IControl2                       db      ?               ; 05h
					db      2 dup (?)       ; 06h
	IProm                           db      8 dup (?)       ; 08h
NICSharedRAM    ends

Message NICLongName     'Novell NE2000 Ethernet'
Message NICShortName    'NE2000'

;----------------------------------------------------------------------------
;       AES ECB used for polling.
;----------------------------------------------------------------------------
	even
AsyncECB                AESECB  <,1000,,offset CGroup:DriverTimeOut>

;----------------------------------------------------------------------------
;       Control variables.
;----------------------------------------------------------------------------
SplitFlag               db      0
TxFlagValue             db      0               ;Use for timeout detection
NextPageWord            db      0               ;Must always be 0
NextPage                db      0               ;Next buffer page, Must be after
						; NextPageWord and is initialized
						; during DriverInit.
CurrentTxPage           db      0
ResetTxConfigFlag       db      0
RetryTxFlag             db      0
EightBitSlot            db      0
MCAddress               db      6 dup (0)
PSTART                  db      4ch             ;Start at 4c00H for I\O Mode
						; and c000H for RAM Mode.
PSTOP                   db      80H             ;Stop at 8000H for I\O Mode
						; and 4000h for 16 bit RAM Mode.

RAMRingSplitStop        dw      0
RAMRingSplitStart       dw      0
RAMSegment              dw      0

TransmitPage0           db      0c0h    ; 1st I/O Transmit Page.
TransmitPage1           db      0c6h    ; 2nd I/O Transmit Page.

Control1                dw      0       ; Shared RAM
ATDetect                dw      0       ; Shared RAM
Control2                dw      0       ; Shared RAM
Prom                    dw      0       ; Shared RAM

RxTestValue             db      4Eh             ;Rx test value for NIC, 8390

	even
TxBufUsageFlags         dw      0
TxRetryCounter          dw      ?

TxPacketSizeTable       dw      0
			dw      0

RAMModeTxBufToUseTable  dw      0
			dw      0600h
			dw      0

TxBufToUseTable         db      TXPAGE
			db      TXPAGE1
			db      TXPAGE

TxBufBitTable           db      01h
			db      02h
			db      01h

NextTxBufTable          db      TXPAGE1
			db      TXPAGE

ClearTxBitTable         db      0feh
			db      0fdh

	ALIGN   4
LookAheadBuf            db      4+14+3+5+1+128 dup (?) ;Max look ahead data=128 bytes

SharedRAMFlag           db      0
Control1Value           db      0               ; For shared RAM.
Control2Value           db      0               ; For shared RAM.

;----------------------------------------------------------------------------
;       I/O Port Variables
;----------------------------------------------------------------------------
	even
Command                 dw      NICCommand
Boundary                dw      NICBoundary
TxStatus                dw      NICTxStatus
TxByteCountLow          dw      NICTxByteCountLow
InterruptStatus         dw      NICInterruptStatus
InterruptMask           dw      NICInterruptMask
RemoteByteCountLow      dw      NICRemoteByteCountLow
StartAddressLow         dw      NICStartAddressLow
DataPort                dw      NICDataPort
DataConfiguration       dw      NICDataConfiguration
TransmitPage            dw      NICTxStatus

NumIOPorts              equ     ($-Command)/2

;******************************************************************************
;             D R I V E R   S T A T I S T I C S   T A B L E
;******************************************************************************

	even
DriverStatTable         label   byte
DriverStatMajorVer      db      01
DriverStatMinorVer      db      01              ;v1.01
DriverStatGenericCnts   dw      14
DriverStatValidCntsMask dd      11111011000011111111111111111111b
DriverStatTotalTx       dw      2 dup (0)
DriverStatTotalRx       dw      2 dup (0)
DriverStatNoECB         dw      2 dup (0)
DriverStatTxTooBig      dw      2 dup (0)               ;Not used
DriverStatTxTooSmall    dw      2 dup (0)               ;Not used
DriverStatRxOverflow    dw      2 dup (0)
DriverStatRxTooBig      dw      2 dup (0)
DriverStatRxTooSmall    dw      2 dup (0)               ;Not used
DriverStatTxMisc        dw      2 dup (0)
DriverStatRxMisc        dw      2 dup (0)
DriverStatTxRetryCount  dw      2 dup (0)
DriverStatRxChecksum    dw      2 dup (0)
DriverStatRxMismatch    dw      2 dup (0)
DriverQueueDepth        dw      2 dup (0)

DriverStatNumCustomVars dw      4

TxAbortTimeOut          dw      2 dup (0)
DMAWouldNotComplete     dw      2 dup (0)
ReinitializeNIC         dw      2 dup (0)
RxBufferError           dw      2 dup (0)

			dw      offset CGroup:DS1
			dw      0
			dw      offset CGroup:DS2
			dw      0
			dw      offset CGroup:DS3
			dw      0
			dw      offset CGroup:DS4
			dw      0

;       Statistics Custom Variable counter strings

Message DS1             'TxAbortTimeOut'
Message DS2             'DMAWouldNotComplete'
Message DS3             'ReinitializeNIC'
Message DS4             'RxBufferError'


;******************************************************************************
;            D R I V E R   C O N F I G U R A T I O N   T A B L E
;******************************************************************************
even
DriverConfigTable       label   byte
Signature               db      'HardwareDriverMLID        '
CfgMajorVersion         db      1
CfgMinorVersion         db      13
NodeAddress             db      6 dup (0FFh)            
ModeFlags               dw      0A009h          ;OBR,Real,Multicast,Promiscuous
BoardNumber             dw      ?
BoardInstance           dw      ?
MaxDataSize             dw      ?
BestRcvSize             dw      ?
WrstRcvSize             dw      ?
LANCardName             dw      offset CGroup:NICLongName,0
LANShortName            dw      offset CGroup:NICShortName,0
LANMediaName            dd      ?
Reserved0               dw      0
LANMediaID              dw      ?
TransportTime           dw      1
RouteHandler            dd      0               ;Not used by Ethernet
LookAheadSize           dw      18              ;Just enough for IPX
LineSpeed               dw      10              ;10 Mbps
Reserved1               db      6 dup (0)
PrioritySup             db      0
BusID                   db      BUS_ID_HSM_DEFAULT      
MLIDMajorVersion        db      MajorVersion
MLIDMinorVersion        db      MinorVersion
OSFlags                 dw      0
SendRetries             dw      SendRetryCount  ;Default retry value
Link                    dd      0
ShareFlags              dw      1               ;Default to Shutdown
PS2Slot                 dw      0               
IOAddress1              dw      300h
IORange1                dw      20h
IOAddress2              dw      0
IORange2                dw      0
MemAddress1             dd      0
MemRange1               dw      0
MemAddress2             dw      2 dup (0)
MemRange2               dw      0
IRQLine1                db      3
IRQLine2                db      -1
DMALine1                db      -1
DMALine2                db      -1

;*******************************************************************************
;       DriverManagement          ---Not used .... Stubbed here for the MSM
;       DriverChangeLookAheadSize ---Not used .... Stubbed here for the MSM
;       DriverPoll                ---Not used .... Stubbed here for the MSM
;*******************************************************************************

DriverManagement:
DriverChangeLookAheadSize:
DriverPoll:
	ret

;******************************************************************************
;                    D R I V E R   R O U T I N E S
;******************************************************************************
;  DriverMulticastChange
;
;       Called by the envelope when changes have occurred to our
;       multicast address list.
;
;       assumes:        ds      is set to CGroup
;                       si      -> Multicast Table
;                       cx      number of entries in table
;                       CLD has been executed
;                       Interrupts are disabled
;
;       returns:        Nothing need be preserved
;****************************************************************************
DriverMulticastChange   proc    near
;----------------------------------------------------------------------------
;       Select page 1 registers.
;----------------------------------------------------------------------------
	mov     bp, IOAddress1

	cmp     cs:SharedRAMFlag, 0             ; Shared RAM Mode?
	je      DriverMulticastChangeIOMode

	add     bp, 10h

DriverMulticastChangeIOMode:
	mov     dx, bp                          ;DX = command register
	mov     al, NICPage1
	out     dx, al

	lea     dx, [bp].MulticastRegister0     ;Get port for MC registers
	mov     bx, 8                           ;There are 8 registers to setup
	xor     al, al                          ;Set all bits to 0.

DisableAllBitsLoop:
	out     dx, al
	inc     dx
	dec     bx
	jnz     DisableAllBitsLoop

;       Set page -> back to page 0 regs.

	mov     dx, bp
	mov     al, NICPage0
	out     dx, al

ScanMCTableForAddressLoop:
	cmp     word ptr [si+6], 0              ;Is entry used?
	je      MCEntryNotUsed

	call    SetupMCHashBit                  ;Setup the address on the NIC

MCEntryNotUsed:
	add     si, 8                           ;Goto next entry
	loop    ScanMCTableForAddressLoop
	ret                                     ;All done

DriverMulticastChange   endp

;----------------------------------------------------------------------------
;       SetupMCHashBit
;
;       This routine will take the specified 6-byte address and calculate
;       which bit should be set in the 8390 multicast hash table to enable
;       reception of that address.  It will then set the appropriate bit
;       in the 8390 hash table.
;
;       assumes:        ds      is set to CGroup.
;                       si      -> buffer holding the 6-byte address to use.
;                       Interrupts unspecified.
;
;       returns:        ds,si,bx,cx preserved.
;                       Interrupt start preserved.
;
;----------------------------------------------------------------------------
SetupMCHashBit  proc    near

	push    si
	push    cx

;----------------------------------------------------------------------------
;       First select the page 1 registers.
;----------------------------------------------------------------------------
	mov     dx, Command                     ;Get command reg port
	mov     al, NICPage1
	out     dx, al

	push    cs
	pop     es

	mov     di, offset CGroup:MCAddress
	movsw
	movsw
	movsw

	mov     bp, 0FFFFh                      ;BP = crcl
	mov     di, bp                          ;DI = crch

;----------------------------------------------------------------------------
;       for (index=0; index < 6; ++index)
;----------------------------------------------------------------------------
	mov     cx, 6                           ;Do 6 bytes of address
	mov     si, offset CGroup:MCAddress

FilterOuterLoop:
	push    cx                              ;Save outer loop counter

;       for (j=0; j<8; ++j)

	mov     cl, 8                           ;Set innerloop count

;       carry = ((crc&0x80000000)?1:0)^(multaddress[i]&01)

FilterInnerLoop:
	mov     ax, 8000h
	and     ax, di

	rcl     ax, 1
	rcl     ax, 1

	mov     dx, 1
	and     dx, [si]                        ;multiaddress[i]&01

	xor     ax, dx

;       crc <<= 1

	shl     bp, 1
	rcl     di, 1

;       multaddress[i] >>= 1

	shr     byte ptr [si], 1

;       if (carry)
;        crc = ((crc^PLYNOMIAL) | carry)

	or      ax, ax                          ;AX has carry in it
	jz      FinishInnerFilterPass
	cwd                                     ;DX:AX = sign extended car y value

;       crc^POLYNOMIAL

	xor     bp, 1DB6h                       ;High end of polynomial.
	xor     di, 04C1h                       ;Low end of polynomial.

;       crc^Polynomial | carry
	
	or      bp, ax                          ;Low end of carry
	or      di, dx                          ;High end

FinishInnerFilterPass:
	loop    FilterInnerLoop

	inc     si                              ;-> next byte of address
	pop     cx                              ;Restore outer counter
	loop    FilterOuterLoop

;       crc = crc & 0x0ff00 0000h

	and     di, 0ff00h

;       crc = crc >> 26

	mov     cl, 26
	xor     ax, ax
	mov     dx, di

MultiShiftForFilter:
	shr     dx, 1
	rcr     ax, 1
	loop    MultiShiftForFilter

;----------------------------------------------------------------------------
;       AX has the filter bit index.
;       Calculate the register and bit offset for this bit position.
;----------------------------------------------------------------------------
	mov     dx, ax
	mov     cx, ax
	and     cl, 07h                         ;Mask off all except last 3 bits
	mov     al, 1
	shl     al, cl                          ;Setup variable

	shr     dx, 1                           ;Figure register to use (/8)
	shr     dx, 1
	shr     dx, 1
	add     dx, Command                     ;Figure real I/O port
	add     dx, MulticastRegister0

;----------------------------------------------------------------------------
;       Write the value to the NIC.
;----------------------------------------------------------------------------
	mov     ah, al                          ;Save AL
	in      al, dx                          ;Get current bit setting

	or      al, ah                          ;Mask in our bit
	out     dx, al

;----------------------------------------------------------------------------
;       Set the page -> back to page 0 regs.
;----------------------------------------------------------------------------
	mov     dx, Command                     ;Get command register
	mov     al, NICPage0
	out     dx, al

	pop     cx
	pop     si
	ret

SetupMCHashBit  endp


;****************************************************************************
;       DriverShutdown
;
;       Called by the MSM when it wants us to shut our board down.
;
;       assumes:        ds      is set to CGroup
;                       cx      has the shutdown flag as doc. in ODI spec.
;                               =0 if perm shutdown
;                               !0 if just temp shutdown
;                       CLD has been executed
;                       Interrupts are disabled
;
;       returns:        Nothing need be preserved
;
;****************************************************************************
DriverShutdown  proc    near

	mov     dx, InterruptMask               ;Disable NIC's interrupt
	xor     al, al
	out     dx, al

	mov     dx, Command                     ;Stop the NIC
	mov     al, 21h
	out     dx, al

;       If this is a perm. shutdown we need to stop our AES event.

	or      cx, cx
	jnz     TempShutDownOnly

	push    cs                              ;Cancel our AES poll
	pop     es
	mov     si, offset CGroup:AsyncECB
	mov     bx, MLIDSUP_CANCEL_AES_EVENT
	call    LSLSupport

TempShutDownOnly:
	ret

DriverShutdown  endp


;****************************************************************************
;       DriverReset
;
;       Called by the MSM when it wants us to reset our board.
;
;       assumes:        ds      is set to CGroup
;                       CLD has been executed
;                       Interrupts are disabled
;
;       returns:        Nothing need be preserved
;
;****************************************************************************
DriverReset     proc    near

	jmp     ResetNICRunTime

DriverReset     endp

;****************************************************************************
;       DriverPromiscuousChange
;
;       Called by the MSM when it wants us to set/reset Promsicuous Mode
;
;       assumes:        
;               CX      Zero            Disable Promiscuous Mode
;                       Non-Zero        Enable Promiscuous Mode
;               DS      is set to CGroup
;                       CLD has been executed
;                       Interrupts are disabled
;
;       returns:
;               DS, SS, SP preserved
;
;****************************************************************************
DriverPromiscuousChange proc    near
	mov     bp, IOAddress1                  ; Get base I/O port
	cmp     SharedRAMFlag, 0                ; Shared RAM Mode?
	je      DriverPromiscuousChangeIOMode

	add     bp, 10h

DriverPromiscuousChangeIOMode:
	lea     dx, [bp].NICRxConfiguration

; enable Unicast, Broadcasts & Multicasts only 
	mov     RxTestValue, 4Eh                ;normal Rx mode

	mov     al, 0Ch                         ;assume Disable
	jcxz    DrvResetPromisMode              ;CX - Zero, disable Prom. Mode


;enable Unicast, Broadcasts, Multicasts, Errors, Runt and Promiscuous Physical
	mov     RxTestValue, 48h                ;allow CRC & CRC/Align Errors
	mov     al, 1Fh                         
	mov     cl, -1                          ;Used to set AL all 1's later

DrvResetPromisMode:
	out     dx, al          

DrvPromNICOutput:
	mov     dx, Command                     ;Get command reg port
	mov     al, NICPage1                    ;Select the page 1 registers.
	out     dx, al

	mov     al, cl                          ;get value to output
	add     dx, MulticastRegister0          ;Figure real I/O port
	mov     cx, 08h                         ;get # of Multicast Regs

;Write the Mutlicast bits to the NIC.

DrvPromNICOutputGrp:
	out     dx, al                          ;set bits
	add     dx, 01h                         ;get next Register
	loop    DrvPromNICOutputGrp             ;loop till all bits set

;Set the page -> back to page 0 regs.

	mov     dx, Command                     ;Get command register
	mov     al, NICPage0
	out     dx, al
	ret

DriverPromiscuousChange endp


;****************************************************************************
;       DriverISR
;
;       This routine is called when the NIC interrupts the system.
;
;       assumes:        ds      CGroup
;                       CLD has been executed
;                       Inside a critical section
;                       Interrupts are disabled
;                       All registers have been preserved
;
;       returns:        ds      CGroup
;                       All other registers may be destroyed
;                       Interrupts disabled
;
;****************************************************************************
DoSlave:
	out     0A0h, al                        ;Issue EOI to the 2nd PIC
	jmp short BackFromEOI

public DriverISR

DriverISR       proc    near

	mov     dx, InterruptMask               ;Disable NIC's interrupt
	xor     al, al
	out     dx, al

	mov     al, 20h                         ;Issue EOI to PIC(s)
	cmp     MSMEOIFlag, 0                   ;EOI needed on slave PIC?
	jne     DoSlave

BackFromEOI:
	out     20h, al                         ;Issue EOI to the 1st PIC

NextInterruptService:
	mov     dx, InterruptStatus             ;Get the status bits
	in      al, dx
	mov     ah, al                          ;Save status for later

	sti                                     ;Must be here

	test    al, 15h                         ;Packet received (with/without errors)?
	jnz     PacketReceived

	test    al, 0ah                         ;Any Tx interrupt bits
	jnz     TxInterruptJmp

ISRExit:
	cli
	mov     dx, InterruptMask               ;Enable NIC's interrupt
	mov     al, NICUnmaskByte
	out     dx, al
	ret


;
;       NEAR TO FAR JUMPS
;

TxInterruptJmp:
	jmp     TxInterrupt

DequeueTxJmp:
	jmp     CheckForQueuedTx

HandleBufferOverflowCaseJmp:
	jmp     HandleBufferOverflowCase

ExitReceptionHandlerJmp:
	jmp     ExitReceptionHandler

UpdateRcvErrorCounter:
	cmp     RxTestValue, 48h                ;Keep stats if receive monitor
	je      CheckIfPacketInBuffer           ; not active (this stat will
						; be handled later when rcv
						; monitor is active)

	inc     DriverStatRxChecksum
	jmp short CheckIfPacketInBuffer

;----------------------------------------------------------------------------
;       PACKET RECEPTION HANDLER
;----------------------------------------------------------------------------
PacketReceivedTop:
	mov     dx, InterruptStatus             ;Read and save receive status
	in      al, dx
	mov     ah, al

	cmp     ResetTxConfigFlag, 1            ;See if already handling buf
	je      DontCheckOverflowBit            ;  overflow

PacketReceived:
	test    al, 10h                         ;Ring buffer overflow?

DontCheckOverflowBit:
	mov     al, 05h                         ;Reset the Rx bits
	out     dx, al
	jnz     HandleBufferOverflowCaseJmp

	test    ah, 04h                         ;Check for receive errors
	jnz     UpdateRcvErrorCounter

CheckIfPacketInBuffer:
	mov     dx, Command                     ;Switch to Page1 registers
	mov     al, NICPage1
	out     dx, al
  
	mov     dx, InterruptStatus             ;Read 'Current' register
	in      al, dx
	cmp     al, NextPage                    ;Packet waiting?
  
	mov     dx, Command                     ;DX gets Command register port
	mov     al, NICPage0
	out     dx, al

	je      ExitReceptionHandlerJmp
	
;----------------------------------------------------------------------------
;       Read in the look ahead data.
;----------------------------------------------------------------------------
	cmp     cs:SharedRAMFlag, 0             ; Shared RAM Mode?
	je      DriverISRIOMode
	jmp     DriverISRSharedRAM

PacketTooLong:
	dec     al
	mov     cl, 3
	shr     al, cl                          ; AL = (AL - 1) / 8.
	mov     cl, 11
	shl     ax, cl                          ; AX = AX * 2048.
	add     bp, ax                          ; Add to total count.
	add     bx, ax
	jmp short PacketTooLongReturn

DriverISRIOMode:
	mov     dx, StartAddressLow             ;-> at the packet
	xor     al, al
	out     dx, al
	inc     dx
	mov     al, NextPage
	out     dx, al

;----------------------------------------------------------------------------
;       Get the Length needed for the Current LookAhead Size
;----------------------------------------------------------------------------
	mov     ax, LookAheadSize       ; Current Lookahead Size (MAX = 128)
	add     al, 4+14+3+5+1          ; allow for Worst case Headers and evenize
	and     al, 0FEh                ; evenize length
	mov     cx, ax

;----------------------------------------------------------------------------
;       Tell the NIC the Low Byte Count
;----------------------------------------------------------------------------
	inc     dx                      ; DX=RemoteByteCountLow
	out     dx, al

	mov     al, ah
	inc     dx                      ; DX=RemoteByteCountHigh
	out     dx, al
  
;----------------------------------------------------------------------------
;       Tell the NIC we want to read.
;----------------------------------------------------------------------------
	mov     dx, Command
	mov     al, NICPage0DMARead             ;Tell NIC we want to read
	out     dx, al

;----------------------------------------------------------------------------
;       Set up to Read
;----------------------------------------------------------------------------
	mov     dx, DataPort                    ; Where we read the packet data from
	mov     ax, ds
	mov     es, ax
	mov     di, offset CGroup:LookAheadBuf  ; ES:DI -> LookaheadBuffer (incl. Buffer mang.)
	lea     si, [di+4]                      ; ES:SI -> look ahead frame data

;----------------------------------------------------------------------------
;       Read the DATA from the NIC
;----------------------------------------------------------------------------
RunTimeCode2:
;;;-> RunTimeModifyCodeBegin
ifdef SHORTSLOT           ; (@Runtime determine correct code, Default 16-bit)
;----------------------------------------------------------------------------
;       Do byte Transfers 8-bit Slot
;----------------------------------------------------------------------------
	in      al, dx                  ; 8088 code (EC)
	stosb                           ; 8088 code (AA)
	loop    RunTimeCode2            ; 8088 code (E2 FC)

else
;----------------------------------------------------------------------------
;       Do word transfer 16-bit Slot
;----------------------------------------------------------------------------
	shr     cx, 1                   ; 386 code (D1 E9) 
.286                                    ; 386 code ()
	rep     insw                    ; 386 code (F3> 6D)
.8086                                   ; 386 code ()
endif
;;;-> RunTimeModifyCodeEnd

	test    LookAheadBuf.RxStatus, NIC_RxCRCError + NIC_RxFrameAlignError
	jz      TestRxConfiguration

	inc     DriverStatRxChecksum            ;at least this

TestRxConfiguration:
	mov     al, LookAheadBuf.RxStatus       ;Buffer memory problems?
	test    RxTestValue, al                 ;get test value
	jnz     HandleBufferError

	mov     bx, LookAheadBuf.RxCount
	lea     bp, [bx-4]                      ;Load BP & take off CRC length

	xor     ax, ax
	mov     al, LookAheadBuf.NextPacket     ;AL = Next Page.
	cmp     al, NextPage                    ;Did Next Page wrap?
	jae     CheckForLong                    ;Jump if not.
	add     al, PSTOP                       ;Set AL higher.
	sub     al, PSTART

CheckForLong:
	sub     al, NextPage                    ;AL = Number of Pages.
	cmp     al, 8                           ;> 1792?
	ja      PacketTooLong                   ;Jump if so.

PacketTooLongReturn:
	add     bx, 103h
	add     bh, NextPage
	cmp     bh, PSTOP
	jb      BackFromWrapAround

	sub     bh, PSTOP                       ;Wrap to beginning of ring?
	add     bh, PSTART

BackFromWrapAround:
	cmp     LookAheadBuf.NextPacket, bh
	je      GetReceiveRCB

;----------------------------------------------------------------------------
;       Since we have the NIC configured so it won't save errored packets,
;       we can safely say that if we get any set error bits in the Rx status
;       byte, there is a buffer problem.  To recover we will reset 
;       the NIC.
;----------------------------------------------------------------------------
HandleBufferError:
	inc     DriverStatRxMisc
	inc     RxBufferError
	cli
	call    ResetNICRunTime 
	jmp     ISRExit

;----------------------------------------------------------------------------
;       Attempt to get a receive buffer for this packet.
;----------------------------------------------------------------------------
GetReceiveRCB:
	mov     al, LookAheadBuf.RxStatus       ;Get RX Status.
	xor     ah, ah
	and     ax, 06h                         ;Get Error Status Bits if any
	shr     ax, 1                           ;these match up with Rcv Mon Modes
	call    MSMGetRCB
	jnz     FinishUpReceive

	or      bp, bp
	jz      NothingToCopy

;----------------------------------------------------------------------------
;       Copy the packet data into the Rx buffers. 
;       Setup the DMA for the packet data move.
;----------------------------------------------------------------------------
	lea     di, [bp+1]      ; Prepare to evenize (di = Number of bytes to move)
	mov     bh, 0           ; Clear odd byte flag
	test    al, 01h         ; If odd then get 1st odd byte

RunTimeCode3:
;;;-> RunTimeModifyCode3Begin
ifdef SHORTSLOT
;----------------------------------------------------------------------------
;       8 bit card
;----------------------------------------------------------------------------
	jmp short OnEvenBoundary        ; 8088 code (EB 0A)
else
;----------------------------------------------------------------------------
;       16 bit card
;----------------------------------------------------------------------------
	jz      OnEvenBoundary          ; 386 code (74 0A)
endif
;;;-> RunTimeModifyCode3End

;----------------------------------------------------------------------------
;       Load BL with First Byte to make transfer even possible
;----------------------------------------------------------------------------
	dec     di                              ; Adjust move count
	mov     bx, ax                          ; Get Skip Value 
	mov     bl, byte ptr [bx].LookAheadBuf+4 ; Get high byte of data 
						; (Skip Ax Value+RxHeaderStruc)
	inc     bh                              ; Set odd byte flag
	inc     ax                              ; Skip past first data byte
						; (The Byte is Stored in BL)
OnEvenBoundary:
	mov     dx, StartAddressLow             ;-> at the real packet data
	add     al, 4                           ; Skip past NIC header also
	out     dx, al                          ; High byte is already set!

	mov     dx, RemoteByteCountLow          ; 
	mov     ax, di                          ; Address of Data
	and     al, 0feh                        ; Evenize it
	out     dx, al
	inc     dx
	mov     al, ah
	out     dx, al
  
	mov     dx, Command
	mov     al, NICPage0DMARead             ;Tell NIC we want to read
	out     dx, al

	mov     dx, DataPort
	mov     ax, es
	mov     ds, ax                          ;DS:SI -> RCB
	mov     ax, bx                          ;AX gets the odd byte stuff
	lea     bx, [si].RCBFrag0Addr           ;DS:SI -> Fragment descriptors
	push    bp                              ; BP = Number of Bytes to move
						; (POP 'd in RunTimeCode4)

FillECBLoop:
	les     di, dword ptr [bx]              ;ES:DI -> frag buffer
	mov     cx, word ptr [bx+4]             ;Get this frags length
	jcxz    NoOddByte                       ;Skip this fragment

	cmp     cx, bp                          ;Which one is greater?
	jb      UseOneFromRCB

	mov     cx, bp                          ;Do remainder

UseOneFromRCB:
	sub     bp, cx

	or      ah, ah                          ;Odd byte?
	jz      NoOddByte

	stosb                                   ;Store the odd byte
	dec     cx
	mov     ah, 0                           ;Clear odd byte flag

NoOddByte:
RunTimeCode4:
;;;-> RunTimeModifyCode4Begin
ifdef SHORTSLOT
	jcxz    MoveWasEven     ; 8088 code (E3 06)

ReadBytePort:
	in      al, dx          ; 8088 code (EC)
	stosb                   ; 8088 code (AA)
	loop    ReadBytePort    ; 8088 code (E2 FC)
	jmp short MoveWasEven   ; 8088 code (EB 00)

MoveWasEven:                    ; 8088 code ()
	add     bx, 6           ; 8088 code (83 C3 06)
	or      bp, bp          ; 8088 code (0B ED)
	jnz     FillECBLoop     ; 8088 code (75 DA)

	pop     cx              ; 8088 code (59)
	shr     cx, 1           ; 8088 code (D1 E9)
	jnc     NoExtraByte     ; 8088 code (73 02)
	in      al, dx          ; 8088 code (EC)        ; Read the extra byte.
	inc     ax              ; 8088 code (40)        ; Not Used.
	
NoExtraByte:                    ; 8088 code ()
else
	jcxz    MoveWasEven     ; 386 code (E3 0C)
	shr     cx, 1           ; 386 code (D1 E9)
.286                            ; 386 code ()
	rep     insw            ; 386 code (F3> 6D)     ; Fill the fragment
.8086                           ; 386 code ()
	jnc     MoveWasEven     ; 386 code (73 06)

	in      ax, dx          ; 386 code (ED) ; Move in the next word
	stosb                   ; 386 code (AA) ; Store the byte
	mov     al, ah          ; 386 code (8A C4)
	mov     ah, 1           ; 386 code (B4 01)      ; Set the odd byte flag
MoveWasEven:                    ; 386 code ()
	add     bx, 6           ; 386 code (83 C3 06)   ; Advance to next frag descriptor
	or      bp, bp          ; 386 code (0B ED)
	jnz     FillECBLoop     ; 386 code (75 D4)

	pop     cx              ; 386 code (59)         ; CX = Clean stack off.
endif
;;;-> RunTimeModifyCode4End

;----------------------------------------------------------------------------
;       Setup for RcvComplete
;----------------------------------------------------------------------------
	mov     ax, ds                  ; ES:SI -> RCB
	mov     es, ax
	mov     ax, cs
	mov     ds, ax                  ; DS = CGroup

NothingToCopy:
	cli
	call    MSMRcvComplete
	sti                             

FinishUpReceive:
	mov     al, LookAheadBuf.NextPacket     ;Get the next page
	mov     NextPage, al

	dec     al
	cmp     al, PSTART                      ;Are we back at the beginning?
	jb      ResetBoundary                   ;Yes, time to reset

ResetBoundaryReentry:
	mov     dx, Boundary
	out     dx, al
	jmp     PacketReceivedTop

ExitReceptionHandler:
	cmp     ResetTxConfigFlag, 0            ;In loopback mode?
	jne     ResetTxConfig
	jmp     NextInterruptService            ;Look for more stuff to do

ResetBoundary:
	mov     al, PSTOP
	dec     al
	jmp short ResetBoundaryReentry

ResetTxConfig:
	mov     dx, InterruptStatus             ;Reset overflow bit
	mov     al, 10h
	out     dx, al

	mov     bp, IOAddress1
	lea     dx, [bp].NICTxConfiguration     ;Put us back in normal Tx mode
	xor     al, al
	out     dx, al
	mov     ResetTxConfigFlag, al           ;Reset this flag

	cmp     RetryTxFlag, al                 ;See if a Tx was active
	je      NoTxActive

	mov     RetryTxFlag, al                 ;Reset the flag

	mov     bl, CurrentTxPage
	and     bx, 0002h
	mov     ax, TxPacketSizeTable[bx]       ;Restart the transmit
	mov     dx, TxByteCountLow              ;Low byte of size
	out     dx, al

	inc     dx                              ;DX = XmitByteCount0
	mov     al, ah
	out     dx, al

	mov     dx, Command
	mov     al, NICTransmit
	out     dx, al

NoTxActive:
	jmp     NextInterruptService            ;Look for more stuff to do


;
;       Handle the Buffer Overwrite Warning Condition.
;

HandleBufferOverflowCase:
	cli
	mov     dx, Command                     ;Save Tx active bit 
	in      al, dx
	mov     cl, al

	mov     al, 21h                         ;Put NIC in STOP mode
	out     dx, al

	mov     dx, RemoteByteCountLow          ;Clear these registers
	xor     al, al
	out     dx, al
	inc     dx
	out     dx, al

;       Wait 1.6ms for Tx or Rx to finish.

	call    ReadTickCounter                 ;Get initial value
	mov     bx, ax                          ;Save start value

WaitForTxOrRxToFinish:
	call    ReadTickCounter
	neg     ax
	add     ax, bx                          ;Get difference

	cmp     ax, 4096                        ;419 nsec per count (1.6ms/419ns = 3818)
	jb      WaitForTxOrRxToFinish           ;Actually give it 1.7ms (nice round value)

	test    cl, 04h                         ;Is Tx active?
	jz      NoTxActive1

	mov     dx, InterruptStatus             ;See if the Tx was cancelled
	in      al, dx
	test    al, 0ah
	jnz     NoTxActive1                     ;Tx was transmitted or error occurred

	inc     RetryTxFlag                     ;Tx was cancelled (Restart it later)

;       Put the NIC in LoopBack mode, it will be reset by the 
;       receive service routine.

NoTxActive1:
	mov     bp, IOAddress1
	lea     dx, [bp].NICTxConfiguration     ;Put NIC in loopback mode
	mov     al, 2
	out     dx, al

	mov     dx, Command
	mov     al, 22h                         ;Start the NIC
	out     dx, al

	inc     DriverStatRxOverflow
	mov     ResetTxConfigFlag, 1            ;Put Tx in normal mode later
	jmp     CheckIfPacketInBuffer

TestRxConfigurationJMP:
	jmp     TestRxConfiguration

HandleBufferErrorJMP:
	jmp     HandleBufferError

RAMModeWrapAround:
	sub     ah, PSTOP                       ;Wrap around for next page.
	add     ah, PSTART
	jmp short RAMModeBackFromWrapAround


SharedPacketTooLong:
	dec     al
	mov     cl, 3
	shr     al, cl                          ; AL = (AL - 1) / 8.
	mov     cl, 11
	shl     ax, cl                          ; AX = AX * 2048.
	add     bp, ax                          ; Add to total count.
	jmp short SharedPacketTooLongReturn


DriverISRSharedRAM:
	mov     es, RAMSegment
	mov     si, word ptr NextPageWord

	test    es:[si].RxStatus, NIC_RxCRCError + NIC_RxFrameAlignError
	jz      RAMModeTestRxConfiguration

	inc     word ptr DriverStatRxChecksum   ;at least this

RAMModeTestRxConfiguration:
	mov     al, es:[si].RxStatus            ;Buffer memory problems?
	test    RxTestValue, al                 ;get test value
	
	jnz     HandleBufferErrorJMP

	mov     bp, es:[si].RxCount             ;Get hardware reported length

	xor     ax, ax
	mov     al, es:[si].NextPacket          ;AL = Next Page.
	cmp     al, NextPage                    ;Did Next Page wrap?
	jae     SharedCheckForLong              ;Jump if not.
	add     al, PSTOP                       ;Set AL higher.
	sub     al, PSTART

SharedCheckForLong:
	sub     al, NextPage                    ;AL = Number of Pages.
	cmp     al, 8                           ;> 1792?
	ja      SharedPacketTooLong             ;Jump if so.

SharedPacketTooLongReturn:
	lea     ax, [si+103h]
	add     ax, bp
	cmp     ah, PSTOP                       ;Wrap to Beginning of ring?
	jae     RAMModeWrapAround

RAMModeBackFromWrapAround:
	cmp     es:[si].NextPacket, ah          ;Make sure pointers are inline
	jne     HandleBufferErrorJMP
	mov     NextPage, ah


;       Attempt to get a receive buffer for this packet.

	mov     al, es:[si].RxStatus            ;get Rx Status again
	add     si, 4                           ;ES:SI -> Look ahead data
	sub     bp, 4                           ;Take off CRC length

	xor     ah, ah
	and     ax, 06h                         ;get Error Status Bits if any
	shr     ax, 1                           ;these match up with Rcv Mon Modes
	push    si                              ;Save packet offset
	call    MSMGetRCB
	pop     cx                              ; CX = Packet Offset
	jnz     RAMModeFinishUpReceive

;       Copy the packet data into the Rx buffers.

	mov     dx, si                          ;Save RCB offset
	lea     bx, [si].RCBFrag0Addr           ;ES:BX -> RCB frag descriptors

	mov     ds, RAMSegment                  ;DS:SI loc to start copying from        
	mov     si, cx
	add     si, ax

;       See if this frame is split across the ring boundary.

	mov     ax, cs:RAMRingSplitStop
	sub     ax, si
	cmp     bp, ax
	ja      RAMModeProcessSplitFrame
		
RAMModeFillECBLoop:

	push    es                              ;Save RCB segment value
	mov     cx, word ptr es:[bx+4]          ;Get this frag's length
	les     di, dword ptr es:[bx]           ;ES:DI -> frag buffer

RAMModeSplitReentry:
	cmp     cx, bp                          ;Which one is greater?
	jb      RAMModeUseOneFromRCB

	mov     cx, bp                          ;Do remainder

RAMModeUseOneFromRCB:
	sub     bp, cx  

	cmp     cx, 0                           ;Zero Length Fragment?
	je      RAMModeZeroLengthFragCont

	test    si, 1                           ;Off by a byte?
	jz      RAMISRNoAdjustmentNeeded        ;Jump if not.
	movsb
	dec     cx

RAMISRNoAdjustmentNeeded:       
	shr     cx, 1                           ;Get Num words
   rep  movsw
	adc     cl, cl                          ;Any odd bytes left?
   rep  movsb                                   ;Move odd byte if it exists.

RAMModeZeroLengthFragCont:
	pop     es                              ;ES:BX -> next frag descriptor
	add     bx, 6
	or      bp, bp                          ;All the data copied?
	jnz     RAMModeFillECBLoop

	cmp     cs:SplitFlag, 0                 ;Is there some at ring start?
	jne     RAMModeDoRestOfSplitFrame

	mov     ax, cs
	mov     ds, ax                          ;DS = CGroup
	mov     si, dx                          ;ES:SI -> RCB
	cli
	call    MSMRcvComplete

RAMModeFinishUpReceive:
	mov     al, NextPage
	dec     al
	cmp     al, PSTART                      ;Are we back at the beginning?
	jb      RAMModeResetBoundary

RAMModeResetBoundaryReentry:
	mov     dx, Boundary
	out     dx, al
	jmp     PacketReceivedTop

RAMModeResetBoundary:
	mov     al, PSTOP
	dec     al
	jmp short RAMModeResetBoundaryReentry

RAMModeProcessSplitFrame:
	mov     cs:SplitFlag, 1                 ;Signal we have some at ring start
	sub     bp, ax                          ;Get amount located at ring start
	push    bp                              ;Save it

	mov     bp, ax                          ;Get amount located at ring end
	jmp     RAMModeFillECBLoop                      ;Fill as much as we can

RAMModeDoRestOfSplitFrame:
	mov     cs:SplitFlag, 0                 ;Reset the flag
	sub     bx, 6                           ;Back up
	pop     bp                              ;Get amount located at ring start
	push    es                              ;Save RCB segment
	mov     cx, word ptr es:[bx+4]          ;Get the frag's length
	les     ax, dword ptr es:[bx]           ;ES:AX -> current frag buffer
	neg     ax
	add     ax, di                          ;Get the amount already used up
	sub     cx, ax                          ;CX has the amount left in buffer

	mov     si, cs:RAMRingSplitStart
	jmp     RAMModeSplitReentry

;
;       TRANSMIT COMPLETE HANDLER
;

TxErrorInterruptJMP:
	jmp short TxErrorInterrupt



TxInterrupt:
	test    al, 08h                         ;Transmit error?
	mov     al, 0ah
	out     dx, al                          ;Reset Tx bits
	jnz     TxErrorInterrupt

TxErrorCheckReentry:
	inc     MSMTxFreeCount                  ;One of our Tx resources is free

	mov     bl, CurrentTxPage               ;Convert page to index value
	and     bx, 0002h                       ;Keep bit 1
	shr     bx, 1

	mov     ah, ClearTxBitTable[bx]         ;Get clear byte to use
	and     byte ptr TxBufUsageFlags, ah    ;Clear out appropriate bit
	jnz     StartNextOneGoing

CheckForQueuedTx:
	cmp     word ptr MSMTxQueueHead+2, 0    ;Anything on the queue?
	jne     GetTheNextSend

TxCompleteExit:
	cli                                     ;Exit our ISR
	mov     dx, InterruptMask
	mov     al, NICUnmaskByte
	out     dx, al
	ret

GetTheNextSend:
	cli
	call    MSMGetNextSend
	jnz     TxCompleteExit                  ;We must be saturated already

	sti                                     ;Back on
	jmp     InternalDriverSend              ;It will reenable our card's IRQ

StartNextOneGoing:
	mov     al, NextTxBufTable[bx]          ;Get other Tx page location
	mov     CurrentTxPage, al
	mov     dx, TransmitPage
	out     dx, al

	shl     bx, 1
	xor     bl, 02h                         ;Toggle index
	mov     ax, TxPacketSizeTable[bx]       ;Get # of bytes for this Tx
	mov     dx, TxByteCountLow
	out     dx, al
	inc     dx
	mov     al, ah
	out     dx, al

	mov     dx, Command                     ;Start this one going
	mov     al, NICTransmit
	out     dx, al

	mov     TxFlagValue, 0                  ;Do Tx timeout detection setup
	mov     TxRetryCounter, SendRetryCount
	jmp     CheckForQueuedTx


TxErrorInterrupt:
	mov     TxFlagValue, 0                  ;Reset timeout detection
	inc     DriverStatTxMisc                ;Save memory, don't decode error bits
	dec     TxRetryCounter                  ;One less try to go
	jnz     RetryTheTransmit

;       Max retries reached. Quit trying.

	inc     TxAbortTimeOut
	jmp     TxErrorCheckReentry

;       Tell the NIC to retry the transmit.

RetryTheTransmit:
	inc     DriverStatTxRetryCount          ;Do diags

	mov     bl, CurrentTxPage
	and     bx, 0002h                       ;Keep bit 1
	mov     ax, TxPacketSizeTable[bx]
	mov     dx, TxByteCountLow              ;Low byte of size
	out     dx, al

	inc     dx                              ;DX = XmitByteCount0
	mov     al, ah
	out     dx, al

	mov     dx, Command
	mov     al, NICTransmit
	out     dx, al
	jmp     NextInterruptService            ;Look for more to do

DriverISR       endp


;
;       ReadTickCounter
;
;       Reads the Timer tick count register which decrements by 2 from 65536 
;       to 0 every 1/36.414 of a second.  Each 2 decrements of the count 
;       represents 838 nsec's.
;
;       assumes:        Interrupts disabled
;
;       returns:        ax      has the current count
;                       All other registers preserved
;                       Interrupts disabled
;

ReadTickCounter proc    near

	mov     al, 06h
	out     43h, al                         ;Command 8254 to latch T0's count

;Now read the count.

	in      al, 40h                         ;Read LSB first
	mov     ah, al
	in      al, 40h                         ;Read MSB
	xchg    al, ah                          ;Get in proper order
	ret

ReadTickCounter endp


;
;       DriverTimeOut
;
;       This routine will be called approximately once every second.
;       It will check to see if the NIC ack'ed the last transmit operation.
;       If the NIC did not ack it, then the NIC will be reset.
;
;       assumes:        es:si   -> our AES ECB.
;                       CLD is in affect.
;                       Interrupts are disabled.
;
;       returns:        ds,ss,sp,bp preserved.
;                       Interrupts disabled.
;

DriverTimeOut   proc    far

	mov     dx, cs:Command
	in      al, dx

	test    al, 04h
	jnz     TxIsInProgress

TxOkayReentry:
	mov     bx, MLIDSUP_SCHEDULE_AES_EVENT      ;Reschedule ourselves
	mov     word ptr es:[si].MSecondValue, 1000 ;Wake up in 1 second.
	jmp     cs:LSLSupport

TxIsInProgress:
	inc     cs:TxFlagValue                  ;Signal a possible timeout
	cmp     cs:TxFlagValue, 1               ;See if we have a timeout.
	je      TxOkayReentry                   ;Not yet

	push    ds
	push    es
	push    bp
	push    si

	push    cs
	pop     ds
	inc     TxAbortTimeOut
	inc     DriverStatTxMisc
	call    ResetNICRunTime

	call    MSMGetNextSend
	jnz     NoMoreToSend

	mov     dx, InterruptMask               ;Disable the NIC's interrupt
	xor     al, al
	out     dx, al
	cld
	call    InternalDriverSend              ;Leave interrupts off!

	mov     bx, MLIDSUP_SERVICE_EVENTS
	call    LSLSupport

NoMoreToSend:
	pop     si
	pop     bp
	pop     es
	pop     ds
	jmp     TxOkayReentry

DriverTimeOut   endp



;
;       DriverSend
;
;       This routine will initiate a packet transmission.
;
;       assumes:        ds      CGroup
;                       si      -> TCB to be transmitted.
;                       cx      hardware frame length
;                       CLD has been executed
;                       Interrupts are disabled.
;                       A critical section has been started
;                       
;       returns:        Can destroy all registers and flags
;                       Interrupts disabled
;

DriverSend      proc    near

	mov     dx, InterruptMask               ;Disable the NIC's interrupt
	xor     al, al
	out     dx, al
	in      al, dx                          ;Flush the OUT before doing
	sti                                     ;the sti.

InternalDriverSend:
	cmp     cs:SharedRAMFlag, 0             ; Shared RAM Mode?
	je      DriverSendIOMode
	jmp     DriverSendSharedRAM

DriverSendIOMode:
	mov     dx, InterruptStatus
	mov     al, 40h                         ;Reset DMA complete bit
	out     dx, al

	inc     dx                              ;StartAddressLow
	xor     al, al                          ;Tell NIC where we want data
	out     dx, al
	inc     dx
	mov     bx, TxBufUsageFlags             ;Get proper Tx buffer to use
	mov     al, TxBufToUseTable[bx]
	out     dx, al

	and     bl, 0FDh                        ;Clear bit 1
	shl     bx, 1                           ;Need word index
	mov     TxPacketSizeTable[bx], cx       ;Save the frame length
	mov     bp, cx                          ;Save it

	inc     dx                              ;DX=RemoteByteCountLow
	mov     ax, [si].TCBDataLen

RunTimeCode:
;;;-> RunTimeModifyCodeBegin
ifdef SHORTSLOT
	jmp short $+3           ; 8088 code     ;Run Time modify code.
	inc     ax              ; 8088 code
else
	inc     ax              ; 386 code (40)    ; Length moved must be even
	and     al, 0FEh        ; 386 code (24 FE)
endif
;;;-> RunTimeModifyCodeEnd

	out     dx, al
	inc     dx
	mov     al, ah
	out     dx, al

;       Setup the NIC to accept the data transfer.
  
	mov     dx, Command
	mov     al, NICPage0DMAWrite
	out     dx, al

;       Output the media header.

	mov     dx, DataPort                    ;Transfer port
	mov     bx, si                          ;Save TCB offset
	mov     cx, [si].TCBMediaHeaderLen
	add     si, TCBMediaHeader              ;DS:SI -> Media header
	mov     ah, cl
	jcxz    RawSend

RunTimeCode0:
;;;-> RunTimeModifyCodeBegin
ifdef SHORTSLOT
	lodsb                   ; 8088 code
	out     dx, al          ; 8088 code
	loop    RunTimeCode0    ; 8088 code
	xor     ax, ax          ; 8088 code
	jmp     $+2             ; 8088 code
else
	and     ah, 01h         ; 386 code (80 E4 01)   ;AL != 0 if add byte to do
	shr     cx, 1           ; 386 code (D1 E9)      ;Do word move
.286
	rep     outsw           ; 386 code (F3> 6F)
.8086
	lodsb                   ; 386 code (AC) ; Pick up left over byte (If any)
endif
;;;-> RunTimeModifyCodeEnd

;       Write out the data fragments.

RawSend:
	push    bx
	les     di, [bx].TCBFragStrucPtr        ;ES:DI -> FragStruc
	mov     bh, byte ptr es:[di]            ;Get the number of fragments

FragmentOutLoop:
	mov     cx, word ptr es:[di+2].FragLen   ; Get fragment length
	jcxz    AdvanceToNextFrag                ; Zero Fragment.
	lds     si, dword ptr es:[di+2].FragAddr ; Get -> fragment data

RunTimeCode1:
;;;-> RunTimeModifyCodeBegin
ifdef SHORTSLOT
	lodsb                           ; 8088 code
	out     dx, al                  ; 8088 code
	loop    RunTimeCode1            ; 8088 code
	xor     ah, ah                  ; 8088 code
	jmp short AdvanceToNextFrag     ; 8088 code
else
	or      ah, ah                  ; 386 code (0A E4)      ;Odd byte to do?
	jnz     OddByteToDo             ; 386 code (75 6C)
					
DoTheCurrentFrag:                       ; 386 code ()
	shr     cx, 1                   ; 386 code (D1 E9)
.286                                    ; 386 code ()
	rep     outsw                   ; 386 code (F3> 6F)
.8086                                   ; 386 code ()
	jnc     AdvanceToNextFrag       ; 386 code (73 03)
endif
;;;-> RunTimeModifyCodeEnd

	lodsb                                   ;Pick up the odd byte
	inc     ah                              ;Set odd byte flag

AdvanceToNextFrag:
	add     di, 6                           ;-> next fragment structure
	dec     bh                              ;One more fragment done
	jnz     FragmentOutLoop

	or      ah, ah                          ;Odd byte still left?
	jnz     DoLastOddByte

CopyFragDone:
	mov     ax, cs                          ;DS:SI -> TCB
	mov     ds, ax
	pop     si

;       Now wait for the NIC to tell us that the remote DMA has completed.
  
	mov     dx, InterruptStatus
	in      al, dx                          ;Normally it will be finished!!
	test    al, 40h                         ;DMA interrupt bit set?
	jz      DMAHasNotCompleted

DMAFinished:
	mov     di, TxBufUsageFlags             ;Get current Tx buffer bits
	mov     bx, di
	or      bl, TxBufBitTable[di]
	mov     byte ptr TxBufUsageFlags, bl

	cmp     bl, 11b                         ;If both set, don't start this one
	je      DontStartThisOne

	mov     dx, TxByteCountLow              ;Tell the NIC how many to Tx
	mov     ax, bp                          ;BP has the frame length
	out     dx, al
	inc     dx                              ;DX = TxByteCountHigh
	mov     al, ah
	out     dx, al

	mov     dx, TransmitPage
	mov     al, TxBufToUseTable[di]         ;Get current Tx buffer location
	mov     CurrentTxPage, al
	out     dx, al

;       Tell the NIC to transmit the packet.

	mov     dx, Command
	mov     al, NICTransmit
	out     dx, al

	mov     TxFlagValue, 0                  ;Do Tx timeout detection setup
	mov     TxRetryCounter, SendRetryCount  ;Reset Tx retry counter

DontStartThisOne:
	cli
	mov     dx, InterruptMask               ;Enable the boards interrupt line
	mov     al, NICUnmaskByte
	out     dx, al
	jmp     MSMSendComplete


DoLastOddByte:
	out     dx, ax                          ;Value in AH doesn't matter!
	jmp     CopyFragDone

OddByteToDo:
	jcxz    AdvanceToNextFrag               ;This fragment is empty
	mov     ah, al                          ;Save odd byte value
	lodsb                                   ;Steal a byte from the fragment
	xchg    al, ah                          ;Get in proper order
	out     dx, ax
	dec     cx                              ;Adjust fragment count
	xor     ah, ah                          ;Clear odd byte flag
	jmp     DoTheCurrentFrag

DMADidntComplete:
	cli
	push    si
	call    ResetNICRunTime
	pop     si
	jmp     DontStartThisOne

DMAHasNotCompleted:
	mov     cx, 0FFFFh/2                    ;Max. wait time

WaitForDMACompleteLoop:
	in      al, dx
	test    al, 40h
	jnz     DMAFinished

	dec     cx
	jnz     WaitForDMACompleteLoop

	inc     DriverStatTxMisc
	inc     DMAWouldNotComplete
	jmp     DMADidntComplete

DriverSend      endp

;
;***********************************************************************\
;
; BEGIN_MANUAL_ENTRY( DriverSendSharedRAM, NE2000/API/DRVSNDSR )
;
; Name:         DriverSendSharedRAM
;
; Description:  This routine is call to transmit a packet when the card 
;               is configured for Shared RAM. It is responsible for setting 
;               up the card, moving the data to the card and initiating 
;               the send.
;
; On Entry:     AX      N/A
;               BX      N/A
;               CX      Hardware frame length
;               DX      N/A
;               BP      N/A
;               SI      -> to TCB
;               DI      N/A
;               DS      CGROUP
;               CS      CGROUP
;
;               Note:   CLD has been executed
;                       Interrupts are disabled.
;                       A critical section has been started
;
; On Return:    AX      Destroyed
;               BX      Destroyed
;               CX      Destroyed
;               DX      Destroyed
;               BP      Destroyed
;               SI      Destroyed
;               DI      Destroyed
;
;               Flags:
;
;               Note:   Interrupts disabled.
;
; Remarks:      The entry point for this routine is via DriverSend.  The 
;               NIC's interrupts were disabled in DriverSend before it 
;               got to this routine.
;
; See Also:     DriverSend
;
; END_MANUAL_ENTRY
;
;***********************************************************************/
;

DriverSendSharedRAM     proc
	mov     bx, TxBufUsageFlags             ;Get proper Tx buffer to use
	shl     bx, 1
	mov     di, RAMModeTxBufToUseTable[bx]

	and     bl, not 04h                     ;Clear bit 2
	mov     TxPacketSizeTable[bx], cx       ;Save hard frame length
	mov     ax, cx                          ;Save it

;       Output the media header.

	push    si                              ;Save TCB offset
	mov     bx, si                          ;DS:BX -> TCB

	mov     es, RAMSegment                  ;ES:DI -> NICs Tx buffer
	add     si, TCBMediaHeader              ;DS:SI -> media header
	mov     cx, [bx].TCBMediaHeaderLen      ;Get the length

	shr     cx, 1                           ;Get Num words
   rep  movsw
	adc     cl, cl                          ;Any odd bytes left?
   rep  movsb                                   ;Move odd byte if it exists.

;       Now write out the data fragments.

	lds     bx, [bx].TCBFragStrucPtr        ;DS:BX -> frag list
	mov     dh, byte ptr [bx]               ;Get the number of fragments

RAMModeFragmentOutLoop:
	push    ds
	mov     cx, word ptr [bx+2].FragLen     ;Get fragment length
	lds     si, dword ptr [bx+2].FragAddr   ;Get -> fragment data

	cmp     cx, 0                           ;Zero Length Fragment?
	je      ZeroLengthFragContinue

	test    di, 1                           ;Off by a byte?
	jz      RAMNoAdjustmentNeeded           ;Jump if not.
	movsb
	dec     cx

RAMNoAdjustmentNeeded:  
	shr     cx, 1                           ;Get Num words
   rep  movsw
	adc     cl, cl                          ;Any odd bytes left?
   rep  movsb                                   ;Move odd byte if it exists.

ZeroLengthFragContinue:
	pop     ds
	add     bx, 6                           ;-> next frag descriptor
	dec     dh                              ;One more fragment done
	jnz     RAMModeFragmentOutLoop

	pop     si
	mov     bx, cs
	mov     ds, bx                          ;DS:SI -> TCB

	mov     di, TxBufUsageFlags             ;Get current Tx buffer bits
	mov     bx, di
	or      bl, TxBufBitTable[di]
	mov     byte ptr TxBufUsageFlags, bl

	cmp     bl, 11b                         ;If both set, don't start this one
	je      RAMModeDontStartThisOne

	mov     dx, TxByteCountLow              ;Program the Tx count register
	out     dx, al
	inc     dx
	mov     al, ah
	out     dx, al

	mov     dx, TransmitPage
	shl     di, 1
	mov     al, byte ptr RAMModeTxBufToUseTable[di+1]
	out     dx, al
	mov     CurrentTxPage, al

;       Tell the NIC to transmit the packet.

	mov     dx, Command
	mov     al, NICTransmit
	out     dx, al

	mov     TxFlagValue, 0                  ;Do Tx timeout detection setup
	mov     TxRetryCounter, SendRetryCount  ;Reset Tx retry counter

RAMModeDontStartThisOne:
	cli
	mov     dx, InterruptMask               ;Enable the boards interrupt line
	mov     al, NICUnmaskByte
	out     dx, al
	jmp     MSMSendComplete

DriverSendSharedRAM     endp

;
;       ResetNIC
;
;       Routine to totally reinitialize the NIC.
;
;       assumes:        ds      is set to CGroup
;                       Interrupt state is disabled
;
;       returns:        bx,ds,ints preserved.
;

ResetNICRunTime:
	inc     ReinitializeNIC

ResetNIC        proc    near

	mov     bp, IOAddress1                  ;Get base I/O port

	cmp     cs:SharedRAMFlag, 0             ; Shared RAM Mode?
	je      IOModeResetNIC

	add     bp, 10h                         ; Adjust I/O address

	mov     dx, cs:Control2                 ; DX = Control 2.
	mov     al, cs:Control2Value            ; Setup control register 2
	out     dx, al

	mov     dx, cs:Control1                 ; DX = Control 1.
	mov     al, cs:Control1Value
	or      al, 10000000b                   ; Set Reset line.
	out     dx, al                          ; Reset adapter.
	slow
	and     al, 01111111b                   ; Release Reset line.
	out     dx, al
	jmp short ResetContinue


IOModeResetNIC:
	lea     dx, [bp].NICReset               ;Reset the NIC
	in      al, dx
	slow
	out     dx, al                          ;Remove the reset

ResetContinue:
	mov     dx, Command                     ;Page 0, stop mode
	mov     al, 21h

	lea     dx, [bp].NICDataConfiguration
	mov     al, 49h
	mov     ah, EightBitSlot
	not     ah
	and     al, ah
	out     dx, al

	mov     dx, RemoteByteCountLow          ;Reset to 0
	xor     al, al
	out     dx, al
	inc     dx
	out     dx, al

	lea     dx, [bp].NICRxConfiguration
	mov     al, 0Ch                         ;Broadcasts & Multicasts enabled
	out     dx, al

	lea     dx, [bp].NICTxConfiguration     ;Put in loopback for a moment
	mov     al, 02h
	out     dx, al

	mov     cl, PSTART
	mov     ch, PSTOP

	mov     dx, Boundary                    ;Set the boundary register
	mov     al, cl
	out     dx, al

	lea     dx, [bp].NICPageStart           ;Set the PageStart register
	out     dx, al

	lea     dx, [bp].NICPageStop            ;Set the PageStop register
	mov     al, ch
	out     dx, al

	mov     dx, InterruptStatus             ;Clear ISR
	mov     al, 0FFh
	out     dx, al

	mov     dx, InterruptMask
	mov     al, NICUnmaskByte               ;Get current mask 
	out     dx, al

	mov     dx, Command                     ;Set to page1 registers
	mov     al, 61h
	out     dx, al

	mov     si, offset CGroup:MSMPhysNodeAddress    ;Address source
	lea     dx, [bp].PhysicalReg0           ;Address destination
	mov     cx, 6                           ;Do six bytes

ProgramNodeAddressLoop:
	lodsb                                   ;Get address byte
	out     dx, al                          ;Give it to the 8390
	inc     dx                              ;Goto next address port
	loop    ProgramNodeAddressLoop

;       While in page 1 setup Current page register.

	lea     dx, [bp].Current
	mov     al, PSTART
	inc     al
	out     dx, al

	mov     NextPage, al                    ;Reset the next page 

	mov     dx, Command                     ;Switch back to page 0 registers
	mov     al, 22h                         ; and start the NIC.
	out     dx, al

;       Setup the multicast hash bits.

	push    bp
	call    MSMUpdateMulticast
	pop     bp

	mov     TxBufUsageFlags, 0              ;No Tx buffers in use
	mov     MSMTxFreeCount, 2

	lea     dx, [bp].NICTxConfiguration     ;Set Tx mode to normal
	xor     al, al
	out     dx, al

	mov     ResetTxConfigFlag, al           ;Reset this stuff
	mov     CurrentTxPage, al
	ret

ResetNIC        endp

Code    ends



Init    segment para public 'CODE'

;       Error messages.

;----------------------------------------------------------------------------
;
;                               Message Center
;
;----------------------------------------------------------------------------

NoNE2000InMachine       db      50,0,'The board cannot be found.',0Dh,0Ah,0
RAMFailureMsg           db      51,0,'Board RAM failed the memory test.',0DH,0AH,0
CommandPortNotAliveMsg  db      54,0,'The board did not respond to the initialization command.',0DH,0AH,0
InvalidBusIDMsg         db      96,0,'The driver does not support the selected BUS ID.',0Dh,0Ah
			db              'Make sure the NET.CFG file contains a valid '
			db              'parameter for the ',0Dh,0Ah, 'BUS ID keyword.',0Dh,0Ah, 0

IRQInvalidMessage       db      91,0,'The hardware configuration conflicts.', 0Dh, 0Ah, 0

DriverNumKeywords       equ     0               ;Stub these out
DriverKeywordText:
DriverKeywordTextLen:
DriverProcessKeywordTab:

SharedRAMINTArray       db      3, 4, 5, 2, 10, 11, 12, 15


;****************************************************************************
;       DriverInit
;
;       This routine is called to intialize and bring the specified NIC
;       into operation.
;
;       assumes:
;               ds=es=cs CGroup
;               CLD has been executed
;               Interrupts are enabled.
;
;       returns:
;               dx      -> error message if error during init.
;               ax      = 0 if init okay (MSMERR_NOT)
;                       = MSMERR_HSM_FAILED if error (dx -> err msg)
;                       = MSMERR_CARDNOTPRESENT if The Card is missing
;                                       (used by PCMCIA Drivers)
;               Flags set according to ax
;               All other registers can be destroyed
;
;****************************************************************************

CheckForSupportedBusIDs:
	cmp     BusID, BUS_ID_EISA                      ; See if we happen to be
	je      CheckForEISABus                         ;   in EISA machine.
	cmp     BusID, BUS_ID_ISA                       ; See if set to ISA.
	jne     InitErrorExitJmp
	mov     al, BUS_ID_ISA
	test    MSMSystemFlags, EISAFlag                ; Do we have EISA Bus?
	jz      ValidBusIDFound
	mov     al, BUS_ID_EISA
	jmp     short ValidBusIDFound
	
CheckForEISABus:
	mov     al, BUS_ID_EISA
	test    MSMSystemFlags, EISAFlag                ; EISA Bus?
	jnz     ValidBusIDFound

InitErrorExitJmp:
	jmp     InitErrorExit

IOModeInitPortsJMP:
	jmp     IOModeInitPorts

DriverInit      proc    near
;----------------------------------------------------------------------------
;       Make sure user did not place invalid Bus ID in the NET.CFG file.
;----------------------------------------------------------------------------
	mov     dx, offset CGroup:InvalidBusIDMsg
	cmp     BusID, BUS_ID_HSM_DEFAULT
	jne     CheckForSupportedBusIDs

	mov     al, BUS_ID_EISA
	test    MSMSystemFlags, EISAFlag                ; EISA Bus?
	jnz     ValidBusIDFound
	mov     al, BUS_ID_ISA                          ; Assume ISA.

;----------------------------------------------------------------------------
;       Take care of the manditory assignments
;----------------------------------------------------------------------------

ValidBusIDFound:
	mov     BusID, al                       ; Set Bus ID.

	mov     MSMLookAheadSegment, cs         ;Tell MSM about our look ahead seg value
	mov     MSMTxFreeCount, 2               ;Tell MSM we can do 2 at a time

;----------------------------------------------------------------------------
;       See if the Adapter is even present, Hold off ints
;----------------------------------------------------------------------------
	cli
	mov     bp, IOAddress1

						;edr 03/10/95 SPD 93236
	lea     dx, [bp].NICCommand             ;command register
	in      al, dx                          ;get currentvalue
	mov     ch, al
	mov     al, NICPage0
	out     dx, al
	slow

	lea     dx, [bp].NICStartAddressLow     ;Location to test.
	in      al, dx                          ;Read the current value.
	mov     ah, al                          ;Save current value.
	inc     al                              ;Create value to test with.
	mov     cl, al                          ;CL = New value.
	out     dx, al                          ;Write new value.
	slow
	in      al, dx
	cmp     al, cl                          ;Will equal if I/O.
	mov     al, ah                          ;Restore old value
	out     dx, al                          ; in case it wasn't
						; really an NE2000.
	je      IOModeInitPortsJMP              ;Jump if I/O.

						;edr 03/10/95 SPD 93236
	lea     dx, [bp].NICCommand             ;command register
	mov     al, ch                          ;restore what ever it was that was there
	out     dx, al
	slow

;----------------------------------------------------------------------------
;       Well we know that if a NE2000 is present it is not in IO mode
;       Now Check to see if it is even present.
;----------------------------------------------------------------------------
	add     dx, 10h                 ; Location to test. (adjust for shared ram ioport locations.)

					;edr 03/10/95 SPD 93236
	in      al, dx                  ;get currentvalue
	mov     ch, al
	mov     al, NICPage0
	out     dx, al
	slow

	lea     dx, [bp].NICStartAddressLow     ;Location to test.
	add     dx, 10h                 ; Location to test. (adjust for shared ram ioport locations.)

	in      al, dx                  ; Read the current value.
	mov     ah, al                  ; Save current value.
	inc     al                      ; Create value to test with.
	mov     cl, al                  ; CL = New value.
	out     dx, al                  ; Write new value.
	slow
	in      al, dx
	cmp     al, cl                  ; Will equal if SharedRam and Card Present.
	mov     al, ah                  ; Restore old value
	out     dx, al                  ; in case it wasn't
					; really an NE2000.
	je      SharedRAMMode

					;edr 03/10/95 SPD 93236
	lea     dx, [bp].NICCommand     ;command register
	mov     al, ch                  ;restore what ever it was that was there
	out     dx, al
	slow

	jmp     NoNE2000Card            ; Jump if not Present.

SharedRAMMode:  
	mov     SharedRAMFlag, -1

	mov     PSTART, 0Ch                     ;12 pages for send (enough for 2 buffers)

	lea     cx, word ptr [bp].IControl1     ; Base + 00h (Control 1).
	mov     Control1, cx

	lea     cx, word ptr [bp].IATDetect     ; Base + 01h (AT Detect).
	mov     ATDetect, cx

	lea     cx, word ptr [bp].IControl2     ; Base + 05h (Control 2).
	mov     Control2, cx            

	lea     cx, word ptr [bp].IProm         ; Base + 08h (PROM).
	mov     Prom, cx        

	add     bp, 10h                         ; Adjust Base I/O for RAM Mode.

	mov     dx, ATDetect                    ; DX = AT Detect.
	in      al, dx
	test    al, 01h                         ; 8 bit slot?
	jnz     Not8BitSlot

	mov     EightBitSlot, 1
	mov     Control2Value, 01h              ; 8 bit mode.
	mov     MemRange1, 200h         ; 8k of RAM available.
	mov     PSTOP, 20h                      ; PSTOP for 8 bit mode.
	jmp short EightBitSlotContinue


Not8BitSlot:
	mov     PSTOP, 40h                      ; PSTOP for 16 bit mode.
	mov     Control2Value, 0c1h             ; 16 bit mode.
	mov     MemRange1, 400h         ; 16k of RAM available.

EightBitSlotContinue:
	xor     bx, bx
	mov     bl, byte ptr PSTOP
	mov     ax, 100h
	mul     bx
	mov     RAMRingSplitStop, ax

	xor     bx, bx
	mov     bl, byte ptr PSTART
	mov     ax, 100h
	mul     bx 
	mov     RAMRingSplitStart, ax

	cmp     word ptr MemAddress1+2, 0       ; Was address found in NET.CFG?7
	jne     AddrSetByNETCFG
	mov     word ptr MemAddress1+0, 0h      ; Set Default as 0d0000h.
	mov     word ptr MemAddress1+2, 0dh     

AddrSetByNETCFG:
	mov     al, byte ptr MemAddress1+2      ; Get RAM segment base from phys addr
	mov     cl, 12
	shl     ax, cl
	mov     dx, word ptr MemAddress1+0
	mov     cl, 4
	shr     dx, cl
	or      dx, ax

	mov     RAMSegment, dx
	mov     MSMLookAheadSegment, dx         ; Tell MSM where our look 

	mov     bx, dx
	mov     cl, 9
	shr     bx, cl
	or      bl, 01000000b                   ; set memory bit
	mov     Control1Value, bl               ; save for reset

	mov     TransmitPage0, 0h               ; Initialize Transmit page variables.
	mov     TransmitPage1, 6h

IOModeInitPorts:
	mov     cx, NumIOPorts                  ; Number of ports to initialize.
	mov     si, offset CGroup:Command       ; Get offset to port where port 
						;  values are stored.
SetupPortLoop:
	add     word ptr [si], bp               ; Add in base I/O port
	add     si, 2
	loop    SetupPortLoop

;       Must issue reset to the NIC before doing anything with
;       its registers.

	cmp     SharedRAMFlag, 0                ; Shared RAM Mode?
	je      ResetNICIOMode

	xor     bx, bx
	mov     dx, RemoteByteCountLow          ; Read Configuration Register A
	in      al, dx                          ;  to get INT card is configured for.
	mov     bl, al
	and     bl, 38h                         ; Mask INT Bits
	mov     cl, 3
	shr     bl, cl

	mov     al, SharedRAMINTArray[bx]       ; Get interrupt number.
	mov     IRQLine1, al                    ; Set Int value in configuration table.

	mov     dx, Control2                    ; DX = Control 2.
	mov     al, Control2Value               ; Setup control register 2
	out     dx, al

	mov     dx, Control1                    ; DX = Control 1.
	mov     al, Control1Value
	or      al, 10000000b                   ; Set Reset line.
	out     dx, al                          ; Reset adapter.
	slow
	and     al, 01111111b                   ; Release Reset line.
	out     dx, al

	jmp short IsPortAlive


ResetNICIOMode:
	lea     dx, [bp].NICReset               ;Get base I/O port
	in      al, dx                          ;Do a reset on the NIC

	out     dx, al                          ;Remove the reset

IsPortAlive:
	mov     dx, Command                     ;Get the command port
	mov     al, 21h
	out     dx, al

	slow

	in      al, dx                          ;Is the port alive?
	cmp     al, 21h
	je      PortIsAlive

	cmp     al, 23h                         ;23h is also possible
	je      PortIsAlive

;       The NIC is not alive.

	mov     dx, offset CGroup:CommandPortNotAliveMsg

ErrorInRAM:
	jmp     InitErrorExit

;       Read in the node address from the NIC unless the user
;       specified one in the NET.CFG file.
;
;       First initialize the NIC far enough to read and write its RAM.

PortIsAlive:
	lea     dx, [bp].NICDataConfiguration   
	mov     al, 49h
	out     dx, al

	lea     dx, [bp].NICTxConfiguration     ;Put in loopback for a moment
	mov     al, 02h
	out     dx, al

	mov     ax, word ptr MSMPhysNodeAddress+0       ;See if override
	and     ax, word ptr MSMPhysNodeAddress+2
	and     ax, word ptr MSMPhysNodeAddress+4

	cmp     ax, -1
	jne     NodeOverride

;       The user did not override the node address so read
;       the one on the NIC.

	mov     di, offset CGroup:MSMPhysNodeAddress    ;Destination

	cmp     cs:SharedRAMFlag, 0             ; Shared RAM Mode?
	je      IOModeGetNodeAddr

	mov     dx, cs:Prom                     ; DX = Prom I/O port.
	mov     cx, 6
RAMModeReadAddressLoop:
	in      al, dx
	inc     dx
	stosb
	loop    RAMModeReadAddressLoop

	jmp short NodeOverride


IOModeGetNodeAddr:

;       Setup to read the NIC's node address.  It is located at offset 0
;       in the NICs address space.

	mov     dx, RemoteByteCountLow
	mov     ax, 12                          ;We want 6 bytes 
	out     dx, al
	inc     dx
	mov     al, ah
	out     dx, al

	mov     dx, StartAddressLow
	xor     ax, ax
	out     dx, al
	inc     dx
	out     dx, al

	mov     dx, Command
	mov     al, NICPage0DMARead
	out     dx, al

	mov     dx, DataPort
	mov     cx, 6                           ;Do 6 bytes

InputNodeAddressLoop:
	in      al, dx
	stosb
	loop    InputNodeAddressLoop

;----------------------------------------------------------------------------
;  Now check to see if the NIC is in a 16-bit slot.
;----------------------------------------------------------------------------
NodeOverride:
	cmp     cs:SharedRAMFlag, 0             ; Shared RAM Mode?
	je      IOModeCheckCardType
	jmp     SlotIsOkay

IOModeCheckCardType:
	mov     dx, RemoteByteCountLow
	mov     ax, 2                           ;We need up to 1 byte
	out     dx, al
	inc     dx
	mov     al, ah
	out     dx, al

	mov     dx, StartAddressLow             ;Load the the Remote Start address to read prom.
	mov     ax, 14h                         ;Offset of Ethernet board type
	out     dx, al
	inc     dx
	mov     al, ah
	out     dx, al

	mov     dx, Command
	mov     al, NICPage0DMARead
	out     dx, al

	mov     dx, DataPort
	in      al, dx                          ;Get the board type

	cmp     al, 0dh
	jne     IOModeChk8or16Bit

	xor     bx, bx
	mov     dx, RemoteByteCountLow          ; Read Configuration Register A
	in      al, dx                          ;  to get INT card is configured for.
	mov     bl, al
	and     bl, 38h                         ; Mask INT Bits
	mov     cl, 3
	shr     bl, cl

	mov     al, SharedRAMINTArray[bx]       ; Get interrupt number.
	mov     IRQLine1, al                    ; Set Int value in configuration table.

	mov     dx, Command                     ; Dummy read to kept next
	in      al, dx                          ;  operation from writing to
						;  configuration register A.

IOModeChk8or16Bit:
	mov     dx, RemoteByteCountLow
	mov     ax, 2                           ;We need up to 1 byte
	out     dx, al
	inc     dx
	mov     al, ah
	out     dx, al

	mov     dx, StartAddressLow             ;Start right after node address
	mov     ax, 1ch                         ;Offset of 'B' or 'W'
	out     dx, al
	inc     dx
	mov     al, ah
	out     dx, al

	mov     dx, Command
	mov     al, NICPage0DMARead
	out     dx, al

	mov     dx, DataPort
	in      al, dx                          ;Get the bus type flag

	cmp     al, 'B'
	je      NE2000In8BitSlot

	cmp     al, 'W'                         ;Is it in 16-bit slot?
	jnz     NoNE2000Card
	jmp     SlotIsOkay

NoNE2000Card:
	mov     dx, offset CGroup:NoNE2000InMachine
	jmp     InitErrorExit

;----------------------------------------------------------------------------
;       Set up NE2000 to run in 8 bit Slot.
;----------------------------------------------------------------------------
NE2000In8BitSlot:
	mov     PSTOP, 60h
	mov     EightBitSlot, 1

	lea     dx, [bp].NICDataConfiguration   
	mov     al, 48h
	out     dx, al

;----------------------------------------------------------------------------
;       Adjust RunTimeCode
;----------------------------------------------------------------------------
	lea     di, RunTimeCode
	mov     ax, 01ebh               ; 
	stosw
	mov     al, 40h                 ; 
	stosb

;----------------------------------------------------------------------------
;       Adjust RunTimeCode0
;----------------------------------------------------------------------------
	lea     di, RunTimeCode0
	mov     ax, 0eeach              ;
	stosw
	mov     ax, 0fce2h              ;
	stosw
	mov     ax, 0c033h              ;
	stosw
	mov     ax, 00ebh               ;
	stosw

;----------------------------------------------------------------------------
;       Adjust RunTimeCode1
;----------------------------------------------------------------------------
	lea     di, RunTimeCode1
	mov     ax, 0eeach              ;
	stosw
	mov     ax, 0fce2h              ;
	stosw
	mov     ax, 0e432h              ;
	stosw
	mov     ax, 05ebh               ;
	stosw

;----------------------------------------------------------------------------
;       Adjust RunTimeCode2 
;----------------------------------------------------------------------------
	lea     di, RunTimeCode2
	mov     ax, 0AAECh              ; in al, dx ; stosb
	stosw           
	mov     ax, 0FCE2h              ; Loop RunTimeCode2
	stosw

	lea     di, RunTimeCode3
	mov     ax, 0AEBh               ; jmp short 
	stosw

;----------------------------------------------------------------------------
;       Adjust RunTimeCode4 
;----------------------------------------------------------------------------
	lea     di, RunTimeCode4
	mov     ax, 06E3h               ; jcxz MoveWasEven
	stosw
	mov     ax, 0AAECh              ; in al, dx; stosb
	stosw
	mov     ax, 0FCE2h              ; Loop ReadBytePort
	stosw
	mov     ax, 00ebh               ; jmp short MoveWasEven
	stosw
	mov     ax, 0c383h              ; add bx. 6
	stosw
	mov     ax, 0b06h               ; or bp, bp
	stosw
	mov     ax, 75edh               ; jnz FillECBLoop
	stosw
	mov     ax, 59dah               ; pop CX
	stosw
	mov     ax, 0e9d1h              ; shr cx, 1
	stosw
	mov     ax, 0273h               ; jnc NoExtraByte
	stosw
	mov     ax, 40ech               ; in al, dx ; inc ax
	stosw

SlotIsOkay:
	mov     al, PSTART
	inc     al
	mov     NextPage, al                    ;Reset the next page 

	call    DoRAMTest
	jnz     InitErrorExit

	call    ResetNIC                        ; Call common init routine
	call    MSMSetIRQ                       ; Hook the interrupt vector
						;  and enable interrupt.

        call    VerifyIRQ                       ; Call and verify IRQ setting
    	jnz     InitErrorExit                   ; if fails error out (dx -> message)

;----------------------------------------------------------------------------
;       Startup our AES polling procedure.
;----------------------------------------------------------------------------
	push    cs
	pop     es
	mov     si, offset CGroup:AsyncECB
	mov     [si+2].AESESR, cs
	mov     bx, MLIDSUP_SCHEDULE_AES_EVENT
	call    LSLSupport
	jmp     MSMMediaConfigUpdate                    ; No error here, pass on

InitErrorExit:
	mov     ax, MSMERR_HSM_FAILED           ; Return Error
	or      ax, ax                          ; Set appropriate flags
	ret
	
DriverInit      endp


;
;       DoRAMTest
;
;       Routine to test out the NICs shared RAM.
;
;       assumes:        ds      is set to CGroup
;                       bp      Base I\O
;                       Interrupts enabled.
;
;       returns:        z flag  if no errors
;                       nz flag if errors
;                       dx      -> error message (only if nz flag)
;                       bx,ds,int preserved.
;

DoRAMTest       proc    near

	lea     dx, [bp].NICTxConfiguration     ;Put NIC in loopback mode
	mov     al, 2
	out     dx, al

	cmp     cs:SharedRAMFlag, 0             ; Shared RAM Mode?
	je      IOModeRAMTest
	jmp     RAMModeRAMTest

IOModeRAMTest:
;       Start at the Tx page memory and go to the end of the Rx memory.

	mov     dx, StartAddressLow
	mov     ax, TXPAGE*100h
	out     dx, al
	inc     dx
	mov     al, ah
	out     dx, al

	xor     ax, ax
	mov     dx, RemoteByteCountLow          ;How many bytes
	mov     ah, PSTOP                       ;Total memory
	sub     ah, TXPAGE
	mov     cx, ax
	push    ax                              ;Save memory length

	out     dx, al
	inc     dx
	mov     al, ah
	out     dx, al

	mov     dx, Command
	mov     al, NICPage0DMAWrite
	out     dx, al

	mov     dx, DataPort                    ;Transfer port
	mov     ax, 0a5a5h                      ;Fill word

	cmp     EightBitSlot, 1
	jz      FillRAMLoopByte

	shr     cx, 1

FillRAMLoop:
	out     dx, ax
	slow                                    ;TNL 2/8/95
	loop    FillRAMLoop

;       Now read the data in and make sure it's good.

ReadRAMData:
	mov     dx, StartAddressLow
	mov     ax, TXPAGE*100h
	out     dx, al
	inc     dx
	mov     al, ah
	out     dx, al

	mov     dx, RemoteByteCountLow          
	pop     ax                              ;Get amount of memory
	mov     cx, ax
	out     dx, al
	inc     dx
	mov     al, ah
	out     dx, al

	mov     dx, Command
	mov     al, NICPage0DMARead
	out     dx, al

	mov     dx, DataPort                    ;Transfer port
	mov     bx, 0a5a5h

	cmp     EightBitSlot, 1
	jz      CheckRAMLoopByte

	shr     cx, 1

CheckRAMLoop:
	in      ax, dx
	slow                                    ;TNL 2/8/95
	cmp     bx, ax                          ;Good data?
	jne     RAMLooksBad

	loop    CheckRAMLoop
	xor     ax, ax                          ;Set z flag
						;z flag will be set correctly
RAMLooksBad:
	mov     dx, offset CGroup:RAMFailureMsg
	ret

FillRAMLoopByte:
	out     dx, al
	loop    FillRAMLoopByte
	jmp short ReadRAMData

CheckRAMLoopByte:
	in      al, dx
	cmp     bl, al                          ;Good data?
	jne     RAMLooksBad

	loop    CheckRAMLoopByte
	xor     ax, ax                          ;Set z flag
	jmp short DoRAMTestExit

RAMModeRAMTest:
	mov     es, RAMSegment                  ;Get our RAM's segment value
	xor     di, di                          ;Offset is 0

	cmp     EightBitSlot, 1
	jz      TestRAMModeLoopByte

	mov     cx, 4000h / 2                   ;Number of words to write
	push    cx
	mov     ax, 0a5a5h                      ;Fill pattern

   rep  stosw                                   ;Do the fill

	xor     di, di                          ;-> at front of RAM
	pop     cx                              ;Number of words to compare

   repe scasw                                   ;Compare the data
						;z flag will be set correctly
	mov     dx, offset CGroup:RAMFailureMsg
	ret

TestRAMModeLoopByte:
	mov     cx, 2000h                       ;Number of bytes to write
	push    cx
	mov     ax, 0a5a5h                      ;Fill pattern

   rep  stosb                                   ;Do the fill

	xor     di, di                          ;-> at front of RAM
	pop     cx                              ;Number of words to compare

   repe scasb                                   ;Compare the data
						;z flag will be set correctly
	mov     dx, offset CGroup:RAMFailureMsg

DoRAMTestExit:
	push    ax
	pushf
	lea     dx, [bp].NICTxConfiguration     ;Take NIC out ofloopback mode
	xor     al, al                  
	out     dx, al
	.popf
	pop     ax

	ret

DoRAMTest       endp


;****************************************************************************
;       VerifyIRQ
;
;       this routin will verify that the IRQ level selected by the 
;       user matches the IRQ that the hardware is configured for.  We 
;       do this by configuring the card for internal loopback then we
;       do a transmit to force an interrupt event.  If our ISR gets 
;       called then everything's fine. otherwise we flag an error
;       condition.
;
;       assumes:        ds is set to CGroup
;                       interrupts Disabled.
;
;       returns:        z flag if no errors
;                       nz flag if errors
;                       dx error message if nz flag
;

public VerifyIRQ

VerifyIRQSharedRAMJmp:
	jmp     VerifyIRQSharedRAM


VerifyIRQ       proc    near

	mov     bp, IOAddress1

;----------------------------------------------------------------------------
; for some reason the loopback test won't work for shared RAM mode.  However 
; we can still verify the IRQ in shared RAM mode by reading the  Mode 
; Configuration Register A.
;----------------------------------------------------------------------------

	cmp     SharedRAMFlag, 0                ; Shared RAM Mode?
	jne     VerifyIRQSharedRAMJmp           ; jump if so
						; else, PIO mode.

;----------------------------------------------------------------------------
; At this point we are mostly initialized and the transmit configuration 
; register is already programmed for internal loopback.  However, we need 
; to adjust the Data Configuration and Receive Configuration registers for
; loopback testing.
;----------------------------------------------------------------------------

	pushf                                   ; save interrupt state
	cli

	mov     dx, DataConfiguration           ; Data Configuration register
	mov     al, 41h                         ; set to loop back mode
	out     dx, al
			    
	lea     dx, [bp].NICRxConfiguration     ; Receive Configuration register
	mov     al, 1Fh                         ; set to prom. mode temporarily.
	out     dx, al

	mov     dx, InterruptStatus             ; Interrupt status register
	mov     al, 0FFh                        ; make sure all ints cleared
	out     dx, al

	lea     dx, [bp].NICTxConfiguration     ;Put in loopback for a moment
	mov     al, 02h
	out     dx, al

;----------------------------------------------------------------------------
; Now we just need to set the Transmit Page and Byte Counts and initiate the 
; transmit.  Since we don't really care about the transmit data we won't 
; bother copying any particular data to the NIC.
;----------------------------------------------------------------------------

	mov     dx, TransmitPage                ; Transmit Page Start
	mov     al, TransmitPage0               ; set to Tx Page 0
	out     dx, al

	inc     dx                              ; TxByteCount0
	mov     ax, 40h                         ; Tx size LSB (use 64 bytes)
	out     dx, al                          ; send LSB of tx byte count

	inc     dx                              ; TxByteCount1
	mov     al, ah                          ; Tx size MSB
	out     dx, al                          ; send MSB of tx byte count

	mov     dx, Command                     ; Command register
	mov     al, NICTransmit                 ; send packet
	out     dx, al

;----------------------------------------------------------------------------
; Idle about 16 ms to allow loopback Tx to finish.  If interrupt occurs
; then DriverStatRxChecksum from the stat table will be non-zero
;----------------------------------------------------------------------------

	sti                                     ; enable interrupts momentarily
	call    ReadTickCounter                 ;Get initial value
	mov     dx, ax

WaitForTxToFinish:
	call    ReadTickCounter
	neg     ax
	add     ax, dx
	cmp     ax, 4000 * 10                   ;16000 ms (419 ns per count) plus somes
	jb      WaitForTxToFinish

	.popf
	cmp     MSMTxFreeCount, 2               ; check counter to see if IRQ fired
	je      VerifyIRQError

VerifyIRQContinue:
	mov     MSMTxFreeCount, 2               ; restore counters

;----------------------------------------------------------------------------
; The IRQ is correctly configured. 
;     Call ResetNIC.
;----------------------------------------------------------------------------

      	mov     dx, InterruptStatus             ; Interrupt status register
      	mov     al, 0FFh                        ; make sure all ints cleared
      	out     dx, al

      	mov     dx, Command                     ; Command Register
      	mov     al, NICPage0Stop                ; set to page 0
      	out     dx, al

      	mov     dx, DataConfiguration           ; DataConfiguration register
      	mov     al, 49h                         ; Set back to normal mode.
      	out     dx, al

      	lea     dx, [bp].NICTxConfiguration     ;Set Tx mode to normal
      	xor     al, al
      	out     dx, al

	lea     dx, [bp].NICRxConfiguration
	mov     al, 0Ch                         ;Broadcasts & Multicasts enabled
	out     dx, al

      	call    MSMUpdateMulticast              ; this will set the rx configuration register to default

VerifyIRQSuccess:
	xor     dx, dx                          ; return successful
	jmp short LeaveVerifyIRQ

;----------------------------------------------------------------------------
; this is a shared RAM card so just check irq value
;----------------------------------------------------------------------------
VerifyIRQSharedRAM:
	add     bp, 10
	xor     ax, ax
	mov     dx, RemoteByteCountLow          ; a read from RBCR0 will give us mode config register A
	in      al, dx
	and     al, INT_MASK                    ; mask off INT bits
	shr     al, 3                           ; make it an index

	mov     ah, SharedRAMINTArray           ; convert to IRQ value
	cmp     ah, IRQLine1                    ; do they match
	je      VerifyIRQSuccess

VerifyIRQError:

	mov     dx, offset CGroup:IRQInvalidMessage ; message to display when error out

LeaveVerifyIRQ:

	or      dx, dx
	ret

VerifyIRQ       endp

Init    ends
	end
