;----------------------------------------------------------------------------
;   (C) Unpublished Work of Novell, Inc.  All Rights Reserved.
;
;	THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL, 
;	PROPRIETARY AND TRADESECRET INFORMATION OF NOVELL, INC.  ACCESS 
;	TO THIS WORK IS RESTRICTED TO (I) NOVELL EMPLOYEES WHO HAVE A
;	NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE OF THEIR ASSIGNMENTS 
;	AND (II) ENTITIES OTHER THAN NOVELL WHO HAVE ENTERED INTO APPROPRIATE 
;	LICENSE AGREEMENTS.  NO PART OF THIS WORK MAY BE USED, PRACTICED,
;	PERFORMED, COPIED, DISTRIBUTED, REVISED, MODIFIED, TRANSLATED,
;	ABBRIDGED, CONDENSED, EXPANDED, COLLECTED, COMPILED, LINKED, RECAST,
;	TRANSFORMED OR ADAPTED WITHOUT THE PRIOR WRITTEN CONSENT OF NOVELL. 
;	ANY USE OR EXPLOITATION OF THIS WORK WITHOUT AUTHORIZATION COULD 
;	SUBJECT THE PERPERTRATOR TO CRIMINAL AND CIVIL LIABILITY.
;
;****************************************************************************
;	
;	82586 Structures and Equates
;	FileName:	82586def.inc
;	Date: 		November 2, 1991
;	Written By:	Jon Rosdahl
;
;	Modification  History:
;----------------------------------------------------------------------------
; 19 May 1992 RTY
;	Changed	RxBufferSize from 1520 to 3040 to allow for big packets.
;	this was done for Receive Monitor.
;----------------------------------------------------------------------------
;****************************************************************************

;******************************************************************************
;	82586  Network Coprocessor Commands, Ports, and Registers
;******************************************************************************

;------------------------------------------------------------------------------
;	System Control Block (SCB)
;------------------------------------------------------------------------------

SCBstruc	struc
	SCBStatus	dw  ?	;  0 - Status of the 82586
	SCBCommand	dw  ?	;  2 - Specifies action to be done @ next CA
	SCBCBL		dw  ?	;  4 - 1st Action Command in Command list if CUC= CU-Start (Offset)
	SCBRFA		dw  ?	;  6 - 1st Receive Frame Descriptor in RFA  (Offset)
	SCBCRCerrs	dw  ?	;  8 - CRC errors - properly aligned frames w/CRC error
	SCBALNerrs	dw  ?	; 10 - Alignment Errors - Misaligned frames w/CRC error
	SCBRSCerrs	dw  ?	; 12 - Resource Errors - # correct frames discarded due to lack of memory
	SCBOVRNerrs	dw  ?	; 14 - Overrun Errors - # received frames lost due to memory bus not available in time
SCBstruc ends

;------------------------------------------------------------------------------
;	SCB_Status word definitions 
;------------------------------------------------------------------------------
;			  STAT  CUS  RUS
;  SCB_Status word =      xxxx 0xxx 0xxx 0000

;STAT
  SCBSTAT_CX		equ	8000h	; Command in the CBL has been executed
  SCBSTAT_FR		equ	4000h	; A Frame has been Received
  SCBSTAT_CNA		equ	2000h	; The Command Unit left the Active State
  SCBSTAT_RNA		equ	1000h	; The Receive Unit left the Ready State

;CUS
  SCBCUS_IDLE		equ	0000h	; CUS = 0 => CU is IDLE
  SCBCUS_SUSPEND	equ	0100h	; CUS = 1 => CU is SUSPENDED
  SCBCUS_ACTIVE		equ	0200h	; CUS = 2 => CU is ACTIVE

;RUS
  SCBRUS_IDLE		equ	0000h	; RUS = 0 => RU is IDLE
  SCBRUS_SUSPEND	equ	0010h	; RUS = 1 => RU is SUSPENDED
  SCBRUS_NoRESOURCE	equ	0020h	; RUS = 2 => RU has NO RESOURCES
  SCBRUS_READY		equ	0040h	; RUS = 4 => RU is READY

;------------------------------------------------------------------------------
;	SCB_COMMAND word definitions 
;------------------------------------------------------------------------------
;			  ACK   CUC  RUC
;  SCB_COMMAND word =     xxxx 0xxx 0xxx 0000

;ACK
  SCBACK_CX		equ	8000h	; Acknowledges the command executed event
  SCBACK_FRx		equ	4000h	; Acknowledges the Frame Received event
  SCBACK_CNA		equ	2000h	; Acknowledges that the Command Unit became not ready
  SCBACK_RNR		equ	1000h	; Acknowledges that the Receive Unit became not ready

;CUC
  SCBCUC_NOP		equ	0000h	; CUC = 0 => CU NOP
  SCBCUC_START		equ	0100h	; CUC = 1 => CU Start
  SCBCUC_RESUME		equ	0200h	; CUC = 2 => CU RESUME
  SCBCUC_SUSPEND	equ	0300h	; CUC = 3 => CU SUSPENDED
  SCBCUC_ABORT		equ	0400h	; CUC = 4 => CU ABORT

;RUC
  SCBRUC_NOP		equ	0000h	; RUC = 0 => RU NOP 
  SCBRUC_START		equ	0010h	; RUC = 1 => RU Start
  SCBRUC_RESUME		equ	0020h	; RUC = 2 => RU RESUME
  SCBRUC_SUSPEND	equ	0030h	; RUC = 3 => RU SUSPENDED
  SCBRUC_ABORT		equ	0040h	; RUC = 4 => RU ABORT


;------------------------------------------------------------------------------
;	SCB Misc Working Definitions
;------------------------------------------------------------------------------

  SCB_CUneedsAttn	equ    SCBACK_CX  or SCBACK_CNA	; Command unit needs attention (Send)
  SCB_RUneedsAttn	equ    SCBACK_FRx or SCBACK_RNR ; Receive unit needs attention
  SCB_IntAckMask	equ    0F000h			; Acknowledge bit mask masks off CUC and RUC
  Reset_OKbits		equ    0A000h			; Reset is OK	
 
;------------------------------------------------------------------------------
;	System Configuration Pointer (SCP) 
;------------------------------------------------------------------------------

SCPstruc struc
	SCPBusSize	dw	?	; system bus size
	SCPRes		dd	?	; reserved space
	SCPISCPoff	dw	?	; iscp offset (16/24 bits physical address)
	SCPISCPseg	dw	?	; iscp base   (last 8/24 bits physical address)
SCPstruc ends 

  BUS_WIDTH_8		equ	1	; 8-bit bus
  BUS_WIDTH_16		equ	0	; 16-bit bus
  SCP_Location		equ	3FF6h	; Offset in window

  
;------------------------------------------------------------------------------
;	Intermediate System Configuration Pointer (ISCP)
;------------------------------------------------------------------------------

ISCPstruc struc 
	ISCPbusy	dw	?	;initialization indicator
	ISCPSCBoff	dw	?	;offset of scb in the memory from base
	ISCPSCBbase	dw	?	;24-bit value giving the starting address 
	ISCPSCBbasehi	dw	?	;of the 64-k segment containing all structures
ISCPstruc ends


;------------------------------------------------------------------------------
;------------------------------------------------------------------------------
; 	Command Block (CmdBlk) -- Generic
;------------------------------------------------------------------------------
;------------------------------------------------------------------------------

CmdBlkstruc struc
	CmdBlkStatus	dw	?	;command completion code
	CmdBlkCommand	dw	?	;command code
	CmdBlkLink	dw	?	;next command block (Offset)
	CmdBlkData	db 36 dup(?)	;command related data
CmdBlkstruc ends

;Status word
  C_BIT		equ	8000H	;completion bit
  B_BIT		equ	4000H	;busy bit
  OK_BIT	equ	2000H	;error-free bit
  A_BIT		equ	1000H	;aborted bit
  CRC_BIT	equ	0800H	;CRC error
  ALLIGN_BIT	EQU	0400h	;Allignment error.

;COMMAND word
  EL_bit		equ	8000H	;last command block on list
  S_bit			equ	4000h	; Suspend after completion of this command
  I_bit			equ	2000H	;interrupt host after completion
  CMD_NOP		equ	0	;no-operation
  CMD_IASETUP		equ	1	;individual address setup
  CMD_CONFIGURE		equ	2	;configuration (options)
  CMD_MULTICAST		equ	3	;multicast address setup
  CMD_TRANSMIT		equ	4	;transmit a packet
  CMD_TDR		equ	5	;Time Domain Reflectometer
  CMD_DUMP		equ	6	;dump chip configuaration
  CMD_DIAGNOSE		equ	7	;internal diagnostic test

;------------------------------------------------------------------------------
; Individual Address command block	(IA)
;------------------------------------------------------------------------------

IAstruc struc
			dw 3 dup(?)	; common fields for all commands
	IANodeAddress	db 6 dup(?)	; ethernet address
IAstruc ends


;------------------------------------------------------------------------------
; 	Configuaration Command Block	(CFG)
;------------------------------------------------------------------------------

CFGstruc struc
			dw 3 dup(?)	;common fields for all commands
        CFGbyte6        db      ?       ;Following parameter Byte Count bits 0-3
        CFGbyte7        db      ?       ;Fifo-Lim bits 0-3 - Value of TX-FIFO threshold
        CFGbyte8        db      ?       ;SRDY/ARDYbar bit 6, Sav-BF bit 7
        CFGbyte9        db      ?       ;Addr-Len bits 0-2, Al-Loc bit 3, Pream-len bits 4-5, Loopback bits 6,7
        CFGbyte10       db      ?       ;Lin-Prio bits 0-2, ACR bits 4-6, BOF_MET bit 7
        CFGbyte11       db      ?       ;InterFrame-Spacing 0-7
        CFGbyte12       db      ?       ;slot_time(l) bits 0-7
        CFGbyte13       db      ?       ;slot_time(h) bits 0-2, Retry-Num bits 4-7 (3 reserved)
        CFGbyte14       db      ?       ;PRM-0, BC_DIS-1, Manch/NRZbar-2, Tono_CRS-3, NCRC_Ins-4, CRC_16-5, BT_STF-6, pad-7
        CFGbyte15       db      ?       ;CRSF 0-2, CRS_SRC 3, CDTF 4-6, CDT-SRC 7
        CFGbyte16       db	?       ;Min_FRM_LEN bits 0-7
        CFGbyte17       db      ?       ;reserved
CFGstruc ends

;------------------------------------------------------------------------------
; 		Configuration  Bit Definitions 
;------------------------------------------------------------------------------

  CFG_ByteCnt		equ	12	; Default parameter byte count ( number bytes in Config)
  CFG_FIFOLIM		equ	8	; FIFO threshold 
  CFG_SRDYbit		equ	40h	; 1=External Synch 0=Internal Synch
  CFG_SavBF		equ	80h	; 1=Bad Frames are saved 0=Not Saved
  CFG_AddrLen		equ	6	; Address length
  CFG_AddrLoc		equ	08h	; 0=82586 inserts source @ 08h=82586 doesn't insert source address in frame
  CFG_Preamble_Len	equ	20h	; preamble length ( xx10b => 8 bytes ) 
  CFG_IntLoopBack	equ	40h	; Internal Loop-Back
  CFG_ExtLoopBack	equ	80h	; External Loop-Back (and if Internal not set)
  CFG_LinPrio		equ	00h	; Linear Priority method (0=802.3 method)
  CFG_ACR		equ	00h	; Accelerated Contention Resolution (0=802.3 method)
  CFG_BackOffMethod	equ	00h	; 0=IEEE-802.3   1=Alternate Method
  CFG_InterFrameSpace	equ	96	; Interframe Spacing 96 bits
  CFG_SlotTime		equ	512	; slot time   512 bits
  CFG_Retry_Num		equ	15*10h	; Maximum Retry on collisions (position bits)
  CFG_PRMbit		equ	01h	; Promiscuous Mode flag
  CFG_BCdis		equ	02h	; Broadcast Disable bit
  CFG_Manch		equ	04h	; 1=Manchester 0=NRZ encoding/decoding
  CFG_TONO		equ	08h	; 1=Continue Transmission even if no Carrier Sense
  CFG_NoCRCinsert	equ	10h	; 1=no Insert  0=default=Append CRC to frame
  CFG_CRCtype16		equ	20h	; 1=16-bit CCITT CRC polynomial   0=32-bit Autodin II CRC polynomial
  CFG_BitStuff		equ	40h	; 1=HDLC like bitstuffing  0=default=End of Carrier mode (IEEE 802.3)
  CFG_Pad		equ	80h	; 1=Perform pad by transmitting Flags for remainder of slot time 0=no pad
  CFG_CRSF		equ	00h	; 0 bits default -- Carrier Sense Filter in bit times
  CFG_CRSsource		equ	08h	; 0=default=External source  1=Internal Carrier Sense Source
  CFG_CDTF		equ	00h	; 0 bits default -- Collision Detect Filter in bit times
  CFG_CDTsource		equ	80h	; 0=default=External source  1=Internal Collision Detect Source
  CFG_MinFrame		equ	64	; minimum frame length
  PRM_MinFrame		equ	14	; minimun frame length in promiscuous mode.


;------------------------------------------------------------------------------
;		Transmit Command Block (TxCmdBlk)
;------------------------------------------------------------------------------
				; All Transmit Structures are Grouped into One Structure
				; Don't Move Blocks around
TxBlockStruc struc
     ;TxCmdBlkstruc struc
	TxCmdBlkStatus	 dw	?	; Extra bits are defined for Tx
	TxCmdBlkCmd	 dw	?	; Same for most CmdBlks
	TxCmdBlkLnk	 dw	?	; Common link -> next Command Block (Offset)
	TxCmdBlkTxBufDes dw	?	; TxBufDes offset 
        TxCmdBlkDest	 db 6 dup(?)	; Destination address 
	TxCmdBlkLen      dw	?	; Length Field 	      
     ;TxCmdBlkstruc ends	

;------------------------------------------------------------------------------
;	Transmit Buffer Descriptor (TxBufDes)
;------------------------------------------------------------------------------

     ;TxBufDesStruc struc
	TxBufDesSize	 dw	?	;size of xmit buffer (xxSSSSSSSSSSSSSSb)
	TxBufDesLnk	 dw	?	;next TxBufDes
	TxBufDesBuffl	 dw	?	;Absolute Addr. to xmit buffer LS 4 hex digits
	TxBufDesBuffh	 dw	?	;Absolute Addr. to xmit buffer MS 2 Hex digits
     ;TxBufDesstruc ends

;------------------------------------------------------------------------------
;	Special Spot for Firmware use Added to the Book's Structure
;------------------------------------------------------------------------------

	TxRetryCount	dw	?		; Counts the number of times I've Asked the chip to send it.

;------------------------------------------------------------------------------
;	BMTCBFragPtrStruc
;------------------------------------------------------------------------------

	BMTCBFragPtrEnd	   db (6*NumBMTCBFrags) dup (0) ; 6 bytes per pointer

;------------------------------------------------------------------------------
;	BMTCBStructure Data Variables
;------------------------------------------------------------------------------

	BMTCB586FrameLen   dw	0	; Value in CX when DriverSend Called
	BMTCB586DataLen	   dw	0	; Value in TCBdataLen
	BMTCB586Frags	   dw	0	; Number of Valid BMTCB Frag Ptrs filled in above

	;BMTCBTxBuffer  db	TxCollSizeDef dup (0) ; Overlaps TBdata to Always Auto move first Fragment.
					; Don't uncomment it... Listed here for comment only.

;------------------------------------------------------------------------------
; Transmit Buffer (TB)
;------------------------------------------------------------------------------
    
   ;TBstruc struc
	TBData		 db 1514 dup(?) ;Enough for one frame
   ;TBstruc ends
TxBlockStruc 	ends

;------------------------------------------------------------------------------
;		Transmit Equates
;------------------------------------------------------------------------------

  TxCmdBlk_NoSenseSig		equ	0400H	;
  TxCmdBlk_NoClrSndSig		equ	0200H	;
  TxCmdBlk_DMAunderRUN		equ	0100H	;
  TxCmdBlk_DeferTx		equ	0080H	;
  TxCmdBlk_HeartBeat		equ	0040H	;
  TxCmdBlk_MAXCOLLSNS		equ	0020H	;
  TxCmdBlk_COLLSNS_CNT		equ	000FH	;
  TxBufDes_EOFbit		equ	8000H	; Last buffer for this frame (End of Frame flag)
  TxBMICTransferLoc		equ	BMTCB586FrameLen ; Spot the BMIC will put the BMTCB 


;------------------------------------------------------------------------------
;		Receive Equates
;------------------------------------------------------------------------------

	RFD_CRCErrbit	equ	0800H	; crc error bit
	RFD_ALNErrbit	equ	0400H	; alignment error bit
	RFD_RSCErrbit	equ	0200H	; out-of-resources bit
	RFD_OVRErrbit	equ	0100H	; over-run error bit
	RFD_Shortbit 	equ	0080H	; short frame
	RFD_EOFErrbit	equ	0040H	; No eof flag detected bit

	RBD_EOFbit	equ	8000H	; last buffer holding the packet
	RBD_F_bit	equ	4000H	; actual count field valid
	RBD_LenMask	equ	3FFFH	; mask for packet length
	RxBufferSize	equ	1520*2    ; Max Rx Packet + MAC Header + 6 bytes just incase
 

;------------------------------------------------------------------------------
; Receive Frame Descriptor (RFD)
;------------------------------------------------------------------------------
				; All the Receive Blocks are combined into 1 Stucture.
				; Don't change the order
RxBlockStruc struc
    ; RFDstruc struc
	RFDStatus	dw	?	;receive packet status
	RFDCommand	dw	?	;
	RFDLink		dw	?	;next rfd on the link-list
	RFDRBD		dw	?	;offset of the 1st RBD
	RFDDestAddr	db	6 dup(?) ; Destination Network address
	RFDSrcAddr	db	6 dup(?) ; Source Network Address
	RFDTyp_Len	dw	?	 ; Type/Length Field
    ; RFDstruc ends

;------------------------------------------------------------------------------
;	For BMIC LookAhead
;------------------------------------------------------------------------------

	BMFrameSize	dw	?	; Size of the Incoming Frame

;------------------------------------------------------------------------------
;		Receive Buffer (RBuff)
;------------------------------------------------------------------------------

    ; RBuffstruc struc
	RBuffData	db RxBufferSize dup(?)	;receive buffer Extra Just incase of error
    ; RBuffstruc ends

;------------------------------------------------------------------------------
; 	Receive Buffer Descriptor (RBD)
;------------------------------------------------------------------------------

    ; RBDstruc struc
	RBDStatus	dw	?	;frame size in the receive buffer
	RBDLink		dw	?	;next rbd on the link-list
	RBDbufflo	dw	?	;receive buffer address (low)
	RBDbuffhi	dw	?	;receive buffer address (high)
	RBDbuffSize	dw	?	;receive buffer length
    ; RBDstruc ends

RxBlockStruc	ends


;------------------------------------------------------------------------------
; Multicast Control Block (mcb)
;------------------------------------------------------------------------------

MAX_MULTICAST_ADDR	equ	16		  ; Driver Defined Maximum (ISO supported Number is 16)

MCBstruc struc
		dw 3 dup(?)			  ; common Fields for all commands
	MCBCnt	dw	0			  ; Number of Addresses being submitted
	MCBlist	db (6 * MAX_MULTICAST_ADDR) dup (0) ; List of MultiCast Addresses
MCBstruc ends


