;****************************************************************************
; BMICMACS.INC
;
;  Copyright Unpublished Work of Novell, Inc. All Rights Reserved
;  
;  THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
;  PROPRIETARY AND TRADE SECRET INFORMATION OF NOVELL, INC.   
;  ACCESS TO THIS WORK IS RESTRICTED TO (i) NOVELL EMPLOYEES  
;  WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE  
;  OF THEIR ASSIGNMENTS AND (ii) ENTITIES OTHER THAN NOVELL   
;  WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.      
;  NO PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED,
;  DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED,      
;  CONDENSED, EXPANDED, COLLECTED, COMPILED, LINKED, RECAST,  
;  TRANSFORMED OR ADAPTED WITHOUT THE PRIOR WRITTEN CONSENT OF
;  NOVELL.  ANY USE OR EXPLOITATION OF THIS WORK WITHOUT AUTHORIZATION
;  COULD SUBJECT THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY.
;****************************************************************************
;	Title:		82355 BMIC Macros ADAPTER SIDE
;	FileName:	BMICMACS.INC
;	Written by:	Don Shultz
;	Date:		4-30-91
;
;	Modification  History:
;	By			Date		Modification Description
;	-------------------	---------	-----------------------------
;	Jon Rosdahl		5-22-91		TASM Macros cannot have comments between
;						the Macro header and the Variable statements
;						Corrected the Problem.	
;
;	Don Shultz		6-25-91		Mask off the Reserved bits on a transfer complete
;
;	Jon Rosdahl		7-09-91		Removed Write to the Status Register Before every Transfer.
;						It is sufficient immediately after the transfer.
;
;****************************************************************************


;TransferDataToHost	macro
;
; On Entry:	ESI -> 7 byte Host Count/Address(Destination).
;		DI -> local address(Source).
;
TransferDataToHost	macro
	local	WaitForTransferToHost

	mov	dx, BMIC_INDEX			; Set index to Transfer
	mov	al, CH0_TBI_BASE_ADDR OR AUTO_INC	; Buffer Interface channel 0.
	out	dx, al
	mov	dx, BMIC_DATA			; DX = BMIC Data port.

	mov	ax, di				; AX -> Source.
	shr	ax, 1				; AX = Source Address/2.

	out	dx, al				; Set Source LSB.
	mov	al, ah				; AL = Source MSB.
	out	dx, al				; Set Source MSB.

	mov	dx, BMIC_INDEX			; Set index to channel 0
	mov	al, CH0_BASE_COUNT OR AUTO_INC	; Base count.
	out	dx, al
	mov	dx, BMIC_DATA			; DX = BMIC Data port.

	mov	cx, 7				; Copy 7 bytes.
 rep	outsb

	mov	dx, BMIC_INDEX			; Set index to Channel 0
	mov	al, CH0_STROBE			; Strobe register.
	out	dx, al
	mov	dx, BMIC_DATA			; DX = BMIC Data Port.
	out	dx, al				; No data required for strobe.

	mov	dx, BMIC_INDEX			; Set index to Channel 0
	mov	al, CH0_STATUS			; Status.
	out	dx, al
	mov	dx, BMIC_DATA			; DX = BMIC Data Port.

WaitForTransferToHost:
	in	al, dx				; AL = current status.
	test	al, TRANSFER_COMPLETE		; Transfer complete?
	jz	WaitForTransferToHost		; Jump if not.
;;;;;
;;	and	al, 00011111b			;# Mask out reserved Bits.
and	al, 00000001b			;# Mask out reserved Bits.
	out	dx, al				; Reset Transfer Complete bit.
	
			endm

;TransferDataFromHost	macro
; Uses 	SI, DI, DX, AX, CX
; Destroys AX, DX, CX, SI
; Preserves DI
; On Entry:	SI -> 7 byte Host Count/Address(Source).
;		DI -> local address(Destination).
;
TransferDataFromHost	macro
	local	WaitForTransferFromHost

	mov	dx, BMIC_INDEX			; Set index to Transfer
	mov	al, CH0_TBI_BASE_ADDR OR AUTO_INC	; Buffer Interface channel 0.
	out	dx, al
	mov	dx, BMIC_DATA			; DX = BMIC Data port.

	mov	ax, di				; AX -> Dest.
	shr	ax, 1				; AX = Dest Address/2.

	out	dx, al				; Set Dest LSB.
	mov	al, ah				; AL = Dest MSB.
	out	dx, al				; Set Dest MSB.

	mov	dx, BMIC_INDEX			; Set index to channel 0
	mov	al, CH0_BASE_COUNT OR AUTO_INC	; Base count.
	out	dx, al
	mov	dx, BMIC_DATA			; DX = BMIC Data port.

	mov	cx, 7
 rep	outsb

	mov	dx, BMIC_INDEX			; Set index to Channel 0
	mov	al, CH0_STROBE			; Strobe register.
	out	dx, al
	mov	dx, BMIC_DATA			; DX = BMIC Data Port.
	out	dx, al				; No data required for strobe.

	mov	dx, BMIC_INDEX			; Set index to Channel 0
	mov	al, CH0_STATUS			; Status.
	out	dx, al
	mov	dx, BMIC_DATA			; DX = BMIC Data Port.

WaitForTransferFromHost:
	in	al, dx				; AL = current status.
	test	al, TRANSFER_COMPLETE		; Transfer complete?
	jz	WaitForTransferFromHost		; Jump if not.

	and	al, 00011111b			;# Mask out reserved bits.
	out	dx, al				; Reset Transfer Complete bit.
	
			endm

;PeekFromHost	macro

;
; On Entry:	ESI -> 7 byte Host Count/Address(Source).
;		DI -> local address(Destination).
;
PeekFromHost	macro
	local	WaitForPeek

	mov	dx, BMIC_INDEX			; Set index to Peek Poke
	mov	al, PEEK_POKE_ADDR OR AUTO_INC	; address.
	out	dx, al
	mov	dx, BMIC_DATA			; DX = BMIC Data Port.

	mov	cx, 4				; Write four bytes.
 rep	outsb

	mov	dx, BMIC_INDEX			; Set index to Peek Poke
	mov	al, PEEK_POKE_CONTROL		; control register.
	out	dx, al

	mov	dx, BMIC_DATA			; Set to peek from memory.
	mov	al, 01011111b
	out	dx, al

	mov	dx, BMIC_STATUS			; DX = BMIC Status/Control.
WaitForPeek:
	in	al, dx				; AL = current status.
	test	al, PEEK_POKE_PENDING		; Peek in process?
	jnz	WaitForPeek			; Jump if so.

	mov	dx, BMIC_INDEX			; Set index to peek/poke
	mov	al, PEEK_POKE_DATA OR AUTO_INC	; data register.
	out	dx, al
	mov	dx, BMIC_DATA			; DX = BMIC Data Port.

	mov	cx, 4				; Read four bytes.
 rep	insb

		endm

InterruptTheHost	macro
				; BL Holds Interrupt Bit to Set.
				; AX, DX Destroyed

	mov	dx, BMIC_INDEX			; Set index to EISA Doorbell
	mov	al, EISA_DOORBELL_STATUS	; interrupt status register.
	out	dx, al

	mov	dx, BMIC_DATA			; DX = BMIC Data Port.
	mov	al, bl				; Set interrupt bit.
	out	dx, al	

			endm

;ResetThe82586		macro
;
;	Sets the 82586 for 3 wait states and then back to 0 wait states.
;		
ResetThe82586		macro
	mov	dx, 0FFA4h			; Configure to 3 wait states.
	mov	ax, 003Fh
	out	dx, ax
	mov	dx, RESET_586_PORT		; Reset 82586.
	in	al, dx
	mov	dx, 0FFA4h			; Configure to 0 wait states.
	mov	ax, 0038h
	out	dx, ax

			endm

Attention		macro
	mov	dx, CA_PORT
	out	dx, al
			endm

StatisticsUpdate	macro	CounterOffset

	push	di
	lea	di, CounterOffset
	add	word ptr [di], 1
	adc	word ptr [di + 2], 0
	inc	CounterUpdateFlag
	pop	di	

	endm

