%PAGESIZE 66,130
;*****************************************************************************
; NE3200.ASM
;	(c) Copyright Unpublished Work of Novell, Inc.  All Rights Reserved.
;
;	THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL, 
;	PROPRIETARY AND TRADESECRET INFORMATION OF NOVELL, INC.  ACCESS 
;	TO THIS WORK IS RESTRICTED TO (I) NOVELL EMPLOYEES WHO HAVE A
;	NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE OF THEIR ASSIGNMENTS 
;	AND (II) ENTITIES OTHER THAN NOVELL WHO HAVE ENTERED INTO APPROPRIATE 
;	LICENSE AGREEMENTS.  NO PART OF THIS WORK MAY BE USED, PRACTICED,
;	PERFORMED, COPIED, DISTRIBUTED, REVISED, MODIFIED, TRANSLATED,
;	ABBRIDGED, CONDENSED, EXPANDED, COLLECTED, COMPILED, LINKED, RECAST,
;	TRANSFORMED OR ADAPTED WITHOUT THE PRIOR WRITTEN CONSENT OF NOVELL. 
;	ANY USE OR EXPLOITATION OF THIS WORK WITHOUT AUTHORIZATION COULD 
;	SUBJECT THE PERPERTRATOR TO CRIMINAL AND CIVIL LIABILITY.
;
;----------------------------------------------------------------------------
;
;	Title:		Novell NE3200 MLID for DOS ODI
;
;	Filename:	NE3200.asm
;
;	Written by:	JWR
;
;	Date:		05/01/91
;
;	ODI Specification Version:	4.00
;
;	Description:   Novell NE3200 HSM Code for DOS ODI
;		       This code must be linked with MSM.obj and Ethernet.obj.
;
;	Note:	The Version Number is only changed when changes are made to 
;		the code after a version has been certified.  Changes made to
;		the code during certification are tracked by the date, time,
;		and file size of the version being certified.
;
; Modification History:
;	By   	Date	 Ver.	   Modification Description
;	------ ---------   -----   --------------------------------------
;	JWR,WTT	01 August 94 2.00    Removed Old History, 
;				   Remade to confirm to ODI SPEC 4.00
;
;	JWR,WTT	22 Sept. 94 2.01   Re-linked to include latest tool kit.
;				
;	JWR,WTT,AD 10 Oct. 94 2.02   Re-linked to include latest tool kit.
;				     Added one byte to LookAheadBuf.  
;				     Buffer was one byte too small.
;
;	JWR	7 Dec. 94 2.03 1    Update Copyright date, New MSM Keep Pad
;				     When in Promiscous Mode.
;
;	JWR	9 Jan. 95 2.03	     Re-Link Latest ToolKit.
;
;	JWR	5 April 95 2.04	     Re-Link Latest ToolKit.
;				     Corrected Interrupt state to call MSMGetRCB.
;
;	JWR	30 Aug 95 2.05	     Re-Link Latest ToolKit.
;				     Corrected Specification Version String.
;
;	JWR	25 Oct 95 2.06	Re-Link Latest ToolKit.
;				Changed Connector is String SPD#106596
;
;	JWR	4 Jan. 96 2.06  Update CopyRight. 
;
;	JWR	9 Feb 96 2.09	Re-Link Latest ToolKit.
;
;	JWR	1 April 96 2.09	Re-Link Latest ToolKit v2.10.
;
;	LON	18 Sep  96 2.11 On some multiprocessor boxes, the system would
;				occasionally hang when sending a frame. This
;                               fix flushes any IO queued up by reading IO port
;                               61h, eliminating any APIC/BMIC timing problems.
;
;	SMN 21 Nov. 97 2.13		Re-Link Latest ToolKit. v2.12, Update Copyright date
;
;*****************************************************************************

Include DRIVER.inc
Include BMTCB.inc

;******************************************************************************

.386
Init	segment	para public 'CODE'
;----------------------------------------------------------------------------
;	Version numbers for sign on message and configuration table
;----------------------------------------------------------------------------
MajorVersion	equ	2
MinorVersion	equ	13		; Last Release November 21,1997 (v2.13)
;AlphaVersion	equ	1		; Comment in and out as needed for
;BetaVersion	equ	0		; Prerelease (Set November 21, 1997)

;----------------------------------------------------------------------------
;	Signon Message
;----------------------------------------------------------------------------
HSMSPEC			db	'HSM_SPEC_VERSION: 4.00', 0
			db	'VeRsIoN='
DriverSignOnMessage	db	'Novell NE3200 Ethernet MLID  v'
			db	'0' OR (MajorVersion MOD 10),'.'
			db	'0' OR (MinorVersion/10)
			db	'0' OR (MinorVersion MOD 10)
ifdef AlphaVersion      
		db      ' Alpha ',((AlphaVersion/10) OR '0')
		db	((AlphaVersion MOD 10) OR '0') ; Only for Alpha builds.
endif

ifdef BetaVersion       
		db      ' BETA ',(BetaVersion + '0')   ; Only for Beta Builds.
endif
		db	' (971121)',0Dh,0Ah      ;FixDate will correctly stamp it.
		db	'(c) Copyright 1991 - 1997, by Novell, Inc. '
		db	'All rights reserved.',0

DriverMainSectionText		db	'LINK DRIVER NE3200',0

Init	ends

;******************************************************************************

Code	segment para public 'CODE'

Message NICLongName  'Novell NE3200 Ethernet'
Message NICShortName 'NE3200'

public MAX_TX_RETRIES			; MAKE PUBLIC to guarantee FirmWare is The Same.
Public DoorBellBits			; Used by the FirmWare...Interrupt Enablers.

MAX_TX_RETRIES		equ	10	; Number of Retries to the Chip.
DoorBellBits		equ	PKT_IN_LOOKAHEAD OR TX_COMPLETE_BIT OR 30h ; Interrupt bits allowed.

;******************************************************************************
;             D R I V E R   S T A T I S T I C S   T A B L E
;******************************************************************************

	even
DriverStatTable		label	byte
DriverStatMajorVer	db	01
DriverStatMinorVer	db	01
DriverStatGenericCnts	dw	14
DriverStatValidCntsMask dd	11111000100011111111111111111111B
DriverStatTotalTx	dw	2 dup (0)		; 1 - Set by TSM
DriverStatTotalRx	dw	2 dup (0)		; 1 - Set by TSM
DriverStatNoECB		dw	2 dup (0)		; 1 - Set by TSM
DriverStatTxTooBig	dw	2 dup (0)		; 1 -Not Used
DriverStatTxTooSmall	dw	2 dup (0)		; 1 -Not Used

FirmWareStatLabel	label	byte

DriverStatRxOverflow	dw	2 dup (0)		; 0 - Set By FirmWare
DriverStatRxTooBig	dw	2 dup (0)		; 0 - Set by FirmWare (MSM Checks for this also)
DriverStatRxTooSmall	dw	2 dup (0)		; 0 - Set by FirmWare
DriverStatTxMisc	dw	2 dup (0)		; 1 - Not Used 
DriverStatRxMisc	dw	2 dup (0)	    	; 0 - Set by FirmWare
DriverStatTxRetryCount	dw	2 dup (0)		; 0 - Set by FirmWare
DriverStatRxChecksum	dw	2 dup (0)   		; 0 - Set by FirmWare
DriverStatRxMismatch	dw	2 dup (0)		; 1 -Not Used
DriverQueueDepth	dw	2 dup (0)		; 1 - Set by TSM

;	Custom Diagnostic counters.		- All Set by FirmWare

     NumberOfCustomCounters	dw	10

	NumberOf82586Resets	dw	2 dup (0)	; number of times 586 reset	
	TxRetryFailure		dw	2 dup (0)	; number of retry failures
	TxCarrierSenseLost	dw	2 dup (0)	; tx carrier lost	
	TxClearToSend		dw	2 dup (0)	; tx error	
	TxUnderRun		dw	2 dup (0)	; tx error	
	TxDefers		dw	2 dup (0)	; tx error	
	TxMaxCollisions		dw	2 dup (0)	; tx error	
	TxNumCollisions		dw	2 dup (0) 	; tx error Actual number of collisions
	RxDMAOverrun		dw	2 dup (0)	; Rx error
	RxNoEOF			dw	2 dup (0) 	; Rx error

					  
;Pointers to the Counter Message Strings

			dw	offset CGroup:DS1
			dw	0
			dw	offset CGroup:DS2
			dw	0
			dw	offset CGroup:DS3
			dw	0		 
			dw	offset CGroup:DS4
			dw	0
			dw	offset CGroup:DS5
			dw	0
			dw	offset CGroup:DS6
			dw	0
			dw	offset CGroup:DS7
			dw	0
			dw	offset CGroup:DS8
			dw	0
			dw	offset CGroup:DS9
			dw	0
			dw	offset CGroup:DS10
			dw	0
	
; Diagnostic Custom Error Counters

Message		DS1	'Num82586Reset'  
Message		DS2	'TxRetryFail'
Message		DS3	'CarrierSenseLost'
Message		DS4	'ClearToSend'
Message		DS5	'UnderRun'	
Message		DS6	'TxDefer'	
Message		DS7	'MaxCollision'
Message		DS8	'NumCollision'
Message		DS9 	'RxDMAOvrn'
Message		DS10	'RxNoEOF'		


;******************************************************************************
;            D R I V E R   C O N F I G U R A T I O N   T A B L E
;******************************************************************************

DriverConfigTable	label	byte
Signature		db	'HardwareDriverMLID',8 dup (' ')
CfgTbleMajorVersion	db	01
CfgTbleMinorVersion	db	13
NodeAddress		db	6 dup (0FFh)		
ModeFlags		dw	1010000000001001b ;Real,Multicast, OBR, Promiscuous.
BoardNumber		dw	?		;set by MSM
BoardInstance		dw	?		;set by MSM
MaxFrameSize		dw	?		;set by MSM
BestDataSize		dw	?		;set by MSM
WrstDataSize		dw	?		;set by MSM
LANCardName		dw	offset CGroup:NICLongName,0
LANShortName		dw	offset CGroup:NICShortName,0
LANFrameName		dd	?		;Don't touch MSM will set
Reserved0		dw	0  		; No longer used RESERVED MUST be 0
LANFrameID		dw	?		;Don't touch MSM will set
TransportTime		dw	1
SourceRouteHandler	dd	0		;Don't Touch
LookAheadSize		dw	18		; Just enough for IPX default
LineSpeed		dw	0Ah 		; 10 Mbs -- Ethernet
Reserved		db	6 dup (0)       ; Saved for future uses
PrioritySup		db	0
BusID			db	BUS_ID_HSM_DEFAULT
DriverMajorVersion	db	MajorVersion	; Set in equate
DriverMinorVersion 	db	MinorVersion	; Set in equate
Flags			dw	0
SendRetries		dw	MAX_TX_RETRIES 	; Default retry value 
ConfigTableLink		dd	0		;Don't Touch it.
SharingFlags		dw	1	    	; Default to Shutdown
Slot			dw	0		; Default is to Scan for First Slot.
IOAddress1		dw	0000h		; Filled in by driver at Init TIme
IORange1		dw	30h		; BMIC Register Ports
IOAddress2		dw	0		;Not Used
IORange2		dw	0		;Not Used
MemAddress1		dd	0h		;Not Used 
MemRange1		dw	0h		;Not Used 
MemAddress2		dd      0h		;Not Used 
MemRange2		dw	0h		;Not Used 
IRQLine1		db	0h		; Read ESIA Slot for Interrupt Level
IRQLine2		db	0FFh		;Not Used
DMALine1		db	0FFh		;Not Used
DMALine2		db	0FFh		;Not Used

;****************************************************************************
;	Receive Variables and Buffers
;****************************************************************************

	ALIGN 4
BMFrameSize		dw	0		; Size of Incoming Frame
LookAheadBuffer		db	1514 dup (0)	; Lookahead to Hold The WHole Packet.
RCBinProcess		dd	0		; Outstanding RCB   Offset - Segment

;****************************************************************************
;      Transmit Variables and Buffers
;****************************************************************************

 even
PhysicalBMTCB		dd	0
PtrBMTCB		dw	offset CGroup:TempTxColBuffer	; Default to the Temporary Buffer.
TCBinProcess		dw	0		; Current TCB Offset
CollBuffBytes		dw	0		; Count leftover bytes 
CollBuffFrags		dw	0		; Count Number of Frags needed.
TxCollSize		dw	1514		; set the TxCollBuffer to MAX size.

;----------------------------------------------------------------------------
;	Paramater Block Allocation
;----------------------------------------------------------------------------

	align 4
PARAMETERBLOCK		label dword
NodeAddressPointer	dd	0	; Physical pointer to DriverConfigTable.NodeAddress
BMFrameSizePtr		dd	0	; Physical Pointer to the Lookahead area
FirmWareStatisticsPtr	dd	0	; Physical Pointer to the Generic Counters
CustomStatCounter	dw	0	; Number of Custom counters
MulticastCount		dw	0	; Max Number of Multicast Entries
MulticastListPtr	dd	0	; Physical Pointer to MulticastTable
AdapterParamBlkSize	equ	$-PARAMETERBLOCK ; Size of The Table

;----------------------------------------------------------------------------
;	Post Office for communication with the NE3200 Adapter
;----------------------------------------------------------------------------
	even
IORegistersDef	label byte
NE3200ResetRegister	dw	0000h	; x000h - Resets the Adapter
SystemIntEnable		dw	0C89h	; xC89h - Enable/Disable Interrupts to the Host
;SemephoreReg1		dw	0C8Ah	; xC8Ah where x is slot number
;SemephoreReg2		dw	0C8Bh	; xC8Bh where x is slot number
;NE3200DoorBellEnable	dw	0C8Ch	; xC8Dh - Enables the Interrupt from Host (Read only)
;NE3200DoorBell		dw	0C8Dh	; xC8Ch - Interrupts the NE3200 (Not Used)
;HostDoorBellEnable	dw	0C8Eh	; xC8Fh - Enables Interrupts to the Host x=slot number
HostDoorBell		dw	0C8Fh	; xC8Eh - Interrupts the Host (Host write Clears Int)
StartMailBoxRegs   label word		; xC90h  where x is slot number
RCBMailBox		dw	0C90h	; xC90h - 0F0h = Ready for next Packet.
Abend_IdleMailBox	dw	0C91h	; xC91h	- 0F0h = Abend  00Fh = Idle adapter  
UpDateParamMailBox	dw	0C92h	; xC92h - 04h = Multicast update, 01H = ReInitParameter table 02h = Reset Adapter
BMTCBMailBox		dw	0C93h	; xC93h	- Number Frags in BMTCBStruc. (including Header Frag)
					; Also tells the adapter to start the transfer 6*Number of FRags back for default location
BMTCBTransMailBox	dw	0C94h	; xC94h & 5 - Number of Bytes to Transfer  (Hi-Lo Format) hi=4   lo=5
;MailBox6		dw	0C96h	; xC96h
;MailBox7		dw	0C97h	; xC97h
;RCBOS2MAILBOX		dw	0C98h	; xC98h - Tells new Address After Every RCB.
;MailBox9		dw	0C99h	; xC99h
;MailBoxA		dw	0C9Ah	; xC9Ah
;MailBoxB		dw	0C9Bh	; xC9Bh 
BMTCBStartPtr	   label word		; xC9Ch - Point to Start Downloading BMTCB
ParameterMailBox	dw	0C9Ch	; xC9Ch - Down Load and Reset Parameter Address

IORegisterSpace   equ	$-IORegistersDef ; For DriverInit.


DDS	struc
	RegionSize	dd	0
	DMAOffset	dd	0
	DMASegment	dw	0
	BufferID	dw	0
	DMAPhysAddr	dd	0
DDS	ends

CodeAddress    		dd	0			;This is used at init time to
							;get physical address of code image
CodeDescriptor	DDS	<>	;allocate DMA descriptor structure.

;*******************************************************************************
;       DriverManagement 	  ---Not used .... Stubbed here for the MSM
;       DriverChangeLookAheadSize ---Not used .... Stubbed here for the MSM
;       DriverPoll       	  ---Not used .... Stubbed here for the MSM
;*******************************************************************************

DriverManagement:
DriverChangeLookAheadSize:
DriverPoll:
	ret

;****************************************************************************
;****************************************************************************
;
; BEGIN_MANUAL_ENTRY ( DriverMulticastChange, NE3200/ODI/API/Multchg )
;
; NAME:		DriverMulticastChange
;
; DESCRIPTION:	
;	This Routine Updates the physical boards list of Multicast Addresses.
;
; ON ENTRY:	AX	N/A
;		BX	N/A
;		CX	Max Num. Entries in Table.
;		DX	N/A
;		BP	N/A
;		SI	N/A
;		DI	N/A
;		CS	DGROUP
;		DS	DGROUP
;		ES	N/A
;	
;		NOTE:	cld in effect, Interrupts are ...abled	
;
; ON RETURN:	AX	
;	    	BX
;		CX
;		DX
;		BP
;		SI
;		DI
;		CS	
;		DS	
;		ES
;
;
;		NOTE: Interrupts are enabled/disabled during function.
;		      Interrupts are returned ...abled
;	
;
; REMARKS:
;
; SEE ALSO:
;
; END_MANUAL_ENTRY
;****************************************************************************
  
;****************************************************************************
;    On Entry:				On Return:
;	DS = CGROUP				No Registers or flags
;	DS:SI -> Multicast Address Table.	need to be preserved.
;	CX = Max. Num of Entries in Table.	Ints may be enabled. 
;	Ints disabled
;	CLD 
;	Call only at process time.
;
;****************************************************************************

DriverMulticastChange	proc	near
	push	eax
	push	esi			; Preserve Registers that the MSM don't
	mov	eax, CodeAddress	;
	and	esi, 0000FFFFh		; Get Offset
	add	eax, esi		; EAX = Physical Address
	mov	MulticastListPtr, eax	; Store it
	mov	MulticastCount, cx	; Store number of Multicast addresses.
	mov	dx, UpDateParamMailBox	; Get MailBox Address
	mov	al, UpDateMulticastCMD	; 
	out	dx, al			; Tell Adapter and forget it.

WaitforChange:
	in	al, dx
	or	al, al
	jnz	WaitforChange

	pop	esi
	pop	eax			; Restore Registers that the MSM don't
	ret
DriverMulticastChange	endp

;****************************************************************************
;  DriverShutdown			
;****************************************************************************
;    On Entry:				On Return:
;	DS = CGROUP				No Registers or flags
;	CX == 0 if Permentant ShutDown		need to be preserved.
;	   != 0 if Temporary ShutDown		Ints may be enabled. 
;	Ints disabled
;	CLD
;	Call only at process time.
;
;****************************************************************************

DriverShutdown	proc	near

;----------------------------------------------------------------------------
;	Disable Adapter's Interrupt
;----------------------------------------------------------------------------

	mov	dx, SystemIntEnable		; System Interrupt Enable Reg.
	xor	ax, ax				; Clear Reg
	out	dx, al				; disabling future Ints

	push	cx				; Save PermencyState
	mov	dx, Abend_IdleMailBox		; Get register Address
	in	al, dx				; Check on CurrentState
   	sti					; Don't Hold everyone else up.
	cmp	al, FIRMWARE_STATE_FLAG 	; Are We Somewhat Active?
	jne  short CheckOnTCBs			; If we are already idle then skip waiting.	

	mov	al, IDLE_AdapterCMD		; Mark as need to idle
	out	dx, al				; Tell Adapter.

WaitForIdleState:
	in 	al, dx				; Get Mail
	cmp	al, ABEND_STATE_FLAG 		; Are we DEAD?
	je   short CheckOnTCBs			; If we are dead then forget waiting.

	cmp	al, IDLE_STATE_FLAG		; IS it IDLE?
	jne  short WaitForIdleState		; Wait till it is

CheckOnTCBs:
	xor	si, si
	xchg	si, TCBinProcess		; Get TCB offset (Maybe none)
	or	si, si				; Is there one?
	jz   short ShutDownNoTCB		; Jump over sendcomplete if not 

	cli
	call	MSMSendComplete			; Return outstanding TCB
	mov	MSMTxFreeCount, 1 		; DriverMaxActiveTx = 1
	mov	bx, MLIDSUP_SERVICE_EVENTS	; Set up for Service Events Call.
	call	LSLSupport

;----------------------------------------------------------------------------
;	Reset all the DoorBell Bits, and See if Permenant or not.
;----------------------------------------------------------------------------

ShutDownNoTCB:
	mov	dx, HostDoorBell	; Get the Door Bell Register
	mov	ax, DoorBellBits	; Clear All of the Active Bits.
	out	dx, al

	pop	cx			; Restore Permency state.
	jcxz short DrvPermShutdown	; Don't get Better.
	ret				; cx != 0 => TempShutDown.

DrvPermShutdown:
	mov	dx, Abend_IdleMailBox	; Mail Box
	mov	al, Abend_AdapterCmd 	; Mark as shut off
	out	dx, al			; Tell Adapter
	ret

DriverShutdown	endp

;****************************************************************************
;  DriverReset
;****************************************************************************
;    On Entry:				On Return:
;	DS = CGROUP				No Registers or flags
;	Ints disabled				need to be preserved.
;	CLD					Ints may be enabled. 
;	Call only at process time.
;
;****************************************************************************

DriverReset	proc	near
	mov	cx, Reset_AdapterCMD	; Mark for a non-Permenant ShutDown
	call	DriverShutdown		; Turns off Ints and Puts in Idle_State.

;----------------------------------------------------------------------------
;	Tell The Card to Go Ahead and Reset.
;----------------------------------------------------------------------------

	mov	dx, UpDateParamMailBox	; Time to update Parameters
	mov	al, Reset_AdapterCMD
	out	dx, al			; Tell the Adapter
	sti

	mov	dx, Abend_IdleMailBox  ; Look in the Abend/Idle MailBox 
WaitTilReset:
	in	al, dx			; Have we reset yet?
	cmp	al, FIRMWARE_STATE_FLAG	; Wait for 'J' => Back in the MainEvent Loop.
	jne  short WaitTilReset		; to come.
	
	mov	MSMTxFreeCount, 1 	; DriverMaxActiveTx = 1	- Reset our Resources

	mov	dx, RCBMailBox
	mov	al, RCBCompleteDOS	; Mark RCB Available.
	out	dx, al			; Signal the FirmWare

;----------------------------------------------------------------------------
;	DriverReset EXIT   --- Enable Interrupt line and ret.
;----------------------------------------------------------------------------

	mov	dx, SystemIntEnable 	; System Int enable port
	mov	al, EISAIntEnableBit	; Enable the Interrupt
	out	dx, al			; Tell the Adapter
	ret

DriverReset	endp



;------------------------------------------------------------------------------
;	DriverPromiscuousChange
;
;	Called by the MSM when it wants us to set/reset Promsicuous Mode
;
;	assumes:	CX	Zero		Disable Promiscuous Mode
;				Non-Zero	Enable Promiscuous Mode
;			ds	is set to CGroup
;				CLD has been executed
;				Interrupts are disabled
;
;	returns:	Nothing need be preserved
;
;------------------------------------------------------------------------------
DriverPromiscuousChange	proc	near

	mov	al, Clear_Promiscuous		;set up mode to Normal operation
	jcxz	NormalMODE
	mov	al, Set_Promiscuous		;Set up mode word in Init Block
						;for promiscuous receives.
NormalMODE:
	mov	dx, UpDateParamMailBox		; Get MailBox Address
	out	dx, al				; Tell Adapter and forget it.

WaitForReset:
	in	al, dx
	or	al, al	    			;Wait for command to be Executed
	jnz	WaitForReset

	ret

DriverPromiscuousChange	endp


;****************************************************************************
;  DriverISR -- Interrupt Service Routine
;****************************************************************************
;    On Entry:				On Return:
;	DS = CGROUP				DS = CGROUP
;	Ints disabled				All Registers or flags Can Destroyed
;	CLD					Ints are disabled.
;	A Critical Section Has been started.	
;	All Registers and Flags have been preserved.
;	
;****************************************************************************

EOISlave:
	out	0A0H, al		; Issue EOI to the SLAVE PIC
	jmp short BackFromEOISlave

DriverISR	proc	near
;----------------------------------------------------------------------------
;	Disable Adapter's Interrupt
;----------------------------------------------------------------------------
	mov	dx, SystemIntEnable		; System Interrupt Enable Reg.
	xor	ax, ax				; Clear Reg
	out	dx, al				; disabling future Ints

;----------------------------------------------------------------------------
;	Issue EOI
;----------------------------------------------------------------------------
	mov	al, 20h			; Issue EOI to PIC(s)
	cmp	MSMEOIFlag, 0		; EOI needed on slave PIC?
	jne  short EOISlave		; Go do slave
		
BackFromEOISlave:	
	out	20h, al			; Issue EOI to the 1st PIC
	
;----------------------------------------------------------------------------
;	Who Called Interrupt
;----------------------------------------------------------------------------
CheckWhoCalled:
	mov	dx, HostDoorBell	; Door Bell Register
	in	al, dx			; You Rang?

	test	al, PKT_IN_LOOKAHEAD	; Packet Received?
	jnz  short PacketReceived	; Have a Packet in the Lookahead buffer.

	test	al, TX_COMPLETE_BIT	; Finished with a TCB?
	jnz   	CompleteTCB		; nz => Done with it.

   	jmp	ISRExit		; No one wants anything.(need long jmp)

;----------------------------------------------------------------------------
;	Receive Handler	
;----------------------------------------------------------------------------
PacketReceived:
	mov	bx, ds			; Get CGROUP and Clear Upper word EAX.
	mov	es, bx			; ES = CGROUP
	shr	ax, 4			; get Error Code in AX
	and	ax, 03h
	mov	si, offset CGroup:LookAheadBuffer
	mov	bp, BMFrameSize		; How large was the Packet
	sti				; Spec calls for Enabled Ints for MSMGetRCB.
	call	MSMGetRCB		; Do You Want it?
	jnz  short ISRNoRCB
	
;----------------------------------------------------------------------------
;	Have an RCB --- Fill it in.
;	AX = offset into Packet to Start moving data.
;	BP = Amount of Data left to move
;	ES:SI -> RCB
;----------------------------------------------------------------------------
	push	fs				; Save it cause the MSM don't
	mov	word ptr RCBinProcess+0, si
	mov	word ptr RCBinProcess+2, es	; Save it for now.
	mov	bx, es				
	mov	fs, bx				; Load FS Segment  FS:SI -> RCB
	mov	bx, si				; FS:BX -> RCB (Dest)
	mov	si, offset CGroup:LookAheadBuffer ; SI -> Lookahead buffer.
	add	si, ax				; SI -> Location in LookAhead buffer 

;----------------------------------------------------------------------------
;		Fill in the Frags till done.
;----------------------------------------------------------------------------
	lea	bx, fs:[bx - 6].RCBFrag0Addr ; BX -> First Fragment (Back up for Add below.)

NextRCBFrag:
	add	bx, 6			; Next Frag Pointer.
	mov	cx, fs:[bx].FragLen	; Size of Fragment
	jcxz short NextRCBFrag		; Skip it if nothing there

	les	di, fs:[bx].FragAddr	; ES:DI -> Fragment 
	cmp	cx, bp			; Nearly Done?
	jb  short UseCX
      	
	mov	cx, bp			; Finish out what was asked for.

UseCX:
	sub	bp, cx			; BP = #Bytes left after move
	mov	dx, cx			; Save Original Byte Count.
	shr	cx, 2			; CX = DWord count
   rep  movsd
   	mov	cx, dx			; Restore Original Byte Count.
	and	cx, 3			; MOD 4
   rep	movsb	
   	or	bp, bp			; Finished?
	jnz  short NextRCBFrag		; Then get the next one.

;----------------------------------------------------------------------------
;	Return the RCB to the MSM
;----------------------------------------------------------------------------
	les	si, RCBinProcess	; Get The RCB
	cli				; hold off the Ints.
	call	MSMRcvComplete		; Return The RCB
	pop	fs			; Restore Registers That MSM doesn't.
	sti				; Allow them to go.

;----------------------------------------------------------------------------
;Fill in the Signaling MailBox.  Tells Adapter that the Host is finished with the Lookahead
;----------------------------------------------------------------------------
ISRNoRCB:	
   	mov	dx, HostDoorBell	; Interrupt Port
	mov	al, PKT_IN_LOOKAHEAD+30h; Clear RCB Interrupt Signals Adapter to go on
	out	dx, al			; byte out
	mov	dx, RCBMailBox 		; Get address of Mail Box
	mov	al, RCBCompleteDOS 	; Mark RCB Status -- We are Done With the LookAhead.
	out	dx, al	       		; Tell adapter...Fall into ChkNewTCBComplete.

;----------------------------------------------------------------------------
;	Check for an RCB or a TCB Completed
;----------------------------------------------------------------------------
ChkNewTCBComplete:
	mov	dx, HostDoorBell	; Look Again
	in	al, dx			; Some one Ready?

	test	al, TX_COMPLETE_BIT	; TCB Ready?
 	jnz  short CompleteTCB		; Complete it then

	test	al, PKT_IN_LOOKAHEAD	; Packet Received?
  	jnz short PacketReceived	; Have a Packet in the Lookahead buffer.


	jmp short ISRExit		; Nothing yet.

;----------------------------------------------------------------------------
;	TCB Completed
;----------------------------------------------------------------------------
CompleteTCB:
	xor	si, si
	xchg	si, TCBinProcess	; Get/clear TCB offset
	or	si, si			; Was there one?
	cli
	jz short LyingSend		; Check just in case the Adapter lied.

	call	MSMSendComplete		; Return TCB

LyingSend:
	mov	dx, HostDoorBell	; MailBox
	mov	al, TX_COMPLETE_BIT	; Get Bit to clear
	out	dx, al			; Clear TCBcomplete INT

;----------------------------------------------------------------------------
;	Set up to get the Next Send from the MSM.
;----------------------------------------------------------------------------
	mov	MSMTxFreeCount, 1	 ; DriverMaxActiveTx = 1
	call	MSMGetNextSend		 ; Get next TCB
	jnz	CheckWhoCalled		 ; NZ = no TCB look for another RCB. (need long jmp)

	jmp  short InternalDriverSend	 ; DriverSend will Enable The Int Line on exiting

;----------------------------------------------------------------------------
;	ISR EXIT   --- Enable Interrupt line and ret.
;----------------------------------------------------------------------------
ISRExit:
	cli				; Hold the Ints off.
	mov	dx, SystemIntEnable 	; System Int enable port
	mov	al, EISAIntEnableBit	; Enable the Interrupt
	out	dx, al			; Tell the Adapter

	ret

DriverISR	endp


;****************************************************************************
;  DriverSend
;****************************************************************************
;    On Entry:				On Return:
;	DS = CGROUP				DS = CGROUP
;	DS:SI -> TCB				All Registers or flags Can Destroyed
;	CX = Hardware Frame Length (Ethernet)	Ints are disabled.
;	Ints disabled
;	CLD		
;	A Critical Section Has been started.	
;	All Registers and Flags have been preserved.
;
;****************************************************************************
DriverSend	proc	near
;----------------------------------------------------------------------------
;	Disable Cards Interrupt
;----------------------------------------------------------------------------
	mov	dx, SystemIntEnable		; Get Doorbell's Master Enabler
	xor	ax, ax				; Get cleared
	out	dx, al				; Tell Adapter

InternalDriverSend:
	push	eax				; Saves the Extended Register Values
	push	ecx				; that the MSM didn't save.
	push	edi
	push	esi
	push	fs				; Save off the Segment pointers (MSM doesn't doit for us).
	push	gs
	sti					; Turn Interrupts back on
        in	al, 61h				; Flush io queue

;----------------------------------------------------------------------------
;	Move the TCB into the BMTCB Structure
;----------------------------------------------------------------------------
	mov	eax, ds
	mov	es, ax				; Get CGROUP / Clear upper Word EAX for later.
	mov	TCBinProcess, si		; Save TCB address for Complete Call Later
	mov	bp, si				; ES:BP -> TCB
	mov	di, PtrBMTCB			; DS:DI -> BMTCB
	mov	CollBuffFrags, 1		; Count Number of Frags, counting TxCollBuffer
	mov	[di].BMTCBFrameLen, cx		; Store Physical Length
	mov	ax, es:[bp].TCBDataLen 		; Get Data Length 
	mov	[di].BMTCBDataLen, ax		; Store Data Length
	mov	ax, TxCollSize			; AX = Size of BMTCBTxBuffer
	add	di, BMTCBTxBuffer		; DS:DI -> Collection Buffer
	mov	cx, es:[bp].TCBMediaHeaderLen	; get size of header
	jcxz short NoHeader			; Skip header

;----------------------------------------------------------------------------
;	Load Header into the Collection Buffer
;----------------------------------------------------------------------------
	lea	si, es:[bp].TCBMediaHeader	; ES:SI -> Media Header
	sub	ax, cx				; Keep track of buffer space left.
	mov	bx, cx				; keep Original Size handy.
   	shr  	cx, 2				; Get Header Size / 4
   rep	movsd					; Move all the Dwords
   	mov	cx, bx				; Restore Original Size.
	and	cx, 3				; MOD 4
   rep	movsb					; move leftovers if any

;----------------------------------------------------------------------------
;	Get Fragment Structure
;----------------------------------------------------------------------------
NoHeader:	
	lfs	bp, es:[bp].TCBFragStrucPtr	; FS:BP -> Frag Stucture
	mov	dx, fs:[bp].FFragmentCount	; DX keep track of Fragments
	add	bp, 2				; BP -> First Frag Pointer desc.

;----------------------------------------------------------------------------
;  GS:SI -> Fragment	FS:BP -> TCBFragStruc	ES:DI -> BMTCBTxBuffer
;----------------------------------------------------------------------------
NextFrag:
	lgs	si, fs:[bp].FragAddr		; GS:SI -> Fragment
	mov	cx, word ptr fs:[bp].FragLen	; Get Length
	jcxz short CheckNextFrag		; If Zero forget it.

	cmp	cx, ax				; Enough room?
	ja  short FillToEnd			; Not enough then go Fill to End of Buffer.

;----------------------------------------------------------------------------
;	Move Data from Fragment into the Collection Buffer.
;----------------------------------------------------------------------------
	sub	ax, cx				 ; Keep track of buffer space left.
	mov	bx, cx				 ; Save Original Size
   	shr  	cx, 2				 ; Get Header Size / 4
   rep	movs   dword ptr [di], dword ptr gs:[si] ; Move all the Dwords
   	mov	cx, bx				 ; Restore Original Size.
	and	cx, 3				 ; MOD 4
   rep	movs   byte ptr [di], byte ptr gs:[si]	 ; move it then

CheckNextFrag:
	dec	dx				; Keep Track of How Many Frags Left.
	jz  short FullyLoadedWithSpace

	add	bp, 6				; Next Frag Address
	jmp short NextFrag			; Get Data from next Fragment.

;----------------------------------------------------------------------------
;	Finish filling Collection Buffer
;----------------------------------------------------------------------------
FillToEnd:
	xchg	ax, cx				 ; AX = Frag size   CX = Buf space remaining
	sub	ax, cx				 ; AX = Left overs in Frag
	mov	bx, cx				 ; Save Original Size.
   	shr  	cx, 2				 ; Get Size / 4
   rep	movs   dword ptr [di], dword ptr gs:[si] ; Move all the words
	mov	cx, bx				 ; Restore Original Size.
   	and	cx, 3				 ; Mod 4
   rep	movs   byte ptr [di], byte ptr gs:[si]	 ; move it then
	mov	CollBuffBytes, cx		 ; Space left in the Collection Buffer. (0) 

;----------------------------------------------------------------------------
;	Create First Physical Pointer
;----------------------------------------------------------------------------
	movzx	edi, PtrBMTCB
	add	di, BMTCBFragsStart	; DS:EDI -> 1st fragment Descriptor
	mov	word ptr [di].FragLen, ax	; AX = Left over amount in TCB frag is new BMTCB frag size
	mov	eax, gs				; EAX = Segment
	shl	eax, 4				; move to position
	and	esi, 0000FFFFh			; Clear upper word
	add	eax, esi			; EAX = Offset
	mov	dword ptr [di].FragAddr, eax	; Store Physical Address
	inc	CollBuffFrags	 		; Count the BMTCB Frags.

;----------------------------------------------------------------------------
;	Convert all the SEG:Offs to Physical Pointer for the BMTCB
;----------------------------------------------------------------------------
NextTCBFrag:
	dec	dx				; Have any TCB Frags Left?
	jz  short FullyLoaded			; z => all frags done.

	add	bp, 6				; FS:BP -> Next TCB Frag Address
	mov	cx, word ptr fs:[bp].FragLen	; Length of TCB Fragment
	jcxz short NextTCBFrag			; Ignore this TCB fragment

	sub	di, 6				; DS:EDI -> Next BMTCB Frag Descriptor
	mov	word ptr [di].FragLen, cx	; store Length of TCBfrag in BMTCB
	movzx	esi, word ptr fs:[bp].FragAddr	; Get Offset
	movzx	eax, word ptr fs:[bp].FragAddr+2; Get Segment
	shl	eax, 4				; Move to Position
	add	eax, esi			; EAX =  Physical address
	mov	dword ptr [di].FragAddr, eax	; Store Physical Address
	inc	CollBuffFrags			; Count the BMTCB Frags
   	jmp short NextTCBFrag			; Get all the Fragment's descriptors.

;----------------------------------------------------------------------------
;	BMTCB is Finally Loaded  -- Determine how many bytes to transfer
;----------------------------------------------------------------------------
FullyLoadedWithSpace:

	mov	CollBuffBytes, ax		; Space left in the Collection Buffer. (may be 0) 

;----------------------------------------------------------------------------
; FullyLoadedWith Space implies that the Collection Buffer holds all the data.
;	Therefore will do a lying send here to help speed things up.
;---------------------------------------------------------------------------
	xor 	si, si			; Set up to Clear TCBinProcess
	xchg	si, TCBinProcess	; Get and Clear TCBinProcess
	cli
	call	MSMSendComplete		; Return TCB to MSM.
	sti
	movzx	edi, PtrBMTCB
	add	di, BMTCBFrameLen	; With Space Left => No Frag Pointers.

;----------------------------------------------------------------------------
;    Collection Buffer is loaded tell the adapter and let him send it on.
;----------------------------------------------------------------------------
FullyLoaded:					; BMTCB is fully loaded.
	movzx	eax, CollBuffFrags		; Get number of frags
	shl	ax, 1				; (Frags * 2)
	lea	eax, [eax * 2 + eax]		; (Frags * 3)  => AX = Frags * 6 bytes
	mov	cx, TxCollSize			; CX = Size of BMTCBTxBuffer
	sub	cx, CollBuffBytes
	add	ax, cx				; AX = Total Number of Bytes to move.
	mov	dx, BMTCBTransMailBox		; Send some mail
	xchg	ah, al				; Swap for Easier Pick-up
	out	dx, ax				; Post the mail to the card. (HI-Lo Format)

;----------------------------------------------------------------------------
;	Tell the Card Where to Start the Transfer
;----------------------------------------------------------------------------
	mov	eax, CodeAddress		; Get Segment Value
	add	eax, edi			; Get Physical Address
	
	mov	dx, BMTCBStartPtr		; Starting Point Mail Box
	out	dx, eax				; Send Physical Address in Mail.

;----------------------------------------------------------------------------
;	Tell the Card to do the Transfer
;----------------------------------------------------------------------------
	mov	dx, BMTCBMailBox	; Mark MailBox to Transmit
	mov	ax, CollBuffFrags	; Number of Frags including Collection Buffer.
	out	dx, al			; Tell Nic

;----------------------------------------------------------------------------
;	Enable Cards Interrupt
;----------------------------------------------------------------------------
	cli					; Ints must be disabled
	mov	dx, SystemIntEnable		; Get Doorbell's Master Enabler
	mov	al, EISAIntEnableBit		; Turn on Bit
	out	dx, al				; Tell Adapter

	pop	gs			; Restore Segment Registers the MSM didn't save.
	pop	fs
	pop	esi			; Restores the Extended Register Values 
	pop	edi			; that the MSM didn't save.
	pop	ecx
	pop	eax
	ret

DriverSend	endp

Code	ends


Init	Segment para public 'CODE'

;Variables to download FirmWare to NE3200
Extrn	FirmWareStart:word
Extrn	FirmWareSize:ABS

;***************************************************************\
;                                                               *
; EISA slot equates.						*
;      EISA Table is at the End of Segment                      *
;***************************************************************/
TOTAL_EISA_SLOTS	equ	8			
MANUFACTURE_CHAR_CODE_1	equ	3Ah			; Computed from the
MANUFACTURE_CHAR_CODE_2	equ	0CCh			; code "NVL 07 01".
PRODUCT_NUMBER_1	equ	07
ISOLATE_INT_MASK	equ	00001111b		
EISA_INT_FUNCTION_BIT	equ	00000100b
StaticRamError		equ	40h		; NE3200 Adapter's Code for MEMORY Failure

;****************************************************************************
;		Message Center
;****************************************************************************
NumberMessage AdapterNotFound,50, 'The board cannot be found',0
NumberMessage IntNotFoundMSG,235, 'No interrupt was selected.  The board must be',0Dh,0Ah
			db  '               reconfigured.',0
NumberMessage AdapterROMErrorMSG,238, 'The firmware cannot be started.',0
NumberMessage StaticRamErrorMSG,51, 'Board RAM failed the memory test.',0
NumberMessage ROMErrorMSG,070,	    'The board has failed.',0
FirmWareAbendCode	db  0,0,'h.',0			; If code not
NumberMessage FirmWareFailedMSG,070,   'The board has failed.',0
NumberMessage FirmWareStartErrMSG,237,'The FirmWare cannot be initialized.',0
PhysicalAddressNotKnown	DB 'VDS failed to return Physical address.  Assuming Logical = Physical.',0,0DH,0AH
InvalidBusIDMsg		db	96,0,'The driver does not support the selected BUS ID.',0Dh,0Ah
			db		'Make sure the NET.CFG file contains a valid '
			db		'parameter for the ',0Dh,0Ah, 'BUS ID keyword.',0Dh,0Ah, 0
WrongBusMsg		db	28,0,'The driver cannot load.',0Dh,0Ah
			db		'This Machine does not have the correct bus for the driver;',0Dh,0Ah
			db		'or the BUS ID specified in the NET.CFG is incorrect.',0Dh,0Ah,0

;----------------------------------------------------------------------------
; Part of the Sign-on Informational Messages.  No Message Numbers will be assigned.
; MicroDyne says we can't tell when TP is used.--tack it on the DIX string.
;----------------------------------------------------------------------------
ConnectorIsDIX		db  'The selected connector is DIX/TP.',0Dh,0Ah, 0  ;Let the MSM print the Cr,Lf
ConnectorIsBNC		db  'The selected connector is BNC.',0Dh,0Ah, 0

;----------------------------------------------------------------------------
;	Custom Defined KeyWords and Parameters
;----------------------------------------------------------------------------
DriverNumKeywords   	equ 1        
DriverKeywordText   	dw  offset CGroup:DblBuffTxt
DriverKeywordTextLen    dw  DblBufTxtLen
DriverProcessKeywordTab dw  offset CGroup:DisableDblBuff
DblBuffTxt	db	'DOUBLE BUFFER OFF'			; Sets up to use a large TxCollBuff.
DblBufTxtLen	equ	$-DblBuffTxt		


;****************************************************************************
;  DriverInit	----	DRIVER INITIALIZATION
;****************************************************************************
;    On Entry:			On Return:
;	DS = CGROUP			AX == 0 if init was succesful
;	ES = CGROUP			   != 0 if an Error Occured.
;	Ints ENABLED			DS:DX -> Zero-Terminated Error Message
;	CLD					 if AX is not 0
;      					All Registers or flags Can Destroyed
;  
;****************************************************************************
PossibleBusIDError:
	cmp	BusID, BUS_ID_EISA		; Do we have valid Bus ID.
	je	short ValidBusIDFound
	jmp	DriverInitErrorExit

DriverInit	proc	near
;----------------------------------------------------------------------------
;	Make sure user did not place invalid Bus ID in the NET.CFG file.
;----------------------------------------------------------------------------
	mov	dx, offset CGroup:InvalidBusIDMsg
	cmp	BusID, BUS_ID_HSM_DEFAULT      	; See if set to Default
	jne	PossibleBusIDError

;----------------------------------------------------------------------------
;	Check to see if we are even on the right bus.
;	NE3200 only supported on a EISA bus.
;----------------------------------------------------------------------------
ValidBusIDFound:
	mov	dx, offset CGroup:WrongBusMsg	; Possible error code
	test	MSMSystemFlags, EISAFlag 	; EISA Bus?
	jz	DriverInitErrorExit		; Exit out.

	mov	BusID, BUS_ID_EISA		; Set Bus ID.

	xor	eax, eax
	mov	ax, cs
	shl	eax, 4
	mov	CodeAddress, eax

;-----------------------------------------------------------------
; Check if Virtual DMA Services are available
; if available Double Buffer Transmits
;-----------------------------------------------------------------
	push	ds
	mov	ax, 40h
	mov	ds, ax
	test	byte ptr ds:7bh, 20h
	pop	ds
	jz	short CheckForEISA   		;pws Z VDS not loaded.

;;VDS is loaded Set up for Double Buffered Transmits.

;;We now default to double buffering
;;	call	SetUpDblBuff			;Set up Double Buffer.


;-----------------------------------------------------------------
; Get physical address of our Code image
;-----------------------------------------------------------------
	pusha					;save registers.
	push	ds
	push	es
						
	lea	di, CodeDescriptor			;Load DMA Descriptor
	mov	word ptr [di].RegionSize, 100h		;save size of region.
	mov	word ptr [di].DMAOffset, 0h	;offset BMTCB.
	mov	word ptr [di].DMASegment, cs	;segment of ram struc.
	
	push	cs
	pop	es
	push	di
	mov	ax, 8103h		;Lock down our RingStructure 		
	mov	dx, 0004h		;do not allocate buffer if region not contiuous
	int	4bh			;doit
	pop	di
	pop	es
	pop	ds

	mov	eax, [di].DMAPhysAddr
	mov	CodeAddress, eax	   ;put phys. address in CodeAddress
	popa

	jnc	short UnlockMemBlock
	mov	dx, offset CGroup:PhysicalAddressNotKnown
	call	MSMPrintStringZero
	jmp	short CheckForEISA


;------------------------------------------------------------------
; Free memory that was just locked
;------------------------------------------------------------------
UnlockMemBlock:
	pusha
	push	es
	mov	ax, 8104h
	xor	dx, dx
	lea	di, CodeDescriptor
	push	ds
	pop	es

	int	4bh
	pop	es
	popa

;-----------------------------------------------------------------
; Check if we are even on an EISA Bus
;-----------------------------------------------------------------
CheckForEISA:

;***************************************************************\
; Find the first NE3200 NIC in the machine, 			*
;     or the one specified in the NET.CFG.			*
;***************************************************************/
	mov	dx, 0C80h			; Set to slot zero, but checks next one first.
	mov	si, Slot			; Get slot number from configtable.
	or	si, si				; Zero?
	jz  short ScanSlots			; Z = Scan till found, or out of slots.

	mov	cx, si
SLOTLOOP:
	add	dx, 1000h			; DX = xC80h. (next Slot please)
	loop	SLOTLOOP			; Get to the Correct? Slot

	cmp	si, TOTAL_EISA_SLOTS		; Net.cfg has given a slot, IS it possible?
	ja  short FinishScanSlots		; if too big forget it.
	jmp short NetCFGSLOT			; Go Check it once

ScanSlots:
	inc	si				; Update slot number.
	cmp	si, TOTAL_EISA_SLOTS		; Are we done?
	ja  short FinishScanSlots		; Jump if so.

	and	dx, 0FFF0h			; Roll back to first id.
	add	dx, 1000h			; DX = xC80h. (next Slot please)

NetCFGSLOT:
	in	al, dx				; Get first ID byte.
	cmp	al, MANUFACTURE_CHAR_CODE_1	; Is this our ID?
	jne short ScanSlotChk			; If not goto next slot if one not given

	inc	dx				; Move to next ID port.
	in	al, dx				; Get next ID byte.
	cmp	al, MANUFACTURE_CHAR_CODE_2	; Is this our ID?
	jne short ScanSlotChk			; If not goto next slot if one not given

	inc	dx				; Move to next id port.
	in	al, dx				; Get next ID byte.
	cmp	al, PRODUCT_NUMBER_1		; Is this our ID?
	jne short ScanSlotChk			; If not goto next slot if one not given

	mov	Slot, si			; Update Slot Number.
	jmp short FinishScanSlots		; Got the Slot

;****************************************************************************
;	Error Out Area
;****************************************************************************

;----------------------------------------------------------------------------
;	An Interrupt Level Not Found
;----------------------------------------------------------------------------
NOIntFound:
	mov	dx, offset CGroup:IntNotFoundMSG

;----------------------------------------------------------------------------
;		DriverInit Error Exit..... DX -> Error MEssage Offset.
;----------------------------------------------------------------------------
DriverInitErrorExit:
	mov	ax, MSMERR_HSM_FAILED		; Return Error
	or	ax, ax				; Set appropriate flags
	ret

;****************************************************************************
;	End of Error Out Area ---- Continue with our regular scheduled Program
;****************************************************************************
ScanSlotChk:
	cmp	Slot, 0				; Was a Slot Given
	je  short ScanSlots			; Z = Look Some More
	mov	si, -1				; Flag Not Found

FinishScanSlots:
	cmp	Slot, si			; Any cards found?
	mov	dx, Offset CGroup:AdapterNotFound ; DX -> Error message.
	jne short DriverInitErrorExit		; NZ => no Card found.

;----------------------------------------------------------------------------
;	Store Base I/O Port in ConfigTable	(Based on Slot Value)
;----------------------------------------------------------------------------
	movzx	eax, si				; Grab Slot Number
	shl	eax, 12				; EAX = Slot Index
	add	word ptr IOAddress1, ax		; AX = Base I/O Port.

;----------------------------------------------------------------------------
;	Setup Post Office
;----------------------------------------------------------------------------
	mov	si, offset CGroup:IORegistersDef	; Pointer to Table of MailBox and Registers
	mov	cx, IORegisterSpace			; Number of Bytes
	shr	cx, 1					; CX = Number of Registers

MarkMailBoxLoop:
	add	word ptr [si], ax	; Add in the Slot Specific Number for all
	add	si, 2			; Next Register
	dec	cx			; Count down
	jnz short MarkMailBoxLoop			

;----------------------------------------------------------------------------
;	Read configuration from EISA BIOS	
;----------------------------------------------------------------------------
	mov	cx, word ptr Slot		; Grab Slot Number again. Start with Block 0.

ReadConfigBlockLoop:
	push	cx			; Save Block/Slot number CH=Block, CL = Slot
	lea	si, EISAcfgTable	; SI = Offset to 320 byte table 
	mov	ax, 0D801h		; AH = 0D8h, AL = 01  =>Read Function Configuration Information.
	int	15h			; EISA BIOS Call

	pop	cx			; Restore Block/Slot Number
	jc  short NOIntFound		; Carry Flag Signals Error.

	inc	ch			; CH = Next Block
	mov	dl, byte ptr EISAcfgTable.EISAInterrupt ; DL = Possible Int.
	and	dl, ISOLATE_INT_MASK	; Strip out interrupt field
	or	dl, dl			; Anything left?
	jz  short ReadConfigBlockLoop	; Jump if nothing is left and try again.

	mov	al, byte ptr EISAcfgTable.EISAFuncValid ; EAX = Function Info Valid Flag
	test	al, EISA_INT_FUNCTION_BIT		; Vailid int?
	jnz short StoreInterruptLevel			; Got it!!

	jmp short ReadConfigBlockLoop	; Try again.

;----------------------------------------------------------------------------
;	Store and Set up Interrupt Line
;----------------------------------------------------------------------------
StoreInterruptLevel:
	mov	IRQLine1, dl		; Store Int to Config Table.
	cli
	call	MSMSetIRQ		; install our Interrupt
	sti

;----------------------------------------------------------------------------
;	Check on the Connector Type 
;----------------------------------------------------------------------------
	mov	si, offset CGroup:EISAcfgTable	; Get pointer to Cfg table
	test	byte ptr [si].EISAConnect, 08h	; Look for Bit 3 to be set
	jz  short MustBeDIX			; z = Not Set -- Must be Dix

	mov	dx, offset CGroup:ConnectorIsBNC
	jmp short	GoOn

MustBeDIX:
	mov	dx, offset CGroup:ConnectorIsDIX

;----------------------------------------------------------------------------
;  Know the Connector, Tell the World and then Set up Parameter Block Values
;----------------------------------------------------------------------------
GoOn:
	call	MSMPrintStringZero			; Tell the World

	mov	eax, CodeAddress			; EAX = Segment
	mov	edx, offset CGroup:MSMPhysNodeAddress	; EDX = Offset
	add	edx, eax			; EDX = Physical Address
	mov	NodeAddressPointer, edx 	; Store Physical Address;

	mov	edx, offset CGroup:BMFrameSize	; EDX = Offset
	add	edx, eax			; EDX = Physical Address
	mov	BMFrameSizePtr, edx		; Store Physical Address

	mov	edx, offset CGroup:FirmWareStatLabel
	add	edx, eax			; EDX = Physical Address
	mov	FirmWareStatisticsPtr, edx	; Store Physical Address

	mov	dx, NumberOfCustomCounters	; How many Custom Counters
	mov	CustomStatCounter, dx		; Store Number

;----------------------------------------------------------------------------
;	Give the Adapter a Kick to Restart the EPROM Resetting the Adapter
;----------------------------------------------------------------------------
	mov	dx, NE3200ResetRegister		; DX -> NIC Reset Port.
	mov	eax, 1				; Set Reset Bit.
	out	dx, eax				; Reset the NIC.

	mov	dx, StartMailBoxRegs		; Clear first 4 MailBoxes.
	xor	eax, eax
	out	dx, eax

	xor	cx, cx				; Allow reset settling time.

ShutdownHoldReset:
	loop short ShutdownHoldReset

	mov	dx, NE3200ResetRegister		; DX -> NIC Reset Port.
	out	dx, eax				; Clear NIC's reset state.
	
;----------------------------------------------------------------------------
;	Check on the Adapter has the Hardware Reset yet?
;----------------------------------------------------------------------------
	mov	dx, StartMailBoxRegs		; Check the Mail.
	mov	cx, -1				; Wait for a while 
	shl	ecx, 3

WaitForHardware:
	in	al, dx				; Read Status
	or	al, al				; Non Zero return?
	jnz  short TimeToDownLoad		

	loopd short WaitForHardware			; Wait for a while

	mov	dx, offset CGroup:AdapterROMErrorMSG ; Adapter never Responded.
	jmp DriverInitErrorExit			     ; (need long jmp)	

;----------------------------------------------------------------------------
;	ROM has given a return code -- what is it?
;----------------------------------------------------------------------------
ROMErrorChk:		
	cmp	al, StaticRamError		    ; Static Ram Error detected?
	mov	dx, offset CGroup:StaticRamErrorMSG ; Message for the World
	je	DriverInitErrorExit		    ; Exit time!! (need long jmp)

	mov	dx, offset CGroup:ROMErrorMSG	; Generic Error Msg 
	jmp	DriverInitErrorExit		; (need long jmp)

;----------------------------------------------------------------------------
;	Time to Tell the Card to Down Load the FirmWare
;----------------------------------------------------------------------------
TimeToDownLoad:
	cmp	al, 80h				; ROM's Ok sign?
	jne short ROMErrorChk			; equal => good sign

	inc	dx				; DX => MailBox 1
	mov	ax, FirmWareSize		; Get the Size of the FirmWare Need upper bytes zeroed
	out	dx, ax				; Output Both Bytes

	add	dx, 3				; EDX => MailBox 4
	mov	eax, CodeAddress
	add	eax, offset CGroup:FirmWareStart ; Pointer to the Start
	out	dx, eax				; Write out 4 bytes

	add	dx, 4				; EDX => MailBox 8
	xor	eax, eax			; CLEAR!
	out	dx, eax				; Send Destination Addr -- Load at 0.

	mov	dx, StartMailBoxRegs		; EDX => MailBox 0
	out	dx, al				; Signal we are Ready

;----------------------------------------------------------------------------
; Hook into INT 21h so that we can gain control when the MSM exists
; to TSR us.  Then we will Allocate the Size of the Tx collection buffers 
; without screwing up the MSM init code, or adding extra Code to the Tx Path.
;----------------------------------------------------------------------------
	cli
	sub	ax, ax
	mov	es, ax
	mov	bx, offset CGroup:Int21ISR	 ; Hooks Int 21 
	mov	cx, cs
	xchg	word ptr es:[21h*4+0], bx
	xchg	word ptr es:[21h*4+2], cx
	mov	OrgInt21Vector+0, bx
	mov	OrgInt21Vector+2, cx
	sti					

;----------------------------------------------------------------------------
; 		Wait for the FirmWare to get going.
;----------------------------------------------------------------------------
	mov	dx, ParameterMailBox	; EDX => MailBox C
	mov	cx, -1
	shl	ecx, 3			; Wait for awhile

WaitForDownLoadLoop:
	in	eax, dx			; EAX = DownLoad Status
	cmp	eax, FirmWareRunOK_FLAG	; Has the FirmWare given the Signal?
	je  short FirmWareIsGood	; got the signal

	loopd short WaitForDownLoadLoop	

;----------------------------------------------------------------------------
;	Has the Code but never Got Running
;----------------------------------------------------------------------------
	mov	dx, offset CGroup:FirmWareStartErrMSG
	jmp	DriverInitErrorExit		; Waited too long => Problems? Timed Out.(need long jmp)

;----------------------------------------------------------------------------
;	Give Adapter the Parameter Address
;----------------------------------------------------------------------------
FirmWareIsGood:
	mov	eax, CodeAddress		  ; Get Segment
	mov	edx, offset CGroup:PARAMETERBLOCK ; offset to table
	add	eax, edx			; Physical Address of Parameter Block.
	mov	dx,  ParameterMailBox
	out	dx, eax				; Tell the Adapter
	
;----------------------------------------------------------------------------
; 	Check For NodeOverRide
;----------------------------------------------------------------------------
	mov	ax, word ptr MSMPhysNodeAddress		; get first word
	and	ax, word ptr MSMPhysNodeAddress + 2	; And it with the 2nd
	and	ax, word ptr MSMPhysNodeAddress + 4	; And it with the 3rd
	cmp	ax, -1					; Still ALL FF's
	je   short NoNodeOverRide

;----------------------------------------------------------------------------
; Signal Mail is here, and if Node Override in effect.
;----------------------------------------------------------------------------
	mov	al, 00Fh			; 00Fh => Node Over Ride Given

NoNodeOverRide:					; If JMPed here then 0FFh => NoNodeOverRide
	mov	dx, UpDateParamMailBox		; To the Adapter
	out	dx, al

;----------------------------------------------------------------------------
;	Wait for the Adapter to get caught up
;----------------------------------------------------------------------------
	mov	dx, Abend_IdleMailBox
	mov	ecx, -1				; Hang Prevention.

WaitForAdapter:
	dec	ecx				; Prevent perment looping
	jz   short AdapterHUNG

	in	al, dx				; Read Status
	cmp	al, ABEND_STATE_FLAG		; Did it Die?  'A'
	je   short AdapterAbended		; Get Code.	

	cmp	al, FIRMWARE_STATE_FLAG		; Got the 'J' yet?
	jne  short WaitForAdapter		; Wait for a while

;----------------------------------------------------------------------------
;	Enable the DoorBells and the System Interrupts
;----------------------------------------------------------------------------
EnableDoorBell:
	mov	dx, HostDoorBell	; Enable door bells
	dec	dx			; DX = HostDoorBellEnable
	mov	al, DoorBellBits
	out	dx, al			; Tell the Card

	mov	dx, SystemIntEnable	; Enable the EINT line
	mov	al, EISAIntEnableBit	
	out	dx, al			; Tell the Card

	mov	MSMLookAheadSegment, cs	; Tell MSM which Segment will Lookahead be in.
	mov	MSMTxFreeCount, 1	; DriverMaxActiveTx=1 

	xor	ax, ax			; Mark of Success!!!
	cli
	jmp	MSMMediaConfigUpdate	; No error here, pass on

AdapterAbended:
	inc	dx			; Code Port
	in	al, dx			; got Code
	mov	ah, al			; save code
	shr	al, 4			; Get Nibble at a time
	call	MSMGetHexChar		; Get ASCII Char
	mov	FirmWareAbendCode, al	; First Digit
	and	ah, 0Fh			; Keep low nibble
	mov	al, ah			; get nibble
	call	MSMGetHexChar		; Get ASCII Char
	mov	FirmWareAbendCode+1, al	; First Digit

AdapterHUNG:
	mov	dx, offset CGroup:FirmWareFailedMSG
	jmp	DriverInitErrorExit	;(need long jmp)

DriverInit	endp

;****************************************************************************
;	EISA Config Table Structure   (320 Bytes)
;****************************************************************************
EISAcfgTableStruc	struc
	EISAInfo	db	022h dup (?)	; 000h
	EISAFuncValid	db	?	    	; 022h Determines if Int for Block is Valid
	EISAInfo1	db	08Fh dup (?)	; 023h
	EISAInterrupt	db 	?		; 0B2h Card's Interrupt Number
	EISAInfo2	db	054h dup (?)	; 0B3h
	EISAConnect	db	?		; 107h Adapter's Output Connector (BNC = 29  DIX = 21)
	EISAInfo3	db	038h dup (?)	; 108h
	EISAEND		dw	?
EISAcfgTableStruc	ends

;****************************************************************************
;	EISA config Table Space.....
;****************************************************************************
	EISAcfgTable	EISAcfgTableStruc <>

;****************************************************************************
;	Set up to do Double Buffering
;****************************************************************************
DisableDblBuff	proc	near
	mov	TxCollSize, TxCollSizeDef  ; Make the Buffer large enough for 1 pkt.
	ret	
DisableDblBuff	endp

;****************************************************************************
;	Int21ISR
;
;	Called by every INT 21 request.  We will look for the MSM
;	termination function code.  Either the MSM will TSR or it will
;	exit with a error code.
;
;	assumes:	Interrupts disabled (Invoked by a INT instruction)
;
;	returns:	All registers and flags preserved
;
       Align 4
TempTxColBuffer	db	1514 dup (0) 		;1500+14 = 1 data Packet + Headers
OrgInt21Vector	dw	2 dup (0)

Int21ISR	proc	far

	cmp	ah, 31h				;TSR function code?
	je  short MSMIsTSRing

	cmp	ah, 4Ch				;MSM exit code?
	je  	short MSMIsExiting

	jmp	dword ptr cs:OrgInt21Vector	;Chain on

;----------------------------------------------------------------------------
;	If we are going to TSR we must allocate our Tx-Rx buffer
;	from the end of our TSR memory.
;----------------------------------------------------------------------------
MSMIsTSRing:
	push	ds			;Save what you destroy
	push	cs			;DS=CGroup
	pop	ds

;----------------------------------------------------------------------------
;	DX has the number of paragraphs that are currently staying resident.
;----------------------------------------------------------------------------
	shl	dx, 4			; Make it an offset
	mov	PtrBMTCB, dx		; Store addr of TxCollection buffer

	add	dx, Size BMTCBstructure	; Allocate Space for BMTCB
	add	dx, TxCollSize		; Add enough for the TxCollBuffer.
	add	dx, 15			; Round up
	shr	dx, 4			; Make a Paragraph again

	push	ax
	push	bx
	mov	eax, CodeAddress
	movzx	ebx, word ptr PtrBMTCB
	add	ebx, BMTCBTxBuffer	; eax -> Collection Buffer
	add	eax, ebx
	mov	PhysicalBMTCB, eax	; save BMTCB's Physical address
	pop	bx
	pop	ax

;----------------------------------------------------------------------------
;	If the MSM is NOT TSRing then all we need to do is unlink
;	us from the INT 21 vector chain.
;----------------------------------------------------------------------------
RestoreVector:
	pop	ds			;FALL THROUGH TO UNCHAIN CODE

MSMIsExiting:
	push	ax
	push	ds
	sub	ax, ax
	mov	ds, ax
	mov	ax, cs:OrgInt21Vector+0
	mov	word ptr ds:[21h*4+0], ax
	mov	ax, cs:OrgInt21Vector+2
	mov	word ptr ds:[21h*4+2], ax
	pop	ds
	pop	ax
	jmp	dword ptr cs:OrgInt21Vector

Int21ISR	endp

Init	ends

end

