/*--------------------------------------------------------------------------*
 *  Copyright Unpublished Work of Novell, Inc. All Rights Reserved
 *
 *  THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 *  PROPRIETARY AND TRADE SECRET INFORMATION OF NOVELL, INC.
 *  ACCESS TO THIS WORK IS RESTRICTED TO (I) NOVELL EMPLOYEES
 *  WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE
 *  OF THEIR ASSIGNMENTS AND (ii) ENTITIES OTHER THAN NOVELL
 *  WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.
 *  NO PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED,
 *  DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED,
 *  CONDENSED, EXPANDED, COLLECTED, COMPILED, LINKED, RECAST,
 *  TRANSFORMED OR ADAPTED WITHOUT THE PRIOR WRITTEN CONSENT OF
 *  NOVELL.  ANY USE OR EXPLOITATION OF THIS WORK WITHOUT AUTHORIZATION
 *  COULD SUBJECT THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY.
 *--------------------------------------------------------------------------*/
/*****************************************************************************
 *
 * Title:			CMSM String Support Module.
 *
 * Filename:		cmsmstr.c
 *
 * ODI Spec Ver:	1.11
 *
 * Description:		Special collection of string and sscanf functions
 *                 	for the LAN CHSM parameter parser.  Note that only
 *                 	the functions needed by the parameter parser are
 *                 	included here and some of the functions have been
 *                 	modified and/or reduced to match the parameter
 *                 	parsers needs.  This is not a general purpose
 *                 	library.
 *
 * Modification History:
 *
 * 01-15-98 KRA		Added #include <ntos.h> for the ifdef OS_NT case.
 *
 ****************************************************************************/

#ifdef _WATCOMC_
#define WATCOM_LIBRARY_COMPATABILITY_MODE
#endif

#ifdef WATCOM_LIBRARY_COMPATABILITY_MODE
#define VA_LIST_TOP_TYPE_IS_ARRAY 1
#else
#define VA_LIST_TOP_TYPE_IS_ARRAY 0
#endif

#ifdef OS_NT
#include <ntos.h>
#endif
#include <odi.h>
#include <odi_nbi.h>
#include <cmsm.h>
#include <cmsmstr.h>
#include <splat.h>

enum			/* all the various types that can be given to va_arg() */
{
	T_INT,	/* default */
	T_LONG,
	T_DBL,
	T_LDBL,
	T_VPTR,
	T_CPTR,
	T_SPTR,
	T_IPTR,
	T_LPTR,
	T_FPTR,
	T_DPTR,
	T_LDPTR,
	T_PPTR
};



#define F_EOF		0x0001	/* GET() returned EOF (or sometimes bad multibyte) */
#define F_HALF		0x0002	/* 'h' */
#define F_LONG		0x0004	/* 'l' */
#define F_LDBL		0x0008	/* 'L' */
#define F_STAR		0x0010	/* '*' */
#define F_UINT		0x0020	/* unsigned integer destination */
#define F_VPTR		0x0040	/* void * "integer" destination */
#define F_NEG		0x0080	/* negate resulting value */
#define F_ALLOC	0x0100	/* result is allocated string */
#define F_SIGN		0x0200	/* seen leading [+-] for numeric value */
#define F_ZERO		0x0400	/* seen leading 0[bBxX] for integer value */

typedef struct
{
	size_t		width;	/* max number of bytes/characters to scan */
	size_t		total;	/* number of bytes/characters processed */
	Uint			flags;	/* the above bits */
	Uint			uival;	/* base for integer; array length for float */
	Uchar			*cur;		/* current source string scan pointer	*/
	union
	{
		Ulong	ul;
		long	l;
		Uchar	*str;
	}				ret;		/* "return value" shapes */
} Scan;



/*-----------[ Non exported proto types. ]----------------------------------*/
int _cdecl  CMSMidoscan(
				const char *src,
				const char *fmt,
				va_list ap);

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/



#define _XDIGT_	_XDIGIT_
#define _______	0
#define ___			0

unsigned char	__ctype[] =
{
/*
** This table is composed of the following elements...
** 000			classification for character EOF
** 001-127		classification of ASCII characters
** 128-256		unused padding corresponding to high-order ASCII
** 257			unused padding
** 258-385		upper/lower casing tables (unused: USL's code was broken)
** 386-513		high-order ASCII part of same (equally broken)
** 514-521		7 values used for CSWIDTH (unknown as yet)
*/
/*					_XDIGT_,_BLANK_,_CNTRL_,_PUNCT_,_SPACE_,_DIGIT_,_LOWER_,_UPPER_*/
/* -1,EOF */	_______|_______|_______|_______|_______|_______|_______|_______,
/* 00,Nul */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 01,SOH */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 02,STX */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 03,ETX */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 04,EOT */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 05,ENQ */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 06,ACK */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 07,Bel */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 08,BS  */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 09,Tab */	_______|_______|_CNTRL_|_______|_SPACE_|_______|_______|_______,
/* 0a,LF  */	_______|_______|_CNTRL_|_______|_SPACE_|_______|_______|_______,
/* 0b,VT  */	_______|_______|_CNTRL_|_______|_SPACE_|_______|_______|_______,
/* 0c,FF  */	_______|_______|_CNTRL_|_______|_SPACE_|_______|_______|_______,
/* 0d,CR  */	_______|_______|_CNTRL_|_______|_SPACE_|_______|_______|_______,
/* 0e,SO  */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 0f,SI  */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 10,DLE */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 11,DC1 */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 12,DC2 */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 13,DC3 */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 14,DC4 */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 15,NAK */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 16,SYN */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 17,ETB */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 18,CAN */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 19,EM  */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 1a,SUB */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 1b,Esc */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 1c,FS  */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 1d,GS  */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 1e,RS  */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 1f,US  */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,
/* 20,' ' */	_______|_BLANK_|_______|_______|_SPACE_|_______|_______|_______,
/* 21, !  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 22, "  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 23, #  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 24, $  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 25, %  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 26, &  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 27, '  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 28, (  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 29, )  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 2a, *  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 2b, +  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 2c, ,  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 2d, -  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 2e, .  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 2f, /  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 30, 0  */	_XDIGT_|_______|_______|_______|_______|_DIGIT_|_______|_______,
/* 31, 1  */	_XDIGT_|_______|_______|_______|_______|_DIGIT_|_______|_______,
/* 32, 2  */	_XDIGT_|_______|_______|_______|_______|_DIGIT_|_______|_______,
/* 33, 3  */	_XDIGT_|_______|_______|_______|_______|_DIGIT_|_______|_______,
/* 34, 4  */	_XDIGT_|_______|_______|_______|_______|_DIGIT_|_______|_______,
/* 35, 5  */	_XDIGT_|_______|_______|_______|_______|_DIGIT_|_______|_______,
/* 36, 6  */	_XDIGT_|_______|_______|_______|_______|_DIGIT_|_______|_______,
/* 37, 7  */	_XDIGT_|_______|_______|_______|_______|_DIGIT_|_______|_______,
/* 38, 8  */	_XDIGT_|_______|_______|_______|_______|_DIGIT_|_______|_______,
/* 39, 9  */	_XDIGT_|_______|_______|_______|_______|_DIGIT_|_______|_______,
/* 3a, :  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 3b, ;  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 3c, <  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 3d, =  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 3e, >  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 3f, ?  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 40, @  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 41, A  */	_XDIGT_|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 42, B  */	_XDIGT_|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 43, C  */	_XDIGT_|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 44, D  */	_XDIGT_|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 45, E  */	_XDIGT_|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 46, F  */	_XDIGT_|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 47, G  */	_______|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 48, H  */	_______|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 49, I  */	_______|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 4a, J  */	_______|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 4b, K  */	_______|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 4c, L  */	_______|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 4d, M  */	_______|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 4e, N  */	_______|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 4f, O  */	_______|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 50, P  */	_______|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 51, Q  */	_______|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 52, R  */	_______|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 53, S  */	_______|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 54, T  */	_______|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 55, U  */	_______|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 56, V  */	_______|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 57, W  */	_______|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 58, X  */	_______|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 59, Y  */	_______|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 5a, Z  */	_______|_______|_______|_______|_______|_______|_______|_UPPER_,
/* 5b, [  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 5c, \  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 5d, ]  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 5e, ^  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 5f, _  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 60, `  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 61, a  */	_XDIGT_|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 62, b  */	_XDIGT_|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 63, c  */	_XDIGT_|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 64, d  */	_XDIGT_|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 65, e  */	_XDIGT_|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 66, f  */	_XDIGT_|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 67, g  */	_______|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 68, h  */	_______|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 69, i  */	_______|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 6a, j  */	_______|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 6b, k  */	_______|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 6c, l  */	_______|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 6d, m  */	_______|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 6e, n  */	_______|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 6f, o  */	_______|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 70, p  */	_______|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 71, q  */	_______|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 72, r  */	_______|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 73, s  */	_______|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 74, t  */	_______|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 75, u  */	_______|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 76, v  */	_______|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 77, w  */	_______|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 78, x  */	_______|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 79, y  */	_______|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 7a, z  */	_______|_______|_______|_______|_______|_______|_LOWER_|_______,
/* 7b, {  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 7c, |  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 7d, }  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 7e, ~  */	_______|_______|_______|_PUNCT_|_______|_______|_______|_______,
/* 7f,Del */	_______|_______|_CNTRL_|_______|_______|_______|_______|_______,

/*
** These are 8-bit ASCII characters between 128 and 255 which everyone uses
** differently and here are simply padding for now--perhaps they'll be used
** someday so we keep them in.
*/
/*              0   1   2   3   4   5   6   7   8   9   a   b   c   d   e   f */
/* 80     */	___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,
/* 90     */	___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,
/* a0     */	___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,
/* b0     */	___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,
/* c0     */	___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,
/* d0     */	___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,
/* e0     */	___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,
/* f0     */	___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,

	0,

/*
** tolower()  and toupper() conversion table. USL used these subscripts with-
** out checking their applicability. For instance, the upper-case value of 'a'
** is correctly found at the expected offset ((__ctype + 258)['a']) just as the
** lower-case value for 'A', but tolower() was coded identically and so the
** lower-case value of 'a' is rendered 'A' and toupper() returns the upper-case
** value of 'A' as being 'a'! It has been recoded to ignore the table. Also,
** high-order ASCII part of the table was initialized to 0 (as it still is)
** which produced less than exciting results when the argument to toupper() or t
** tolower() falls in this range.
*/
	0,		1,		2,		3,		4,		5,		6,		7,
	8,		9,		10,	11,	12,	13,	14,	15,
	16,	17,	18,	19,	20,	21,	22,	23,
	24,	25,	26,	27,	28,	29,	30,	31,
	32,	33,	34,	35,	36,	37,	38,	39,
	40,	41,	42,	43,	44,	45,	46,	47,
	48,	49,	50,	51,	52,	53,	54,	55,
	56,	57,	58,	59,	60,	61,	62,	63,
	64,	97,	98,	99,	100,	101,	102,	103,
	104,	105,	106,	107,	108,	109,	110,	111,
	112,	113,	114,	115,	116,	117,	118,	119,
	120,	121,	122,	91,	92,	93,	94,	95,
	96,	65,	66,	67,	68,	69,	70,	71,
	72,	73,	74,	75,	76,	77,	78,	79,
	80,	81,	82,	83,	84,	85,	86,	87,
	88,	89,	90,	123,	124,	125,	126,	127,

/*              0   1   2   3   4   5   6   7   8   9   a   b   c   d   e   f */
/* 80     */	___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,
/* 90     */	___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,
/* a0     */	___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,
/* b0     */	___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,
/* c0     */	___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,
/* d0     */	___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,
/* e0     */	___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,
/* f0     */	___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,___,

/* CSWIDTH information */
	1,	0,	0,	1,	0,	0,	1
};

int _cdecl  CMSMisalpha( int c )
{
	return((__ctype + 1)[c] & (_UPPER_ | _LOWER_));
}

int _cdecl  CMSMisupper( int c )
{
	return((__ctype + 1)[c] & _UPPER_);
}

int _cdecl  CMSMislower( int c )
{
	return((__ctype + 1)[c] & _LOWER_);
}

int _cdecl  CMSMisdigit( int c )
{
	return((__ctype + 1)[c] & _DIGIT_);
}

int _cdecl  CMSMisxdigit( int c )
{
	return((__ctype + 1)[c] & _XDIGT_);
}

int _cdecl  CMSMisalnum( int c )
{
	return((__ctype + 1)[c] & (_UPPER_ | _LOWER_ | _DIGIT_));
}

int _cdecl  CMSMisspace( int c )
{
	return((__ctype + 1)[c] & _SPACE_);
}

int _cdecl  CMSMispunct( int c )
{
	return((__ctype + 1)[c] & _PUNCT_);
}

int _cdecl  CMSMisprint( int c )
{
	return((__ctype + 1)[c] & (_PUNCT_ | _UPPER_ | _LOWER_ | _DIGIT_ | _BLANK_));
}

int _cdecl  CMSMisgraph( int c )
{
	return((__ctype + 1)[c] & (_PUNCT_ | _UPPER_ | _LOWER_ | _DIGIT_));
}

int _cdecl  CMSMiscntrl( int c )
{
	return((__ctype + 1)[c] & _CNTRL_);
}

int _cdecl  CMSMisascii( int c )
{
	return(!(c & ~0x7f));
}

int _cdecl  CMSMtoascii(int c)
{
	return (c & 0x7f);
}

int _cdecl  CMSMtoupper(int c)
{
	return (CMSMislower(c)) ? c - ' ' : c;
}

int _cdecl  CMSMtolower(int c)
{
	return (CMSMisupper(c)) ? c + ' ' : c;
}


/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/


MEON_STRING * _cdecl  CMSMstrcat(
				register MEON_STRING			*t,
				register const MEON_STRING	*s)
{
	MEON_STRING	*base = t;

	while (*t++)
		;

	t--;

	while (*t++ = *s++)
		;

	return base;
}


/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/



int _cdecl  CMSMsscanf(
				const MEON_STRING	*src,
				const MEON_STRING	*fmt,
				...)
{
	int		res;
	va_list	ap;

	va_start(ap, fmt);

	res = CMSMidoscan((char *)src, (char *)fmt, ap);

	va_end(ap);
	return res;
}




/*==========================================================================*/
/*==========================================================================*/


/* altered version of getc() */
static int _cdecl  dogetb(
				Scan *sp)
{
	sp->flags &= ~F_EOF;	/* assume that there are more */

	if (*(sp->cur) != '\0')	/* still are characters */
	{
		return *sp->cur++;
	}

	sp->flags |= F_EOF;
	return EOF;

}

#define GET(sp)		(dogetb(sp))
#define UNGET(sp)	(--(sp->cur))




/*==========================================================================*/
/*==========================================================================*/

/* process integer field */
static int _cdecl  scanint(
				register Scan *sp)
{
	register Ulong ul;
	register Uint ch;

/*
* Gather any sign and base-specifying prefix
* until the first digit of the integer has been scanned.
*/
	for (;;)
	{
		switch (ch = GET(sp))
		{
ErrorBack :
			UNGET(sp);
			return 0;
ReturnBack :
			UNGET(sp);
			sp->ret.ul = 0;
			return 1;
		default:
			if (sp->flags & (F_SIGN | F_ZERO))
				goto ErrorBack;
			if (!CMSMisspace(ch))
				goto ErrorBack;
			sp->total++;
			continue;
		case '-':
			sp->flags |= F_NEG;
			/*FALLTHROUGH*/
		case '+':
			if (sp->flags & (F_SIGN | F_ZERO))
				goto ErrorBack;
			if (--sp->width == 0)
				return 0;
			sp->total++;
			sp->flags |= F_SIGN;
			continue;
		case '0':
			if (sp->uival == 10 || sp->uival == 8
				|| (sp->flags & F_ZERO))
			{
				ul = 0;
				break;
			}
			sp->total++;
			if (--sp->width == 0)
			{
				sp->ret.ul = 0;
				return 1;
			}
			switch (ch = GET(sp))
			{
			default:
				if (sp->uival == 0)
					sp->uival = 8;

				ul = 0;
				goto gotdig;

			case 'b':
			case 'B':
				if (sp->uival == 16)
				{
					ul = 0;
					goto gotdig;
				}

				if (sp->width == 1 && sp->uival == 0)
					goto ReturnBack;

				sp->uival = 2;
				break;

			case 'x':
			case 'X':
				if (sp->uival == 2 || sp->width == 1 && sp->uival == 0)
					goto ReturnBack;

				sp->uival = 16;
				break;
			}
			if (--sp->width == 0)
				return 0;
			sp->total++;
			sp->flags |= F_ZERO;
			continue;
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			ul = ch - '0';
			if (sp->uival == 0)
				sp->uival = 10;
			else if (ul >= sp->uival)
				goto ErrorBack;
			break;
		case 'a':
		case 'b':
		case 'c':
		case 'd':
		case 'e':
		case 'f':
			ul = ch - 'a' + 10;
			if (sp->uival != 16)
				goto ErrorBack;
			break;
		case 'A':
		case 'B':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
			ul = ch - 'A' + 10;
			if (sp->uival != 16)
				goto ErrorBack;
			break;
		}
		break;
	}
	/*
	* Any leading sign, base-specific prefix, and the first digit
	* of the value have been scanned and handled.
	* Accumulate the value in a Ulong, ignoring any overflows.
	*/
	while (++sp->total, --sp->width != 0)
	{
		ch = GET(sp);
	gotdig:;
		if (CMSMisdigit(ch))
			ch -= '0';
		else if (!CMSMisxdigit(ch))
			ch = sp->uival;	/* bad value below */
		else if (CMSMisupper(ch))
			ch -= 'A' - 10;
		else
			ch -= 'a' - 10;
		if (ch >= sp->uival)
		{
			UNGET(sp);
			break;
		}
		ul *= sp->uival;
		ul += ch;
	}
	sp->ret.ul = ul;	/* this covers signed values, too */
	return 1;
}

#define NANSIZE	0

#define LP	'('
#define RP	')'


#define TABSIZE	(1 << CHAR_BIT)	/* size of the %[...] scanning table */

/* fill for '[' scan */
static const Uchar * _cdecl  maketab(
				register const Uchar *fmt,
				register Uchar *tab)
{
	register int ch, prev, fill, next;

	fill = 1;
	if (fmt[1] == '^')	/* complements the set */
	{
		fill = 0;
		fmt++;
	}
/* WTT 1/13/95 - start */
/*	CSetB(fill ^ 1, (void *)tab, (size_t)TABSIZE);*/
	CMSMmemset((void *)tab, fill ^ 1, (size_t)TABSIZE);
/* WTT 1/13/95 - end */
	if ((ch = fmt[1]) == ']' || ch == '-')	/* first gets special handling */
	{
		tab[ch] = fill;
		fmt++;
	}
	while ((ch = *++fmt) != ']')
	{
		if (ch == '\0')		/* unexpected end of format */
			return 0;
		if (ch == '-' && (next = fmt[1]) != ']'
			&& (prev = fmt[-1]) < next)
		{
			fmt++;
			do
				tab[++prev] = fill;
			while (prev < next);
			continue;
		}
		tab[ch] = fill;
	}
	return fmt;
}

#define NARGS	30	/* cache positions of this many arguments */

#define BUFLEN	(TABSIZE < NANSIZE ? NANSIZE : TABSIZE)








/*==========================================================================*/
/*==========================================================================*/


int _cdecl  CMSMidoscan
(
	const char *src,
	const char *fmt,
	va_list ap
)
{
	const char *orig_fmt = fmt;	/* saved start of format */
	Uchar buf[BUFLEN];		/* scan table & floating strings */
	RA_va_list arglist[NARGS];	/* cached arg positions */
	RA_va_list curpos;		/* current va_list position */
	int nmatch = 0;			/* number of conversions matched */
	int pos_arg_flag = 1;		/* no positional arguments so far */
	Scan field;			/* current field processing */

	field.total = 0;		/* no bytes/characters yet */
	field.flags = 0;		/* so that F_EOF is clear */
	field.cur   = (Uchar *)src;
	
#if VA_LIST_TOP_TYPE_IS_ARRAY
/*
** Since an array type gets converted to a pointer type when passed as a function
** argument, taking it's address (as in the else clause below), will not have the
** desired effect of obtaining a type compatible with (RA_va_list *) for use in
** an union assignment.  Rather, we should get the value  *(RA_val_list)ap which
** will be assignment compatible with curpos.
*/
	curpos = *(RA_va_list *) ap;	/* ap is type pointer, not type array */
#else
	curpos = *(RA_va_list *) &ap;
#endif
	for (;; fmt++)
	{
		/*
		* Scan format string for matching directives.
		* Eat white space given a white-space character.
		* A conversion specification starts with % and
		* is handled below.  Otherwise, the input must
		* match the format byte/character.
		*
		* CMSMidoscan assumes that bytes with the value %
		* and the single-byte white-space characters
		* cannot be part of a multibyte character.
		*/
		for (;; fmt++) /* loop until return or reach % */
		{
			switch (*fmt)
			{
			default:
				if (GET(&field) != *fmt)
				{
					UNGET(((Scan *)&field));
					goto mismatch;
				}
				field.total++;
				continue;
			case '\0':
				return nmatch;	/* successful */
			case '%':
				break;
			case ' ':
			case '\t':
			case '\n':
			case '\r':
			case '\f':
			case '\v':
				{
					register int ch;

					while (ch = GET(&field), CMSMisspace(ch))
						field.total++;
				}
				UNGET(((Scan *)&field));
				continue;
			}
			break;	/* only here for % */
		}
		/*
		* Must be at a %.  Process conversion specification.
		*/
		field.flags &= F_EOF;
		field.width = ~(size_t)0;
	again:;
		switch (*++fmt)
		{
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
		{
			register Uint n = *fmt - '0';

			while (++fmt, CMSMisdigit(*fmt))	/* ignore any overflow */
			{
				n *= 10;
				n += *fmt - '0';
			}
			if (*fmt != '$')	/* not position spec. */
			{
				fmt--;
				if ((field.width = n) == 0)	/* no zeros */
					field.width = ~(size_t)0;
				goto again;
			}
			if (pos_arg_flag)	/* first time */
			{
				pos_arg_flag = 0;
#if VA_LIST_TOP_TYPE_IS_ARRAY
/*
** Since an array type gets converted to a pointer type when passed as a function
** argument, taking it's address (as in the else clause below), will not have the
** desired effect of obtaining a type compatible with (RA_va_list *) for use in
** an union assignment.  Rather, we should get the value  *(RA_val_list)ap which
** will be assignment compatible with curpos.
*/
				arglist[0] = *(RA_va_list *)ap;	/* ap is type pointer, not type array */
#else
				arglist[0] = *(RA_va_list *)&ap;
#endif
				_parglist(orig_fmt, arglist, NARGS, 0, 1);
			}
			if (n > NARGS)
			{
				curpos = arglist[NARGS - 1];
				_parglist(orig_fmt, &curpos, NARGS, n, 1);
			}
			else
			{
				if (n == 0)
					n++;
				curpos = arglist[n - 1];
			}
			goto again;
		}
		/*
		* The following cases each end a conversion specification.
		* ('[' still needs to reach its matching ']', however.)
		*/
		default:
			return nmatch;	/* invalid format character */
		case '%':
			if (GET(&field) != '%')
			{
				UNGET(((Scan *)&field));
				goto mismatch;
			}
			field.total++;
			continue;
		case 'c':
		{
			register Uchar *p;
			register int ch;

			if (field.width == ~(size_t)0)
				field.width = 1;
			if ((ch = GET(&field)) == EOF)
				goto mismatch;
			if (field.flags & F_STAR)
				p = 0;
			else
				p = (Uchar *)va_arg(curpos.ap, char *);
			do
			{
				if (p != 0)
					*p++ = ch;
				field.total++;
				if (--field.width == 0)
					break;
			} while ((ch = GET(&field)) != EOF);
			if (p != 0)
				nmatch++;
			continue;
		case 's':
			while (ch = GET(&field), CMSMisspace(ch))
				field.total++;
			if (ch == EOF)
				goto mismatch;
			if (field.flags & F_STAR)
				p = 0;
			else
				p = (Uchar *)va_arg(curpos.ap, char *);
			for (;;)
			{
				if (p != 0)
					*p++ = ch;
				field.total++;
				if (--field.width == 0)
					break;
				if ((ch = GET(&field)) == EOF)
					break;
				if (CMSMisspace(ch))
				{
					UNGET(((Scan *)&field));
					break;
				}
			}
			if (p != 0)
			{
				*p = '\0';
				nmatch++;
			}
			continue;
		case '[':
			if ((fmt = (const char *)maketab((Uchar *)fmt, buf)) == 0)
				goto mismatch;
			if ((ch = GET(&field)) == EOF)
				goto mismatch;
			if (!buf[ch])
			{
				UNGET(((Scan *)&field));
				goto mismatch;
			}
			if (field.flags & F_STAR)
				p = 0;
			else
				p = (Uchar *)va_arg(curpos.ap, char *);
			for (;;)
			{
				if (p != 0)
					*p++ = ch;
				field.total++;
				if (--field.width == 0)
					break;
				if ((ch = GET(&field)) == EOF)
					break;
				if (!buf[ch])
				{
					UNGET(((Scan *)&field));
					break;
				}
			}
			if (p != 0)
			{
				*p = '\0';
				nmatch++;
			}
			continue;
		}
		case 'd':
		case 'D':
			field.uival = 10;
		integer:;
			if (!scanint(&field))
				goto mismatch;
			if (field.flags & F_STAR)
				continue;
			if (field.flags & F_UINT)	/* unsigned target */
			{
				if (field.flags & F_NEG)
					field.ret.l = -field.ret.l;
				if (field.flags & F_VPTR)
				{
					*va_arg(curpos.ap, void **)
						= (void *)field.ret.ul;
				}
				else if (field.flags & F_LONG)
				{
					*va_arg(curpos.ap, Ulong *)
						= field.ret.ul;
				}
				else if (field.flags & F_HALF)
				{
					*va_arg(curpos.ap, Ushort *)
						= *(Ushort *)&field.ret.ul;
				}
				else
				{
					*va_arg(curpos.ap, Uint *)
						= field.ret.ul;
				}
			}
			else	/* signed target */
			{
				if (field.flags & F_NEG)
				{
#if -SHRT_MAX > SHRT_MIN /* two's complement: don't negate most negative */
					if (field.ret.l != LONG_MIN)
						field.ret.l = -field.ret.l;
#else
					field.ret.l = -field.ret.l;
#endif
				}
				if (field.flags & F_LONG)
				{
					*va_arg(curpos.ap, long *)
						= field.ret.l;
				}
				else if (field.flags & F_HALF)
				{
					*va_arg(curpos.ap, short *)
						= *(short *)&field.ret.l;
				}
				else
				{
					*va_arg(curpos.ap, int *)
						= field.ret.l;
				}
			}
			nmatch++;
			continue;
		case 'u':
		case 'U':
			field.flags |= F_UINT;
			field.uival = 10;
			goto integer;
		case 'x':
		case 'X':
			field.flags |= F_UINT;
			field.uival = 16;
			goto integer;
		}
	}
mismatch:;
	if (nmatch != 0)
		return nmatch;
	if (field.flags & F_EOF)
		return EOF;
	return 0;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/


static void _cdecl  CMSMadv	/* advance "p->ap" by specified type */
(
	RA_va_list	*p,
	int			type
)
{
	switch (type)
	{
	default:
		(void) va_arg(p->ap, int);
		break;
	case T_LONG:
		(void) va_arg(p->ap, long);
		break;
	case T_VPTR:
		(void) va_arg(p->ap, void *);
		break;
	case T_CPTR:
		(void) va_arg(p->ap, char *);
		break;
	case T_SPTR:
		(void) va_arg(p->ap, short *);
		break;
	case T_IPTR:
		(void) va_arg(p->ap, int *);
		break;
	case T_LPTR:
		(void) va_arg(p->ap, long *);
		break;
	case T_PPTR:
		(void) va_arg(p->ap, void **);
		break;
	}
}

/*
** Manage positional argument list conversion specifications.
**
** Invoke either with printf-family format (scan == 0) or
** with scanf-family format (scan != 0).
**
** Invoked with (pos <= len) to fill an array with up to "len"
** RA_va_list's with cached "ap" locations.  In this case, "lp"
** points to the first element of the array to fill.  The first
** element must contain the original "ap" value for the ... args.
**
** Invoked with (pos > len) to change "*lp" to be ready for a
** va_arg() invocation for the "pos"-th argument.  "lp" must
** hold the value of the "len"-th element of the previously
** filled-in array.
*/
void _cdecl  _parglist
(
	const char	*fmt,
	RA_va_list	*lp,
	int			len,
	int			pos,
	int			scan
)
{
	RA_va_list tmp;
	int high, maxpos, argno;

	tmp = *lp;
	if (pos > len)		/* position "*lp" for arg "pos" */
	{
		high = len;
	}
	else if (len > 0)	/* first time: fill array with "int" */
	{
		register RA_va_list *p = &lp[len];
		register size_t n = len;

		do
			(--p)->flag = T_INT;
		while (--n != 0);
		high = 0;
		maxpos = 0;
	}
	argno = 1;	/* default to initial argument */
	/*
	* Outer loop only taken when positioning "*lp" for arg "pos".
	*/
	for (;;)
	{
		register const char *s = fmt;
		register int n, type;
		const char *x;

		/*
		* This code assumes that a byte with value % cannot be part
		* of a multibyte character.
		*/
		while ((s = (char *)CMSMstrchr((MEON_STRING *)s, '%')) != 0) /* each conversion spec. */
		{
			x = s;
			while (++s, CMSMisdigit(*s))
				;
			if (*s == '$')	/* move to specified argument */
			{
				if ((n = CMSMatoi((MEON_STRING *)(x + 1))) <= 0) /* bad position */
					n = 1;
				argno = n;
			}
			n = '\0';
		again:;
			switch (*s++)
			{
			default:	/* no argument (e.g. %%) */
				continue;
			case 'h':
			case 'l':
			case 'L':
				n = s[-1];
				/*FALLTHROUGH*/
			case '.':
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
			case '-':
			case '+':
			case '#':
			case '\'':
			case ' ':
			case '$':
				goto again;
			case '*':
				if (scan)	/* no argument */
					continue;
				if (CMSMisdigit(*s))
				{
					/*
					* Positional width or precision.
					* Always has "int" type:  Default
					* behavior does the right stuff.
					*/
					if (high == 0 && (n = CMSMatoi((MEON_STRING *)s)) > 0
						&& n <= len && n > maxpos)
					{
						maxpos = n;
					}
					while (++s, CMSMisdigit(*s))
						;
					goto again; /* *s should be '$' */
				}
				n = '*';
				type = T_INT;
				break;
			case 'c':
				if (scan)
					type = T_CPTR;
				else
					type = T_INT;
				break;
			case 's':
				type = T_CPTR;
				break;
			case '[':
				if (!scan)
					continue;	/* bogus spec. */
				if (*s == '^')
					s++;
				while (*s != '\0' && *++s != ']')
					;
				type = T_CPTR;
				break;
#ifndef NO_CI4
			case 'B':
			case 'D':
			case 'I':
			case 'O':
			case 'U':
				if (!scan)
					continue;	/* bogus spec. */
				type = T_LPTR;
				break;
#endif
			case 'X':
			case 'b':
			case 'd':
			case 'i':
			case 'o':
			case 'u':
			case 'x':
				if (scan)
				{
					if (n == 'l')
						type = T_LPTR;
					else if (n == 'h')
						type = T_SPTR;
					else
						type = T_IPTR;
				}
				else if (n == 'l')
					type = T_LONG;
				else
					type = T_INT;
				break;
			case 'p':
				if (scan)
					type = T_PPTR;
				else
					type = T_VPTR;
				break;
#ifndef NO_NCEG_FPE
			case 'A':
#endif
			case 'E':
			case 'F':
			case 'G':
#ifndef NO_NCEG_FPE
			case 'a':
#endif
			case 'e':
			case 'f':
			case 'g':
				if (scan)
				{
					if (n == '\0')
						type = T_FPTR;
					else if (n == 'L')
						type = T_LDPTR;
					else
						type = T_DPTR;
				}
				else if (n == 'L')
					type = T_LDBL;
				else
					type = T_DBL;
				break;
			case 'n':
				if (n == 'l')
					type = T_LPTR;
				else if (n == 'h')
					type = T_SPTR;
				else
					type = T_IPTR;
				break;
			}
			if (high == 0)	/* filling array mode */
			{
				if (argno <= len)
				{
					lp[argno - 1].flag = type;
					if (argno > maxpos)
						maxpos = argno;
				}
			}
			else if (argno == high)
			{
				CMSMadv(&tmp, type);
				if (++high >= pos)	/* reached target */
				{
					*lp = tmp;
					return;
				}
			}
			argno++;	/* default to next */
			if (n == '*')
			{
				n = '\0';
				goto again;
			}
		}
		/*
		* Scanned entire format string.
		* If array filling, do so (below) and return.
		* Otherwise, take "high" as an "int" and restart scan.
		*/
		if (high == 0)
			break;		/* go fill array */
		CMSMadv(&tmp, T_INT);
		if (++high >= pos)	/* reached target */
		{
			*lp = tmp;
			return;
		}
	}
	if (maxpos != 0) /* fill array, overwriting flag with ap value */
	{
		register RA_va_list *p = lp;
		register int type, n = maxpos;

		do
		{
			type = p->flag;
			*p++ = tmp;
			CMSMadv(&tmp, type);
		} while (--n != 0);
	}
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* compare two strings to see if equal. */
int	_cdecl  CMSMstrcmp
(
	register const MEON_STRING	*s1,
	register const MEON_STRING	*s2
)
{

	while (*s1 != '\0' && *s2 != '\0' && *s1 == *s2)
   {
      ++s1;
      ++s2;
   }

	if (*s1 == *s2)
		return(0);
	else
		return(-1);
}




/* returns the number of non-NULL bytes in string argument. */
size_t _cdecl  CMSMstrlen
(
	register const MEON_STRING	*s
)
{
	const MEON_STRING	*base = s;

	while (*s++)			/* when done, 's' will point at null terminator */
		;

	return (s - base - 1);
}






/*-------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/

/*
** Return the point in 'sp' at which the character ch appears or nil if
** not found.
*/
MEON_STRING * _cdecl  CMSMstrchr
(
	register const MEON_STRING	*sp,
	register int			ch
)
{
	do
	{
		if (*sp == ch)
			return (MEON_STRING *) sp;
	} while (*sp++);

	return (MEON_STRING *) NULL;
}






/*-------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/

/*
** Return the number of characters in the maximum leading segment
** of string which consists solely of characters NOT from charset.
*/
size_t _cdecl  CMSMstrcspn
(
	const MEON_STRING	*string,
	const MEON_STRING	*charset
)
{
	register const MEON_STRING	*p, *q;

	for (q = string; *q; ++q)
	{
		for (p = charset; *p && *p != *q; ++p)
			;

		if (*p)
			break;
	}

	return (q - string);
}






/*-------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/

/*
** Return ptr to first occurance of any character from `brkset'
** in the character string `string'; NULL if none exists.
*/
MEON_STRING * _cdecl  CMSMstrpbrk
(
	register const MEON_STRING	*string,
	const MEON_STRING				*brkset
)
{
	register const MEON_STRING	*p;

	do
	{
		for (p = brkset; *p && *p != *string; ++p)
			;

		if (*p)
			return (MEON_STRING *) string;
	}
	while (*string++);

	return (MEON_STRING *) NULL;
}






/*-------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/

/*
** Return the ptr in sp at which the character c last appears; NULL if
** not found.
*/
MEON_STRING * _cdecl  CMSMstrrchr
(
	register const MEON_STRING	*sp,
	register int			c
)
{
	register const MEON_STRING	*r = (const MEON_STRING *) NULL;

	do
	{
		if (*sp == c)
			r = sp;
	} while (*sp++);

	return (MEON_STRING *) r;
}






/*-------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/

/*
** Return the number of characters in the maximum leading segment
** of string which consists solely of characters from charset.
*/
size_t _cdecl  CMSMstrspn
(
	const MEON_STRING	*string,
	const MEON_STRING	*charset
)
{
	register const MEON_STRING	*p, *q;

	for (q = string; *q; ++q)
	{
		for (p = charset; *p && *p != *q; ++p)
			;

		if (!*p)
			break;
	}
	return (q - string);
}






/*-------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/

/*
** Locates the first occurrence in the string as1 of the sequence of
** characters (excluding the terminating null character) in the string as2.
** Returns a pointer to the located string, or a null pointer if the string
** is not found. If as2 is "", the function returns as1.
*/
MEON_STRING * _cdecl  CMSMstrstr
(
	const MEON_STRING	*as1,
	const MEON_STRING	*as2
)
{
	register MEON_STRING	c;
	register const MEON_STRING	*s1, *s2, *tptr;

	s1 = as1; s2 = as2;

	if (!s2 || !*s2)
		return (MEON_STRING *) s1;

	c = *s2;

	while (*s1)
	{
		if (*s1++ == c)
		{
			tptr = s1;

			while ((c = *++s2) == *s1++ && c)
				;

			if (!c)
				return ((MEON_STRING *) tptr - 1);

			s1 = tptr;
			s2 = as2;
			c  = *s2;
		}
	}

	return (MEON_STRING *) NULL;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/


MEON_STRING * _cdecl  CMSMstrcpy
(
	register MEON_STRING			*t,
	register const MEON_STRING	*s
)
{
	MEON_STRING *base = t;

	while (*t++ = *s++)
		;

	return base;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/


MEON_STRING * _cdecl  CMSMstrncpy
(
	register MEON_STRING			*t,
	register const MEON_STRING	*s,
	register size_t		n
)
{
	register MEON_STRING	*base = t;

	n++;				

	while (--n > 0 && (*t++ = *s++))
		;

	if (n > 0)
	{
		while (--n > 0)
			*t++ = '\0';
	}

	return base;
}


/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/


int _cdecl  CMSMatoi
(
	register const MEON_STRING	*p
)
{
	register	int	c,
						neg = 0;
	register	int	n;

	if (!CMSMisdigit(c = *p))
	{
		while (CMSMisspace(c))
			c = *++p;

		switch (c)
		{
			case '-':
				neg++;
			/* FALLTHROUGH */
			case '+':
				c = *++p;
		}

		if (!CMSMisdigit(c))
			return 0;
	}

	for (n = 0; CMSMisdigit(c = *p); ++p)
	{
		n *= 10;			/* two steps to avoid unnecessary overflow */
		n += '0' - c;	/* accumulate neg to avoid surprises at MAX */
	}

	return (neg) ? n : -n;
}



/*-----------------------------------------------------------------------------
**
** not ANSI standard
**
**-----------------------------------------------------------------------------
*/


int _cdecl  CMSMatox
(
	register const MEON_STRING *p
)
{
	register	int	c;
	register	int	n;

	if (!CMSMisxdigit(c = *p))
	{
		/* Eat white space if exists. */
		while (CMSMisspace(c))
			c = *++p;

		if (!CMSMisxdigit(c))
			return 0;
	}

	for (n = 0; CMSMisxdigit(c = *p); ++p)
	{
		n *= 16;			/* two steps to avoid unnecessary overflow */
		if (CMSMisdigit(c))
			n += c - '0';
		else
			n+= CMSMtoupper(c) - 'A' + 10;
	}

	return n;
}


void _cdecl  CMSMitoa(MEON_STRING *s, int n)
{
	int	i, j, c, sign;

	/* Check and save sign. */
	if ((sign = n) < 0)
	{
		/* If negative make positive. */
		n = -n;
	}

	/* Now lets convert to ascii in reverse order. */
	i = 0;
	do
	{
		s[i++] = n % 10 + '0';
	} while ((n /= 10) > 0);

	if (sign < 0)
		s[i++] = '-';
	s[i] = '\0';

	/* Now lets reverse the string. */
	for (i = 0, j = CMSMstrlen(s) - 1; i < j; i++, j--)
	{
		c = s[i];
		s[i] = s[j];
		s[j] = c;
	}
}


void _cdecl  CMSMxtoa(MEON_STRING *s, unsigned int n)
{
	int		i, j, c;

	/* Lets convert to ascii in reverse order. */
	i = 0;
	do
	{
		j = n % 16;
		if (j > 9)
			s[i++] = j + 'A' - 10;
		else
			s[i++] = j + '0';
	} while ((n /= 16) > 0);

	s[i] = '\0';

	/* Now lets reverse the string. */
	for (i = 0, j = CMSMstrlen(s) - 1; i < j; i++, j--)
	{
		c = s[i];
		s[i] = s[j];
		s[j] = c;
	}
}



/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

MEON_STRING * _cdecl  CMSMstrlwr					/* lower-case a string--single-byte ASCII only */
(
	MEON_STRING	*base
)
{
	register MEON_STRING	*s = base - 1;

	while (*++s)
	{
		if (inrange('A', *s, 'Z'))
			*s += ' ';
	}

	return base;
}



/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

MEON_STRING * _cdecl  CMSMstrupr					/* upper-case a string--single-byte ASCII only */
(
	MEON_STRING	*base
)
{
	register MEON_STRING	*s = base - 1;

	while (*++s)
	{
		if (inrange('a', *s, 'z'))
			*s -= ' ';
	}

	return base;
}



