/*****************************************************************************
 *  (c) 1995, Copyright Unpublished Work of Novell, Inc. All Rights Reserved
 *
 *  THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 *  PROPRIETARY AND TRADE SECRET INFORMATION OF NOVELL, INC.
 *  ACCESS TO THIS WORK IS RESTRICTED TO (i) NOVELL EMPLOYEES
 *  WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE
 *  OF THEIR ASSIGNMENTS AND (ii) ENTITIES OTHER THAN NOVELL
 *  WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.
 *  NO PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED,
 *  DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED,
 *  CONDENSED, EXPANDED, COLLECTED, COMPILED, LINKED, RECAST,
 *  TRANSFORMED OR ADAPTED WITHOUT THE PRIOR WRITTEN CONSENT OF
 *  NOVELL.  ANY USE OR EXPLOITATION OF THIS WORK WITHOUT AUTHORIZATION
 *  COULD SUBJECT THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY.
 *
 *---------------------------------------------------------------------------
 *
 * Title:            NWCACS Find Message Enabling Messages
 *
 * Filename:         findmsgs.c
 *
 * Date Created:     06 Sept 95
 *
 * Version:          1.0
 *
 * Programmed by:    JRT
 *
 * Files used:
 *
 * Last Modified:
 *
 * Description:	     During the process of loading and initializing of either
 *                   the NWCACS shim or a CHSM that runs under the shim this
 *                   function is called to locate the messages attached to
 *                   the load file by the linker.  During initialization the
 *                   the load file is already open, this function makes use
 *                   of the fact so that it does not need to chase down the
 *                   load file on the media.  Therefore, unlike A3112.NLM this
 *                   function can find messages in NLM and LAN modules that
 *                   do NOT reside on the default load path.
 *
 * Origin:           A3112.NLM module NUTSHIM.C was consulted as was LOADER.C.
 *
 * Modification History:
 *
 * 10-04-95 jrt      Version 1.00
 *                   Original release to source control.
 *
 * 10-10-95 jrt      Version 1.01
 *                   Function name changed from FindMessages to
 *                   CACSFindMessages in preparation of exporting the symbol.
 *                   This modification allow for seperation of common
 *                   portions of the shim from topology specific portions.
 *
 * 11-13-96 jrt      CMSM.C strings usable by CHSMs added.
 *
 * 02-27-97 jrt      Power Cycle HW string removed.
 *
 * 06-02-97 jrt      Added handling for multi-language enabling.
 *
 ****************************************************************************/

#define BYTE unsigned char
#define LONG unsigned long
#define BUFFER_SIZE 4096
#define MESSAGE_POINTER_OFFSET 118
#define TRUE (1==1)
#define FALSE !TRUE
#define NULL (void *)0
#define TxtMSG(s,id) s
#define OFFSET_OF_MESSAGE_BUFFER_POINTER 15
#define NO_ASSOCIATED_BUFFER (void *)1

#include <cmsm.txt>

struct ExtendedHeaderStructure
  {
  BYTE  stamp[8] ;
  LONG  languageID ;
  LONG  messageFileOffset ;
  LONG  messageFileLength ;
  LONG  messageCount ;
  LONG  helpFileOffset ;
  LONG  helpFileLength ;
  LONG  RPCDataOffset ;
  LONG  RPCDataLength ;
  LONG  sharedCodeOffset ;
  LONG  sharedCodeLength ;
  LONG  sharedDataOffset ;
  LONG  sharedDataLength ;
  LONG  sharedRelocationFixupOffset ;
  LONG  sharedRelocationFixupCount ;
  LONG  sharedExternalReferenceOffset ;
  LONG  sharedExternalReferenceCount ;
  LONG  sharedPublicsOffset ;
  LONG  sharedPublicsCount ;
  LONG  sharedDebugRecordOffset ;
  LONG  sharedDebugRecordCount ;
  LONG  (*SharedInitializationOffset)() ;
  void  (*SharedExitProcedureOffset)() ;
  LONG  productID ;
  LONG  reserved0 ;
  LONG  reserved1 ;
  LONG  reserved2 ;
  LONG  reserved3 ;
  LONG  reserved4 ;
  LONG  reserved5 ;
  } ;


void *CACSAllocC( LONG ) ;
void  CACSFreeC( void * ) ;
void *MSMReturnMessageInformation( LONG ) ;

extern LONG MSMModuleHandle ;
extern LONG *MSMMessageBuffer ;
extern BYTE MSMFileServerMajorVersionNumber ;
extern BYTE MSMFileServerMinorVersionNumber ;

extern LONG MSMIsLoading ;

extern LONG **DriverStackPointer ;
extern LONG DriverModuleHandle ;


static BYTE stamp[] = TxtMSG("MeSsAgEs", 117) ;


/****************************************************************************
                      C A C S F i n d M e s s a g e s
 ****************************************************************************

 INPUT:	 fileHandle - handle to open file (.NLM or .LAN)
         FileRead   - address of function that reads the file

 OUTPUT: messagePtrArrayPtr - where to store the pointer to the array of
		                          pointers to message character strings

 RETURN: pointer to the message load file.  This buffer must be freed when
         no longer in use.

 ****************************************************************************/

void *CACSFindMessages( void *fileHandle,
                        LONG (*FileRead)( void *, LONG, void *, LONG),
                        void **messagePtrArrayPtr )
  {
  BYTE                           *bufferPtr ;
  struct ExtendedHeaderStructure *extendedHeaderPtr ;
  BYTE                           *searchPtr ;
  LONG                           *ptrArrayPtr ;
  LONG                            found ;
  LONG                            i ;
  LONG                            messageFileOffset ;
  LONG                            messageFileLength ;
  LONG                            messageCount ;

  if( NULL == FileRead || NULL == messagePtrArrayPtr )
    {
    return( NULL ) ;
    }
  *messagePtrArrayPtr = NULL ;

  /* If NetWare 4 or newer than can get messages from OS */
  if( 4 <= MSMFileServerMajorVersionNumber )
    {
    /* For NetWare 4.0 through NetWare 4.10 use ReturnMessageInformation */
    if( 4 == MSMFileServerMajorVersionNumber &&
        11 > MSMFileServerMinorVersionNumber )
      {
      ptrArrayPtr = MSMReturnMessageInformation( 
        (0 != MSMIsLoading) ? MSMModuleHandle : DriverModuleHandle ) ;
      if( NULL != ptrArrayPtr )
        {
        *messagePtrArrayPtr = ptrArrayPtr ;
        return( NO_ASSOCIATED_BUFFER ) ;
        }
      else
        { /* Call to MSMReturnMessageInformation failed. */
        ; /* Fall through into NetWare 3 handling code.  */
        }
      }
    else
      {
      /* For NetWare 4.11 and newer get from stack on init call. */
      *messagePtrArrayPtr = ( 0 != MSMIsLoading ) ? MSMMessageBuffer :
        *(DriverStackPointer + OFFSET_OF_MESSAGE_BUFFER_POINTER) ;
      return( NO_ASSOCIATED_BUFFER ) ;
      }
    }

  /* NetWare 3 message handling, read the load file and find messages. */
  /* Get a buffer to read load file fragments into */
  bufferPtr = CACSAllocC( BUFFER_SIZE ) ;
  if( NULL == bufferPtr )
    {
    return( 0 ) ;
    }

  /* Read in the first fragment which will contain the extended header */
  if( (*FileRead)( fileHandle, 0, bufferPtr, BUFFER_SIZE ) )
    {
    CACSFreeC( bufferPtr ) ;
    return( NULL ) ;
    }

  /* Search for the MESSAGES extended header */
  for( searchPtr = bufferPtr ;
       searchPtr < searchPtr+BUFFER_SIZE-sizeof( struct ExtendedHeaderStructure )-1 ;
       searchPtr++ )
    {
    for( i=0; i<sizeof( stamp )-1; i++ )
      {
      found = TRUE ;
      if( *(searchPtr+i) != stamp[i] )
        {
        found = FALSE ;
        break ;
        }
      }
    if( found ) break ;
    }

  if( !found )
    {
    CACSFreeC( bufferPtr ) ;
    return( NULL ) ;
    }

  /* Get message information from the extended header */
  extendedHeaderPtr = (struct ExtendedHeaderStructure *)searchPtr ;
  messageFileOffset = extendedHeaderPtr->messageFileOffset ;
  messageFileLength = extendedHeaderPtr->messageFileLength ;
  messageCount      = extendedHeaderPtr->messageCount      ;

  /* Free the buffer we have been using for fragments */
  CACSFreeC( bufferPtr ) ;

  /* Allocate memory to hold the messages, then read the messages */
  bufferPtr = CACSAllocC( messageFileLength ) ;
  if( NULL == bufferPtr )
    {
    return( NULL ) ;
    }
  if( (*FileRead)( fileHandle, messageFileOffset, bufferPtr, messageFileLength ) )
    {
    CACSFreeC( bufferPtr ) ;
    return( NULL ) ;
    }

  /* Calculate pointer to array of pointers to messages then adjust the pointers
     to messages to point to memory.  Currently they are simply offsets */
  ptrArrayPtr = (LONG *)(bufferPtr + MESSAGE_POINTER_OFFSET) ;
  for( i=0; i<messageCount; i++ ) *(ptrArrayPtr+i) += (LONG)ptrArrayPtr ;

  /* Store the function's output value */
  *messagePtrArrayPtr = ptrArrayPtr ;

  return( bufferPtr ) ;
  }


/*====================================================================**

   Following are strings usable by the CHSM. Must be in sync with
   CMSM.C.

**====================================================================*/


/***********************************************************************
*	Class Names
***********************************************************************/

BYTE NESL_Service_Suspend[]      = NESL_SERVICE_SUSPEND ;
BYTE NESL_Service_Resume[]       = NESL_SERVICE_RESUME ;
BYTE NESL_ServiceStatus_Change[] = NESL_SERVICESTATUS_CHANGE ;
BYTE NESL_Suspend_Request[]      = NESL_SUSPEND_REQUEST ;

/***********************************************************************
*	Event Names for Service Suspend Class
***********************************************************************/

BYTE NESL_MLID_ShutDown[]            = NESL_MLID_SHUTDOWN ;
BYTE NESL_MLID_Card_Removal[]        = NESL_MLID_CARD_REMOVAL ;
BYTE NESL_MLID_Out_Range[]           = NESL_MLID_OUT_RANGE ;
BYTE NESL_MLID_HW_Failure[]          = NESL_MLID_HW_FAILURE ;
BYTE NESL_MLID_Cable_Disconnect[]    = NESL_MLID_CABLE_DISCONNECT ;
BYTE NESL_MLID_Media_Access_Denied[] = NESL_MLID_MEDIA_ACCESS_DENIED ;

/***********************************************************************
*	Event Names for Service Resume Class
***********************************************************************/

BYTE NESL_MLID_Reset[]                   = NESL_MLID_RESET ;
BYTE NESL_MLID_Card_Insertion_Complete[] = NESL_MLID_CARD_INSERTION_COMPLETE ;
BYTE NESL_MLID_In_Range[]                = NESL_MLID_IN_RANGE ;
BYTE NESL_MLID_Cable_Reconnect[]         = NESL_MLID_CABLE_RECONNECT ;

/***********************************************************************
*	Event Names for Service/Status Class
***********************************************************************/

BYTE NESL_MLID_Access_Point_Change[]   = NESL_MLID_ACCESS_POINT_CHANGE ;
BYTE NESL_MLID_Speed_Change[]          = NESL_MLID_SPEED_CHANGE ;
BYTE NESL_MLID_Config_Table_Change[]   = NESL_MLID_CONFIG_TABLE_CHANGE ;

BYTE NESL_MLID_DeRegister_Resource_Change[] =
	 NESL_MLID_DEREGISTER_RESOURCE_CHANGE ;

BYTE NESL_MLID_ReRegister_Hardware_Options_Change[] =
	 NESL_MLID_REREGISTER_HARDWARE_OPTIONS_CHANGE ;

/***********************************************************************
*	Event Names for Service Request Class
***********************************************************************/

/* Currently there are no defined events for this class */
