/***********************************************************************\
*
*   (c) Unpublished Work of Novell, Inc.  All rights reserved.
*
*   THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
*   PROPRIETARY AND TRADE SECRET INFORMATION OF NOVELL, INC.  ACCESS
*   TO THIS WORK IS RESTRICTED TO (I) NOVELL EMPLOYEES WHO HAVE A
*   NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE OF THEIR ASSIGNMENTS
*   AND (II) ENTITIES OTHER THAN NOVELL WHO HAVE ENTERED INTO
*   APPROPRIATE LICENSE AGREEMENTS.  NO PART OF THIS WORK MAY BE USED,
*   PRACTICED, PERFORMED, COPIED, DISTRIBUTED, REVISED, MODIFIED,
*   TRANSLATED, ABRIDGED, CONDENSED, EXPANDED, COLLECTED, COMPILED,
*   LINKED,RECAST, TRANSFORMED OR ADAPTED WITHOUT THE PRIOR WRITTEN
*   CONSENT OF NOVELL.  ANY USE OR EXPLOITATION OF THIS WORK WITHOUT
*   THE PRIOR WRITTEN CONSENT OF NOVELL.  ANY USE OR EXPLOITATION OF
*   THIS WORK WITHOUT AUTHORIZATION COULD SUBJECT THE PERPETRATOR TO
*   CRIMINAL AND CIVIL LIABILITY.
*
***********************************************************************
*
*  Name:			C MSM Structure Definitions
*
*  FileName:		CMSM.H
*
*  ODI Spec Ver:	1.11
*
*  Description:   	C version of the MSM Structures
*
*  Modification History:
*
*  08-27-96 WTT		Added prototype for CMSMGetConfigInfo.  Added defines for
*					CMSM_TABLE_MAJOR_VER, and CMSM_TABLE_MINOR_VER. Added
*					structure definition for _CMSM_CONFIG_TABLE_.
*
*  08-28-96 WTT		Added defines for CTSM_TABLE_MAJOR_VER, and
*					CTSM_TABLE_MINOR_VER. Added structure definition
*					for _CTSM_CONFIG_TABLE_.
*
*  09-03-96 KRA		Added prototype for CMSMUpdateConfigTables.
*
*  09-03-96 WTT		Added define in CMSMStatusFlags for POLLING_SUSPENDED.
*
*  09-04-96 KRA		Changed struct DRIVER_DATA to struct _DRIVER_DATA_
*					for the CMSMNESLProduceMLIDEvent function prototype.
*
*  09-23-96 KRA		Added prototypes for CMSMDeRegisterResource and
*					CMSMReRegisterHardwareOptions.
*
*  09-26-96 KRA		Added prototypes for CMSMResumePolling and CMSMSuspendPolling.
*
*  10-04-96 KRA		Removed oldConfigTable parameter from
*					CMSMReRegisterHardwareOptions.
*
*  10-08-96 MPK		Added defines HSM_PERMANENT_SHUTDOWN and
*					HSM_TEMPORARY_SHUTDOWN for the CMSMShutdownMLID routine.
*					Although these have been added on a value basis, the
*					define is part of a bit mapping scheme where bit one
*					active indicates MLID initiated shutdown (either temp
*					or permanent)
*
*  10-09-96 MPK		Added prototypes for CMSMShutdownMlid and CMSMResetMlid
*					And IOCTL routine MLIDReset
*
*  11-04-96 KRA		Added #defines for possitions within the
*					"HSM_CSPEC_VERSION: X.XX" and the "HSM_SPEC_VERSION: X.XX".
*
*  11-05-96 MPK		Added prototype for CMSMHardwareFailure.
*
*  11-19-96 MPK		Added UNLOADING bit definition for MLID_status.
*
*  02-14-97 AYD		Added prototype for CMSMParseParameter.
*
*  02-18-97 WTT		Added prototypes for CMSMGetHINFromHINName and
*					CMSMGetHINNameFromHIN.
*
*  02-19-97 KRA		Added prototype for CMSMInitParser.
*
*  02-20-97 MPK		Renamed prototype for CMSMParseParameter to
*					CMSMParseSingleParameter.
*
*  02-25-97 WTT		Changed prototype for CMSMParseSingleParameter to
*					match final spec.  The parameter
*					DRIVER_PARM_BLOCK *hsmParamBlock was not used or
*					needed and was removed.
*
*  02-26-97 MPK		Changed FAILURE_CRITICAL and FAILURE_FATAL to
*					NOTIFY_CRITICAL and NOTIFY_FAILURE, and
*					addes a NOTIFY_DEGRADED define to the file.
*
*  07-29-97 WTT		Removed define for SPARE_ECB_STATUS.
*					SPD 162312
*
*  10-08-97 WTT		Changed proto type for CMSMGetBusSpecificInfo from
*					ODI_NBI	_cdecl	CMSMGetBusSpecificInfo (
*										void    *busTag,
*										UINT32	size,
*										UINT32	*busSpecificInfo);
*
*					to
*					ODI_NBI	_cdecl	CMSMGetBusSpecificInfo (
*										void    *busTag,
*										UINT32	size,
*										void	*busSpecificInfo);
*					This change was made to conform with the specification.
*
*	01-06-98 WTT	Added CTSMCFG_Reserved definition to the comment section
*						for CTSM_CONFIG_TABLE.
*
* 01-12-98 KRA		Removed prototype for CMSMGetNBIConfiguration.  This function
*						is not listed in the spec.
*
* 03-05-98 mcd      Added CMSMFlushWriteCombine and removed "ifdef HPPA"
*                   (HP Precision Architecture)
*
*	04-03-98 WTT	Modified CMSMFlushWriteCombine pragma to be a nop for 
*						clients.
*
*	04-07-98 WTT	Reversed the change made to CMSMFlushWriteCombine pragma
*						on 04/03/98 and put the correct set of instructions
*						in the pragma as outlined in the ODI Design document for
*						Write Combine.
*
\***********************************************************************/
#ifndef CMSM_H

#include <odi_nesl.h>

/***********************************************************************\
*                                                                       *
* MSM/TSM/HSM common structures.                                        *
*                                                                       *
\***********************************************************************/
struct   _DRIVER_PARM_BLOCK_;
struct   _TSM_PARM_BLOCK_;
struct   _GENERIC_STATS_;
struct   _MEDIA_DATA_;
struct   _FRAGMENT_LIST_STRUCT_;
struct   _RTAG_HANDLE_;
struct   _MODULE_HANDLE_;
struct   _MSTATS_;


typedef  MLID_CONFIG_TABLE   CONFIG_TABLE;

#ifndef  NULL
#define  NULL 0
#endif

#ifndef  ALL_ONES
#define  ALL_ONES   -1
#endif

#ifndef  DRIVER_DATA
struct   _DRIVER_DATA_;
#endif

#ifndef  MAX_FRAME_TYPES
#define  MAX_FRAME_TYPES 4
#endif

/**************************************************************\
*                                                              *
* Portable alias definitions.                                  *
*                                                              *
\**************************************************************/

#define   COPY_UINT16(dest_addr, src_addr)   COPY_WORD(src_addr, dest_addr)
#define   COPY_UINT32(dest_addr, src_addr)   COPY_LONG(src_addr, dest_addr)
#define   COPY_ADDR(dest_addr, src_addr)     COPY_NODE(src_addr, dest_addr)

#define   GET_UINT16(addr)          GET_WORD(addr)
#define   GET_UINT32(addr)          GET_LONG(addr)
#define   PUT_UINT16(addr, value)   PUT_WORD(value, addr)
#define   PUT_UINT32(addr, value)   PUT_LONG(value, addr)

#define   GET_HILO_UINT16(addr)   GET_HILO_WORD(addr)
#define   GET_HILO_UINT32(addr)   GET_HILO_LONG(addr)
#define   GET_LOHI_UINT16(addr)   GET_LOHI_WORD(addr)
#define   GET_LOHI_UINT32(addr)   GET_LOHI_LONG(addr)

#define   PUT_HILO_UINT16(addr, value)   PUT_HILO_WORD(value, addr)
#define   PUT_HILO_UINT32(addr, value)   PUT_HILO_LONG(value, addr)
#define   PUT_LOHI_UINT16(addr, value)   PUT_LOHI_WORD(value, addr)
#define   PUT_LOHI_UINT32(addr, value)   PUT_LOHI_LONG(value, addr)

#define   VALUE_TO_HILO_UINT16(value)   VALUE_TO_HILO_WORD(value)
#define   VALUE_TO_HILO_UINT32(value)   VALUE_TO_HILO_LONG(value)
#define   VALUE_TO_LOHI_UINT16(value)   VALUE_TO_LOHI_WORD(value)
#define   VALUE_TO_LOHI_UINT32(value)   VALUE_TO_LOHI_LONG(value)

#define   VALUE_FROM_HILO_UINT16(value)   VALUE_FROM_HILO_WORD(value)
#define   VALUE_FROM_HILO_UINT32(value)   VALUE_FROM_HILO_LONG(value)
#define   VALUE_FROM_LOHI_UINT16(value)   VALUE_FROM_LOHI_WORD(value)
#define   VALUE_FROM_LOHI_UINT32(value)   VALUE_FROM_LOHI_LONG(value)

#define   COPY_TO_HILO_UINT16(dest_addr, src_addr)   COPY_TO_HILO_WORD(src_addr, dest_addr)
#define   COPY_TO_HILO_UINT32(dest_addr, src_addr)   COPY_TO_HILO_LONG(src_addr, dest_addr)
#define   COPY_TO_LOHI_UINT16(dest_addr, src_addr)   COPY_TO_LOHI_WORD(src_addr, dest_addr)
#define   COPY_TO_LOHI_UINT32(dest_addr, src_addr)   COPY_TO_LOHI_LONG(src_addr, dest_addr)

#define   COPY_FROM_HILO_UINT16(dest_addr, src_addr)   COPY_FROM_HILO_WORD(src_addr, dest_addr)
#define   COPY_FROM_HILO_UINT32(dest_addr, src_addr)   COPY_FROM_HILO_LONG(src_addr, dest_addr)
#define   COPY_FROM_LOHI_UINT16(dest_addr, src_addr)   COPY_FROM_LOHI_WORD(src_addr, dest_addr)
#define   COPY_FROM_LOHI_UINT32(dest_addr, src_addr)   COPY_FROM_LOHI_LONG(src_addr, dest_addr)

#define   HOST_TO_HILO_UINT16(addr)   HOST_TO_HILO_WORD(addr)
#define   HOST_TO_HILO_UINT32(addr)   HOST_TO_HILO_LONG(addr)
#define   HOST_TO_LOHI_UINT16(addr)   HOST_TO_LOHI_WORD(addr)
#define   HOST_TO_LOHI_UINT32(addr)   HOST_TO_LOHI_LONG(addr)

#define   HOST_FROM_HILO_UINT16(addr)   HOST_FROM_HILO_WORD(addr)
#define   HOST_FROM_HILO_UINT32(addr)   HOST_FROM_HILO_LONG(addr)
#define   HOST_FROM_LOHI_UINT16(addr)   HOST_FROM_LOHI_WORD(addr)
#define   HOST_FROM_LOHI_UINT32(addr)   HOST_FROM_LOHI_LONG(addr)

#define   UINT16_EQUAL(addr1, addr2)   WORD_EQUAL(addr1, addr2)
#define   UINT32_EQUAL(addr1, addr2)   LONG_EQUAL(addr1, addr2)
#define   ADDR_EQUAL(addr1, addr2)     NODE_EQUAL(addr1, addr2)

/* Possitions within the "HSM_CSPEC_VERSION: X.XX" string. */
#define C_MAJOR_VERSION_POS			19
#define C_MINOR_VERSION_TENS_POS		21
#define C_MINOR_VERSION_ONES_POS		22

/* Possitions within the "HSM_SPEC_VERSION: X.XX" string. */
#define A_MAJOR_VERSION_POS			18
#define A_MINOR_VERSION_TENS_POS		20
#define A_MINOR_VERSION_ONES_POS		21

#if defined _WATCOMC_ || defined OS_NT
/* Set PRAGMA to pack these structures */
#pragma   pack(1)
#endif

typedef   struct   _FRAGMENT_LIST_STRUCT_
{
   UINT32            FragmentCount;
   FRAGMENT_STRUCT   FragmentStruct[1];
}
FRAGMENT_LIST_STRUCT;

/********************************************************************
		C MSM Configuration Table Definitions.
*********************************************************************/

#define CMSM_TABLE_MAJOR_VER 1
#define CMSM_TABLE_MINOR_VER 0

/********************************************************************
		C MSM Configuration Table System Flags definitions.
*********************************************************************/

#define	CMSM_CFG_SERVER_BIT	0x40000000
#define	CMSM_CFG_CLIENT_BIT	0x80000000

/********************************************************************
		C MSM Configuration Table Structure definition.

           	typedef struct _CMSM_CONFIG_TABLE
			{
	            UINT32	   CMSMCFG_TableSize;
                UINT8     CMSMCFG_TableMajorVersion;
                UINT8     CMSMCFG_TableMinorVersion;
                UINT8     CMSMCFG_ModuleMajorVersion;
                UINT8     CMSMCFG_ModuleMinorVersion;
                UINT8     CMSMCFG_ODISpecMajorVersion;
                UINT8     CMSMCFG_ODISpecMinorVersion;
                UINT32	   CMSMCFG_MaxNumberOfBoards;
				UINT32	   CMSMCFG_SystemFlags;
            }CMSM_CONFIG_TABLE;


            CMSMCFG_TableSize
               	This field contains the actual size of the
				configuration table.  The value of this field should
				not be confused with the number of bytes requested
				or copied (i.e., nBytes).

            CMSMCFG_TableMajorVersion
               	This field contains the major version of the
				configuration table.  The current major version is 1.

            CMSMCFG_TableMinorVersion
				This field contains the minor version of the
				configuration table.  The current minor version is 0.

            CMSMCFG_ModuleMajorVersion
				This field contains the major version of the C MSM
				binary (i.e., CMSM.NLM).

            CMSMCFG_ModuleMinorVersion
				This field contains the minor version of the C MSM
				binary (i.e., CMSM.NLM).

            CMSMCFG_ODISpecMajorVersion
				This field contains the major version of the ODI
				Specifiaction that this version of the C MSM is written
				too.  For example, if the version of the ODI
				specification 1.11, the value of this field is 1.

            CMSMCFG_ODISpecMinorVersion
				This field contains the minor version of the ODI
				Specifiaction that this version of the C MSM is written
				too.  For example, if the version of the ODI
				specification 1.11, the value of this field is 11.

            CMSMCFG_MaxNumberOfBoards
               	The value of this field represents the maximum number
				of boards the C MSM supports.

			CMSMCFG_SystemFlags
				The bits in this field are defined below.
				
				Bit 0-29	These bits are reserved and must be
							set to zero.

				Bit 30		CMSM_CFG_SERVER_BIT

							When set to 1 this bit indicates the C MSM
							is running in a server environment.  This
							bit is exclusive with bit 31.

				Bit 31		CMSM_CFG_CLIENT_BIT

							When set to 1 this bit indicates the C MSM
							is running in a client environment.  This
							bit is exclusive with bit 30.

*********************************************************************/

typedef struct _CMSM_CONFIG_TABLE_
{
	UINT32	CMSMCFG_TableSize;
	UINT8   CMSMCFG_TableMajorVersion;
	UINT8   CMSMCFG_TableMinorVersion;
	UINT8   CMSMCFG_ModuleMajorVersion;
    UINT8   CMSMCFG_ModuleMinorVersion;
    UINT8   CMSMCFG_ODISpecMajorVersion;
    UINT8   CMSMCFG_ODISpecMinorVersion;
	UINT16	CMSMCFG_Reserved;
    UINT32	CMSMCFG_MaxNumberOfBoards;
	UINT32	CMSMCFG_SystemFlags;
}CMSM_CONFIG_TABLE;

/********************************************************************
		C TSM Configuration Table Definitions.
*********************************************************************/

#define CTSM_TABLE_MAJOR_VER 1
#define CTSM_TABLE_MINOR_VER 0

/********************************************************************
		C TSM Configuration Table System Flags definitions.
*********************************************************************/

#define	CTSM_CFG_SERVER_BIT	0x40000000
#define	CTSM_CFG_CLIENT_BIT	0x80000000

/********************************************************************
		C TSM Configuration Table Structure definition.

				typedef struct _CTSM_CONFIG_TABLE
				{
					UINT32	CTSMCFG_TableSize;
					UINT8		CTSMCFG_TableMajorVersion;
					UINT8		CTSMCFG_TableMinorVersion;
					UINT8		CTSMCFG_ModuleMajorVersion;
					UINT8		CTSMCFG_ModuleMinorVersion;
					UINT8		CTSMCFG_ODISpecMajorVersion;
					UINT8		CTSMCFG_ODISpecMinorVersion;
					UINT16	CTSMCFG_Reserved;
					UINT32   CTSMCFG_MaxFrameSize;
					UINT32	CTSMCFG_SystemFlags;
				}CTSM_CONFIG_TABLE;


           CTSMCFG_TableSize
				This field contains the actual size of the
				configuration table.  The value of this field should
				not be confused with the number of bytes requested
				or copied (i.e., nBytes).

           CTSMCFG_TableMajorVersion
               	This field contains the major version of the
				configuration table.  The current major version is 1.

           CTSMCFG_TableMinorVersion
				This field contains the minor version of the
				configuration table.  The current minor version is 0.

           CTSMCFG_ModuleMajorVersion
				This field contains the major version of the C TSM
				binary (i.e., ETHERTSM.NLM).

           CTSMCFG_ModuleMinorVersion
				This field contains the minor version of the C TSM
				binary (i.e., ETHERTSM.NLM).

           CTSMCFG_ODISpecMajorVersion
				This field contains the major version of the ODI
				Specifiaction that this version of the C TSM is written
				too.  For example, if the version of the ODI
				specification 1.11, the value of this field is 1.

           CTSMCFG_ODISpecMinorVersion
				This field contains the minor version of the ODI
				Specifiaction that this version of the C TSM is written
				too.  For example, if the version of the ODI
				specification 1.11, the value of this field is 11.

			  CTSMCFG_Reserved
			   This field is reserved.

           CTSMCFG_MaxFrameSize
           	The value of this field represents the maximum frame
				size that the C TSM supports.

		     CTSMCFG_SystemFlags
		     	The bits in this field are defined below.
				
				Bit 0-29	These bits are reserved and must be
							set to zero.

				Bit 30		CTSM_CFG_SERVER_BIT

							When set to 1 this bit indicates the C TSM
							is running in a server environment.  This
							bit is exclusive with bit 31.

				Bit 31		CTSM_CFG_CLIENT_BIT

							When set to 1 this bit indicates the C TSM
							is running in a client environment.  This
							bit is exclusive with bit 30.

*********************************************************************/

typedef struct _CTSM_CONFIG_TABLE_
{
	UINT32	CTSMCFG_TableSize;
	UINT8   CTSMCFG_TableMajorVersion;
	UINT8   CTSMCFG_TableMinorVersion;
	UINT8   CTSMCFG_ModuleMajorVersion;
    UINT8   CTSMCFG_ModuleMinorVersion;
    UINT8   CTSMCFG_ODISpecMajorVersion;
    UINT8   CTSMCFG_ODISpecMinorVersion;
	UINT16	CTSMCFG_Reserved;
    UINT32	CTSMCFG_MaxFrameSize;
	UINT32	CTSMCFG_SystemFlags;
}CTSM_CONFIG_TABLE;

/********************************************************************
		MEDIA HEADER Definitions used at the MLI interface, and in CHSMs.
		Useful for MLID, HSM, and TxMonitor Developers
*********************************************************************/
typedef struct	_ETHER_MEDIA_HEADER_
{
	NODE_ADDR		MH_Destination;
	NODE_ADDR		MH_Source;
	UINT16			MH_Length;
	UINT8				MH_DSAP;
	UINT8				MH_SSAP;
	UINT8				MH_Ctrl0;
	UINT8				MH_SNAP[5];
} ETHER_MEDIA_HEADER;

typedef struct	_TOKEN_MEDIA_HEADER_
{
	UINT8		   MH_AccessControl;
	UINT8		   MH_FrameControl;
	NODE_ADDR   MH_Destination;
	NODE_ADDR	MH_Source;
	UINT8		   MH_DSAP;
	UINT8		   MH_SSAP;
	UINT8		   MH_Ctrl0;
	UINT8		   MH_SNAP [5];
} TOKEN_MEDIA_HEADER;

typedef struct	_FDDI_MEDIA_HEADER_
{
	UINT8		   MH_FrameControl;
	NODE_ADDR   MH_Destination;
	NODE_ADDR	MH_Source;
	UINT8		   MH_DSAP;
	UINT8		   MH_SSAP;
	UINT8		   MH_Ctrl0;
	UINT8		   MH_SNAP [5];
} FDDI_MEDIA_HEADER;

typedef struct	_FIBRE_MEDIA_HEADER_
{
	UINT8		   MH_FrameControl;
	NODE_ADDR   MH_Destination;
	NODE_ADDR	MH_Source;
	UINT8		   MH_DSAP;
	UINT8		   MH_SSAP;
	UINT8		   MH_Ctrl0;
	UINT8		   MH_SNAP [5];
} FIBRE_MEDIA_HEADER;

/**************************************************************\
*
*	RX-Net Short Packet TCB Structure.
*
\**************************************************************/

typedef struct	_RXNET_SHORT_PACKET_TCB_
{
	UINT8	SourceAddress;
	UINT8	DestAddress;
	UINT8	ByteOffset;
	UINT8	TCBSecondHeaderLen;
	UINT8	ProtocolType;
	UINT8	SplitFlag;
	UINT16	PacketSequenceNumber;
} RXNET_SHORT_PACKET_TCB;

/**************************************************************\
*
*	RX-Net Long Packet TCB Structure.
*
\**************************************************************/

typedef struct	_RXNET_LONG_PACKET_TCB_
{
	UINT8	SourceAddress;
	UINT8	DestAddress;
	UINT8	LongPacketFlag;
	UINT8	ByteOffset;
	UINT8	TCBSecondHeaderLen;
	UINT8	ProtocolType;
	UINT8	SplitFlag;
	UINT16	PacketSequenceNumber;
} RXNET_LONG_PACKET_TCB;

/**************************************************************\
*
*	RX-Net Exception Packet TCB Structure.
*
\**************************************************************/

typedef struct	_RXNET_EXCEPTION_PACKET_TCB_
{
	UINT8	SourceAddress;
	UINT8	DestAddress;
	UINT8	LongPacketFlag;
	UINT8	ByteOffset;
	UINT8	TCBSecondHeaderLen;
	UINT8	Pad1;
	UINT8	Pad2;
	UINT8	Pad3;
	UINT8	Pad4;
	UINT8	ProtocolType;
	UINT8	SplitFlag;
	UINT16	PacketSequenceNumber;
} RXNET_EXCEPTION_PACKET_TCB;


/**************************************************************\
*
*	TCB Fragment Structure
*
\**************************************************************/

typedef struct _TCB_FRAGMENT_BLOCK_STRUCT_
{
	UINT32	TCB_FragmentCount;
	FRAGMENT_STRUCT TCB_Fragment[1];
} TCB_FRAGMENT_BLOCK;

/**************************************************************\
*
*	TCB Structure
*
\**************************************************************/

typedef struct _TCB_
{
	void 		  		*TCB_Reserved;
	UINT32		  		TCB_BoardNumber;
	UINT32		  		TCB_DriverWS[3];
	UINT32		  		TCB_DataLen;
	TCB_FRAGMENT_BLOCK	*TCB_FragBlockPtr;
	UINT32		  		TCB_MediaHeaderLen;
	union
		{
			UINT8						TCB_Media[MAX_MEDIA_HEADER_SIZE];
			ETHER_MEDIA_HEADER			EtherMedia;
			TOKEN_MEDIA_HEADER			TokenMedia;
			FDDI_MEDIA_HEADER			FDDIMedia;
			FIBRE_MEDIA_HEADER			FibreMedia;
			RXNET_SHORT_PACKET_TCB		RXNetShortMedia;
			RXNET_LONG_PACKET_TCB		RXNetLongMedia;
			RXNET_EXCEPTION_PACKET_TCB	RXNetExceptionMedia;
		} TCB_MediaHeader;
} TCB;

typedef enum _CHSM_COMPLETE_
{
   CHSM_COMPLETE_STATISTICS,
   CHSM_COMPLETE_MULTICAST,
   CHSM_COMPLETE_SHUTDOWN,
   CHSM_COMPLETE_RESET,
   CHSM_COMPLETE_LOOK_AHEAD,
   CHSM_COMPLETE_PROMISCUOUS,
	CHSM_COMPLETE_MANAGEMENT,
	CTSM_COMPLETE_REGISTER_MONITOR		/* mark Reserved in HSM SPEC. */
}
CHSM_COMPLETE;

typedef enum _AES_TYPE_
{
   AES_TYPE_PRIVILEGED_ONE_SHOT,
   AES_TYPE_PRIVILEGED_CONTINUOUS,
   AES_TYPE_PROCESS_ONE_SHOT,
   AES_TYPE_PROCESS_CONTINUOUS
}
AES_TYPE;

typedef struct _MLID_AES_ECB_
{
   struct _MLID_AES_ECB_   *NextLink;
   void                    (_cdecl * DriverAES)(struct _DRIVER_DATA_ *, CONFIG_TABLE *);
   AES_TYPE                AesType;
   UINT32                  TimeInterval;
   void                    *AesContext;
   UINT8                   AesReserved[30];
}
MLID_AES_ECB;

typedef struct   _EXTRA_CONFIG_ISR_
{
   UINT32                  (_cdecl * ISRRoutine)(void *MagicNumber);
   void                    *ISRReserved0;
   void                    *ISRReserved1;
   void                    *ISRReserved2;
   void                    *ISRReserved3;
}
EXTRA_CONFIG_ISR;

typedef struct _EXTRA_CONFIG_
{
   struct _EXTRA_CONFIG_   *NextLink;
   UINT32                  (_cdecl * ISRRoutine0)(void *MagicNumber);
   void                    *ISR0Reserved0;
   void                    *ISR0Reserved1;
   void                    *ISR0Reserved2;
   void                    *ISR0Reserved3;
   UINT32                  (_cdecl * ISRRoutine1)(void *MagicNumber);
   void                    *ISR1Reserved0;
   void                    *ISR1Reserved1;
   void                    *ISR1Reserved2;
   void                    *ISR1Reserved3;
   IO_CONFIG               IOConfig;
}
EXTRA_CONFIG;

/**************************************************************\
*                                                              *
* ScreenStruct - Fields accessable from ScreenHandle, which    *
* is passed in to the NLMs/LANs initialization procedure. It   *
* is normally passed to OS routines during NLM/LAN             *
* initialization, like OutputToScreen.  Used internally by the              *
*                                                              *
\**************************************************************/

typedef struct ScreenStruct
{
   struct         ScreenStruct *previousScreen;
   struct         ScreenStruct *nextScreen;
   struct         ScreenStruct *popUpOriginalScreen;

   UINT32         *CLIBScreenStructure;

   MEON           currentPalette;
   MEON           popUpCount;
   MEON           screenList;             /* Used in multi-processor version only */
   MEON           activeCount;
   struct _RTAG_HANDLE_    *resourceTag;

   MEON           *screenName;
   MEON           *screenMemory;
   UINT32         flags;
   UINT32         state;
   UINT16         outputCursorPosition;   /* Offset from screenAddress * 2 */
   UINT16         inputCursorPosition;    /* Offset from 0, 0 */
}
SCREEN_HANDLE;

/****************************************************************\
*                                                                *
* LoadDefinitionStructure - Fields accessable from               *
* ModuleHandle, which is passed in to the NLMs/LANs              *
* initialization procedure. It is normally passed to OS          *
* routines during NLM/LAN initialization, like                   *
* AllocateResourceTag.                                           *
* The HSM is not to modify any of the fields in this structure.  *
*                                                                *
\****************************************************************/

/* LoadDefinitionStructure */

typedef struct _MODULE_HANDLE_
{
   struct         _MODULE_HANDLE_ *LDLink;
   struct         _MODULE_HANDLE_ *LDKillLink;
   struct         _MODULE_HANDLE_ *LDScanLink;
   struct         _RTAG_HANDLE_ **LDResourceList;
   UINT32         LDIdentificationNumber;
   UINT8          *LDCodeImageOffset;
   UINT32         LDCodeImageLength;
   UINT8          *LDDataImageOffset;
   UINT32         LDDataImageLength;
   UINT32         LDUninitializedDataLength;
   UINT8          *LDCustomDataOffset;
   UINT32         LDCustomDataSize;
   UINT32         LDFlags;
   UINT32         LDType;
   UINT32         (_cdecl * LDInitializationProcedure)(
                     struct   _MODULE_HANDLE_ *LoadRecord,
                     struct   ScreenStruct *screenID,
                     MEON       *CommandLine,
                     MEON      *loadDirectoryPath,
                     UINT32   uninitializedDataLength,
                     UINT32   fileHandle,
                     UINT32   (*ReadRoutine)(
                                 UINT32    fileHandle,
                                 UINT32    offset,
                                 void      *buffer,
                                 UINT32   numberOfBytes),
                     UINT32 customDataOffset,
                     UINT32 customDataSize,
                     UINT32 NumMsgs,
                     MEON    **Msgs);
   void           (_cdecl * LDExitProcedure)(void);
   UINT32         (_cdecl * LDCheckUnloadProcedure)(
                     struct ScreenStruct *screenID);
   void           *LDPublics;
   MEON           LDFileName[36];
   MEON           LDName[128];
   UINT32         *LDCLIBLoadStructure;
   UINT32         *LDNLMDebugger;
   UINT32         LDParentID;
   UINT32         LDReservedForCLIB;
   UINT32         LDNumberOfReferencedModules;                /* Used in NT. */
   struct         _MODULE_HANDLE_   **LDReferencedModules;   /* Used in NT. */
   void           *LDModuleObjectHandle;
   UINT32         LDMajorVersion;
   UINT32         LDMinorVersion;
   UINT32         LDRevision;
   UINT32         LDYear;
   UINT32         LDMonth;
   UINT32         LDDay;
   MEON           *LDCopyright;
   UINT32         LDAllocAvailBytes;
   UINT32         LDAllocFreeCount;
   UINT32         LDLastGarbCollect;
   UINT32         LDAlloc16Lists[64];
   UINT32         LDAlloc256Lists[12];
   UINT32         LDAlloc4kList;
   void           *LDDomainID;
   struct         _MODULE_HANDLE_ *LDEnvLink;
   void           *LDAllocPagesListHead;
   void           *LDTempPublicList;
   UINT32         LDMessageLanguage;       /* for enabling */
   MEON           **LDMessages;            /* for enabling */
   UINT32         LDMessageCount;          /* for enabling */
   MEON           *LDHelpFile;             /* for enabling */
   UINT32         LDMessageBufferSize;     /* for enabling */
   UINT32         LDHelpBufferSize;        /* for enabling */
   UINT32         LDSharedCodeOffset;      /* for protection */
   UINT32         LDSharedCodeLength;      /* for protection */
   UINT32         LDSharedDataOffset;      /* for protection */
   UINT32         LDSharedDataLength;      /* for protection */
   UINT32         (_cdecl * LDSharedInitProcedure)(
                     struct   _MODULE_HANDLE_ *LoadRecord,
                     struct   ScreenStruct *screenID,
                     MEON      *CommandLine);
   void            (_cdecl * LDSharedExitProcedure)(void);
   UINT32         LDRPCDataTable;
   UINT32         LDRealRPCDataTable;
   UINT32         LDRPCDataTableSize;
   UINT32         LDNumberOfReferencedPublics;
   void           **LDReferencedPublics;
   UINT32         LDNumberOfReferencedExports;
   void           *LDUnresolvedExternals;
   void           *LDSharedUnresolvedExternals;
}
MODULE_HANDLE;

typedef enum _REG_TYPE_
{
      REG_TYPE_NEW_ADAPTER,
      REG_TYPE_NEW_FRAME,
      REG_TYPE_NEW_CHANNEL,
      REG_TYPE_FAIL
}
REG_TYPE;

typedef  struct _RCB_
{
   union          {
                     UINT8    RWs_i8val[8];
                     UINT16   RWs_i16val[4];
                     UINT32   RWs_i32val[2];
                     UINT64   RWs_I64val;
                  }
                  RCBDriverWS;
   UINT8          RCBReserved[(UINT32)&(((ECB *)0)->ECB_FragmentCount) -
                              (UINT32)&(((ECB *)0)->ECB_Status)];
   UINT32         RCBFragCount;
   FRAGMENT_STRUCT RCBFragStruct;
}
RCB;

typedef  struct   _SHARED_DATA_
{
   volatile UINT32            CMSMPriorityTxFreeCount;
#ifdef IAPX386
   UINT32                     CMSMReservedForCHSM[16];
	UINT32							CMSMTCB_ECBptrAdjust;
 	UINT32                     CMSMAsmHSMUpdate;
  	struct _MSTATS_            *CMSMAsmStatsPtr;
#else
  	UINT32                     CMSMReservedForCHSM[19];
#endif

   struct _DRIVER_PARM_BLOCK_ *CMSMAsmParmBlock;
   UINT32                     CMSMAdapterIntDisabled;
   UINT32                     CMSMNeedContiguousECB;
   struct _GENERIC_STATS_     *CMSMGenericStatsPtr;
   void                       *CMSMProtectionObject;
   UINT32                     CMSMMaxFrameHeaderSize;
   UINT8                      CMSMPhysNodeAddress[8];
   void                       *CMSMSendListHead;
   void                       *CMSMSendListTail;
   struct _DRIVER_PARM_BLOCK_ *CMSMDPBlockPtr;
   void                       *CMSMMediaDataPtr;
   CONFIG_TABLE               *CMSMDefaultVirtualBoard;
   UINT32                     CMSMDriverInterrupt;
   UINT32                     CMSMInCriticalSection;
   MLID_STATS_TABLE           *CMSMStatsPtr;
   CONFIG_TABLE               *CMSMVirtualBoardLink[4];
   UINT32                     CMSMStatusFlags;
   volatile UINT32            CMSMTxFreeCount;
}
SHARED_DATA;

#define   DADSP_TO_CMSMADSP(n) ((SHARED_DATA *)n - 1)

/* CMSMStatusFlags bit definitions      */
/* Bit 2 (0x04) and bit 3 (0x08) are reserved for compatability with */
/* assembly HSMs. */
#define   SHUTDOWN            0x01
#define   TXQUEUED            0x02
#define	  POLLING_SUSPENDED	  0x10
#define   UNLOADING           0x20

typedef   struct _FRAME_DATA_
{
   CONFIG_TABLE               ConfigTable;
   void                       *Reserved[20];
   void                       *DriverData;
   UINT32                     InternalMediaID;
   UINT32                     MediaHeaderSize;
   UINT8                      PacketType[8];
   UINT32                     BitSwapAddressFlag;
   struct _DRIVER_PARM_BLOCK_ *DPBlock;
   struct _TSM_PARM_BLOCK_    *MPBlock;
   void                       *FirmwareBuffer;
   UINT32                     HardwareIsRegistered;
	void								(_cdecl * TransmitRoutine)(
                                 TCB			   *tcb);
   void                       (_cdecl * SendEntry)(
                                 ECB            *ecb,
                                 CONFIG_TABLE   *configTable);
   UINT32                     (_cdecl * SendRoutine)(
                                 SHARED_DATA    *sharedDATA,
                                 ECB            *ecb,
                                 TCB			   *tcb);
   UINT32                     (_cdecl * Driver_Control)(void);
   void                          *Nothing;
}
FRAME_DATA;

#define CONFIG_TO_MSM(n) (FRAME_DATA *)((UINT8 *)n - (UINT32)&((FRAME_DATA *)0)->ConfigTable)
#define AES_TO_MLID(n) (MLID_AES_ECB *)((UINT8 *)n - (UINT32)&((MLID_AES_ECB *)0)->AesReserved)
#define MILL_TO_TICKS(n)   (n/55)

typedef  struct _CHSM_STACK_
{
   struct _MODULE_HANDLE_  *ModuleHandle;
   SCREEN_HANDLE           *ScreenHandle;
   MEON                    *CommandLine;
   MEON                    *ModuleLoadPath;
   UINT32                  UnitializedDataLength;
   void                    *CustomDataFileHandle;
   UINT32                  (* FileRead)(
                              void      *FileHandle,
                              UINT32   FileOffset,
                              void      *FileBuffer,
                              UINT32   FileSize);
   UINT32                  CustomDataOffset;
   UINT32                  CustomDataSize;
   UINT32                  NumMsgs;
   MEON                    **Msgs;
}
CHSM_STACK;

typedef struct   _DRIVER_PARM_BLOCK_
{
   UINT32                  DriverParameterSize;
   CHSM_STACK              *DriverInitParmPointer;
   struct _MODULE_HANDLE_  *DriverModuleHandle;
   void                    *DPB_Reserved0;
   void                    *DriverAdapterPointer;
   CONFIG_TABLE            *DriverConfigTemplatePtr;
   UINT32                  DriverFirmwareSize;
   void                    *DriverFirmwareBuffer;
   UINT32                  DPB_Reserved1;
   void							*DPB_Reserved2;
   void							*DPB_Reserved3;
   void							*DPB_Reserved4;
   UINT32						DriverAdapterDataSpaceSize;
   struct _DRIVER_DATA_		*DriverAdapterDataSpacePtr;
   UINT32                  DriverStatisticsTableOffset;
   UINT32                  DriverEndOfChainFlag;
   UINT32                  DriverSendWantsECBs;
   UINT32                  DriverMaxMulticast;
   UINT32                  DriverNeedsBelow16Meg;
   void                    *DPB_Reserved5;
   void                    *DPB_Reserved6;

   void                    (_cdecl * DriverISRPtr)(
                              struct _DRIVER_DATA_    *driverData);

   ODISTAT                 (_cdecl * DriverMulticastChangePtr)(
                              struct _DRIVER_DATA_    *driverData,
                              CONFIG_TABLE            *configTable,
                              GROUP_ADDR_LIST_NODE    *mcTable,
                              UINT32                  numEntries,
                              UINT32                  functionalTable);

   void                    (_cdecl * DriverPollPtr)(
                              struct _DRIVER_DATA_    *driverData,
                              CONFIG_TABLE            *configTable);

   ODISTAT                 (_cdecl * DriverResetPtr)(
                              struct _DRIVER_DATA_    *driverData,
                              CONFIG_TABLE            *configTable,
										OPERATION_SCOPE			operationScope);

    void                   (_cdecl * DriverSendPtr)(
                              struct _DRIVER_DATA_    *driverData,
                              CONFIG_TABLE            *configTable,
                              TCB			            *tcb,
                              UINT32                  pktSize,
                              void                    *PhysTcb);

   ODISTAT                 (_cdecl * DriverShutdownPtr)(
                              struct _DRIVER_DATA_    *driverData,
                              CONFIG_TABLE            *configTable,
                              UINT32                  shutDownType,
										OPERATION_SCOPE			operationScope);

   void                    (_cdecl * DriverTxTimeoutPtr)(
                              struct _DRIVER_DATA_    *driverData,
                              CONFIG_TABLE            *configTable);

   ODISTAT                 (_cdecl * DriverPromiscuousChangePtr)(
                              struct _DRIVER_DATA_    *driverData,
                              CONFIG_TABLE            *configTable,
                              UINT32                  promiscuousMode);

   ODISTAT                 (_cdecl * DriverStatisticsChangePtr)(
                              struct _DRIVER_DATA_    *driverData,
                              CONFIG_TABLE            *configTable);

   ODISTAT                 (_cdecl * DriverRxLookAheadChangePtr)(
                              struct _DRIVER_DATA_    *driverData,
                              CONFIG_TABLE            *configTable);

   ODISTAT                 (_cdecl * DriverManagementPtr)(
                              struct _DRIVER_DATA_    *driverData,
                              CONFIG_TABLE            *configTable,
                              ECB                     *ecb);

   void                    (_cdecl * DriverEnableInterruptPtr)(
                              struct _DRIVER_DATA_    *driverData);

   BOOLEAN                 (_cdecl * DriverDisableInterruptPtr)(
                              struct _DRIVER_DATA_    *driverData,
                              BOOLEAN                 switchVal);

   void                    (_cdecl * DriverISR2Ptr)(
                              struct _DRIVER_DATA_    *driverData);

   MEON                    ***DriverMessagesPtr;

	MEON_STRING					*HSMSpecVersionStringPtr;

   ODISTAT                 (_cdecl * DriverPriorityQueue)(
                              struct _DRIVER_DATA_    *driverData,
                              CONFIG_TABLE            *configTable,
										ECB						   *ecb);
   BOOLEAN                 (_cdecl * DriverDisableInterrupt2Ptr)(
                              struct _DRIVER_DATA_    *driverData,
                              BOOLEAN                 switchVal);

}
DRIVER_PARM_BLOCK;

#define	ASM_DPBLOCK_SIZE		136

#define  NOTIFY_FATAL        0
#define  NOTIFY_CRITICAL     1
#define  NOTIFY_DEGRADED     2

typedef  struct   _TSM_CONFIG_LIMITS_
{
   UINT8       *MinNodeAddress;
   UINT8       *MaxNodeAddress;
   UINT32      MinRetries;
   UINT32      MaxCRetries;
   UINT32      NumberFrames;
}
TSM_CONFIG_LIMITS;

typedef  struct   _TSM_PARM_BLOCK_
{
   UINT32            MediaParameterSize;
   TSM_CONFIG_LIMITS *MediaConfigLimits;
   MEON_STRING       **MediaFrameDescriptTable;
   UINT8             (*MediaProtocolIDArray)[6];
   UINT8             *MediaIDArray;
   UINT8             *MediaHeaderSizeArray;
   UINT32            (_cdecl ** MediaSendRoutineArray)(
                        SHARED_DATA   *sharedDATA,
                        ECB *,
                        TCB *tcb);
   UINT32            MediaAdapterDataSpaceSize;
   struct _MEDIA_DATA_ *MediaAdapterPtr;
   ODISTAT           (_cdecl * MediaAdjustPtr)(
					         FRAME_DATA							*frameData);

   ODISTAT           (_cdecl * MediaInitPtr)(
                        const   struct _DRIVER_DATA_	*driverData,
                        FRAME_DATA							*frameData);


   ODISTAT           (_cdecl * MediaResetPtr)(
                        struct _DRIVER_DATA_    *driverData,
                        FRAME_DATA              *frameData,
								OPERATION_SCOPE			operationScope);

   ODISTAT           (_cdecl * MediaShutdownPtr)(
                        struct _DRIVER_DATA_    *driverData,
                        FRAME_DATA              *frameData,
                        UINT32                  shutdownType,
								OPERATION_SCOPE			operationScope);

   void              (_cdecl * MediaSendPtr)(
                        ECB                     *ecb,
                        CONFIG_TABLE            *configTable);

   void              *(_cdecl * MediaGetNextSendPtr)(
                        struct _DRIVER_DATA_    *driverData,
                        CONFIG_TABLE            **configTable,
                        UINT32                  *PacketSize,
                        void                    **PhysTcb);

   ODISTAT           (_cdecl * MediaAddMulticastPtr)(
                        struct _DRIVER_DATA_    *driverData,
								NODE_ADDR					*McAddress);


   ODISTAT           (_cdecl * MediaDeleteMulticastPtr)(
                        struct _DRIVER_DATA_    *driverData,
								NODE_ADDR					*McAddress);

   ODISTAT           (_cdecl * MediaNodeOverridePtr)(
                        FRAME_DATA              *frameData,
                        MEON                    mode);

   ODISTAT           (_cdecl * MediaAdjustNodeAddressPtr)(
                        FRAME_DATA              *frameData);

   ODISTAT           (_cdecl * MediaSetLookAheadSizePtr)(
                        struct _DRIVER_DATA_    *driverData,
                        FRAME_DATA              *frameData,
                        UINT32                  size);

   ODISTAT           (_cdecl * MediaPromiscuousChangePtr)(
                        struct _DRIVER_DATA_    *driverData,
                        FRAME_DATA              *frameData,
                        UINT32                  PromiscuousState,
                        UINT32                  *PromiscuousMode);

   ODISTAT           (_cdecl * MediaRegisterMonitorPtr)(
                        FRAME_DATA              *frameData,
                        void                    (_cdecl * TXRMonRoutine)(TCB *),
                        BOOLEAN                 MonitorState);

   ODISTAT           (_cdecl * MediaGetParametersPtr)(
                        CONFIG_TABLE            *configTable);

	ODISTAT           (_cdecl * MediaGetMulticastInfo)(
                        struct _DRIVER_DATA_    *driverData,
                        ECB                     *MulticastInfoECB);
}
TSM_PARM_BLOCK;

typedef struct _GENERIC_STATS_
{
   UINT32   Gen_TotalTxPackets;
   UINT32   Gen_TotalRxPackets;
   UINT32   Gen_NoECBs;
   UINT32   Gen_TxTooBig;
   UINT32   Gen_TxTooSmall;
   UINT32   Gen_RxOverflow;
   UINT32   Gen_RxTooBig;
   UINT32   Gen_RxTooSmall;
   UINT32   Gen_TxMiscError;
   UINT32   Gen_RxMiscError;
   UINT32   Gen_TxRetryCount;
   UINT32   Gen_RxCheckSumError;
   UINT32   Gen_RxMisMatchError;
   UINT32   Gen_TotalTxOkByteLow;
   UINT32   Gen_TotalTxOkByteHigh;
   UINT32   Gen_TotalRxOkByteLow;
   UINT32   Gen_TotalRxOkByteHigh;
   UINT32   Gen_TotalGroupAddrTx;
   UINT32   Gen_TotalGroupAddrRx;
   UINT32   Gen_AdapterReset;
   UINT32   Gen_AdapterOPRTimeStamp;
   UINT32   Gen_Qdepth;
   UINT32   Gen_TxOKSingleCollisions;
   UINT32   Gen_TxOKMultipleCollisions;
   UINT32   Gen_TxOKButDeferred;
   UINT32   Gen_TxAbortLateCollision;
   UINT32   Gen_TxAbortExcessCollision;
   UINT32   Gen_TxAbortCarrierSense;
   UINT32   Gen_TxAbortExcessiveDeferral;
   UINT32   Gen_RxAbortFrameAlignment;
}
GENERIC_STATS;

#ifdef   IAPX386
/* v3.00 Assembly HSMs stats table.   */
typedef struct _MSTATS_
{
   UINT8           MSTATS_MajorVer;
   UINT8           MSTATS_MinorVer;
   UINT16          MSTATS_NumGenericCounters;
   UINT32          MSTATS_ValidMask0;
   GENERIC_STATS   MSTATS_Generic;
}
MSTATS;
#endif

#define  AES_PROCESS_SIG                  0x50534541         /* 'PSEA' */
#define  ALLOC_SIG                        0x54524C41         /* 'TRLA' */
#define  ALLOC_STACK_SIG                  0x53524C41         /* 'SRLA' */
#define  CACHE_BELOW_16_MEG_MEMORY_SIG    0x36314243         /* '61BC' */
#define  ECB_SIG                          0x53424345         /* 'SBCE' */
#define  EVENT_SIG                        0x544E5645         /* 'TNVE' */
#define  FAST_WORK_TO_DO_SIG              0x44545746         /* 'DTWF' */
#define  INTERRUPT_SIG                    0x50544E49         /* 'PTNI' */
#define  IO_REGISTRATION_SIG              0x53524F49         /* 'SROI' */
#define  LOADABLE_ROUTER_SIG              0x5452444C         /* 'TRDL' */
#define  LSL_AES_EVENT_SIG                0x5345414C         /* 'SEAL' */
#define  LSL_DEFAULT_STACK_SIG            0x444C534C         /* 'DLSL' */
#define  LSL_PRESCAN_STACK_SIG            0x504C534C         /* 'PLSL' */
#define  LSL_STACK_SIG                    0x534C534C         /* 'SLSL' */
#define  LSL_TX_PRESCAN_STACK_SIG         0x544C534C         /* 'TLSL' */
#define  MLID_SIG                         0x44494C4D         /* 'DILM' */
#define  POLLING_PROCEDURE_SIG            0x52504C50         /* 'RPLP' */
#define  PROCESS_SIG                      0x53435250         /* 'SCRP' */
#define  TIMER_SIG                        0x524D4954         /* 'RMIT' */

/***********************************************************************\
*                                                                       *
* MSM Specific declarations.                                            *
*                                                                       *
\***********************************************************************/
void     _cdecl   CMSMAddToCounter(
            STAT_TABLE_ENTRY           *statTableEntryPtr,
            UINT32                     value);

void*    _cdecl   CMSMAlloc(
            const struct _DRIVER_DATA_ *driverData,
            UINT32                     nBytes);

void*    _cdecl   CMSMAllocPages(
            const struct _DRIVER_DATA_ *driverData,
            UINT32                     nBytes);

RCB*     _cdecl   CMSMAllocateRCB(
            const struct _DRIVER_DATA_ *driverData,
            UINT32                     nBytes,
            void                       **PhysicalRCB);

RCB*     _cdecl    CMSMAllocateMultipleRCBs (
            const struct _DRIVER_DATA_    *driverData,
            UINT32               nbytes,
            UINT32               *nRCBs,
            void                 **physicalRCB);

void  	_cdecl	CMSMAcquireProtection(
				void		*pObject,
            void    	*state);

void     _cdecl   CMSMControlComplete(
            struct _DRIVER_DATA_       *driverData,
            CHSM_COMPLETE              controlFunction,
            ODISTAT                    completionStatus);

UINT32   _cdecl   CMSMConvertLinAddrLen(
            void 						*linearAddress,
            UINT32 					linearAddressLength,
            FRAGMENT_STRUCT			**outputFragmentList,
     			struct _DRIVER_DATA_ *driverData);

ODISTAT _cdecl		CMSMDeRegisterResource(
				struct _DRIVER_DATA_ *driverData,
				EXTRA_CONFIG *extraConfig,
				ECB *pAsyncECB);

ODISTAT _cdecl		CMSMResetMLID(
				struct _DRIVER_DATA_ *driverData);

ODISTAT _cdecl		CMSMShutdownMLID(
				struct _DRIVER_DATA_ *driverData,
				UINT32 shutdownType);

void     _cdecl   CMSMDisableHardwareInterrupt(
            const struct _DRIVER_DATA_ *driverData);

void     _cdecl   CMSMDriverRemove(
            struct _MODULE_HANDLE_     *moduleHandle);

/* These functions are used to get the original ECB (with Logical addresses in */
/* its frag list) back if the driver has set the MM_FRAGS_PHYS_BIT and needs   */
/* Logical fragments.  The original ECB is stored in the ECB_ESR field.        */

ECB*     _cdecl   CMSMECBPhysToLogFrags(
            ECB *ecb);

ECB*     _cdecl   CMSMTCBPhysToLogFrags(
            TCB *tcb);

void     _cdecl   CMSMEnableHardwareInterrupt(
            const struct _DRIVER_DATA_ *driverData);

ODISTAT  _cdecl   CMSMEnablePolling(
            const struct _DRIVER_DATA_ *driverData);

void     _cdecl   CMSMEnqueueSend(
            FRAME_DATA                 *frameData,
            const struct _DRIVER_DATA_ *driverData,
            ECB                        *ecb);

void     _cdecl   CMSMFree(
            const struct _DRIVER_DATA_ *driverData,
            void                       *dataPtr);

void     _cdecl   CMSMFreePages(
            const struct _DRIVER_DATA_ *driverData,
            void                       *dataPtr);

ODI_NBI  _cdecl   CMSMGetCardConfigInfo(
            void                       *busTag,
            UINT32                     uniqueIdentifier,
            UINT32                     size,
            UINT32                     parm1,
            UINT32                     parm2,
            void                       *configInfo);

ODISTAT	_cdecl	CMSMGetConfigInfo (
			        void 	*configInfo,
			        UINT32 	*nBytes);

UINT32   _cdecl   CMSMGetCurrentTime(
            void);

ODI_NBI _cdecl	CMSMGetHINFromHINName(
										MEON_STRING *hinName,
										UINT16 		*hin);
ODI_NBI	_cdecl CMSMGetHINNameFromHIN(
										UINT16 		hin,
										MEON_STRING *hinName );

UINT32   _cdecl   CMSMGetMicroTimer (
            void);

void *   _cdecl   CMSMGetPhysical(
            void                       *logicalAddr);

ODISTAT	_cdecl	CMSMGetPhysList(
				UINT32							inputFragCount,
				FRAGMENTSTRUCT					*inputFragList,
				UINT32							*outputFragCount,
				FRAGMENTSTRUCT					*outputFragList,
				struct _DRIVER_DATA_			*driverData);


ECB *    _cdecl   CMSMGetPhysicalECB(
            struct _DRIVER_DATA_       *driverData,
            ECB                        *ecb);

UINT32   _cdecl   CMSMGetPollSupportLevel(
                  	void);

void     _cdecl   CMSMIncrCounter(
            STAT_TABLE_ENTRY           *statTableEntryPtr);

void *   _cdecl   CMSMInitAlloc(
            UINT32                     nbytes);

StatTableEntry *  _cdecl     CMSMBuildGenericStatStrings(
            StatTableEntry             *tableEntry);


#ifdef DEBUG

#define  MONO         0xb0000
#define  COLOR        0xb8000
#define  UPPER_LINE   10
#define  LOWER_LINE   22

void     CMSMOutUINT16(
            UINT16                     value,
            UINT32                     ScreenBase);

void     CMSMOutUINT32(
            UINT32                     value,
            UINT32                     ScreenBase);

void     CMSMOutUINT8(
            UINT8                      value,
            UINT32                     ScreenBase);

void     CMSMOutMEON(
            MEON                       Char,
            UINT32                     screen);
#endif

ODISTAT  _cdecl CMSMInitParser(
            DRIVER_PARM_BLOCK          *hsmParmBlock);

ODISTAT  _cdecl CMSMParseDriverParameters(
            DRIVER_PARM_BLOCK          *hsmParmBlock,
            struct _DRIVER_OPTION_     *driverOption);

ODISTAT  _cdecl CMSMParseSingleParameter(
            struct _DRIVER_OPTION_     *driverOption);

ODISTAT  _cdecl CMSMParseCTSMParameters(
            CONFIG_TABLE               *configTable,
            struct _DRIVER_OPTION_     *driverOption);

ODISTAT  CMSMASCIINodeToHex(
            CONFIG_TABLE               *configTable,
            MEON_STRING                *asciiNode);

typedef  enum   _MSG_TYPE_
{
   MSG_TYPE_INIT_INFO,
   MSG_TYPE_INIT_WARNING,
   MSG_TYPE_INIT_ERROR,
   MSG_TYPE_RUNTIME_INFO,
   MSG_TYPE_RUNTIME_WARNING,
   MSG_TYPE_RUNTIME_ERROR
}
MSG_TYPE;

void     _cdecl   CMSMPrintString(
            const CONFIG_TABLE         *configTable,
            MSG_TYPE                   msgType,
            MEON_STRING                *msg,
            void                       *parm1,
            void                       *parm2);

void     _cdecl   CMSMReadPhysicalMemory(
            UINT32                     nbytes,
            void                       *destAddr,
            void                       *srcBusTag,
            const void                 *physSrcAddr);

REG_TYPE _cdecl   CMSMRegisterHardwareOptions(
            CONFIG_TABLE               *configTable,
            struct _DRIVER_DATA_       **driverData);

ODISTAT  _cdecl   CMSMRegisterMLID(
            const struct _DRIVER_DATA_ *driverData,
            CONFIG_TABLE               *configTable);

ODISTAT  _cdecl   CMSMRegisterResource(
            const struct _DRIVER_DATA_ *driverData,
            CONFIG_TABLE               *configTable,
            EXTRA_CONFIG               *extraConfig);

ODISTAT  _cdecl   CMSMRegisterTSM(
            DRIVER_PARM_BLOCK          *DriverParameterBlock,
            TSM_PARM_BLOCK             *MediaParameterBlock,
            CONFIG_TABLE               **configTable);

void  	_cdecl	CMSMReleaseProtection(
				void		*pObject,
            UINT32  	state);

ODISTAT _cdecl		CMSMReRegisterHardwareOptions(
				struct _DRIVER_DATA_ *driverData,
				IO_CONFIG *newIOConfig,
				ECB *pAsyncECB);

ODISTAT	_cdecl	CMSMResumePolling (
            struct _DRIVER_DATA_       *driverData);

void     _cdecl   CMSMReturnDriverResources(
            const CONFIG_TABLE         *configTable);

ECB*     _cdecl   CMSMReturnPhysicalECB(
            const struct _DRIVER_DATA_ *driverData,
            ECB                        *ecb);

void     _cdecl   CMSMReturnRCB(
            const struct _DRIVER_DATA_ *driverData,
            RCB                        *rcb);

void     _cdecl    CMSMReturnMultipleRCBs (
            RCB                         *rcbp);

ODISTAT  _cdecl   CMSMScheduleAES(
            const struct _DRIVER_DATA_ *driverData,
            MLID_AES_ECB               *mlidAESECB);

ODISTAT	_cdecl	CMSMSuspendPolling (
            struct _DRIVER_DATA_       *driverData);

ODISTAT  _cdecl   CMSMCancelAES (
            const struct _DRIVER_DATA_  *driverData,
            MLID_AES_ECB                *mlidAESECB);


void     _cdecl   CMSMServiceEvents (
            void);

ODISTAT	_cdecl	CMSMUpdateConfigTables (
            struct _DRIVER_DATA_       *driverData,
            MLID_CONFIG_TABLE          *configTable);

void     _cdecl   CMSMYieldWithDelay (void);

void junk (const struct _DRIVER_DATA_ *driverData);

/***********************************************************************\
*                                                                       *
* NOTE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!                                     *
*   driverData in CMSMSetHardwareInterrupt is purposely not set to const*
*  since setting the last prototype in this file using _DRIVER_DATA_    *
*  cannot be set that way for some reason. It causes the HSM to not     *
*   compile, as if _DRIVER_DATA_ gets stuck as being const from here on.*
*                                                                       *
\***********************************************************************/

ODISTAT  _cdecl   CMSMSetHardwareInterrupt(
            struct _DRIVER_DATA_       *driverData,
            const CONFIG_TABLE         *configTable);

void     _cdecl   CMSMWritePhysicalMemory(
            UINT32                     nbytes,
            void                       *destBusTag,
            void                       *physDestAddr,
            const void                 *srcAddr);

ODISTAT  _cdecl   CMSMEnableWriteCombine(
            struct _DRIVER_DATA_       *driverData,
            void                       *firstPageAddr,
            UINT32                     numPages);

ODISTAT  _cdecl   CMSMDisableWriteCombine(
            struct _DRIVER_DATA_       *driverData,
            void                       *firstPageAddr,
            UINT32                     numPages);

ODISTAT  _cdecl   CMSMFlushWriteCombine(void);

#ifdef _WATCOMC_

#pragma aux CMSMFlushWriteCombine = \
	"push	eax"								\
	"mov	eax, CR2"						\
	"mov	CR2, eax"						\
	"pop	eax";

#endif

/***********************************************************************\
*                                                                       *
* CMSM NetWare Bus Interface support declarations.                      *
*                                                                       *
\***********************************************************************/

UINT32	_cdecl	CMSMGetAlignment (UINT32 type );

ODI_NBI	_cdecl	CMSMGetBusInfo (
	  			void   	*busTag,
	  			UINT32	*physicalMemAddressSize,
	  			UINT32	*ioAddrSize );

ODI_NBI	_cdecl	CMSMGetBusName (
				void	    *busTag,
				MEON_STRING **busName );

ODI_NBI	_cdecl	CMSMGetBusTag (
				const	MEON_STRING	*busName,
				void	**busTag );

ODI_NBI	_cdecl	CMSMGetBusType (
	  			void	*busTag,
	  			UINT32	*busType );

ODI_NBI	_cdecl	CMSMGetBusSpecificInfo (
				void    *busTag,
				UINT32	size,
				void	*busSpecificInfo);


ODI_NBI  _cdecl	CMSMGetCardConfigInfo (
                void       *busTag,
                UINT32     uniqueIdentifier,
                UINT32     size,
                UINT32     parm1,
                UINT32     parm2,
                void       *configInfo);

ODI_NBI	_cdecl	CMSMGetInstanceNumber (
				void    *busTag,
				UINT32	uniqueIdentifier,
				UINT16	*instanceNumber);


ODI_NBI	_cdecl	CMSMGetInstanceNumberMapping (
				UINT16	instanceNumber,
				void    **busTag,
				UINT32	*uniqueIdentifier);

ODI_NBI	_cdecl	CMSMGetSlot(
				void		*busTag,
				MEON_STRING	*slotName,
				UINT16		*slot);

ODI_NBI	_cdecl	CMSMGetSlotName(
				void		*busTag,
				UINT16		slot,
				MEON_STRING	**slotName);

ODI_NBI	_cdecl	CMSMGetUniqueIdentifier (
           	void     *busTag,
			  	UINT32   *parameters,
			  	UINT32   parameterCount,
			  	UINT32   *uniqueIdentifier);

ODI_NBI	_cdecl	CMSMGetUniqueIdentifierParameters (
           void     *busTag,
		   UINT32   uniqueIdentifier,
		   UINT32   parameterCount,
		   UINT32   *parameters);

UINT8		_cdecl	CMSMRdConfigSpace8 (
				void		*busTag,
				UINT32	uniqueIdentifier,
				UINT32	offset);

UINT16	_cdecl	CMSMRdConfigSpace16 (
				void		*busTag,
				UINT32	uniqueIdentifier,
				UINT32	offset);

UINT32	_cdecl	CMSMRdConfigSpace32 (
				void		*busTag,
				UINT32	uniqueIdentifier,
				UINT32	offset);

ODI_NBI	_cdecl	CMSMScanBusInfo (
				UINT32	    *scanSequence,
				void		   **busTag,
				UINT32		*busType,
				MEON_STRING	**busName );

ODI_NBI  _cdecl   CMSMSearchAdapter(
            UINT32      *scanSequence,
            UINT32      busType,
            UINT32      productIDLen,
            const MEON  *productID,
            void	  	**busTag,
            UINT32      *uniqueIdentifier);

void	_cdecl	CMSMWrtConfigSpace8 (
  				void	    *busTag,
				UINT32	uniqueIdentifier,
				UINT32	offset,
  				UINT8		writeVal );

void	_cdecl	CMSMWrtConfigSpace16 (
				void		*busTag,
				UINT32	uniqueIdentifier,
				UINT32	offset,
	  			UINT16 	writeVal );

void 	_cdecl	CMSMWrtConfigSpace32 (
	  			void	*busTag,
				UINT32	uniqueIdentifier,
				UINT32	offset,
	  			UINT32	writeVal );


/***********************************************************************\
*                                                                       *
* CMSM NESL support declarations.                  						   *
*                                                                       *
\***********************************************************************/

UINT32 	_cdecl	CMSMNESLDeRegisterConsumer(
				NESL_ECB	*Consumer);

UINT32	_cdecl	CMSMNESLDeRegisterProducer(
				NESL_ECB	*Producer);

UINT32 	_cdecl	CMSMNESLProduceEvent(
				NESL_ECB	*ProducerNecb,
				NESL_ECB	**ConsumerNecb,
				EPB		  	*EventParmBlock);

UINT32 	_cdecl	CMSMNESLProduceMLIDEvent(
				NESL_ECB	*ProducerNecb,
				NESL_ECB	**ConsumerNecb,
				EPB		  	*EventParmBlock,
				struct _DRIVER_DATA_	*driverData);


UINT32	_cdecl	CMSMNESLRegisterConsumer(
				NESL_ECB	*Consumer);

UINT32	_cdecl	CMSMNESLRegisterProducer(
				NESL_ECB	*Producer);

/***********************************************************************\
*                                                                       *
* OS Specific, non portable declarations.                               *
*                                                                       *
\***********************************************************************/

void   _cdecl   EnterDebugger(
            void);

/***********************************************************************\
*                                                                       *
* Message enabling definitions.                                         *
*                                                                       *
\***********************************************************************/

#define  MSG(s,id) DriverMessages[id]
#define  TxtMSG(s,id) s

#define  CMP_NODE(a,b) ( \
   ((UINT32 *)(a))[0] == ((UINT32 *)(b))[0] &&\
   ((UINT16 *)(a))[2] == ((UINT16 *)(b))[2] \
)

/***********************************************************************\
*                                                                       *
* The following structures and definitions are temporary.               *
*                                                                       *
\***********************************************************************/

struct IOOptionStructure

{
   UINT32   NumberOfOptions;
   UINT32   OptionData[1];
};

typedef   struct AdapterOptionDefinitionStructure
{
   struct IOOptionStructure *IOSlot;
   struct IOOptionStructure *IOPort0;
   struct IOOptionStructure *IOLength0;
   struct IOOptionStructure *IOPort1;
   struct IOOptionStructure *IOLength1;
   struct IOOptionStructure *MemoryDecode0;
   struct IOOptionStructure *MemoryLength0;
   struct IOOptionStructure *MemoryDecode1;
   struct IOOptionStructure *MemoryLength1;
   struct IOOptionStructure *Interrupt0;
   struct IOOptionStructure *Interrupt1;
   struct IOOptionStructure *DMA0;
   struct IOOptionStructure *DMA1;
   struct IOOptionStructure *Channel;
}
HSM_OPTIONS;

#if defined _WATCOMC_ || defined OS_NT
/* Reset PRAGMA to normal after packing above structures */
#pragma   pack()
#endif

/* Flags that can be set on the OptionData */

#define OD_IS_RANGE              0x80000000
#define OD_HAS_INCREMENT         0x40000000

/* ParseIOParameters bits */

#define  NeedsSlotBit            0x00000001
#define  NeedsIOPort0Bit         0x00000002
#define  NeedsIOLength0Bit       0x00000004
#define  NeedsIOPort1Bit         0x00000008
#define  NeedsIOLength1Bit       0x00000010
#define  NeedsMemoryDecode0Bit   0x00000020
#define  NeedsMemoryLength0Bit   0x00000040
#define  NeedsMemoryDecode1Bit   0x00000080
#define  NeedsMemoryLength1Bit   0x00000100
#define  NeedsInterrupt0Bit      0x00000200
#define  NeedsInterrupt1Bit      0x00000400
#define  NeedsDMA0Bit            0x00000800
#define  NeedsDMA1Bit            0x00001000
#define  NeedsChannelBit         0x00002000

/* ParseLANParameters needFlags defines */

#define  CAN_SET_NODE_ADDRESS    0x40000000
#define  MUST_SET_NODE_ADDRESS   0x80000000

/* RegisterForEventNotification defines */

#define  EVENT_DOWN_SERVER             4
#define  EVENT_CHANGE_TO_REAL_MODE     5
#define  EVENT_RETURN_FROM_REAL_MODE   6
#define  EVENT_EXIT_TO_DOS             7
#define  EVENT_MODULE_UNLOAD           8

#define  EVENT_PRIORITY_OS             0
#define  EVENT_PRIORITY_APPLICATION    20
#define  EVENT_PRIORITY_DEVICE         40

/********************************************************************
* MSMBitSwapTable is defined in cmsm.c.  It is used by the TSMs and
* possibly the HSMs.  It is made public in the msm to remove duplicate
* tables.
*********************************************************************/
extern UINT8 MSMBitSwapTable[];


/***********************************************************************\
*                                                                       *
* The following declarations are temporary.                              *
*                                                                       *
************************************************************************/
ODISTAT   _cdecl CMSMOldParseDriverParameters(
            CONFIG_TABLE      *configTable,
            HSM_OPTIONS         *hsmOptions,
            UINT32             driverNeedsBitMask);

ODISTAT   _cdecl CMSMParseCustomKeywords(
            DRIVER_PARM_BLOCK   *parameters,
            CONFIG_TABLE      *configTable,
            MEON_STRING         *commandLine);

ODISTAT   _cdecl CMSMHardwareFailure(
	     struct _DRIVER_DATA_ *driverData,
             UINT32        failureType,
             MEON_STRING  *failMsgString);

#define CMSM_H
#endif /* CMSM_H */

