/***********************************************************************\
*                                                                       *
*   (c) Unpublished Work of Novell, Inc.  All rights reserved.
*
*   THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
*   PROPRIETARY AND TRADE SECRET INFORMATION OF NOVELL, INC.  ACCESS
*   TO THIS WORK IS RESTRICTED TO (I) NOVELL EMPLOYEES WHO HAVE A
*   NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE OF THEIR ASSIGNMENTS
*   AND (II) ENTITIES OTHER THAN NOVELL WHO HAVE ENTERED INTO
*   APPROPRIATE LICENSE AGREEMENTS.  NO PART OF THIS WORK MAY BE USED,
*   PRACTICED, PERFORMED, COPIED, DISTRIBUTED, REVISED, MODIFIED,
*   TRANSLATED, ABRIDGED, CONDENSED, EXPANDED, COLLECTED, COMPILED,
*   LINKED,RECAST, TRANSFORMED OR ADAPTED WITHOUT THE PRIOR WRITTEN
*   CONSENT OF NOVELL.  ANY USE OR EXPLOITATION OF THIS WORK WITHOUT
*   THE PRIOR WRITTEN CONSENT OF NOVELL.  ANY USE OR EXPLOITATION OF
*   THIS WORK WITHOUT AUTHORIZATION COULD SUBJECT THE PERPETRATOR TO
*   CRIMINAL AND CIVIL LIABILITY.
*                                                                      
***********************************************************************
*
*  Name:          ETHERTSM.h
*
*  Description:   C version of the Ethernet Structures
*
*  ODI Spec Ver:  1.11
*
*  Modification History:
*
*  08-29-96 WTT		Added proto type for CEtherTSMGetConfigInfo.
*
*  12-05-96 WTT		Changed proto types for MediaAddMulticast, 
*					MediaDeleteMulticast and FindAddressInMCTable to 
*					conform to spec.  They were using UINT8 pointer and 
*					should have been using NODE_ADDR pointer for
*					the mcAddress.
*					SPD #143895
*
\***********************************************************************/
#ifndef ETHERTSM_H

/**************************************************************************\
**
**      Defines
**
\**************************************************************************/
/* This defines the size of each Frame-Types MAC Header.	*/
#define	SIZE_E8023			14
#define	SIZE_E8022			17
#define	SIZE_EII				14
#define	SIZE_ESNAP			22

/*These are used for indexing into a specific Frame-Type.		*/
#define E8022_INTERNALID	0
#define E8023_INTERNALID	1
#define EII_INTERNALID		2
#define ESNAP_INTERNALID	3

/*These are used for supporting Assembly HSMs.					*/
#define	E8022_DESCRIPT_LEN	14
#define	E8023_DESCRIPT_LEN	14
#define	EII_DESCRIPT_LEN		11
#define	ESNAP_DESCRIPT_LEN	13

/*These are the NOVELL assigned Frame IDs.					*/
#define	E8023_ID				5
#define	EII_ID				2
#define	E8022_ID				3
#define	ESNAP_ID				10

/* This is the Maximum size a packet can be	*/
#define MAX8023LENGTH		1500
#define MAX_PACKET_LENGTH	1514
#define MIN_PKT_SIZE			60
#define MIN_RUNT_SIZE		22

#define MULTICASTBIT			0x01
#define GROUPBIT				0x01
#define LOCALBIT				0x02


/* These bits define 802.2 types.				*/
#define S_OR_U_FORMAT		0x01
#define U_FORMAT				0x02

/* ECB_DataLength bit assignmets for pipeline and ecb aware adapters. */
#define HSM_HAS_BEEN_IN_GETRCB	0x80000000		/* bit 31 */
#define PSTK_WANTS_FRAME			0x40000000		/* bit 30 */
#define HSM_TOLD_TO_SKIP			0x007F0000		/* bits 23 through 16 */
#define HSM_TOLD_TO_COPY			0x0000FFFF		/* bits 15 through 0 */

/* v3.00 Assembly Statistics Table */
#define	STATISTICSMASK				0x0D300003
#define	ETHERNETSTATSCOUNT		8
#define	MAX_MULTICAST				32

/***********************************************************************\
*                                                                       *
* EtherTSM Specific declarations.													*
*                                                                       *
************************************************************************/
#define	NUMBER_OF_PROMISCUOUS_COUNTERS	32

typedef struct	_MEDIA_DATA_	{
	UINT32			MaxMulticastAddresses;
	GROUP_ADDR_LIST_NODE	*MulticastAddressList;
	UINT32			MulticastAddressCount;
	TCB			*TCBHead;    
	TCB			*TCBList;    
	MLID_STATS_TABLE        *NewStatsPtr;
	UINT32			PromiscuousMode;
	UINT8		PromiscuousCounters[NUMBER_OF_PROMISCUOUS_COUNTERS];
	UINT32			RxStatus;
} MEDIA_DATA;


ODISTAT	CEtherTSMRegisterHSM(
				DRIVER_PARM_BLOCK		*DriverParameterBlock,
				CONFIG_TABLE			**configTable);

RCB 		*CEtherTSMFastProcessGetRCB(
				DRIVER_DATA				*driverData,
				RCB			 			*rcb,
				UINT32					pktSize,
				UINT32					rcvStatus,
				UINT32					newRcbSize);

void		CEtherTSMFastSendComplete(
				DRIVER_DATA				*driverData,
				TCB						*tcb,
				UINT32					transmitStatus);

TCB		*CEtherTSMGetNextSend(
				DRIVER_DATA				*driverData,
				CONFIG_TABLE			**configTable,
				UINT32					*lengthToSend,
				void						**TCBPhysicalPtr);

RCB		*CEtherTSMGetRCB(
				DRIVER_DATA				*driverData,
				UINT8						*lookAheadData,
				UINT32 					pktSize,
				UINT32					rcvStatus,
				UINT32 					*startByte,
				UINT32 					*numBytes);

RCB 		*CEtherTSMProcessGetRCB(
				DRIVER_DATA				*driverData,
				RCB			 			*rcb,
				UINT32 					pktSize,
				UINT32					rcvStatus,
				UINT32					newRcbSize);

void		CEtherTSMFastRcvComplete(
				DRIVER_DATA				*driverData,
				RCB						*rcb);

void		CEtherTSMFastRcvCompleteStatus(
				DRIVER_DATA				*driverData,
				RCB 						*rcb,
				UINT32					packetLength,
				UINT32					packetStatus);

UINT32	CEtherTSMGetHSMIFLevel();

void		CEtherTSMRcvComplete(
				DRIVER_DATA				*driverData,
				RCB						*rcb);

void		CEtherTSMRcvCompleteStatus(
				DRIVER_DATA				*driverData,
				RCB 						*rcb,
				UINT32					packetLength,
				UINT32					packetStatus);

void		CEtherTSMSendComplete(
				DRIVER_DATA				*driverData,
				TCB						*tcb,
				UINT32					transmitStatus);

void		CEtherTSMECBRcvComplete(
				DRIVER_DATA				*driverData,
				ECB 						*ecb);

void		CEtherTSMPipelineRcvComplete(
				DRIVER_DATA				*driverData,
				ECB 						*ecb,
				UINT32					packetLength,
				UINT32					packetStatus);

ODISTAT  CEtherTSMUpdateMulticast(
				DRIVER_DATA				*driverData);

UINT32 CMediaSendRaw8023(
            SHARED_DATA          *sharedData,
				ECB 						*ecb,
				TCB 						*tcb);

UINT32 CMediaSend8022Over8023(
            SHARED_DATA          *sharedData,
				ECB 						*ecb,
				TCB 						*tcb);

UINT32 CMediaSend8022Snap(
            SHARED_DATA          *sharedData,
				ECB 						*ecb,
				TCB 						*tcb);

UINT32 CMediaSendEthernetII(
            SHARED_DATA          *sharedData,
				ECB 						*ecb,
				TCB 						*tcb);

GROUP_ADDR_LIST_NODE *FindAddressInMCTable(
				SHARED_DATA				*sharedData,
				NODE_ADDR				*mcAddress);

ODISTAT	MediaAdjust(
				FRAME_DATA				*frameData);

ODISTAT	MediaInit(
				DRIVER_DATA				*driverData,
				FRAME_DATA				*frameData);

ODISTAT	MediaReset(
				DRIVER_DATA				*driverData,
				FRAME_DATA				*frameData,
				OPERATION_SCOPE		operationScope);

ODISTAT	MediaShutdown(
				DRIVER_DATA				*driverData,
				FRAME_DATA				*frameData,
				UINT32					shutdownType,
				OPERATION_SCOPE		operationScope);

void	MediaSend(
				ECB						*ecb,
				CONFIG_TABLE			*configTable);

ODISTAT	MediaAddMulticast(
				DRIVER_DATA				*driverData,
				NODE_ADDR				*McAddress);

ODISTAT	MediaDeleteMulticast(
				DRIVER_DATA				*driverData,
				NODE_ADDR				*McAddress);

ODISTAT	MediaNodeOverride(
				FRAME_DATA				*frameData,
				MEON						mode);


ODISTAT	MediaAdjustNodeAddress(
				FRAME_DATA				*frameData);


ODISTAT	MediaSetLookAheadSize(
				DRIVER_DATA				*driverData,
				FRAME_DATA				*frameData,
				UINT32					size);

ODISTAT	MediaPromiscuousChange(
				DRIVER_DATA				*driverData,
				FRAME_DATA				*frameData,
				UINT32					PromiscuousState,
				UINT32					*PromiscuousMode);

ODISTAT	MediaRegisterMonitor(
				FRAME_DATA				*frameData,
				void                 (*TXRMonRoutine)(TCB *),
				BOOLEAN					MonitorState);

ODISTAT	MediaGetParameters(
            CONFIG_TABLE      *configTable);

ODISTAT	MediaGetMulticastInfo(
             DRIVER_DATA    *driverData,
             ECB            *MulticastInfoECB);

ODISTAT CEtherTSMBuildTransmitControlBlock(
						DRIVER_DATA		*driverData,
						ECB				*ecb,
						TCB				**tcb,
						UINT32			*pktSize);


void CEtherTSMCancelPrioritySend(
							DRIVER_DATA		*driverData,
							ECB				*ecb);

ODISTAT	CEtherTSMGetConfigInfo (
         void 	*configInfo,
         UINT32 *nBytes);


#ifdef IAPX386
void BuildASMStatStrings(
				StatTableEntry			*tableEntry);
#endif


#define ETHERTSM_H
#endif /* ETHERTSM_H */

