/*++===========================================================================
//
// Program Name:	ODILOAD
//
// File Name:	msmstruc.h
//
// Version:	v0.00
//
// Author:	KRA
//
// Abstract:
//		Contains structures used specifically for this version of the cmsm.
//
// Notes:	
//
// Revision History:	
// 961023 KRA
// Modified PUNICODE_STRING *msgUc field to MEON_STRING *msg in MESSAGE_ARGS.
// MESSAGE_ARGS modifications fix the CMSMPringString/CMSMPrintWorker problems.
//
// 970224 KRA
// Removed dead code.
//
// 970505 KRA
// Added HSM_BASED_SHUTDOWN and made its value the same as
// HSM_PERMANENT_SHUTDOWN.  Removed HSM_PERMANENT_SHUTDOWN.
//
// 971218 KRA v1.36
// CMSMScheduleIntTimeCallBack and CMSMScheduleAesCallBack need to return
//	and ODISTAT value.
//
//	(C) Unpublished Work of Novell, Inc.  All Rights Reserved.
//
//	THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
//	PROPRIETARY AND TRADESECRET INFORMATION OF NOVELL, INC.  ACCESS
//	TO THIS WORK IS RESTRICTED TO (I) NOVELL EMPLOYEES WHO HAVE A
//	NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE OF THEIR ASSIGNMENTS
//	AND (II) ENTITIES OTHER THAN NOVELL WHO HAVE ENTERED INTO APPROPRIATE
//	LICENSE AGREEMENTS.  NO PART OF THIS WORK MAY BE USED, PRACTICED,
//	PERFORMED, COPIED, DISTRIBUTED, REVISED, MODIFIED, TRANSLATED,
//	ABBRIDGED, CONDENSED, EXPANDED, COLLECTED, COMPILED, LINKED, RECAST,
//	TRANSFORMED OR ADAPTED WITHOUT THE PRIOR WRITTEN CONSENT OF NOVELL.
//	ANY USE OR EXPLOITATION OF THIS WORK WITHOUT AUTHORIZATION COULD
//	SUBJECT THE PERPERTRATOR TO CRIMINAL AND CIVIL LIABILITY.
//=========================================================================--*/
#define	MAX_ECB_FRAGS					16
#define	NUM_ISR_DPCS					2
#define	BUS_NOT_DEFINED				0xffffffff

#define  HSM_BASED_SHUTDOWN   2
#define  HSM_TEMPORARY_SHUTDOWN   3

#define SPANS_PAGE_BOUNDRY(address, size)   \
      (UINT32)(((size - 1L) >> PAGE_SHIFT) + \
      (((UINT32)((size - 1)&(PAGE_SIZE - 1)) + \
      (UINT32)((UINT32)address & (PAGE_SIZE - 1))) >> PAGE_SHIFT))


typedef	struct	_MESSAGE_ARGS_
{
	NTSTATUS						event;
	MEON_STRING	   			*msg;
	UINT32						logType;
	UINT32						outputDestination;
	WORK_QUEUE_ITEM			*WorkItem;
}MESSAGE_ARGS;

struct	_ADAPTER_DATA_;

typedef	struct	_ISR_CONTEXT_
{
	void	(_cdecl * ISRRoutine)(struct _DRIVER_DATA_ *);
	void		                  *AdapterData;
	struct _DRIVER_DATA_			*DriverData;
	KDPC		                  *IsrDpc;
	UINT32	                  NextDpc;
}ISR_CONTEXT;


typedef	struct	_ADAPTER_DATA_
{
	struct _ADAPTER_DATA_		*InstanceLink;
	DRIVER_PARM_BLOCK		DPBlock;						  /*Needed OK*/
	TSM_PARM_BLOCK			MPBlock;						  /*Needed OK*/

	/* The following ECBs are here for Asyncronous call back of IOCTLs. */
	ECB						*GetStatsECB;
	ECB						*ShutDownECB;
	ECB						*ResetECB;
	ECB						*MulticastECB;
	ECB						*PromiscuousECB;
	ECB						*LookAheadSizeECB;

	AESECB					MSMPollECB;					  /*Needed OK*/
	UINT16					AES0Pad;
	UINT32					PollInterval;					/*Needed OK*/
	MLID_AES_ECB			MlidAES;
	MLID_AES_ECB			MlidIntAES;
	UINT32					HSMInC;							/*Needed OK*/
   LAN_ADAPTER_OBJECT	*LanObject;						/*Needed OK*/
   ISR_CONTEXT				Isr1Context;					/*Needed OK*/
	ISR_CONTEXT				Isr2Context;					/*Needed OK*/
	UINT32					MaxTxFreeCount;				/*Needed OK*/
	ECB						*TxECBList;						/*Needed OK*/
	ECB						*TxECBHead;						/*Needed OK*/
	UINT32					Bus;								/*Needed OK*/
	UINT32					BusType;
	EXTRA_CONFIG			*ExtraConfigList;				/*Needed OK*/
	MLID_AES_ECB			*AESList;

/* The following variables will be accessed by TSM's and HSM's and
   must remain in the following order.											*/

	volatile UINT32		PriorityTxFreeCount;
   UINT32               ReservedForCHSM[16];                    /* Single-Source TSM */
   UINT32               TCB_ECBptrAdjust;
	UINT32					AsmHSMUpdate;
	MSTATS					*AsmStatsPtr;
	DRIVER_PARM_BLOCK		*AsmParmBlock;					/* Single-Source TSM */
	UINT32					AdapterIntDisabled;			/* Single-Source TSM */
	UINT32					NeedContiguousECB;			/* Single-Source TSM */
	GENERIC_STATS			*GenericStatsPtr;				/* Single-Source TSM */
	void						*AdapterSpinLock;				/* Single-Source TSM */
	UINT32					MaxFrameHeaderSize;
	UINT8						PhysNodeAddress[ADDR_SIZE];
	UINT8						PhysNodeFill[2];
	void						*SendListHead;					/* Single-Source TSM */
	void						*SendListTail;					/* Single-Source TSM */
	DRIVER_PARM_BLOCK		*DPBlockPtr;						/* Single-Source TSM */
	void						*MediaDataPtr;					/* Single-Source TSM */
	CONFIG_TABLE			*DefaultVirtualBoard;
	UINT32					DriverInterrupt;
	UINT32					InCriticalSection;
	MLID_StatsTable		*StatsPtr;						/* Single-Source TSM */
	CONFIG_TABLE			*VirtualBoardLink[4];
	UINT32					StatusFlags;
	UINT32					TxFreeCount;
} ADAPTER_DATA;


/* This is a template to put over the top of an C DRIVER_PARM_BLOCK */
/* at the DPB_Reserved1 field. This makes easy access to the ASM fields. */
typedef struct _ASM_DRIVER_KEWYORD_STRUCT_
{
   UINT32                  DriverNumKeywords;
   MEON                    **DriverKeywordText;
   UINT32                  *DriverKeywordTextLen;
   UINT32                  (_cdecl ** DriverProcesskeywordTab)(
                              CONFIG_TABLE   *configTable,
                              PMEON          string,
                              UINT32         value);
} ASM_DRIVER_KEYWORD_STRUCT;

// Equates for Custom Keyword Types.
#define	T_REQUIRED				0x8000			// Keyword must be entered.
#define	T_STRING					0x0100			// String input(default).
#define	T_NUMBER					0x0200			// Dword Decimal input.
#define	T_HEX_NUMBER			0x0300			// Dword Hex input.
#define	T_HEX_STRING			0x0400			// 6-byte hex string input.
#define	T_LENGTH_MASK			0x00FF			//	Length Mask
#define  T_TYPE_MASK          0x7F00         // Type Mask.


/* Fake type definition to keep .h files happy. _DRIVER_DATA_ is defined	*/
/* in the HSM and not accessed directly by the MSM or TSM.						*/
typedef	struct	_DRIVER_DATA_
{
	void		*FakeData;
} DRIVER_DATA;

/* Fake type definition of MEDIA_DATA defined in TSM's, kept in		*/
/* adapter data space, but not touched by MSM.								*/
typedef	struct	_MEDIA_DATA_	{
	void		*FakeData;
} MEDIA_DATA, *PMEDIA_DATA;

typedef struct	RESOURCE_OBJECT_CONTEXT
{
	void					*pWorkItem;
	void					*pResObj;
	DRIVER_DATA			*pDriverData;
	MLID_CONFIG_TABLE	*pConfigTable;
	IO_CONFIG			*pIOConfig;
	EXTRA_CONFIG		*pExtraConfig;
	ECB					*pAsyncECB;
} RESOURCE_OBJECT_CONTEXT;

typedef struct	MLID_SHUTDOWN_CONTEXT
{
	void					*pWorkItem;
	UINT32				boardNumber;
} MLID_SHUTDOWN_CONTEXT;

typedef struct	CMSM_FREE_CONTEXT
{
	void					*pWorkItem;
	const DRIVER_DATA	*pDriverData;
	void					*pDataPtr;
} CMSM_FREE_CONTEXT;

typedef struct	LAN_OBJECT_CONTEXT
{
	void						*pWorkItem;
   LAN_ADAPTER_OBJECT	*pLanObject;
} LAN_OBJECT_CONTEXT;


void	_cdecl CMSMRegisterSharedMemory(CONFIG_TABLE	*configTable);

void	_cdecl CMSMDeRegisterSharedMemory(CONFIG_TABLE	*configTable);

void	*	_cdecl CMSMReadEISAConfig(
				UINT32   slot,
				UINT32	block);

ODISTAT  _cdecl CMSMScheduleAesCallBack(
				DRIVER_DATA	*driverData,
				UINT32		waitTime);

ODISTAT  _cdecl CMSMScheduleIntTimeCallBack(
				DRIVER_DATA	*driverData,
				UINT32		waitTime);

ECB	*  _cdecl LSLGetSizedEcb(UINT32 EcbDataSize, BOOLEAN below16);

UINT32   _cdecl LSLReturnEcb(ECB *ecb);

UINT32   _cdecl LSLCancelEvent(ECB *ecb);

void  _cdecl LSLScheduleAesEvent(AESECB *aesEcb);

UINT32   _cdecl LSLCancelAesEvent(AESECB *aesEcb);

UINT32   _cdecl LSLGetIntervalMarker();


UINT32  _cdecl LSLGetStackEcb(LOOKAHEAD	*LookAheadBuf);

void	_cdecl LSLFastSendComplete(ECB *ecb);

void	_cdecl LSLSendComplete(ECB *ecb);

void	_cdecl LSLHoldEvent(ECB *ecb);

void	_cdecl LSLFastHoldEvent(ECB	*ecb);

UINT32   _cdecl LSLGetMaxEcbBufferSize();

UINT32   _cdecl LSLControlStackFilter(
            UINT32   BoardNumber,
            UINT32   Function,
            UINT32   Mask,
            PVOID    Parameter1,
            PVOID    Parameter2);
