/*--------------------------------------------------------------------------*
 *  Copyright Unpublished Work of Novell, Inc. All Rights Reserved
 *
 *  THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 *  PROPRIETARY AND TRADE SECRET INFORMATION OF NOVELL, INC.
 *  ACCESS TO THIS WORK IS RESTRICTED TO (I) NOVELL EMPLOYEES
 *  WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE
 *  OF THEIR ASSIGNMENTS AND (ii) ENTITIES OTHER THAN NOVELL
 *  WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.
 *  NO PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED,
 *  DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED,
 *  CONDENSED, EXPANDED, COLLECTED, COMPILED, LINKED, RECAST,
 *  TRANSFORMED OR ADAPTED WITHOUT THE PRIOR WRITTEN CONSENT OF
 *  NOVELL.  ANY USE OR EXPLOITATION OF THIS WORK WITHOUT AUTHORIZATION
 *  COULD SUBJECT THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY.
 *--------------------------------------------------------------------------*/
/*****************************************************************************
 *
 * Program Name:	C ODI Header File
 *
 * Filename:		ODI.H
 *
 * ODI Spec Ver:	1.11
 *
 * Description:		This file is the main source for the C ODI SPECIFICATION.
 *					Any structures needed by the MLI or MPI interface are defined here.
 *
 * Modification History:
 *							
 * 08-27-96 WTT		Added ODI_SPEC_MAJOR_VER, and ODI_SPEC_MINOR_VER.
 *
 * 09-10-96 MPK		Added prototype for CLSL_GetMultipleECBs
 *
 * 09-13-96 WTT		Added Shutdown type defines
 *
 * 09-23-96 KRA		Added #defines for UNUSED_IO_PORT, UNUSED_MEMORY_ADDRESS,
 *					UNUSED_INTERRUPT, and UNUSED_DMA_LINE for use in setting
 *					MLID_CONFIG_TABLE resource fields to not in use.
 *
 * 09-25-96 WTT		Added #define for PROM_MODE_RMC.
 *
 * 10-02-96 KRA		Changed CALLCNV to _cdecl for the ESRs in the ECB and
 *					AES_ECB structures.
 *
 * 10-15-96 KRA		Added #defines for UNUSED_IO_RANGE, UNUSED_MEMORY_SIZE,
 *					UNUSED_SLOT.
 *
 * 10-17-96 WTT		Added #defines for MLIDCntl_RemoveNetworkInterface,
 *					MLIDCntl_ShutdownNetworkInterface, and
 *					MLIDCntl_ResetNetworkInterface.
 *
 * 11-15-96 KRA		Set OPERATION_SCOPE values to begin at default of 0
 *					rather than at 1 to preserve the CSPEC110 driver
 *					functionality.
 *
 * 11-20-96 WTT		Added defines for indexes into the LSL entry point array.
 *
 * 11-27-96 WTT		Added Macro definitions for the LSL APIs.
 *
 * 12-09-96 KRA		Added #define UNASSIGNED_BOARD_NUMBER 0xFFFF to check if
 *					the board number has been assigned by the LSL.
 *
 * 12-10-96 WTT		Added to conditional declarations for typedefs a define for
 *					a preprocessor symbol as itself with the same name as the
 *					typedef. The purpose behind this is to protect the typedef
 *					against multiple declarations. The definition of a
 *					preprocessor definition as itself is allowed by the ANSI
 *					language specification.
 *					SPD #144346
 *
 * 01-22-97 WTT		Change prototype for CLSL_ScheduleAESEvent to match the
 *					1.11 specification.  The return type of void is now ODISTAT.
 *
 * 02-12-97 KRA		Added ifdef OS_WIN95 to the ifdef OS_NT for packing
 *					structures.  Added '((' to the CLSLEntry_GetLSLConfiguration
 *					and CLSLEntry_ReSubmitDefault macros so they compile.
 *
 * 07-22-97 WTT		The definition of MS_MEM_PAGE_BIT was not to spec.  It has
 *					been changed from 0x0800 to 0x8000 to match the specification.
 *
 * 08-29-97 WTT		Added defines for NON_STACKID_BIT and NON_STACKID_RAW_SEND_MASK.
 *
 * 09-10-97 WTT		The following changes were made to sync up the header file
 *					with the spec.
 *
 *					Added defines for PSTK_GET_BOUND_NET_INFO and
 *					PStkCntl_GetBoundNetInfo and changed define PSTK_NUM_API
 *					from 9L to 10L.
 *
 *					Added NETWORK_ADDRESS_INFO structure.
 *
 *					Removed define	PStkCntl_SFTIIIExchange and changed define
 *					PSTK_SFTIIIEXCHANGE to PSTK_RESERVED.  These defines were
 *					never used.
 *
 *	09-22-97 AYD	Changed declaration of CLSL_ReturnECB to _cdecl.
 *
 * 10-01-97 KRA		Changed CLSLEntry_ReturnECB to be _cdecl.
 *
 * 10-06-97 WTT		Added prototype for CLSL_GetPhysicalAddressOfECB.  Define
 *					CLSL_GET_PHYSICAL_ADDRESS_OF_ECB and
 *					CLSLEntry_GetPhysicalAddressOfECB
 *
 * 05-15-00 mcd     Added PA_TRANS_PROT_CHKSUM, PA_NET_PROT_CHKSUM, and PA_RX_AUX_DATA
 *
 * 06-02-00 mcd     Added Auxiliary Data Block structure definitions: AUX_DATA_BLOCK_HDR,
 *                  AUX_DATA_START_BLOCK, AUX_DATA_END_BLOCK, and AUX_DATA_NULL_BLOCK
 *                  and various ancillary values.  Also changed the spec minor version
 *                  number.      
 *
 ****************************************************************************/

#ifndef	_ODI_Include_
#define _ODI_Include_

/* C ODI Specification Version Numbers */

#define	ODI_SPEC_MAJOR_VER	1
#define ODI_SPEC_MINOR_VER	20

/*State Definitions  TRUE / FALSE*/

#ifndef OS_NT
#ifndef OS_WIN95
#ifndef _cdecl
#define	_cdecl		
#endif
#endif
#endif
#ifdef OS_WIN95
#define CALLCNV _cdecl
#else
#define CALLCNV
#endif

#ifdef	TRUE
#undef	TRUE
#endif

#ifdef	FALSE
#undef	FALSE
#endif

#ifdef	UNUSED
#undef	UNUSED
#endif

#ifdef	BOOLEAN
#undef	BOOLEAN
#endif

#define	FALSE		0x0
#define	TRUE 		0x1
#define	UNUSED		-1
#define	BOOLEAN		unsigned char


/********************************************************************
	C ODI Type Definitions

	The following are typedef definitions for parameters used by the
	ANSI C ODI Interface. The conditional declarations define both a
	preprocessor symbol as itself and a typedef	of the same name.
	The purpose behind this is to protect the typedef against multiple
	declarations. The definition of a preprocessor definition as
	itself is allowed by the ANSI language specification.
*********************************************************************/

#ifndef	MEON
#define	MEON MEON
typedef	unsigned char	MEON;
#endif

/* Definition for MEON Strings, NB. MEON_STRING is really used as mnemonic. */

#ifndef	MEON_STRING
#define	MEON_STRING MEON_STRING
typedef	unsigned char	MEON_STRING;	/* by convention MEON_STRINGS NULL terminated */
#endif


#ifndef	UINT8
#define	UINT8 UINT8
typedef	unsigned char	UINT8;
#endif

#ifndef	UINT16
#define	UINT16 UINT16
typedef	unsigned short	UINT16;
#endif


#ifndef	UINT32
#define UINT32 UINT32
typedef	unsigned int	UINT32;
#endif

#ifndef	UINT64
#define	UINT64 UINT64
typedef	struct	_UINT64_ {

		UINT32	Low_UINT32;
		UINT32	High_UINT32;
		} UINT64;
#endif

#ifndef	VOID
#define VOID VOID
typedef	void	VOID;
#endif


/* declare the pointer for the ODI definitions */

#ifndef	PMEON
#define	PMEON PMEON
typedef	MEON	*PMEON;
#endif

#ifndef	PUINT8
#define	PUINT8 PUINT8
typedef	UINT8	*PUINT8;
#endif

#ifndef	PUINT16
#define	PUINT16 PUINT16
typedef	UINT16	*PUINT16;
#endif

#ifndef	PUINT32
#define	PUINT32 PUINT32
typedef	UINT32	*PUINT32;
#endif

#ifndef	PUINT64
#define	PUINT64 PUINT64
typedef	UINT64	*PUINT64;
#endif

#ifndef	PVOID
#define	PVOID PVOID
typedef	VOID	*PVOID;
#endif


/* Well Defined Sizes */

#define	PID_SIZE 	0x06			/* Number of Octets in Protocol Identifier */
#define	ADDR_SIZE	0x06			/* Number of Octets in Address */

/* define assumed Maximum Media Header Size that we'll encounter 		*/
/*																 		*/
/* assume that it will be Token-Ring (with SRT)							*/
/*																		*/
/* AC, FC, Dest[6], Source[6], SRFields[30], 802.2UI[3], SNAP[5] = 52	*/
/*  																	*/
#define	MAX_MEDIA_HEADER_SIZE	52L

#define	DefaultNumECBs	0x00
#define	DefaultECBSize	1518L							/* not including ECB Structure */
#define	MinECBSize		(512+74+MAX_MEDIA_HEADER_SIZE)L	/* Max. ECB size < 64K */

#define	MAXLOOKAHEADSIZE		128L			/* Max. LookAhead Data Size */

#define	MAXSTACKNAMELENGTH	65L
#define	MAXNAMELENGTH			64L


#define MAXMULTICASTS	16L		/* MAXIMUM number of Multicast addresses */
#define MAX_FRAG_COUNT	16L   	/* MAXIMUM number of Fragment Count */

typedef struct _PROT_ID_
{
	UINT8	protocolID[PID_SIZE];
}
PROT_ID;

typedef struct	_NODE_ADDR_
{
	UINT8	nodeAddress[ADDR_SIZE];
}
NODE_ADDR;


/* Chained Protocol Stack position values */

typedef enum	_CHNPOS_
{
	CHNPOS_FIRST_MUST = 0,	/* Load at very first position  */
	CHNPOS_FIRST_NEXT = 1,	/* Load at next front position  */
	CHNPOS_LOAD_ORDER = 2,	/* Load dependent on load order */
	CHNPOS_LAST_NEXT =  3,	/* Load next end position       */
	CHNPOS_LAST_MUST =  4	/* Load at very end of chain    */
} CHNPOS;


#define	CHNPOS_MAX_POSIT		0x0004	/* Maximum possible Chain position */

#define	CHAINTYPE_TX			0x0000	/* Transmit type Chain Protocol Stack */
#define	CHAINTYPE_RX			0x0001	/* Receive type Chain Protocol Stack */


/*	SFT III Status values	*/

typedef	enum	_SFTIII_STAT_	{

	SFTIII_STAT_SUCCESSFUL			= 0,
	SFTIII_STAT_MIRROR_NOT_ACTIVE	= 1,
	SFTIII_STAT_NO_PARTNER 			= 2,
	SFTIII_STAT_OUT_OF_RESOURCES	= 3,
	SFTIII_STAT_NOT_SUPPORTED		= -1

}	SFTIII_STAT;


/* Look Ahead Definitions */
/* MLIDCFG_LookAheadSize Values*/

#define	DEFAULT_LOOK_AHEAD_SIZE	0x12	/* Default of 18 look ahead bytes	*/
#define	MAX_LOOK_AHEAD_SIZE		0x80	/* Maximum 128 look ahead bytes		*/

/* Rx Packet Attributes	(ie. LkAhd_PktAttr) */
			
#define	PAE_CRC_BIT          0x00000001	/* CRC Error */
#define	PAE_CRC_ALIGN_BIT 	 0x00000002	/* CRC/Frame Alignment Error */
#define	PAE_RUNT_PACKET_BIT	 0x00000004	/* Runt Packet */
#define	PAE_TOO_BIG_BIT		 0x00000010	/* Packet Too Large for Media */
#define	PAE_NOT_ENABLED_BIT	 0x00000020	/* Unsupported Frame */
#define	PAE_MALFORMED_BIT    0x00000040	/* Malformed Packet */
#define	PA_NO_COMPRESS_BIT	 0x00004000	/* Do not compress received packet	*/
#define	PA_NONCAN_ADDR_BIT	 0x00008000	/* Set if Addr. in Immediate Address field is noncanonical */
#define PA_TRANS_PROT_CHKSUM 0x00010000
#define PA_NET_PROT_CHKSUM   0x00020000
#define	PA_RX_AUX_DATA       0x00800000	/* Set if received auxiliary data is present */

#define	PAE_ERROR_MASK		(PAE_CRC_BIT | PAE_CRC_ALIGN_BIT | PAE_RUNT_PACKET_BIT | \
									 PAE_TOO_BIG_BIT | PAE_NOT_ENABLED_BIT | PAE_MALFORMED_BIT)


/* Rx Packet Destination Address Types	(LkAhd_DestType, Modify Stack Filter) */
/* Low order 16 bits of the Rx Packet Destination Address Types are copied in ECB_DriverWorkspace.DWs_i16val[0] */

#define	DT_MULTICAST	   	0x00000001	/* Multicast Dest. address (Group Address) */
#define	DT_BROADCAST	   	0x00000002	/* Broadcast Dest. address */
#define	DT_REMOTE_UNICAST		0x00000004	/* Remote Unicast Dest. address */
#define	DT_REMOTE_MULTICAST	0x00000008	/* Unsupported Multicast address */
#define	DT_SOURCE_ROUTE		0x00000010	/* Source Routed packet	*/
#define	DT_ERRORED		 		0x00000020	/* Global Error, exculsive bit. */
#define	DT_MAC_FRAME	 		0x00000040	/*MAC/SMT frames. (ie. NON-DATA Frame)	*/
#define	DT_DIRECT		 		0x00000080	/* Unicast for this workstation */

#define	DT_8022_TYPE_I	 		0x00000100	/*	Set if packet is 802.2 Type I  */
#define	DT_8022_TYPE_II		0x00000200	/*	Set if packet is 802.2 Type II */
#define	DT_8022_BYTES_BITS	0x00000300
#define	DT_RX_PRIORITY			0x00000400	/* Set if packet has priority value other than base values	*/
#define	DT_PROMISCUOUS	 	(DT_ERRORED | DT_DIRECT | DT_MULTICAST | DT_BROADCAST |\
									 DT_REMOTE_UNICAST | DT_REMOTE_MULTICAST | DT_SOURCE_ROUTE |\
									 DT_MAC_FRAME | DT_RX_PRIORITY)

/* mask off allowable Destination Address Types */

#define	DT_MASK					(DT_MULTICAST | DT_BROADCAST | DT_REMOTE_UNICAST |\
										 DT_REMOTE_MULTICAST | DT_SOURCE_ROUTE | DT_ERRORED |\
										 DT_MAC_FRAME | DT_DIRECT | DT_RX_PRIORITY)


/* ECB Definitions */

/* Stack ID Definitions */
/* An explaination of the encoding of StackID on transmit */

/* When the ECB is prepared by the Protocol to send a packet, the Protocol typically places 
   its Protocol Stack ID that was assigned by the LSL, when the protocol stack registered, 
   in the ECB_StackID field. The ECB_StackID field of the ECB may hold the Protocol stack ID 
   given when the stack registered with the LSL, or a special set of values to communicate 
   raw sends, priority packet transmission, checksumming generation, or auxiliary data.

	The following values are valid for the ECB_StackID field.

0x0000-0x00FF (0-255)	Protocol Identification Number assigned by the LSL 


0xznFy	The ECB_StackID field may contain information concerning the transmit request as 
        defined by the "z", "n" and "y" fields defined below.
 
0xFy 	Priority Transmission Support. ( See Priority Tx Support, page 5-25, ODI 
        Specification: C HSM ODI Spec, and ECB_StackID definition, ODI Specification: 
        Protocol Stacks and MLIDS (C Language) on page A-5) The Fy values are the same 
        as the last byte of the Priority Send values.

Raw Sends

The Raw Send designation indicates that the Protocol has generated the full MAC header. 
The MLID will not change the MAC headers on a Raw Send. Raw Send packets may be checksummed.
Fy values for raw sends are defined below.

0xFFFy	Values from 0xFFF0 - 0xFFFF (zn = 0xFF) indicate legacy raw and normal send 
        definitions for the ECB_StackID.


0xzn 	The z and n nibbles of the zn byte provide additional information about the transmit
        request. The z and n nibbles are defined as follows:

0xz	    The z nibble of the zn byte has a pattern of 10xx. The following values are defined for the z:

TX_NO_CHECKSUM	        0x8  No Checksum generation on this packet.
TX_TRANSPORT_CHECKSUM	0x9  Generate the Transport Layer Checksum (TCP, UDP, ICMP, RSVP)
TX_NETWORK_CHECKSUM	    0xA  Generate the Network Layer Checksum (Ipv4)

0xn 	The n nibble of the zn byte indicates a value of 0x0 - 0xF. The following values are defined for the n nibble of the zn byte:

Reserved        0x0         Reserved 
TX_AUX_DATA     0x1         Auxiliary Data is present
Reserved        0x2 - 0xF   Reserved
*/


#define	ECB_RAWSEND		 	0xFFFF	/* Raw Send, ie. ECB includes MAC Header */
									/*				implies MLID should not */
									/*				build MAC Header for this Tx.*/

#define	TX_RAW_SEND_PRIORITY_0	0xFF   /* 0 = No (lowest) Priority */
#define	TX_RAW_SEND_PRIORITY_1  0xFE   /* 1 = Low Priority */
#define	TX_RAW_SEND_PRIORITY_2	0xFD	
#define	TX_RAW_SEND_PRIORITY_3	0xFC
#define	TX_RAW_SEND_PRIORITY_4	0xFB
#define	TX_RAW_SEND_PRIORITY_5	0xFA
#define	TX_RAW_SEND_PRIORITY_6	0xF9
#define	TX_RAW_SEND_PRIORITY_7	0xF8   /* 7 = Highest Priority */

#define	TX_SEND_PRIORITY_0		0xF7	/* 0 No Priority */
#define	TX_SEND_PRIORITY_1		0xF6	/* 1 low priority */
#define	TX_SEND_PRIORITY_2		0xF5
#define	TX_SEND_PRIORITY_3		0xF4
#define	TX_SEND_PRIORITY_4		0xF3
#define	TX_SEND_PRIORITY_5		0xF2
#define	TX_SEND_PRIORITY_6		0xF1
#define	TX_SEND_PRIORITY_7		0xF0	/* 7  Highest Priority. */

#define	RAW_SEND		0xFFFF	/* If in ECB.ECB_StackID raw send (Old Def.) */
#define	RAW_SEND_PRIORITY_0	0xFFFF	/* Same as old Def. No Priority */
#define	RAW_SEND_PRIORITY_1  0xFFFE	/* Scale 1-7; 1 being lowest */
#define	RAW_SEND_PRIORITY_2	0xFFFD	
#define	RAW_SEND_PRIORITY_3	0xFFFC
#define	RAW_SEND_PRIORITY_4	0xFFFB
#define	RAW_SEND_PRIORITY_5	0xFFFA
#define	RAW_SEND_PRIORITY_6	0xFFF9
#define	RAW_SEND_PRIORITY_7	0xFFF8	/* Scale 1-7 7 being Highest Priority */
#define	SEND_PRIORITY_0		0xFFF7	/* Scale 1-7 0 No Priority */
#define	SEND_PRIORITY_1		0xFFF6	/* Scale 1-7 1 being low priority */
#define	SEND_PRIORITY_2		0xFFF5
#define	SEND_PRIORITY_3		0xFFF4
#define	SEND_PRIORITY_4		0xFFF3
#define	SEND_PRIORITY_5		0xFFF2
#define	SEND_PRIORITY_6		0xFFF1
#define	SEND_PRIORITY_7		0xFFF0	/* Scale 1-7 7 being Highest Priority. */

/* Receive ECB_DriverWorkspace.DWs_i16val[0] */
													
#define	ECB_MULTICAST			0x0001	/* Multicast Dest. address (Group Address) */
#define	ECB_BROADCAST			0x0002	/* Broadcast Dest. address */
#define	ECB_UNICASTREMOTE		0x0004	/* Remote Unicast Dest. address */
#define	ECB_MULTICASTREMOTE		0x0008	/* Unsupported Multicast address */
#define	ECB_SOURCE_ROUTE		0x0010	/* Source Routed packet	*/
#define  ECB_GLOBALERROR    	0x0020 	/* Set if packet contains errors (exculsive)	*/
														/* NB. If set all other bits should be reset.*/
#define  ECB_MACFRAME       	0x0040 	/* Packet is not a data packet. */
														/* NB. If set all other bits should be reset. */
#define	ECB_UNICASTDIRECT		0x0080	/* Unicast for this workstation */

#define	ECB_MASK				0xFFFF	/* mask off allowable Destination Address Types */

#define	ECB_TYPE_I				0x0100	/* Set if packet is 802.2 Type I  */
#define	ECB_TYPE_II				0x0200	/* Set if packet is 802.2 Type II */
#define	ECB_RX_PRIORITY			0x0400	/* Set if packet has priority value other than base values	*/
#define	ECB_PROMISCUOUS		(ECB_ERRORED | ECB_MULTICAST | ECB_BROADCAST | ECB_UNICASTREMOTE |\
										 ECB_MULTICASTREMOTE | ECB_SOURCE_ROUTE | ECB_MAC_FRAME | ECB_RX_PRIORITY)

/* PromiscuousChange state and mode values.*/
#define	PROM_STATE_OFF			0x00		/* Disable Promiscuous Mode	*/
#define	PROM_STATE_ON			0x01		/* Enable Promiscuous Mode 	*/

#define	PROM_MODE_QUERY      0x00	    /* Query as to promiscuous mode	*/
#define	PROM_MODE_MAC			0x01		/* MAC frames			  		*/
#define	PROM_MODE_NON_MAC		0x02		/* Non-MAC frames		  		*/
#define	PROM_MODE_MACANDNON	0x03		/* Both MAC and Non-MAC frames	*/
#define  PROM_MODE_SMT			0x04		/* FDDI SMT Type MAC frames. */
#define	PROM_MODE_RMC			0x08	/* Remote Multicast frames */

/* Operation Scope Definitions */

typedef  enum  _OPERATION_SCOPE_
{
	OP_SCOPE_ADAPTER       = 0,
	OP_SCOPE_LOGICAL_BOARD = 1
} OPERATION_SCOPE;


/* System Return Code Definitions */

typedef  enum  _ODISTAT_
{
   ODISTAT_SUCCESSFUL				= 0,
   ODISTAT_RESPONSE_DELAYED		= 1,
   ODISTAT_SUCCESS_TAKEN			= 2,
   ODISTAT_BAD_COMMAND				= -127,
   ODISTAT_BAD_PARAMETER			= -126,
   ODISTAT_DUPLICATE_ENTRY			= -125,
   ODISTAT_FAIL						= -124,
   ODISTAT_ITEM_NOT_PRESENT		= -123,
   ODISTAT_NO_MORE_ITEMS			= -122,
   ODISTAT_MLID_SHUTDOWN			= -121,
   ODISTAT_NO_SUCH_HANDLER			= -120,
   ODISTAT_OUT_OF_RESOURCES		= -119,
   ODISTAT_RX_OVERFLOW				= -118,
   ODISTAT_IN_CRITICAL_SECTION	= -117,
   ODISTAT_TRANSMIT_FAILED			= -116,
   ODISTAT_PACKET_UNDELIVERABLE	= -115,
   ODISTAT_CANCELED					= -4
}ODISTAT;

#define	ODISTAT_NO_SUCH_DRIVER	ODISTAT_MLID_SHUTDOWN


/* MLID Configuration Table Bit Defintions. */

/* MLID 'Flags' Bit Definitions. */
#define	MF_HUB_MANAGEMENT_BIT	0x0100
#define	MF_SOFT_FILT_GRP_BIT		0x0200
#define	MF_GRP_ADDR_SUP_BIT		0x0400
#define	MF_MULTICAST_TYPE_BITS	0x0600
#define  MF_RECONFIG_BIT			0x0800
#define	MF_PRIORITYSUP_BIT		0x1000

/* MLID 'ModeFlags' Bit Definitions. */
#define	MM_REAL_DRV_BIT			0x0001
#define	MM_USES_DMA_BIT			0x0002
#define	MM_DEPENDABLE_BIT			0x0004	/*	Should only be set if MM_POINT_TO_POINT_BIT */
														/*	set, for hardware that is normally          */
														/*	dependable but is not 100% guaranteed	    */
#define	MM_MULTICAST_BIT			0x0008

/*#redef'ed define	MM_POINT_TO_POINT_BIT	0x0010*/	/* Set if point-to-point link, dynamic */
														/* call setup and tear down, eg. X.25  */
#define	MM_CSL_COMPLIANT_BIT		0x0010	/* Set if MLID is CSL compliant	*/
#define	MM_PREFILLED_ECB_BIT		0x0020	/* MLID supplies pre-filled ECBs */
#define	MM_RAW_SENDS_BIT			0x0040
#define	MM_DATA_SZ_UNKNOWN_BIT	0x0080
#define	MM_SMP_BIT					0x0100	/* Set if MLID is SMP enabled.	*/
#define	MM_FRAG_RECEIVES_BIT		0x0400	/* MLID can handle Fragmented Receive ECB. */
#define	MM_C_HSM_BIT				0x0800	/* Set if HSM written in C. */
#define	MM_FRAGS_PHYS_BIT			0x1000	/* Set if HSM wants Frags with Physical Addresses. */
#define	MM_PROMISCUOUS_BIT		0x2000	/* Set if supports Promiscuous Mode. */
#define	MM_NONCANONICAL_BIT		0x4000	/* Set if Config Node Address Non-Canonical. */
#define	MM_PHYS_NODE_ADDR_BIT	0x8000	/* Set if MLID utilizes Physical Node Address. */
#define	MM_CANONICAL_BITS			0xC000


/* MLID 'SharingFlags' Bit Defintions */
#define	MS_SHUTDOWN_BIT			0x0001
#define	MS_SHARE_PORT0_BIT		0x0002
#define	MS_SHARE_PORT1_BIT		0x0004
#define	MS_SHARE_MEMORY0_BIT 	0x0008
#define	MS_SHARE_MEMORY1_BIT	0x0010
#define	MS_SHARE_IRQ0_BIT		0x0020
#define	MS_SHARE_IRQ1_BIT		0x0040
#define	MS_SHARE_DMA0_BIT		0x0080
#define	MS_SHARE_DMA1_BIT		0x0100
#define	MS_HAS_CMD_INFO_BIT		0x0200
#define	MS_NO_DEFAULT_INFO_BIT	0x0400
#define MS_MEM_PAGE_BIT			0x8000

/* MLID 'LineSpeed' Bit Definitions. */
#define	MLS_MASK						0x7FFF
#define	MLS_KILO_IND_BIT			0x8000

/* MLID unused resource definitions. */
#define UNUSED_SLOT				0xFFFF
#define UNUSED_IO_PORT			0
#define UNUSED_IO_RANGE			0
#define UNUSED_MEMORY_ADDRESS 0
#define UNUSED_MEMORY_SIZE		0
#define UNUSED_INTERRUPT		0xFF
#define UNUSED_DMA_LINE			0xFF

#define UNASSIGNED_BOARD_NUMBER 0xFFFF


/* STAT_TABLE_ENTRY Definitons. */

#define	ODI_STAT_UNUSED			0xFFFFFFFF	/* Statistics Table Entry not in use.	*/
#define	ODI_STAT_UINT32			0x00000000	/* Statistics Table Entry UINT32 Counter. */
#define	ODI_STAT_UINT64			0x00000001	/* Statistics Table Entry UINT64 Counter. */
#define	ODI_STAT_MEON_STRING	0x00000002	/* Statistics Table Entry Counter is a MEON_STRING. */
#define	ODI_STAT_UNTYPED		0x00000003	/* Statistics Table Entry Counter is a
												UINT32 length preceded array UINT8	*/
#define	ODI_STAT_RESETABLE		0x80000000	/* Statistics Table Entry Counter is resetable
												by external entity.  This define is only used
												by management agents and other modules that
												are using STAT_TABLE definintions from this
												header file.  ODI_STAT_RESETABLE is not a
												part of the ODI Specification and is not allowed
												for MLID stats counters.  ODI considers
												0x80000000 a reserved bit to avoid conflicts with
												other modules in the system. */


#if defined	OS_NT || defined OS_WIN95
/* Set PRAGMA to pack these structures */

#pragma	pack(1)

#endif

typedef	struct _STAT_TABLE_ENTRY_	
{
	UINT32			StatUseFlag;
	void				*StatCounter;
	MEON_STRING		*StatString;
}
StatTableEntry, *PStatTableEntry, STAT_TABLE_ENTRY;

/*    StatUseFlag	Determines how StatCounter is defined. 								*/
/*		*StatCounter	pointer as defined by the StatUseFlag.								*/
/*		*StatCounter	pointer to a UINT32 or UINT64 counter.								*/
/*		*StatString		pointer to a MEON String, describing the statistics counter.*/


/* Definitions for Information Block for passing API's, eg. Function Lists */

typedef	struct _INFO_BLOCK_
{
	UINT32	NumberOfAPIs;
	void		(**SupportAPIArray)();
}
INFO_BLOCK, *PINFO_BLOCK;

/* Definitions for Link Support Layer (LSL) */

typedef	struct _LOG_BRD_STAT_TABLE_ENTRY_
{
	UINT32	LogBrd_TransmittedPackets;
	UINT32	LogBrd_ReceivedPackets;
	UINT32	LogBrd_UnclaimedPackets;
	UINT32	LogBrd_TxOverloaded;
}
LogBrdStatTableEntry, *PLogBrdStatTableEntry, LOG_BRD_STAT_TABLE_ENTRY;

/********************************************************************
		C LSL Configuration Table Definitions.
*********************************************************************/

#define CLSL_CFG_TABLE_MAJOR_VER 2
#define CLSL_CFG_TABLE_MINOR_VER 1

/********************************************************************
		C LSL Configuration Table System Flags definitions.
*********************************************************************/

#define	CLSL_CFG_SERVER_BIT	0x40000000
#define	CLSL_CFG_CLIENT_BIT	0x80000000

/********************************************************************
		C LSL Configuration Table Structure definition.
*********************************************************************/

typedef	struct _LSL_CONFIG_TABLE_
{
	UINT16			LConfigTableMajorVer;
	UINT16			LConfigTableMinorVer;
	MEON_STRING		*LSLLongName;
	MEON_STRING		*LSLShortName;
	UINT16			LSLMajorVer;
	UINT16			LSLMinorVer;
	UINT32			LMaxNumberOfBoards;
	UINT32			LMaxNumberOfStacks;
	UINT32			LConfigTableReserved0;
	UINT32			LConfigTableReserved1;
	UINT32			LConfigTableReserved2;
	UINT8			LSLCFG_ODISpecMajorVer;
	UINT8			LSLCFG_ODISpecMinorVer;
	UINT16			LConfigTableReserved3;
	UINT32			LSLCFG_SystemFlags;
	UINT32			LSLCFG_SmallECBCount;
	UINT32			LSLCFG_MediumECBCount;
	UINT32			LSLCFG_LargeECBCount;
	UINT32			LSLCFG_XLargeECBCount;
	UINT32			LSLCFG_HugeECBCount;
	UINT32			LSLCFG_SmallECBBelow16Count;
	UINT32			LSLCFG_MediumECBBelow16Count;
	UINT32			LSLCFG_LargeECBBelow16Count;
	UINT32			LSLCFG_XLargeECBBelow16Count;
	UINT32			LSLCFG_HugeECBBelow16Count;
	UINT32			LSLCFG_SmallECBMinCount;
	UINT32			LSLCFG_MediumECBMinCount;
	UINT32			LSLCFG_LargeECBMinCount;
	UINT32			LSLCFG_XLargeECBMinCount;
	UINT32			LSLCFG_HugeECBMinCount;
	UINT32			LSLCFG_SmallECBMaxCount;
	UINT32			LSLCFG_MediumECBMaxCount;
	UINT32			LSLCFG_LargeECBMaxCount;
	UINT32			LSLCFG_XLargeECBMaxCount;
	UINT32			LSLCFG_HugeECBMaxCount;
	UINT32			LSLCFG_SmallECBSize;
	UINT32			LSLCFG_MediumECBSize;
	UINT32			LSLCFG_LargeECBSize;
	UINT32			LSLCFG_XLargeECBSize;
	UINT32			LSLCFG_HugeECBSize;
}
LSL_ConfigTable, *PLSL_ConfigTable, LSL_CONFIG_TABLE;

/********************************************************************
		C LSL Statistics Table Structure definition.
*********************************************************************/

typedef	struct _LSL_STATS_TABLE_
{
	UINT16					LStatTableMajorVer;
	UINT16					LStatTableMinorVer;
	UINT32					LNumGenericCounters;
	STAT_TABLE_ENTRY			(*LGenericCountersPtr)[];
	UINT32					LNumLogicalBoards;
	LOG_BRD_STAT_TABLE_ENTRY	(*LogicalBoardStatTablePtr)[];
	UINT32					LNumCustomCounters;
	STAT_TABLE_ENTRY			(*LCustomCountersPtr)[];
}
LSL_StatsTable, *PLSL_StatsTable, LSL_STATS_TABLE;


#define	NUM_GENERIC_LSL_COUNTERS		10

#define	LSL_TOTAL_TX_PACKET_COUNT		0
#define	LSL_GET_ECB_REQUESTS			1
#define	LSL_GET_ECB_FAILURES			2
#define	LSL_AES_EVENTS_COUNT			3
#define	LSL_POSTPONED_EVENTS			4
#define	LSL_CANCEL_EVENT_FAILURES		5
#define	LSL_VALID_BUFFERS_REUSED		6
#define	LSL_RESERVED					7
#define	LSL_TOTAL_RX_PACKETS			8
#define	LSL_UNCLAIMED_PACKETS			9

/* Definitions for LookAhead and Event Control Blocks (ECB). */

typedef	struct	_FRAGMENT_STRUCT_	
{
	void	*FragmentAddress;
	UINT32	FragmentLength;
}
FRAGMENTSTRUCT, FRAGMENT_STRUCT, *PFRAGMENTSTRUCT;


/********************************************************************

   With compilers that implement strict alignement the location of
   were padding is assumed in the structure has been indicated below.

*********************************************************************/
typedef	struct _ECB_	
{
	struct	_ECB_	*ECB_NextLink;
	struct	_ECB_	*ECB_PreviousLink;
	UINT16			ECB_Status;
/*	UINT8			ECB_Pad1[2];*/							/* Compiler padding */
  	void			(_cdecl *ECB_ESR)(struct _ECB_ *);
  	UINT16			ECB_StackID;
  	PROT_ID			ECB_ProtocolID;
  	UINT32			ECB_BoardNumber;
  	NODE_ADDR		ECB_ImmediateAddress;
/*	UINT8			ECB_Pad2[2];*/							/* Compiler padding */
  	union			{
  		 				UINT8	DWs_i8val[4];
  		 				UINT16	DWs_i16val[2];
  		 				UINT32	DWs_i32val;
		 				void	*DWs_pval;
		 			} ECB_DriverWorkspace;
							
	union 			{
		  				UINT8	PWs_i8val[8];
		  				UINT16	PWs_i16val[4];
		  				UINT32	PWs_i32val[2];
		  				UINT64	PWs_i64val;
		  				void	*PWs_pval[2];
		  			} ECB_ProtocolWorkspace;
		UINT32			ECB_DataLength;
		UINT32			ECB_FragmentCount;
		FRAGMENT_STRUCT	ECB_Fragment[1];
}
ECB, *PECB;

#define  MAX_ECB_FRAGS	16

/********************************************************************
		ECB Extension Auxiliary Data Block Structures
*********************************************************************/

#define TX_AUX_DATA 0x1                /* Auxiliary Data is present */

#define	AUX_DATA_BLOCK_HDR_VERSION	1

/* Auxiliary Data Block Types */

#define	AUX_DATA_START 			1
#define	AUX_DATA_END			2
#define	AUX_DATA_NULL			3

typedef struct _AUX_DATA_BLOCK_HDR_
{
  UINT32  auxType;
  UINT32  auxVersion;
  UINT32  auxLength;
  UINT32  auxPayloadVersion;
}AUX_DATA_BLOCK_HDR;

#define	AUX_DATA_START_VERSION		1

typedef struct _AUX_DATA_START_BLOCK_
{
  AUX_DATA_BLOCK_HDR  hdr;
  UINT32              auxDataPhysAddr;
  UINT32              auxDataSize;
}AUX_DATA_START_BLOCK;

#define	AUX_DATA_END_VERSION		1

typedef struct _AUX_DATA_END_BLOCK_
{
  AUX_DATA_BLOCK_HDR  hdr;
  UINT32              terminator;
}AUX_DATA_END_BLOCK;

#define	AUX_DATA_NULL_VERSION		1

typedef struct _AUX_DATA_NULL_BLOCK_
{
  AUX_DATA_BLOCK_HDR  hdr;
}AUX_DATA_NULL_BLOCK;

#define NON_STACKID_BIT				0x8000	/* Used to filter Non-Stack IDs */
#define NON_STACKID_RAW_SEND_MASK	0x80F8	/* Used to filter raw send packets */

/********************************************************************
		CTCB is used at the MLI interface
*********************************************************************/
typedef struct _CTCB_FRAGMENT_BLOCK_STRUCT_
{
	UINT32	CTCB_FragmentCount;
	FRAGMENT_STRUCT CTCB_Fragment[1];
} CTCB_FRAGMENT_BLOCK;

typedef struct _CTCB_
{
	void 						*CTCB_Reserved;
	UINT32					CTCB_BoardNumber;
	UINT32					CTCB_DriverWS[3];
	UINT32					CTCB_DataLen;
   CTCB_FRAGMENT_BLOCK	*CTCB_FragBlockPtr;
	UINT32					CTCB_MediaHeaderLen;
	UINT8						CTCB_MediaHeader[MAX_MEDIA_HEADER_SIZE];
} CTCB;

/********************************************************************

   With compilers that implement strict alignement the location of
   were padding is assumed in the structure has been indicated below.

*********************************************************************/
typedef	struct	_AES_ECB_	
{
  	struct	_AES_ECB_	*AES_Link;
  	UINT32				AES_MSecondValue;
  	UINT16				AES_Status;
/*	UINT8				AES_Pad1[2];*/				/* Compiler padding */
	void 				(_cdecl *AES_ESR)(struct _AES_ECB_ *);
	UINT32				AES_Reserved;
	void 				*AES_ResourceObj;
	void 				*AES_Context;
}
AESECB, *PAESECB, AES_ECB;

typedef	struct	_LOOKAHEAD_	
{
	ECB			*LkAhd_PreFilledECB;
	UINT8			*LkAhd_MediaHeaderPtr;
	UINT32		LkAhd_MediaHeaderLen;
	UINT8			*LkAhd_DataLookAheadPtr;
	UINT32		LkAhd_DataLookAheadLen;
	UINT32		LkAhd_BoardNumber;
	UINT32		LkAhd_PktAttr;			
	UINT32		LkAhd_DestType;
	UINT32		LkAhd_FrameDataSize;
	UINT16		LkAhd_PadAlignBytes1;
	PROT_ID		LkAhd_ProtocolID;
	UINT16		LkAhd_PadAlignBytes2;
	NODE_ADDR	LkAhd_ImmediateAddress;
	UINT32		LkAhd_FrameDataStartCopyOffset;
	UINT32		LkAhd_FrameDataBytesWanted;
	ECB			*LkAhd_ReturnedECB;
	UINT32		LkAhd_PriorityLevel;
	void			*LkAhd_Reserved;
}
LOOKAHEAD, *PLOOKAHEAD;

/* Definitions for Protocol Stack Configuration And Statistics Tables */

/********************************************************************
		C Protocol Stack Configuration Table Definitions.
*********************************************************************/

#define PSTK_CONFIG_TABLE_MAJOR_VER 2
#define PSTK_CONFIG_TABLE_MINOR_VER 1

/********************************************************************
		C Protocol Stack Configuration Table System Flags definitions.
*********************************************************************/

#define PSTK_CFG_AUTO_NETWORK_RESOLUTION_BIT	0x08000000
#define PSTK_CFG_AUTO_BIND_ACTIVE_BIT			0x10000000
#define PSTK_CFG_ROUTER_ACTIVE_BIT				0x20000000
#define	PSTK_CFG_SERVER_BIT						0x40000000
#define	PSTK_CFG_CLIENT_BIT						0x80000000

/********************************************************************
		C Protocol Stack Configuration Table Structure definition.
*********************************************************************/

typedef	struct _PS_CONFIG_TABLE_
{
	UINT16			PConfigTableMajorVer;
	UINT16			PConfigTableMinorVer;
	MEON_STRING		*PProtocolLongName;
	MEON_STRING		*PProtocolShortName;
	UINT16			PProtocolMajorVer;
	UINT16			PProtocolMinorVer;
    UINT8     		PConfigTable_ODISpecMajorVersion;
    UINT8     		PConfigTable_ODISpecMinorVersion;
    UINT8     		PConfigTable_ProtocolAPIMajorVersion;
    UINT8     		PConfigTable_ProtocolAPIMinorVersion;
    UINT32   		PConfigTable_SystemFlags;
    UINT32   		PConfigTable_ProtocolFlags;
    UINT32   		PConfigTable_ProtocolReserved;
}
PS_ConfigTable, *PPS_ConfigTable, PS_CONFIG_TABLE;


typedef	struct _PS_STATS_TABLE_
{
	UINT16				PStatTableMajorVer;		/* Config Table Version 2.00 */
	UINT16				PStatTableMinorVer;		/* For the CODI 1.10 Spec */
	UINT32				PNumGenericCounters;
	STAT_TABLE_ENTRY	(*PGenericCountersPtr)[];
	UINT32				PNumCustomCounters;
	STAT_TABLE_ENTRY	(*PCustomCountersPtr)[];
}
PS_StatsTable, *PPS_StatsTable, PS_STATS_TABLE;


#define	NUM_GENERIC_PS_COUNTERS			3

#define	PS_TOTAL_TX_PACKETS				0
#define	PS_TOTAL_RX_PACKETS				1
#define	PS_IGNORED_RX_PACKETS			2

/********************************************************************
	Network Address Information structure used by Protocol Stack IOCTL 9
*********************************************************************/

typedef struct _NETWORK_ADDRESS_INFO_
{
	UINT32	addressType;
	UINT32	size;
	UINT8	address[32];
}NETWORK_ADDRESS_INFO;


/********************************************************************

   With compilers that implement strict alignement the location of
   were padding is assumed in the structure has been indicated below.

*********************************************************************/

typedef	struct _Lan_Memory_Configuration_
{
	void			*MemoryAddress;
	UINT16			MemorySize;
/*	UINT8			Lan_Mem_Pad1[2];*/			/* Compiler padding */
}
Lan_Memory_Configuration;


/* *********************************************************************** */
/* Definitions for MLID Configuration, Statistics Tables and Misc. structures */
/* *********************************************************************** */


/********************************************************************

   With compilers that implement strict alignement the location of
   were padding is assumed in the structure has been indicated below.

*********************************************************************/

typedef	struct _MLID_CONFIG_TABLE_	
{
	MEON  			MLIDCFG_Signature[26];
	UINT8 			MLIDCFG_MajorVersion;		/* Config Table Version 1.21 */
	UINT8 			MLIDCFG_MinorVersion;	   /* for the CODI 1.11 Spec    */
	NODE_ADDR		MLIDCFG_NodeAddress;
	UINT16			MLIDCFG_ModeFlags;
	UINT16			MLIDCFG_BoardNumber;
	UINT16			MLIDCFG_BoardInstance;
	UINT32			MLIDCFG_MaxFrameSize;
	UINT32			MLIDCFG_BestDataSize;
	UINT32			MLIDCFG_WorstDataSize;
	MEON_STRING		*MLIDCFG_CardName;
	MEON_STRING		*MLIDCFG_ShortName;
	MEON_STRING		*MLIDCFG_FrameTypeString;
	UINT16			MLIDCFG_Reserved0;
	UINT16			MLIDCFG_FrameID;
	UINT16			MLIDCFG_TransportTime;
/*	UINT8 			MLIDCFG_Pad1[2];*/				/* Compiler padding */
	UINT32			(_cdecl *MLIDCFG_SourceRouting)(UINT32, void*, void**, BOOLEAN);
	UINT16			MLIDCFG_LineSpeed;
	UINT16			MLIDCFG_LookAheadSize;
	UINT8 			MLIDCFG_SGCount;
	UINT8 			MLIDCFG_Reserved1;
	UINT16			MLIDCFG_PrioritySup;
	void           *MLIDCFG_Reserved2;
	UINT8 			MLIDCFG_DriverMajorVer;
	UINT8 			MLIDCFG_DriverMinorVer;
	UINT16			MLIDCFG_Flags;
	UINT16			MLIDCFG_SendRetries;
/*	UINT8 			MLIDCFG_Pad2[2];*/				/* Compiler padding */
	void  			*MLIDCFG_DriverLink;
	UINT16			MLIDCFG_SharingFlags;
	UINT16			MLIDCFG_Slot;
	UINT16			MLIDCFG_IOPort0;
	UINT16			MLIDCFG_IORange0;
	UINT16			MLIDCFG_IOPort1;
	UINT16			MLIDCFG_IORange1;
	Lan_Memory_Configuration
						LAN_MEMORY_CONFIGURATION[2];

#define MLIDCFG_MemoryAddress0 LAN_MEMORY_CONFIGURATION[0].MemoryAddress
#define MLIDCFG_MemorySize0 LAN_MEMORY_CONFIGURATION[0].MemorySize
#define MLIDCFG_MemoryAddress1 LAN_MEMORY_CONFIGURATION[1].MemoryAddress
#define MLIDCFG_MemorySize1 LAN_MEMORY_CONFIGURATION[1].MemorySize

	UINT8 			MLIDCFG_Interrupt0;
	UINT8 			MLIDCFG_Interrupt1;
	UINT8 			MLIDCFG_DMALine0;
	UINT8 			MLIDCFG_DMALine1;
	void  			*MLIDCFG_ResourceTag;
	void  			*MLIDCFG_Config;
	void  			*MLIDCFG_CommandString;
	MEON_STRING		MLIDCFG_LogicalName[18];
/*	UINT8 			MLIDCFG_Pad3[2];*/				/* Compiler padding */
	void			*MLIDCFG_LinearMemory0;
	void			*MLIDCFG_LinearMemory1;
	UINT16			MLIDCFG_ChannelNumber;
/*	UINT8 			MLIDCFG_Pad4[2];*/				/* Compiler padding */
	void  			*MLIDCFG_DBusTag;
	UINT8 			MLIDCFG_DIOConfigMajorVer;
	UINT8 			MLIDCFG_DIOConfigMinorVer;
/*	UINT8 			MLIDCFG_Pad5[2];*/				/* Compiler padding */
}
MLID_ConfigTable, *PMLID_ConfigTable, MLID_CONFIG_TABLE;

typedef	struct	_IO_CONFIG_
{
	struct _IO_CONFIG_	*IO_DriverLink;
	UINT16				IO_SharingFlags;
	UINT16				IO_Slot;
	UINT16				IO_IOPort0;
	UINT16				IO_IORange0;
	UINT16				IO_IOPort1;
	UINT16				IO_IORange1;
    Lan_Memory_Configuration
                          LAN_MEMORY_CONFIGURATION[2];

#define IO_MemoryAddress0 LAN_MEMORY_CONFIGURATION[0].MemoryAddress
#define IO_MemorySize0 LAN_MEMORY_CONFIGURATION[0].MemorySize
#define IO_MemoryAddress1 LAN_MEMORY_CONFIGURATION[1].MemoryAddress
#define IO_MemorySize1 LAN_MEMORY_CONFIGURATION[1].MemorySize

	UINT8						IO_Interrupt0;
	UINT8						IO_Interrupt1;
	UINT8						IO_DMALine0;
	UINT8						IO_DMALine1;
	struct ResourceTagStructure	*IO_ResourceTag;
	void						*IO_Config;
	void				 		*IO_CommandString;
	MEON_STRING					IO_LogicalName[18];
/*	UINT8						IO_Pad1[2];*/				/* Compiler padding */
  	void						*IO_LinearMemory0;
  	void						*IO_LinearMemory1;
  	UINT16						IO_ChannelNumber;
/*	UINT8						IO_Pad2[2];*/				/* Compiler padding */
  	void						*IO_DBusTag;
	UINT8					    IO_DIOConfigMajorVer;
	UINT8					    IO_DIOConfigMinorVer;
/*	UINT8						IO_Pad3[2];*/				/* Compiler padding */
}
IO_CONFIG;



typedef	struct _MLID_STATS_TABLE_
{
	UINT16				MStatTableMajorVer;
	UINT16				MStatTableMinorVer;
	UINT32				MNumGenericCounters;
	STAT_TABLE_ENTRY	(*MGenericCountsPtr)[];
	UINT32				MNumMediaCounters;
	STAT_TABLE_ENTRY	(*MMediaCountsPtr)[];
	UINT32				MNumCustomCounters;
	STAT_TABLE_ENTRY	(*MCustomCountersPtr)[];
}
MLID_StatsTable, *PMLID_StatsTable, MLID_STATS_TABLE;

#define	NUM_GENERIC_MLID_COUNTERS		20

#define	MLID_TOTAL_TX_PACKET_COUNT			0
#define	MLID_TOTAL_RX_PACKET_COUNT			1
#define	MLID_NO_ECB_AVAILABLE_COUNT		2
#define	MLID_PACKET_TX_TOO_BIG_COUNT		3
#define	MLID_PACKET_TX_TOO_SMALL_COUNT	4
#define	MLID_PACKET_RX_OVERFLOW_COUNT		5
#define	MLID_PACKET_RX_TOO_BIG_COUNT		6
#define	MLID_PACKET_RX_TOO_SMALL_COUNT	7
#define	MLID_PACKET_TX_MISC_ERROR_COUNT	8
#define	MLID_PACKET_RX_MISC_ERROR_COUNT	9
#define	MLID_RETRY_TX_COUNT					10
#define	MLID_CHECKSUM_ERROR_COUNT			11
#define	MLID_HARDWARE_RX_MISMATCH_COUNT	12
#define	MLID_TOTAL_TX_OK_BYTE_COUNT		13
#define	MLID_TOTAL_RX_OK_BYTE_COUNT		14
#define	MLID_TOTAL_GROUP_ADDR_TX_COUNT	15
#define	MLID_TOTAL_GROUP_ADDR_RX_COUNT	16
#define	MLID_ADAPTER_RESET_COUNT			17
#define	MLID_ADAPTER_OPR_TIME_STAMP		18
#define	MLID_Q_DEPTH							19


#define NUM_TOKEN_SPECIFIC_COUNTERS		13

#define TRN_AC_ERROR_COUNT					0
#define TRN_ABORT_DELIMITER_COUNTER		1
#define TRN_BURST_ERROR_COUNTER			2
#define TRN_FRAME_COPIED_ERROR_COUNTER	3
#define TRN_FREQUENCY_ERROR_COUNTER		4
#define TRN_INTERNAL_ERROR_COUNTER		5
#define TRN_LAST_RING_STATUS				6
#define TRN_LINE_ERROR_COUNTER			7
#define TRN_LOST_FRAME_COUNTER			8
#define TRN_TOKEN_ERROR_COUNTER			9
#define TRN_UPSTREAM_NODE_ADDRESS		10
#define TRN_LAST_RING_ID					11
#define TRN_LAST_BEACON_TYPE				12


#define NUM_ETHERNET_SPECIFIC_COUNTERS 		8

#define ETH_TX_OK_SINGLE_COLLISIONS_COUNT 	0
#define ETH_TX_OK_MULTIPLE_COLLISIONS_COUNT	1
#define ETH_TX_OK_BUT_DEFERRED					2
#define ETH_TX_ABORT_LATE_COLLISION				3
#define ETH_TX_ABORT_EXCESS_COLLISION			4
#define ETH_TX_ABORT_CARRIER_SENSE				5
#define ETH_TX_ABORT_EXCESSIVE_DEFERRAL		6
#define ETH_RX_ABORT_FRAME_ALIGNMENT			7

#define NUM_FDDI_SPECIFIC_COUNTERS	10

#define	FDDI_CONFIGURATION_STATE	0
#define	FDDI_UPSTREAM_NODE			1
#define	FDDI_DOWNSTREAM_NODE			2
#define	FDDI_FRAME_ERROR_COUNT		3
#define	FDDI_FRAMES_LOST_COUNT		4
#define	FDDI_RING_MANAGEMENT_STATE	5
#define	FDDI_LCT_FAILURE_COUNT		6
#define	FDDI_LEM_REJECT_COUNT		7
#define	FDDI_LEM_COUNT					8
#define	FDDI_L_CONNECTION_STATE		9


typedef	struct _MLID_REG_
{
	void 		(*MLIDSendHandler)(ECB*, void*);
	INFO_BLOCK	*MLIDControlHandler;
	void 		*MLIDSendContext;
	void 		*MLIDResourceObj;
	void 		*MLIDModuleHandle;
}
MLID_Reg, *PMLID_Reg, MLID_REG;

/* Definitions for Bound Protocol Stacks */

typedef	struct	_PS_BOUND_NODE_
{
	MEON_STRING	*ProtocolName;
	ODISTAT		(CALLCNV *ProtocolReceiveHandler)(LOOKAHEAD*);
	INFO_BLOCK	*ProtocolControlHandler;
	void		*ProtocolResourceObj;
}
PS_BoundNode, *PPS_BoundNode, PS_BOUND_NODE;


/* Definitions for PreScan Rx and Default Chained Protocol Stacks */

typedef	struct _PS_CHAINED_RX_NODE_
{
	struct	_PS_CHAINED_RX_NODE_	*StackChainLink;
	UINT32							StackChainBoardNumber;
	CHNPOS							StackChainPositionRequested;
	ODISTAT							(CALLCNV *StackRxChainHandler)(LOOKAHEAD*, struct _PS_CHAINED_RX_NODE_ *);
	INFO_BLOCK 						*StackChainControl;
	UINT32	   						StackChainFilter;
	void		   					*StackChainContext;
	void		   					*StackChainResourceObj;
}
PS_ChainedRxNode, *PPS_ChainedRxNode, PS_CHAINED_RX_NODE;

/* Definitions for PreScan Tx Chained Protocol Stacks */

typedef	struct _PS_CHAINED_TX_NODE_
{
	struct	_PS_CHAINED_TX_NODE_	*StackChainLink;
	UINT32							StackChainBoardNumber;
	CHNPOS							StackChainPositionRequested;
	ODISTAT							(CALLCNV *StackTxChainHandler)(ECB*, struct _PS_CHAINED_TX_NODE_ *);
	INFO_BLOCK						*StackChainControl;
	UINT32							StackChainFilter;
	void							*StackChainContext;
	void							*StackChainResourceObj;
}
PS_ChainedTxNode, *PPS_ChainedTxNode, PS_CHAINED_TX_NODE;


/*Definitions for SFT III Exchange Protocol Control Service
*/

typedef	struct	_SFTIII_EXCHANGE_NODE_
{

	UINT32	SubFunction;
	void		*Parameter1;
	void		*Parameter2;
} SFTIIIExchangeNode, SFTIII_EXCHANGE_NODE;


#if defined	OS_NT || defined OS_WIN95

/* Reset PRAGMA to normal after packing above structures */

#pragma	pack()

#endif

/*========[ Function Prototypes ]================*/

ODISTAT	CLSL_AddProtocolID		(	PROT_ID	*ProtocolID,
									MEON_STRING	*ProtocolName,
									MEON_STRING	*FrameName);

ODISTAT	CLSL_BindProtocolToBoard(	UINT32		ProtocolNumber,
								 	UINT32		BoardNumber,
								 	MEON_STRING	*UserParmString);

ODISTAT	CLSL_BindStack	         (	UINT32		ProtocolNumber,
									UINT32		BoardNumber);

ODISTAT	CLSL_CancelAESEvent		(	AES_ECB		*TimerAESECB);

ODISTAT	CLSL_CancelEvent		(	ECB		      *ECBBuffer);

ODISTAT	CLSL_ControlStackFilter	(	UINT32      BoardNumber,
									UINT32		Function,
									UINT32		Mask,
									void  		*Parameter1,
									void  		*Parameter2);

ODISTAT	CLSL_Dummy				(void);
ODISTAT	CLSL_DeRegisterDefaultChain	(PS_CHAINED_RX_NODE	*StackChainNode);
ODISTAT	CLSL_DeRegisterMLID		(UINT32				BoardNumber);
ODISTAT	CLSL_DeRegisterPreScanChain   (PS_CHAINED_RX_NODE *PStkChainRxNode,
												   PS_CHAINED_TX_NODE *PStkChainTxNode);

ODISTAT	CLSL_DeRegisterStack	   (UINT32		ProtocolNumber);
void		CLSL_FastHoldEvent		(ECB			*ECBBuffer);
void		CLSL_FastSendComplete	(ECB			*SendECB);
ODISTAT	CLSL_GetBoundBoardInfo	(UINT32		BoardNumber,
											 UINT32		*StackBuffer);

UINT32				CLSL_GetIntervalMarker		(void);
LSL_CONFIG_TABLE*	CLSL_GetLSLConfiguration	(void);
LSL_STATS_TABLE*	CLSL_GetLSLStatistics		(void);
UINT32				CLSL_GetMaxECBBufferSize	(void);
INFO_BLOCK*			CLSL_GetMLIDControlEntry	(UINT32	BoardNumber,
															ODISTAT	*ErrorStatus);
ECB *CLSL_GetPhysicalAddressOfECB(ECB *ecb);

PROT_ID	* CLSL_GetPIDFromStackIDBoard	(UINT32	ProtocolNumber,
			 										UINT32	BoardNumber,
			 										ODISTAT	*ErrorStatus);

INFO_BLOCK*		CLSL_GetProtocolControlEntry	(UINT32	ProtocolNumber,
															ODISTAT	*ErrorStatus);

ECB*		CLSL_GetSizedECB		      (UINT32			ECBDataSize,
                                     void				*pResourceObj,
										      BOOLEAN	      Below16Meg);

ECB*    CLSL_GetMultipleECBs   (UINT32		ECBDataSize,
                                void		*pResourceObj,
                                UINT32		*nECBs);

ODISTAT	CLSL_GetStackECB				(LOOKAHEAD	      *LookAheadBuf);
ODISTAT	CLSL_GetStackIDFromName		(MEON_STRING   *Name,
                                     UINT32			*ProtocolNumber);

ODISTAT	CLSL_GetStartofChain  	  (UINT32		BoardNumber,
												PS_CHAINED_RX_NODE	**DefaultChainStartNode,
												PS_CHAINED_RX_NODE	**PreScanRxChainStartNode,
												PS_CHAINED_TX_NODE	**PreScanTxChainStartNode);

void		CLSL_HoldEvent		         (ECB			*HoldECB);
ODISTAT	CLSL_ModifyStackFilter		(void					*StackIdentifier,
												 UINT32				BoardNumber,
												 UINT32				NewMask,
												 UINT32				*pCurrentMask);

ODISTAT	CLSL_RegisterDefaultChain	(PS_CHAINED_RX_NODE	*StackChainNode);
ODISTAT	CLSL_RegisterMLID	         (MLID_REG			*MLIDHandlers,
												 MLID_CONFIG_TABLE	*MLIDConfigTable,
												 UINT32				*BoardNumber);

ODISTAT	CLSL_RegisterPreScanChain  ( PS_CHAINED_RX_NODE *PStkChnPreRxNode,
												  PS_CHAINED_TX_NODE *PStkChnPreTxNode);
												
ODISTAT	CLSL_RegisterStack		(PS_BOUND_NODE		*ProtocolNode,
											 UINT32				*ProtocolNumber);

ODISTAT	CLSL_ReSubmitDefault		(PS_CHAINED_RX_NODE	*StackChainnode,
											 LOOKAHEAD				*LookAheadBuf);

ODISTAT	CLSL_ReSubmitPreScanRx	(PS_CHAINED_RX_NODE	*StackChainnode,
											 LOOKAHEAD				*LookAheadBuf);

ODISTAT	CLSL_ReSubmitPreScanTx	(PS_CHAINED_TX_NODE	*StackChainnode,
											 ECB			         *TransmitECB);
ODISTAT	_cdecl CLSL_ReturnECB			(ECB		*ReturnedECB);
ODISTAT	CLSL_ScheduleAESEvent	(AES_ECB	*TimerAESECB);
void   	CLSL_SendComplete		(ECB		*SendECB);
ODISTAT	CLSL_SendPacket			(ECB		*SendECB);

SFTIII_STAT	CLSL_SendProtocolInfoToPartner( UINT32	ProtocolNumber,
														  UINT8	*ProtocolInfo,
														  UINT32	Length,
														  void	(*InfoSendCallBack)
																	(UINT32 Reserved,
																	 UINT8 *ProtocolInfo));
SFTIII_STAT	CLSL_SendProtocolInfoToOtherEngine(	UINT32	ProtocolNumber,
																UINT8	*ProtocolInfo,
																UINT32	Length,
																void	(*InfoSendCallBack)
																		(UINT8 *ProtocolInfo));
void CLSL_ServiceEvents	(void);

ODISTAT	CLSL_UnbindStack		(UINT32	ProtocolNumber,
										 UINT32	BoardNumber);

void CLSL_Reserved( void );

/*========[ Macro Defintions ]================*/

/*	Macro Definitions to ease access to Control Procedures for
	Protocol Stacks and MLIDs
*/


/*	Protocol Stack Control Functions	*/

#define PSTK_NUM_API	10L

#define	PSTK_GET_CONFIGURATION	0x0000
#define	PSTK_GET_STATISTICS		0x0001
#define	PSTK_BIND				0x0002
#define	PSTK_UNBIND				0x0003
#define	PSTK_MLID_DEREGISTER 	0x0004
#define	PSTK_PROMISCUOUS_STATE	0x0005
#define	PSTK_RESERVED			0x0006
#define	PSTK_GET_PROTOCOL_STRG	0x0007
#define	PSTK_PROT_MANAGE	 	0x0008
#define PSTK_GET_BOUND_NET_INFO	0x0009


/*	w is ptr. to INFO_BLOCK defining API Array for Protocol Stack Control	*/

#define	PStkCntl_GetConfig(w, x)			  	\
		((PS_CONFIG_TABLE* (CALLCNV *)(void*))	\
		w->SupportAPIArray[PSTK_GET_CONFIGURATION])(x)

#define	PStkCntl_GetStats(w, x)				  	\
		((PS_STATS_TABLE* (CALLCNV *)(void*)) 	\
		w->SupportAPIArray[PSTK_GET_STATISTICS])(x)

#define	PStkCntl_Bind(w, x, y, z)			  				\
		((ODISTAT (CALLCNV *)(UINT32, MEON_STRING*, void*))	\
		w->SupportAPIArray[PSTK_BIND])(x, y, z)

#define	PStkCntl_MLIDDeReg(w, x, y)			\
		((void (CALLCNV *)(UINT32, void*))	\
		w->SupportAPIArray[PSTK_MLID_DEREGISTER])(x, y)

#define	PStkCntl_Unbind(w, x, y, z)							\
		((ODISTAT (CALLCNV *)(UINT32, MEON_STRING*, void*))	\
		w->SupportAPIArray[PSTK_UNBIND])(x, y, z)

#define	PStkCntl_PromiscState(w, x, y, z)				\
		((ODISTAT (CALLCNV *)(UINT32, UINT32, void*))	\
		w->SupportAPIArray[PSTK_PROMISCUOUS_STATE])(x, y, z)

#define	PStkCntl_GetProtocolString(w, x, y, z)				\
		((ODISTAT (CALLCNV *)(UINT32, MEON_STRING*, void*))	\
		w->SupportAPIArray[PSTK_GET_PROTOCOL_STRG])(x, y, z)

#define	PStkCntl_ProtManage(w, x, y)			\
		((ODISTAT (CALLCNV *)(ECB *, void*))	\
		w->SupportAPIArray[PSTK_PROT_MANAGE])(x, y)

#define	PStkCntl_GetBoundNetInfo(w, x, y, z)							\
		((ODISTAT (CALLCNV *)(UINT32, NETWORK_ADDRESS_INFO*, void*))	\
		w->SupportAPIArray[PSTK_GET_BOUND_NET_INFO])(x, y, z)

/*	MLID Control Functions	*/

#define MLID_NUM_API	19L

#define	MLID_GET_CONFIGURATION			0x0000	/*  0 - GetMLIDConfiguration */
#define	MLID_GET_STATISTICS				0x0001	/*  1 - GetMLIDStatistics */
#define	MLID_ADD_MULTICAST				0x0002	/*  2 - AddMulticastAddress */
#define	MLID_DELETE_MULTICAST			0x0003	/*  3 - DeleteMulticastAddress */
#define MLID_RESERVED					0x0004	/*  4 - Reserved */
#define	MLID_SHUTDOWN					0x0005	/*  5 - MLIDShutdown */
#define	MLID_RESET						0x0006	/*  6 - MLIDReset */
#define MLID_RESERVED1					0x0007	/*  7 - Reserved1 */
#define MLID_RESERVED2					0x0008	/*  8 - Reserved2 */
#define	MLID_SET_LOOK_AHEAD				0x0009	/*  9 - SetLookAheadSize */
#define	MLID_PROMISCUOUS_CHANGE			0x000A	/* 10 - PromiscuousChange */
#define	MLID_REGISTER_TX_MONITOR		0x000B	/* 11 - RegisterMonitor */
#define MLID_RESERVED3					0x000C	/* 12 - Reserved3 */
#define MLID_RESERVED4					0x000D	/* 13 - Reserved4 */
#define	MLID_MANAGEMENT					0x000E	/* 14 - MLIDManagement */
#define	MLID_GET_MULTICAST_INFO			0x000F	/* 15 - GetMulticastInfo */
#define	MLID_REMOVE_NETWORK_INTERFACE	0x0010	/* 16 - RemoveNetworkInterface */
#define	MLID_SHUTDOWN_NETWORK_INTERFACE	0x0011	/* 17 - ShutdownNetworkInterface */
#define	MLID_RESET_NETWORK_INTERFACE	0x0012	/* 18 - ResetNetworkInterface */


/* Shutdown type defines */

#define SHUTDOWN_PERMANENT			0x0000
#define SHUTDOWN_PARTIAL			0x0001


/*	w is ptr. to INFO_BLOCK defining API Array for MLID Control	*/
		
#define	MLIDCntl_GetConfig(w, x, y, z)								\
		((ODISTAT (CALLCNV *)(UINT32, MLID_CONFIG_TABLE**, ECB*))	\
		w->SupportAPIArray[MLID_GET_CONFIGURATION])(x, y, z)

#define	MLIDCntl_GetStats(w, x, y, z)								\
		((ODISTAT (CALLCNV *)(UINT32, MLID_STATS_TABLE**, ECB*))	\
		w->SupportAPIArray[MLID_GET_STATISTICS])(x, y, z)

#define	MLIDCntl_AddMulti(w, x, y, z)						\
		((ODISTAT (CALLCNV *)(UINT32, NODE_ADDR*, ECB*))	\
		w->SupportAPIArray[MLID_ADD_MULTICAST])(x, y, z)

#define	MLIDCntl_DelMulti(w, x, y, z)						\
		((ODISTAT (CALLCNV *)(UINT32, NODE_ADDR*, ECB*))	\
		w->SupportAPIArray[MLID_DELETE_MULTICAST])(x, y, z)

#define	MLIDCntl_Shutdown(w, x, y, z)					\
		((ODISTAT (CALLCNV *)(UINT32, UINT32, ECB*))	\
		w->SupportAPIArray[MLID_SHUTDOWN])(x, y, z)

#define	MLIDCntl_Reset(w, x, y)					\
		((ODISTAT (CALLCNV *)(UINT32, ECB*))	\
		w->SupportAPIArray[MLID_RESET])(x, y)

#define	MLIDCntl_SetLookAhead(w, x, y, z)				\
		((ODISTAT (CALLCNV *)(UINT32, UINT32, ECB*))	\
		w->SupportAPIArray[MLID_SET_LOOK_AHEAD])(x, y, z)

#define	MLIDCntl_PromisChange(w, x, y, z, aa)					\
		((ODISTAT (CALLCNV *)(UINT32, UINT32, UINT32*, ECB*))	\
		w->SupportAPIArray[MLID_PROMISCUOUS_CHANGE])(x, y, z, aa)

#define	MLIDCntl_RegMon(w, x, y, z, aa)							\
		((ODISTAT (CALLCNV *)(UINT32, void*, ECB*, BOOLEAN))	\
		w->SupportAPIArray[MLID_REGISTER_TX_MONITOR])(x, y, z, aa)
					
#define	MLIDCntl_Management(w, x, y)			\
		((ODISTAT (CALLCNV *)(UINT32, ECB*))	\
		w->SupportAPIArray[MLID_MANAGEMENT])(x, y)

#define	MLIDCntl_GetMulticastInfo(w, x, y) 		\
		((ODISTAT (CALLCNV *)(UINT32, ECB*))	\
		w->SupportAPIArray[MLID_GET_MULTICAST_INFO])(x, y)

#define	MLIDCntl_RemoveNetworkInterface(w, x, y)	\
		((ODISTAT (CALLCNV *)(UINT32, ECB*))		\
		w->SupportAPIArray[MLID_REMOVE_NETWORK_INTERFACE])(x, y)

#define	MLIDCntl_ShutdownNetworkInterface(w, x, y)	\
		((ODISTAT (CALLCNV *)(UINT32, ECB*))		\
		w->SupportAPIArray[MLID_SHUTDOWN_NETWORK_INTERFACE])(x, y)

#define	MLIDCntl_ResetNetworkInterface(w, x, y)	\
		((ODISTAT (CALLCNV *)(UINT32, ECB*))	\
		w->SupportAPIArray[MLID_RESET_NETWORK_INTERFACE])(x, y)

typedef struct _GROUP_ADDR_LIST_NODE_ {
		NODE_ADDR	GRP_ADDR;
		UINT16		GRP_ADDR_COUNT;
	} GROUP_ADDR_LIST_NODE;



/*	LSL Function Indexes */

#define LSL_NUM_API 49L

#define	CLSL_GET_SIZED_ECB						0x0000	/*  0 - CLSL_GetSizedECB */
#define	CLSL_RETURN_ECB							0x0001	/*  1 - CLSL_ReturnECB */
#define	CLSL_CANCEL_EVENT						0x0002	/*  2 - CLSL_CancelEvent */
#define	CLSL_SCHEDULE_AES_EVENT					0x0003	/*  3 - CLSL_ScheduleAESEvent */
#define	CLSL_CANCEL_AES_EVENT					0x0004	/*  4 - CLSL_CancelAESEvent */
#define	CLSL_GET_INTERVAL_MARKER				0x0005	/*  5 - CLSL_GetIntervalMarker */
#define	CLSL_REGISTER_STACK						0x0006	/*  6 - CLSL_RegisterStack */
#define	CLSL_DEREGISTER_STACK					0x0007	/*  7 - CLSL_DeRegisterStack */
#define	CLSL_RESERVED							0x0008	/*  8 - Reserved */
#define	CLSL_RESERVED1							0x0009	/*  9 - Reserved1 */
#define	CLSL_RESERVED2							0x000A	/* 10 - Reserved2 */
#define	CLSL_GET_STACK_ECB						0x000B	/* 11 - CLSL_GetStackECB */
#define	CLSL_SEND_PACKET						0x000C	/* 12 - CLSL_SendPacket */
#define	CLSL_FAST_SEND_COMPLETE					0x000D	/* 13 - CLSL_FastSendComplete */
#define	CLSL_SEND_COMPLETE						0x000E	/* 14 - CLSL_SendComplete */
#define	CLSL_REGISTER_MLID						0x000F	/* 15 - CLSL_RegisterMLID */
#define	CLSL_GET_STACK_ID_FROM_NAME				0x0010	/* 16 - CLSL_GetStackIDFromName */
#define	CLSL_GET_PID_FROM_STACK_ID_BOARD		0x0011	/* 17 - CLSL_GetPIDFromStackIDBoard */
#define	CLSL_GET_MLID_CONTROL_ENTRY				0x0012	/* 18 - CLSL_GetMLIDControlEntry */
#define	CLSL_GET_PROTOCOL_CONTROL_ENTRY			0x0013	/* 19 - CLSL_GetProtocolControlEntry */
#define	CLSL_GET_LSL_STATISTICS					0x0014	/* 20 - CLSL_GetLSLStatistics */
#define	CLSL_BIND_STACK							0x0015	/* 21 - CLSL_BindStack */
#define	CLSL_UNBIND_STACK						0x0016	/* 22 - CLSL_UnbindStack */
#define	CLSL_ADD_PROTOCOL_ID					0x0017	/* 23 - CLSL_AddProtocolID */
#define	CLSL_GET_BOUND_BOARD_INFO				0x0018	/* 24 - CLSL_GetBoundBoardInfo */
#define	CLSL_GET_LSL_CONFIGURATION				0x0019	/* 25 - CLSL_GetLSLConfiguration */
#define	CLSL_DEREGISTER_MLID					0x001A	/* 26 - CLSL_DeRegisterMLID */
#define	CLSL_REGISTER_DEFAULT_CHAIN				0x001B	/* 27 - CLSL_RegisterDefaultChain */
#define	CLSL_REGISTER_PRESCAN_CHAIN				0x001C	/* 28 - CLSL_RegisterPreScanChain */
#define	CLSL_RESERVED3							0x001D	/* 29 - Reserved3 */
#define	CLSL_DEREGISTER_DEFAULT_CHAIN			0x001E	/* 30 - DeRegisterDefaultChain */
#define	CLSL_DEREGISTER_PRESCAN_CHAIN			0x001F	/* 31 - DeRegisterPreScanChain */
#define	CLSL_RESERVED4							0x0020	/* 32 - Reserved4 */
#define	CLSL_GET_START_OF_CHAIN					0x0021	/* 33 -	CLSL_GetStartofChain */
#define	CLSL_RESUBMIT_DEFAULT					0x0022	/* 34 - CLSL_ReSubmitDefault */
#define	CLSL_RESUBMIT_PRESCAN_RX				0x0023	/* 35 - CLSL_ReSubmitPreScanRx */
#define	CLSL_RESUBMIT_PRESCAN_TX				0x0024	/* 36 - CLSL_ResubmitPreScanTx */
#define	CLSL_HOLD_EVENT							0x0025	/* 37 - CLSL_HoldEvent */
#define	CLSL_FAST_HOLD_EVENT					0x0026	/* 38 - CLSL_FastHoldEvent */
#define	CLSL_GET_MAX_ECB_BUFFER_SIZE			0x0027	/* 39 - CLSL_GetMaxECBBufferSize */
#define	CLSL_RESERVED5							0x0028	/* 40 - Reserved5 */
#define	CLSL_SERVICE_EVENTS						0x0029	/* 41 - CLSL_ServiceEvents */
#define	CLSL_MODIFY_STACK_FILTER				0x002A	/* 42 - CLSL_ModifyStackFilter */
#define	CLSL_CONTROL_STACK_FILTER				0x002B	/* 43 - CLSL_ControlStackFilter */
#define	CLSL_SEND_PROTOCOL_INFO_TO_OTHER_ENGINE	0x002C 	/* 44 - CLSL_SendProtocolInfoToOtherEngine (Server ONLY) */
#define	CLSL_SEND_PROTOCOL_INFO_TO_PARTNER		0x002D	/* 45 - CLSL_SendProtocolInfoToPartner (Server ONLY) */
#define	CLSL_BIND_PROTOCOL_TO_BOARD				0x002E	/* 46 - CLSL_BindProtocolToBoard */
#define	CLSL_GET_MULTIPLE_ECBS					0x002F	/* 47 - CLSL_GetMultipleECBs */
#define	CLSL_GET_PHYSICAL_ADDRESS_OF_ECB		0x0030	/* 48 - CLSL_GetPhysicalAddressOfECB */

/*	w is ptr. to INFO_BLOCK defining API Array for C LSL APIs */


#define	CLSLEntry_GetSizedECB(w, x, y, z)			\
		((ECB* (CALLCNV *)(UINT32, void*, BOOLEAN))	\
		w->SupportAPIArray[CLSL_GET_SIZED_ECB])(x, y, z)

#define	CLSLEntry_ReturnECB(w, x)		\
		((ODISTAT (_cdecl *)(ECB*))  	\
		w->SupportAPIArray[CLSL_RETURN_ECB])(x)

#define	CLSLEntry_CancelEvent(w, x)		\
		((ODISTAT (CALLCNV *)(ECB*))	\
		w->SupportAPIArray[CLSL_CANCEL_EVENT])(x)

#define	CLSLEntry_ScheduleAESEvent(w, x)	\
		((void (CALLCNV *)(AES_ECB*))		\
		w->SupportAPIArray[CLSL_SCHEDULE_AES_EVENT])(x)

#define	CLSLEntry_CancelAESEvent(w, x)		\
		((ODISTAT (CALLCNV *)(AES_ECB*))	\
		w->SupportAPIArray[CLSL_CANCEL_AES_EVENT])(x)

#define	CLSLEntry_GetIntervalMarker(w)	\
		((UINT32 (CALLCNV *)(void))		\
		w->SupportAPIArray[CLSL_GET_INTERVAL_MARKER])()

#define	CLSLEntry_RegisterStack(w, x, y)				\
		((ODISTAT (CALLCNV *)(PS_BOUND_NODE*, UINT32*))	\
		w->SupportAPIArray[CLSL_REGISTER_STACK])(x, y)

#define	CLSLEntry_DeRegisterStack(w, x)	\
		((ODISTAT (CALLCNV *)(UINT32))	\
		w->SupportAPIArray[CLSL_DEREGISTER_STACK])(x)

#define	CLSLEntry_GetStackECB(w, x)			\
		((ODISTAT (CALLCNV *)(LOOKAHEAD*))	\
		w->SupportAPIArray[CLSL_GET_STACK_ECB])(x)

#define	CLSLEntry_SendPacket(w, x)		\
		((ODISTAT (CALLCNV *)(ECB*))	\
		w->SupportAPIArray[CLSL_SEND_PACKET])(x)

#define	CLSLEntry_FastSendComplete(w, x)	\
		((void (CALLCNV *)(ECB*))			\
		w->SupportAPIArray[CLSL_FAST_SEND_COMPLETE])(x)

#define	CLSLEntry_SendComplete(w, x)	\
		((void (CALLCNV *)(ECB*))		\
		w->SupportAPIArray[CLSL_SEND_COMPLETE])(x)

#define	CLSLEntry_RegisterMLID(w, x, y, z)								\
		((ODISTAT (CALLCNV *)(MLID_REG*, MLID_CONFIG_TABLE*, UINT32*))	\
		w->SupportAPIArray[CLSL_REGISTER_MLID])(x, y, z)

#define	CLSLEntry_GetStackIDFromName (w, x, y)			\
		((ODISTAT (CALLCNV *)(MEON_STRING*, UINT32*))	\
		w->SupportAPIArray[CLSL_GET_STACK_ID_FROM_NAME])(x, y)


#define	CLSLEntry_GetPIDFromStackIDBoard(w, x, y, z)		\
		((PROT_ID* (CALLCNV *)(UINT32,	UINT32,	ODISTAT*))	\
		w->SupportAPIArray[CLSL_GET_PID_FROM_STACK_ID_BOARD])(x, y, z)

#define	CLSLEntry_GetMLIDControlEntry(w, x, y)			\
		((INFO_BLOCK* (CALLCNV *)(UINT32, ODISTAT*))	\
		w->SupportAPIArray[CLSL_GET_MLID_CONTROL_ENTRY])(x, y)


#define	CLSLEntry_GetProtocolControlEntry(w, x, y)		\
		((INFO_BLOCK* (CALLCNV *) (UINT32, ODISTAT*))	\
		w->SupportAPIArray[CLSL_GET_PROTOCOL_CONTROL_ENTRY])(x, y)


#define	CLSLEntry_GetLSLStatistics(w)			\
		((LSL_STATS_TABLE* (CALLCNV *)(void))	\
		w->SupportAPIArray[CLSL_GET_LSL_STATISTICS])()

#define	CLSLEntry_BindStack(w, x, y)			\
		((ODISTAT (CALLCNV *)(UINT32, UINT32))	\
		w->SupportAPIArray[CLSL_BIND_STACK])(x, y)

#define	CLSLEntry_UnbindStack(w, x, y)			\
		((ODISTAT (CALLCNV *)(UINT32, UINT32))	\
		w->SupportAPIArray[CLSL_GET_UNBIND_STACK])(x, y)

#define	CLSLEntry_AddProtocolID(w, x, y, z)							\
		((ODISTAT (CALLCNV *)(PROT_ID*, MEON_STRING*, MEON_STRING*))	\
		w->SupportAPIArray[CLSL_ADD_PROTOCOL_ID])(x, y, z)

#define	CLSLEntry_GetBoundBoardInfo(w, x, y)		\
		((ODISTAT (CALLCNV *)( UINT32, UINT32*))	\
		w->SupportAPIArray[CLSL_GET_BOUND_BOARD_INFO])(x, y)

#define	CLSLEntry_GetLSLConfiguration(w) 		\
		((LSL_CONFIG_TABLE* (CALLCNV *)(void))	\
		w->SupportAPIArray[CLSL_GET_LSL_CONFIGURATION])()

#define	CLSLEntry_DeRegisterMLID(w, x)	\
		((ODISTAT (CALLCNV *)( UINT32))	\
		w->SupportAPIArray[CLSL_DEREGISTER_MLID])(x)

#define	CLSLEntry_RegisterDefaultChain(w, x)			\
		((ODISTAT (CALLCNV *)( PS_CHAINED_RX_NODE*))	\
		w->SupportAPIArray[CLSL_REGISTER_DEFAULT_CHAIN])(x)

#define	CLSLEntry_RegisterPreScanChain(w, x, y)							\
		((ODISTAT (CALLCNV *)( PS_CHAINED_RX_NODE*, PS_CHAINED_TX_NODE*))	\
		w->SupportAPIArray[CLSL_REGISTER_PRE_SCAN_CHAIN])(x, y)

#define	CLSLEntry_DeRegisterDefaultChain(w, x)			\
		((ODISTAT (CALLCNV *)( PS_CHAINED_RX_NODE*))	\
		w->SupportAPIArray[CLSL_DEREGISTER_DEFAULT_CHAIN])(x)

#define	CLSLEntry_DeRegisterPreScanChain(w, x, y)							\
		((ODISTAT (CALLCNV *)( PS_CHAINED_RX_NODE*, PS_CHAINED_TX_NODE*))	\
		w->SupportAPIArray[CLSL_DEREGISTER_PRE_SCAN_CHAIN])(x, y)

#define	CLSLEntry_GetStartofChain(w, x, y, z, aa)				\
		((ODISTAT (CALLCNV *)( UINT32, PS_CHAINED_RX_NODE**,	\
		PS_CHAINED_RX_NODE	**,	PS_CHAINED_TX_NODE	**))		\
		w->SupportAPIArray[CLSL_GET_START_OF_CHAIN])(x, y, z, aa)

#define	CLSLEntry_ReSubmitDefault(w, x, y)					   	\
		((ODISTAT	(CALLCNV *)( PS_CHAINED_RX_NODE	*, LOOKAHEAD*))	\
		w->SupportAPIArray[CLSL_RESUBMIT_DEFAULT])(x, y)

#define	CLSLEntry_ReSubmitPreScanRx(w, x, y)						\
		((ODISTAT (CALLCNV *)( PS_CHAINED_RX_NODE*, LOOKAHEAD*))	\
		w->SupportAPIArray[CLSL_RESUBMIT_PRESCAN_RX])(x, y)

#define	CLSLEntry_ResubmitPreScanTx(w, x, y)				\
		((ODISTAT (CALLCNV *)( PS_CHAINED_TX_NODE*, ECB*))	\
		w->SupportAPIArray[CLSL_RESUBMIT_PRESCAN_TX])(x, y)

#define	CLSLEntry_HoldEvent(w, x)	\
		((void (CALLCNV *)(ECB*))	\
		w->SupportAPIArray[CLSL_HOLD_EVENT])(x)

#define	CLSLEntry_FastHoldEvent(w, x)	\
		((void (CALLCNV *)(ECB*))		\
		w->SupportAPIArray[CLSL_FAST_HOLD_EVENT])(x)

#define	CLSLEntry_GetMaxECBBufferSize(w)	\
		((UINT32 (CALLCNV *)(void))			\
		w->SupportAPIArray[CLSL_GET_MAX_ECB_BUFFER_SIZE])()

#define	CLSLEntry_ServiceEvents(w)	\
		((void (CALLCNV *)(void))	\
		w->SupportAPIArray[CLSL_SERVICE_EVENTS])()

#define	CLSLEntry_ModifyStackFilter(w, x, y, z, aa)			\
		((ODISTAT (CALLCNV *)(void*, UINT32, UINT32, UINT32*))	\
		w->SupportAPIArray[CLSL_MODIFY_STACK_FILTER])(x, y, z, aa)

#define	CLSLEntry_ControlStackFilter(w, x, y, z, aa, bb)				\
		((ODISTAT (CALLCNV *)( UINT32, UINT32, UINT32, void*, void*))	\
		w->SupportAPIArray[CLSL_CONTROL_STACK_FILTER])(x, y, z, aa, bb)

#define	CLSLEntry_SendProtocolInfoToOtherEngine(w, x, y, z, aa)	\
		((SFTIII_STAT (CALLCNV *)( UINT32, UINT8*, UINT32, 			\
		void	(*InfoSendCallBack)(UINT8*)))						\
		w->SupportAPIArray[CLSL_SEND_PROTOCOL_INFO_TO_OTHER_ENGINE])(x, y, z, aa)

#define	CLSLEntry_SendProtocolInfoToPartner(w, x, y, z, aa)	\
		((SFTIII_STAT (CALLCNV *)( UINT32, UINT8*, UINT32,		\
		void	(*InfoSendCallBack)(UINT32, UINT8*)))			\
		w->SupportAPIArray[CLSL_SEND_PROTOCOL_INFO_TO_PARTNER])(x, y, z, aa)

#define	CLSLEntry_BindProtocolToBoard(w, x, y, z)				\
		((ODISTAT (CALLCNV *)( UINT32, UINT32, MEON_STRING*))	\
		w->SupportAPIArray[CLSL_BIND_PROTOCOL_TO_BOARD])(x, y, z)

#define	CLSLEntry_GetMultipleECBs(w, x, y ,z)			\
		((ECB* (CALLCNV *)( UINT32, void*, UINT32*))	\
		w->SupportAPIArray[CLSL_GET_MULTIPLE_ECBS])(x, y, z)

#define	CLSLEntry_GetPhysicalAddressOfECB(w, x)	\
		((ECB* (CALLCNV *)(ECB*))	\
		w->SupportAPIArray[CLSL_GET_PHYSICAL_ADDRESS_OF_ECB])(x)

#endif			/*	_ODI_Include_ */
