/***********************************************************************\
*
*   (c) Unpublished Work of Novell, Inc.  All rights reserved.
*
*   THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
*   PROPRIETARY AND TRADE SECRET INFORMATION OF NOVELL, INC.  ACCESS
*   TO THIS WORK IS RESTRICTED TO (I) NOVELL EMPLOYEES WHO HAVE A
*   NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE OF THEIR ASSIGNMENTS
*   AND (II) ENTITIES OTHER THAN NOVELL WHO HAVE ENTERED INTO
*   APPROPRIATE LICENSE AGREEMENTS.  NO PART OF THIS WORK MAY BE USED,
*   PRACTICED, PERFORMED, COPIED, DISTRIBUTED, REVISED, MODIFIED,
*   TRANSLATED, ABRIDGED, CONDENSED, EXPANDED, COLLECTED, COMPILED,
*   LINKED,RECAST, TRANSFORMED OR ADAPTED WITHOUT THE PRIOR WRITTEN
*   CONSENT OF NOVELL.  ANY USE OR EXPLOITATION OF THIS WORK WITHOUT
*   THE PRIOR WRITTEN CONSENT OF NOVELL.  ANY USE OR EXPLOITATION OF
*   THIS WORK WITHOUT AUTHORIZATION COULD SUBJECT THE PERPETRATOR TO
*   CRIMINAL AND CIVIL LIABILITY.
*
***********************************************************************
*
*  Name:			ODI NESL Definitions
*
*  FileName:		ODI_NESL.H
*
*  ODI Spec Ver:	1.11
*
*  Description:		C version of the NESL header file
*
*  Modification History:
*
*	11-13-96	WTT		Added NecbContext field to the NESL ECB.
*
*	01-22-97	WTT		Added define for NESL_NEB_DATA_TYPE.
*
\***********************************************************************/
#ifndef _ODI_NESL_EVENT_MOD_
#define _ODI_NESL_EVENT_MOD_

/* Include incase forgotten */
#include <odi.h>

#define EPBMajVer 1
#define EPBMinVer 00

#define EPB_GLOBAL_EVENT 1
#define EPB_SPECIFIC_EVENT 2

typedef struct EPB_tag
{
	UINT32	EPBMajorVersion;
	UINT32	EPBMinorVersion;
	void    *EPBEventName;
	void	  *EPBEventType;
	void	  *EPBModuleName;
	void	  *EPBDataPtr0;
	void	  *EPBDataPtr1;
	UINT32	EPBEventScope;
	UINT32	EPBReserved;
} EPB;

/***********************************************************************\
*	NESL_MAX_NAME_LENGTH 
*
*		is the maximum length (including the NULL terminator) 
*		of a NecbEventName. A buffer of this size must have been allocated 
*		to receive the return value from a NESLEnumerateEvents() function call.
\***********************************************************************/

#define	NESL_MAX_NAME_LENGTH 	81

/***********************************************************************\
* NESL_ECB Structure
* 
*	NecbNext			a link pointer used by NESL to link NESL_ECBs together into
*						processing lists
*	NecbVersion		identifies the support level expected by the application. 
*						This field allows the interface to be expanded in the future
*						while still providing full backwards compatibility
*						This should be set to NESL_VERSION1 (2).
*	NecbOsiLayer	a value used in sorting linked lists of consumer NESL_ECBs.
*						This field is used to sort the consumers according to the 
*						producer's specifications (either top-down or bottom-up 
*						order). The high order nibble should be the OSI layer number 
*						(0 - 7). The remaining 3 nibbles should define the relative
*						order with other consumers registered at the same OSI level.
*						(ie 0xLRRR)  This relative ordering is usefull for several
*						event consumers on the same OSI level which require a
*						specific processing order for the event.
*	NecbEventName	a pointer to an ASCIIZ string defining the name of the event
*						or class of events. This name may have a maximum length of
*						NESL_MAX_NAME_LENGTH.
*	NecbRefData		flags specified by producer NESL_ECBs defining whether the 
*						produced event is consumable or broadcast, and whether the 
*						producer is to be a unique producer or if other producers 
*						can register.  If the NESL_ECB is a consumer, this field is
*						not used by NESL.
*	PNecbNotifyProc the callback routine registered by a consumer of an event.
*						If the NESL_ECB is a producer, this field is not used by 
*						NESL.
*	NecbOwner		a platform specific field which identifies the owner module
*						of the NESL_ECB.  The DOS/Windows platforms use the module
*						handle of the NLM to specify the owner.
*	NecbWorkSpace	used by NESL to keep track of processing lists. This field
*						must not be used by the owner module while the NESL_ECB is
*						registered with NESL.
*
*	NecbContext			This field is available for use by the owner of
*						the NESL_ECB structure.  It will not be modified 
*						by anyone else in the system.  It may be used by
*						the owner to pass context or other data to the 
*						notification proceedure.  If the owner is not using 
*						this field it must be set to a NULL.
*
\***********************************************************************/

typedef struct NECBStruct
{
	struct NECBStruct	*NecbNext;
	UINT16				NecbVersion;
	UINT16				NecbOsiLayer;
	MEON_STRING			*NecbEventName;
	UINT32				NecbRefData;
	UINT32				(*PNecbNotifyProc)(
								struct NECBStruct *consumerNecb,
								struct NECBStruct *producerNecb,
								void   *eventData );
	void					*NecbOwner;
	void					*NecbWorkSpace;
	void					*NecbContext;
} NESL_ECB, *PNESL_ECB;


/***********************************************************************\
* PNecbNotifyPROC 
*
*		A typedef which defines the callback function for consumer NESL_ECBs.
*
\***********************************************************************/

typedef UINT32 (*PNecbNotifyPROC)(
	NESL_ECB *consumerNecb,
	NESL_ECB *producerNecb,
	void   *eventData );

/***********************************************************************\
*	NESL Return Codes
*
*		The return codes are all enumerated with the exception of 
*	NESL_EVENT_CONSUMED which is simply an alias for NESL_OK. See the 
*	function descriptions for a discussion of the meaning of any particular
*	return code.
\***********************************************************************/

#define NESL_EVENT_CONSUMED 	NESL_OK

enum {
	NESL_OK = 0,
	NESL_EVENT_NOT_CONSUMED,
	NESL_EVENT_BROADCAST,
	NESL_EVENT_NOT_REGISTERED,
	NESL_EVENT_TABLE_FULL,
	NESL_EVENT_IS_CONSUMABLE,
	NESL_EVENT_IS_NOT_CONSUMABLE,
	NESL_NO_MORE_EVENTS,
	NESL_PRODUCER_NOT_FOUND,
	NESL_CONSUMER_NOT_FOUND,
	NESL_INVALID_CONTEXT_HANDLE,
	NESL_INVALID_DESTINATION,
	NESL_REGISTERED_UNIQUE,
	NESL_REGISTERED_NOT_UNIQUE,
	NESL_REGISTERED_CONSUMABLE,
	NESL_REGISTERED_BROADCAST,
	NESL_REGISTERED_SORT_TOP_DOWN,
	NESL_REGISTERED_SORT_BOTTOM_UP,
	NESL_DUPLICATE_NECB,
	NESL_INVALID_NOTIFY_PROC,
   NESL_FIRST_ALREADY_HOOKED
};

/***********************************************************************\
* NESL NecbRefData Flags
*
*	This list of flags are used by producers to define their NecbRefData.
*
\***********************************************************************/

#define NESL_NOT_UNIQUE_PRODUCER       0x00000000
#define NESL_BROADCAST_EVENT           0x00000000
#define NESL_SORT_CONSUMER_BOTTOM_UP   0x00000001
#define NESL_SORT_CONSUMER_TOP_DOWN    0x00000000
#define NESL_CONSUME_EVENT			   0x00000002
#define NESL_UNIQUE_PRODUCER		   0x00000004
#define NESL_NEB_DATA_TYPE    		   0x00000008


/***********************************************************************\
* NESL Version flags
*
*	This flags define NESL_ECB version number
*
\***********************************************************************/

#define NESL_VERSION1                  1
#define NESL_VERSION2                  2

/***********************************************************************\
* NESL OSI layer definitions
*
*	These definitions are used for the NESL_ECB structure element NecbOsiLayer
*
\***********************************************************************/

#define  NESL_APPLICATION_LAYER        0x7000 
#define  NESL_PRESENTATION_LAYER       0x6000
#define  NESL_SESSION_LAYER            0x5000
#define  NESL_TRANSPORT_LAYER          0x4000
#define  NESL_NETWORK_LAYER            0x3000
#define  NESL_DATALINK_LAYER           0x2000
#define  NESL_PHYSICAL_LAYER           0x1000
#define  NESL_NIOS_LAYER               0x0000

/***********************************************************************\
*
* exported function prototypes
*
\***********************************************************************/

UINT32 NESLRegisterProducer( NESL_ECB *Producer );
UINT32 NESLDeRegisterProducer( NESL_ECB *Producer );
UINT32 NESLRegisterConsumer( NESL_ECB *Consumer );
UINT32 NESLDeRegisterConsumer( NESL_ECB *Consumer );
UINT32 NESLProduceEvent( NESL_ECB *Producer, NESL_ECB **Consumer, void *EventData );
UINT32 NESLProduceMLIDEvent( NESL_ECB *Producer, NESL_ECB **Consumer, 
                             void *EventData, struct _DRIVER_DATA_ *driverData);

#endif				/* End of File */
